/**
 * 
 */
package de.uni_stuttgart.fmi.szs.jmoped.eclipse;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;

import org.apache.log4j.Logger;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceProxyVisitor;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.resources.ResourceAttributes;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.QualifiedName;
import org.eclipse.core.runtime.jobs.ISchedulingRule;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.ui.texteditor.MarkerUtilities;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageAdapter;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Transition;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.marker.MarkerManager;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.marker.MarkerUtils;

public class TestCoverageListener extends CoverageAdapter
{
	private IResource resource;
	
	/**
	 * Marker for previously failed resource lookups.
	 */
	private static IResource NOT_FOUND_RESOURCE = new NotFoundResource();
	
	private Map<String, IResource> resourcesByClassName = 
		new HashMap<String, IResource>();
	
	private String launchConfigurationMemento;
	
	private CoverageDataQueue workQueue = new CoverageDataQueue();
	
	private QueueRunner runner = new QueueRunner();
	
	private static Logger logger = Logger.getLogger(TestCoverageListener.class);

	public TestCoverageListener(IMethod method, String launchConfigurationMemento)
	{
		this(method.getResource(), launchConfigurationMemento);
		String className = method.getDeclaringType().getFullyQualifiedName().replace('.', '/');
		resourcesByClassName.put(className, resource);
	}
	
	public TestCoverageListener(IResource resource, String launchConfigurationMemento) 
	{
		if (resource == null) {
			throw new NullPointerException("resource must not be null");
		}
		if (launchConfigurationMemento == null) {
			throw new NullPointerException("launchConfigurationMemento must not be null");
		}
		this.resource = resource;
		this.launchConfigurationMemento = launchConfigurationMemento;
	}

	/**
	 * Deletes all existing coverage marker of the project this resource 
	 * belongs to.
	 */
	@Override
	public void coverageStarted(PDS pds) 
	{
		MCPlugin.log("coverage started");
		if (launchConfigurationMemento != null) {
			MarkerManager.getInstance().removeMarkerMap(launchConfigurationMemento);
			MarkerManager.getInstance().addMarkerMap(launchConfigurationMemento, new HashMap());
			MarkerManager.getInstance().clearCoverageResults();
		}
		startRunner(true);
	}
	
	protected void startRunner(boolean deleteMarkers)
	{
		runner.start(deleteMarkers);
	}
	
	/**
	 * Returns null if resource could not be found
	 * @param className
	 * @return
	 */
	protected IResource getResource(String className)
	{
		IResource resource = resourcesByClassName.get(className);
		if (resource == null) {
			resource = SearchUtils.searchResource(className);
			if (resource != null) {
				resourcesByClassName.put(className, resource);
			}
			else {
				resource = NOT_FOUND_RESOURCE;
				resourcesByClassName.put(className, resource);
			}
		}
		return resource != NOT_FOUND_RESOURCE ? resource : null;
	}

	/**
	 * Write the result to the lower part of the plug-in at the end of
	 * the coverage analysis.
	 * 
	 * @param result the CoverageResult 
	 */
	@Override
	public void coverageEnded(CoverageResult result)
	{
		MCPlugin.log("coveraged ended");
		workQueue.drain();
		MarkerManager mManager = MarkerManager.getInstance();
		mManager.addCoverageResult(launchConfigurationMemento, result);
		
		final CoverageMethod[] methods = result.getMethods();
		
		final IWorkspaceRunnable runnable = new IWorkspaceRunnable() {
			public void run(IProgressMonitor monitor) throws CoreException {
				for (int i = 0; i < methods.length; i++) {
					CoverageMethod method = methods[i];

					IResource resource = getResource(method.getClassName());
					if (resource == null) {
						continue;
					}
					
					IDocument document = SearchUtils.getDocument(resource.getFullPath());
					
					for (int j = 0; j < method.getLineTable().length; j++) {
						if (method.isEffectiveLine(j) && !method.clinitIsAssertionOnly()
								&& method.getLineState(j) == State.NOT_COVERED) {
							IMarker marker =
								resource.createMarker(MCPlugin.MARKER_NOT_COVERED_ID);
							int line = method.getLineNumber(j);
							HashMap<String, Object> attributes = new HashMap<String, Object>();
							attributes.put(IMarker.LINE_NUMBER, line);
							attributes.put(MCPlugin.ATTR_METHOD_NAME, 
									method.getMethod().getFormattedName());
							attributes.put(MCPlugin.ATTR_LAUNCH_CONFIGURATION, 
									launchConfigurationMemento);
							attributes.put(MCPlugin.ATTR_LINE_OFFSET, j);
							marker.setAttributes(attributes);
							Map coveredLines = MarkerManager.getInstance().getMarkerMap
								(launchConfigurationMemento);
							//Remy
							//coveredLines.put(new CoverageKey(method, j),
							coveredLines.put(new CoverageKey(method, line),
									new MarkerStatePair(marker, State.NOT_COVERED));
							if (document != null && line > 0) {
								try {
									int start = document.getLineOffset(line - 1);
									int end = start + document.getLineLength(line - 1);
									MarkerUtilities.setCharStart(marker, start);
									MarkerUtilities.setCharEnd(marker, end);
								}
								catch (BadLocationException e) {
									e.printStackTrace();
								}
							}
						}
					}
				}
			}
		};
		
			new Thread(new Runnable() {
				public void run() {
					try {
						resource.getWorkspace().run(runnable, null);
					}
					catch (CoreException e) {
						MCPlugin.log(e);
					}
				}
			}).start();
	}

	@Override
	public void coverageFailed(String message, Throwable t) 
	{
		MCPlugin.log("coverage failed " + t.getMessage(), t);
		workQueue.drain();
	}

	@Override
	public void coveredMethodOffset(CoverageMethod method, int lineOffset, 
			int pcOffset, State state, Transition transition)
	{
		workQueue.add(new CoverageData(method, lineOffset, pcOffset, state));
	}
	
	/**
	 * Mark the line with state.
	 * 
	 * @param method		the coverage method
	 * @param lineOffset	the entry of LineNumberTable from jclasslib
	 * @param pcOffset		the bytecode offset
	 * @param state			the state of the line
	 */
	private void  markLineCovered(final CoverageMethod method, final int lineOffset, 
			final int pcOffset, State state) 
	{
//		if (!method.isEffectiveLine(lineOffset)) {
//			return;
//		}
		// don't mark assertion only clinits, since there is no corresponding source code
		if (method.clinitIsAssertionOnly()) {
			return;
		}
		
		try {
			//logger.debug("method: " + method.toString() 
			//		+ ", pcOffset: " + pcOffset + ", lineOffset: " 
			//		+ lineOffset + ", state: " + state.toString());
			int line = method.getLineNumber(lineOffset);
			//Remy
			//CoverageKey key = new CoverageKey(method, lineOffset);
			CoverageKey key = new CoverageKey(method, line);
			MarkerManager manager = MarkerManager.getInstance();
			Map coveredLines = manager.getMarkerMap(launchConfigurationMemento);
			MarkerStatePair pair = (MarkerStatePair)coveredLines.get(key);
			if (pair != null) {
				//logger.debug("method: " + method.toString() 
				//		+ ", pcOffset: " + pcOffset + ", lineOffset: " 
				//		+ lineOffset + ", pair.state: "
				//		+ pair.state.toString() + ", state: " + state.toString());
				if (pair.state == state 
						|| (pair.state  == State.ASSERTION_FAILED 
								&& state != State.HEAP_OVERFLOW) 
						|| pair.state == State.HEAP_OVERFLOW) {
					return;
				}
			}
			else {
				pair = new MarkerStatePair(state);
				coveredLines.put(key, pair);
			}
			MarkerRunnable runnable =
				new MarkerRunnable(method, lineOffset, pcOffset, pair, state);
			try {
				runnable.run(new NullProgressMonitor());
			} 
			catch (CoreException e) {
				MCPlugin.log(e);
			}
		}
		catch (Throwable t) {
			t.printStackTrace();
		}
	}
	
	/**
	 * The thread that put markers.
	 * 
	 * @author suwimodh
	 *
	 */
	private class MarkerRunnable implements IWorkspaceRunnable
	{
		private CoverageMethod method;
		private int lineOffset;
		private int pcOffset;
		private MarkerStatePair pair;
		private State state;
		
		public MarkerRunnable(CoverageMethod method, int lineOffset, int pcOffset,
				MarkerStatePair pair, State state)
		{
			this.method = method;
			this.lineOffset = lineOffset;
			this.pcOffset = pcOffset;
			this.pair = pair;
			this.state = state;
		}

		public void run(IProgressMonitor monitor) throws CoreException 
		{	
			// delete old marker, if there
			if (pair.marker != null) {
				pair.marker.delete();
			}
			
			String id = MCPlugin.getMarkerId(state);
			int line = method.getLineNumber(lineOffset);
			
			IResource resource = getResource(method.getClassName());
			if (resource == null) {
				return;
			}
			IDocument document = SearchUtils.getDocument(resource.getFullPath());
			
			// set new marker
			pair.marker = resource.createMarker(id);
			HashMap<String, Object> attributes = new HashMap<String, Object>();
			attributes.put(IMarker.LINE_NUMBER, line);
			attributes.put(MCPlugin.ATTR_LAUNCH_CONFIGURATION, 
					launchConfigurationMemento);
			attributes.put(IMarker.PRIORITY, IMarker.PRIORITY_NORMAL);
			attributes.put(IMarker.SEVERITY, IMarker.SEVERITY_INFO);
			attributes.put(MCPlugin.ATTR_METHOD_NAME, 
					method.getMethod().getFormattedName());
			attributes.put(MCPlugin.ATTR_PC_OFFSET, pcOffset);
			attributes.put(MCPlugin.ATTR_LINE_OFFSET, lineOffset);
			pair.marker.setAttributes(attributes);
			
			logger.debug("method: " + method.toString() 
					+ ", pcOffset: " + pcOffset + ", line: " 
					+ line + ", pair.state: "
					+ pair.state.toString() + ", state: " + state.toString());
			
			if (document != null && line > 0) {
				try {
					int start = document.getLineOffset(line - 1);
					int end = start + document.getLineLength(line - 1);
					MarkerUtilities.setCharStart(pair.marker, start);
					MarkerUtilities.setCharEnd(pair.marker, end);
				}
				catch (BadLocationException e) {
					e.printStackTrace();
				}
			}
			
			// set new state
			pair.state = state;
		}
	}
	
	/**
	 * A private class used as keys in a {@link java.util.Map<key,value>}.
	 * The class MarkerStatePair is supposed to be used as values.
	 * 
	 * History: the field <code>line</code> was originally implemented
	 * as line_number (or lineOffset) obtained from LineNumberTable
	 * in jclasslib. However, since a line in source code can correspond to
	 * more than one groups in the bytecode, line_number can have the
	 * same value for different parts of the bytecode. This resulted in
	 * more than one key for the same (source code) line. As a consequence, 
	 * more than one marker could be produced for a line. The result might
	 * be invisible, but when the markers were clicked, they generated
	 * resolutions two times.
	 * 
	 * @author suwimodh
	 *
	 */
	private static class CoverageKey
	{
		CoverageMethod method;
		int line;
		int hash;
		
		public CoverageKey(CoverageMethod method, int line)
		{
			this.method = method;
			this.line = line;
			int mHash = method.getMethod().getFormattedName().hashCode();
			hash = ((line + mHash) * (line + mHash + 1)) / 2 + line;
		}
		
		public boolean equals(Object obj) 
		{
			if (hash != obj.hashCode()) {
				return false;
			}
			CoverageKey other = (CoverageKey)obj;
			return line == other.line 
			&& method.getMethod().getFormattedName().equals(other.method.getMethod().getFormattedName());
		}
		
		@Override
		public int hashCode() 
		{
			return hash;
		}
		
	}
	
	/**
	 * A private class used as value in a {@link java.util.Map<key,value>}.
	 * The class CoverageKey is supposed to be used as keys.
	 * 
	 * @author suwimodh
	 *
	 */
	private static class MarkerStatePair
	{
		IMarker marker;
		State state;
		
		public MarkerStatePair(State state)
		{
			this.state = state;
		}
		
		public MarkerStatePair(IMarker marker, State state)
		{
			this.marker = marker;
			this.state = state;
		}
	}
	
	private static class CoverageData
	{
		CoverageMethod method; 
		int lineOffset;
		int pcOffset;
		State state;
		
		public CoverageData(CoverageMethod method, int offset, int offset2, State state) 
		{
			this.method = method;
			lineOffset = offset;
			pcOffset = offset2;
			this.state = state;
		}
	}
	
	private static class CoverageDataQueue
	{
		Queue<CoverageData> queue = new LinkedList<CoverageData>();
		private boolean drain = false;
		
		public synchronized void add(CoverageData data)
		{
			queue.add(data);
			notify();
		}
		
		public synchronized void drain()
		{
			drain = true;
			notify();
		}
		
		public synchronized CoverageData[] get() throws InterruptedException
		{
			while (!drain && queue.isEmpty()) {
				wait();
			}
			CoverageData[] data = queue.toArray(new CoverageData[queue.size()]);
			queue.clear();
			return data;
		}
	}
	
	private class QueueRunner implements Runnable
	{
		private boolean die = false;		
		private Thread t;
		private boolean deleteMarkers;

		public void start(boolean deleteMarkers)
		{
			this.deleteMarkers = deleteMarkers;
			die = false;
			t = new Thread(this, "MarkerRunner");
			t.start();
		}

		public void stop()
		{
			die = true;
			if (t != null) {
				t.interrupt();
			}
		}

		public void run() 
		{
			if (deleteMarkers) {
				try {
					// TODO put this in queue as well
					MarkerUtils.deleteAllCoverageMarkers(resource);
				}
				catch (CoreException ce) {
					MCPlugin.log(ce);
				}
			}
			
			while (!die) {
				CoverageData[] data = new CoverageData[0];
				try {
					data = workQueue.get();
				}
				catch (InterruptedException ie) {
					continue;
				}
				if (data.length == 0) {
					return;
				}
				
				if (resource != null) {
					CoverageDataRunnable runnable = new CoverageDataRunnable(data);
					try {
						resource.getWorkspace().run(runnable, null);
					} 
					catch (CoreException e) {
						MCPlugin.log(e);
					}
				}
			}
		}
	}
	
	private class CoverageDataRunnable implements IWorkspaceRunnable
	{
		private CoverageData[] data;
		
		public CoverageDataRunnable(CoverageData[] data)
		{
			this.data = data;
		}

		public void run(IProgressMonitor monitor) throws CoreException 
		{
			if (monitor == null) {
				monitor = new NullProgressMonitor();
			}
			monitor.beginTask("Setting markers", data.length);
			for (CoverageData d : data) {
				markLineCovered(d.method, d.lineOffset, d.pcOffset, d.state);
				monitor.worked(1);
			}
			monitor.done();
		}
		
	}
	
	private static class NotFoundResource implements IResource
	{

		public void accept(IResourceProxyVisitor visitor, int memberFlags) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void accept(IResourceVisitor visitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void accept(IResourceVisitor visitor, int depth, boolean includePhantoms) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void accept(IResourceVisitor visitor, int depth, int memberFlags) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void clearHistory(IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void copy(IPath destination, boolean force, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void copy(IPath destination, int updateFlags, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void copy(IProjectDescription description, boolean force, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void copy(IProjectDescription description, int updateFlags, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public IMarker createMarker(String type) throws CoreException {
			// TODO Auto-generated method stub
			return null;
		}

		public void delete(boolean force, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void delete(int updateFlags, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void deleteMarkers(String type, boolean includeSubtypes, int depth) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public boolean exists() {
			// TODO Auto-generated method stub
			return false;
		}

		public IMarker findMarker(long id) throws CoreException {
			// TODO Auto-generated method stub
			return null;
		}

		public IMarker[] findMarkers(String type, boolean includeSubtypes, int depth) throws CoreException {
			// TODO Auto-generated method stub
			return null;
		}

		public String getFileExtension() {
			// TODO Auto-generated method stub
			return null;
		}

		public IPath getFullPath() {
			// TODO Auto-generated method stub
			return null;
		}

		public long getLocalTimeStamp() {
			// TODO Auto-generated method stub
			return 0;
		}

		public IPath getLocation() {
			// TODO Auto-generated method stub
			return null;
		}

		public IMarker getMarker(long id) {
			// TODO Auto-generated method stub
			return null;
		}

		public long getModificationStamp() {
			// TODO Auto-generated method stub
			return 0;
		}

		public String getName() {
			// TODO Auto-generated method stub
			return null;
		}

		public IContainer getParent() {
			// TODO Auto-generated method stub
			return null;
		}

		public String getPersistentProperty(QualifiedName key) throws CoreException {
			// TODO Auto-generated method stub
			return null;
		}

		public IProject getProject() {
			// TODO Auto-generated method stub
			return null;
		}

		public IPath getProjectRelativePath() {
			// TODO Auto-generated method stub
			return null;
		}

		public IPath getRawLocation() {
			// TODO Auto-generated method stub
			return null;
		}

		public ResourceAttributes getResourceAttributes() {
			// TODO Auto-generated method stub
			return null;
		}

		public Object getSessionProperty(QualifiedName key) throws CoreException {
			// TODO Auto-generated method stub
			return null;
		}

		public int getType() {
			// TODO Auto-generated method stub
			return 0;
		}

		public IWorkspace getWorkspace() {
			// TODO Auto-generated method stub
			return null;
		}

		public boolean isAccessible() {
			// TODO Auto-generated method stub
			return false;
		}

		public boolean isDerived() {
			// TODO Auto-generated method stub
			return false;
		}

		public boolean isLocal(int depth) {
			// TODO Auto-generated method stub
			return false;
		}

		public boolean isLinked() {
			// TODO Auto-generated method stub
			return false;
		}

		public boolean isPhantom() {
			// TODO Auto-generated method stub
			return false;
		}

		public boolean isReadOnly() {
			// TODO Auto-generated method stub
			return false;
		}

		public boolean isSynchronized(int depth) {
			// TODO Auto-generated method stub
			return false;
		}

		public boolean isTeamPrivateMember() {
			// TODO Auto-generated method stub
			return false;
		}

		public void move(IPath destination, boolean force, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void move(IPath destination, int updateFlags, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void move(IProjectDescription description, boolean force, boolean keepHistory, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void move(IProjectDescription description, int updateFlags, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void refreshLocal(int depth, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void revertModificationStamp(long value) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void setDerived(boolean isDerived) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void setLocal(boolean flag, int depth, IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public long setLocalTimeStamp(long value) throws CoreException {
			// TODO Auto-generated method stub
			return 0;
		}

		public void setPersistentProperty(QualifiedName key, String value) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void setReadOnly(boolean readOnly) {
			// TODO Auto-generated method stub
			
		}

		public void setResourceAttributes(ResourceAttributes attributes) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void setSessionProperty(QualifiedName key, Object value) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void setTeamPrivateMember(boolean isTeamPrivate) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public void touch(IProgressMonitor monitor) throws CoreException {
			// TODO Auto-generated method stub
			
		}

		public Object getAdapter(Class adapter) {
			// TODO Auto-generated method stub
			return null;
		}

		public boolean contains(ISchedulingRule rule) {
			// TODO Auto-generated method stub
			return false;
		}

		public boolean isConflicting(ISchedulingRule rule) {
			// TODO Auto-generated method stub
			return false;
		}
		
	}
	
}
