package de.uni_stuttgart.fmi.szs.jmoped.eclipse.gui;

import java.lang.ref.WeakReference;
import java.util.EnumMap;
import java.util.Map;

import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.swt.graphics.Image;
import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.attributes.LineNumberTableEntry;

import de.uni_stuttgart.fmi.szs.jmoped.BytecodeUtils;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInst;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageResult;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.DescriptorParser;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.JavaAssertionExpressionResolver;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.JavaHeapOverflowResolver;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Line;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.Parameter;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;
import de.uni_stuttgart.fmi.szs.jmoped.eclipse.MCPlugin;

public class CoverageMethodTreeLabelProvider extends LabelProvider 
	implements ITableLabelProvider
{
	
	private Map<State, Image> imagesByState = new EnumMap<State, Image>(State.class);
	
	private WeakReference<CoverageResult> result;
	
	public CoverageMethodTreeLabelProvider()
	{
		for (State state : State.values()) {
			imagesByState.put(state, MCPlugin.getImage(state).createImage());
		}
	}

	@Override
	public String getText(Object element) 
	{
		if (element instanceof CoverageMethod) {
			return ((CoverageMethod)element).getHumanReadableName();
		}
		else if (element instanceof Line) {
			return element.toString();
		}
		else if (element instanceof InstructionState) {
			InstructionState is = (InstructionState)element;
			AbstractInstruction inst = is.getInstruction().getInstruction();
			StringBuilder builder = new StringBuilder(is.getState().toString());
			builder.append(": ");
			builder.append(inst.getOpcodeVerbose());
			try {
				switch (inst.getOpcode()) {
				case Opcodes.OPCODE_NEW:
					ClassFile classFile = is.getLine().getMethod().getClassFile();
					String className = PDSInst.getNewedClassName(classFile, inst);
					className = className.substring(className.lastIndexOf('/') + 1);
					builder.append(" ").append(className);
					break;
				case Opcodes.OPCODE_INVOKESTATIC:
				case Opcodes.OPCODE_INVOKESPECIAL:
				case Opcodes.OPCODE_INVOKEINTERFACE:
				case Opcodes.OPCODE_INVOKEVIRTUAL:
					classFile = is.getLine().getMethod().getClassFile();
					builder.append(" ");
					builder.append(concatMethodName(BytecodeUtils.getReferencedName(classFile, inst))); 
					break;
				}
			}
			catch (InvalidByteCodeException ive) {
			}
			return builder.toString();
		}
		else if (element instanceof TransitionState) {
			TransitionState trs = (TransitionState)element;
			if (trs.getState() == State.HEAP_OVERFLOW) {
				CoverageResult res = getResult();
				if (res != null) {
					try {
						int missing = JavaHeapOverflowResolver.getMissingHeapFields(res,
								trs.getTransition());
						return "Heap was " + missing + " fields short";
					}
					catch (InvalidByteCodeException e) {
						MCPlugin.log(e);
					}
				}
			}
			else if (trs.getState() == State.ASSERTION_FAILED) {
				CoverageResult res = getResult();
				if (res != null) {
					try {
						String failedExpression = 
							JavaAssertionExpressionResolver.getFailedAssertionExpression
							(res, trs.getTransition());
						if (failedExpression != null) {
							return trs.getState().toString() + ": " + failedExpression;
						}
					}
					catch (InvalidByteCodeException e) {
						MCPlugin.log(e);
					}
				}
			}
			return trs.getState().toString();				
		}
		return "no label???";
	}
	
	private String concatMethodName(String[] methods)
	{
		StringBuilder builder =
			new StringBuilder(PDSMethod.createHumanReadableName(methods[0], methods[1]));
		try {
			Parameter[] params = DescriptorParser.parse(methods[2]);
			builder.append("(");
			for (int i = 0; i < params.length; i++) {
				builder.append(params[i]);
				if (i < params.length - 1) {
					builder.append(", ");
				}
			}
			builder.append(")");
		}
		catch (RuntimeException re) {
			builder.append("()");
		}
		return builder.toString();
	}
	
	public void setResult(CoverageResult result)
	{
		this.result = new WeakReference<CoverageResult>(result);
	}
	
	public CoverageResult getResult()
	{
		return result == null ? null : result.get();
	}
	
	@Override
	public Image getImage(Object element) 
	{
		if (element instanceof Line) {
			Line line = (Line)element;
			State state = line.getState();
			return  imagesByState.get(state);
		}
		else if (element instanceof CoverageMethod) {
			CoverageMethod method = (CoverageMethod)element;
			if (method.isFullyCovered()) {
				return imagesByState.get(State.COVERED);
			}
			else {
				return imagesByState.get(State.PARTIALLY_COVERED);
			}
		}
		else if (element instanceof TransitionState) {
			return imagesByState.get(((TransitionState)element).getState());
		}
		else if (element instanceof InstructionState) {
			return imagesByState.get(((InstructionState)element).getState());
		}
		return null;
	}

	@Override
	public void dispose() 
	{
		for (Image image : imagesByState.values()) {
			image.dispose();
		}
		imagesByState.clear();
		if (result != null) {
			result.clear();
		}
	}

	public Image getColumnImage(Object element, int columnIndex) 
	{
		if (columnIndex == 2) {
			return getImage(element);
		}
		return null;
	}

	public String getColumnText(Object element, int columnIndex) 
	{
		if (columnIndex == 0) {
			return getText(element);
		}
		else if (columnIndex == 1) {
			if (element instanceof CoverageMethod) {
				CoverageMethod method = (CoverageMethod)element;
				LineNumberTableEntry[] lines = method.getLineTable();
				return lines[0].getLineNumber() + "-" + lines[lines.length - 1].getLineNumber();
			}
			else if (element instanceof Line) {
				return String.valueOf(((Line)element).getLineNumber());
			}
			else if (element instanceof TransitionState) {
				return String.valueOf(((TransitionState)element).getLine().getLineNumber());
			}
			else if (element instanceof InstructionState) {
				return String.valueOf(((InstructionState)element).getLine().getLineNumber());
			}
		}
		return null;
	}

}
