package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.BranchInstruction;
import org.gjt.jclasslib.bytecode.ImmediateShortInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;
import org.gjt.jclasslib.io.ClassFileReader;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.constants.ConstantClassInfo;
import org.gjt.jclasslib.structures.constants.ConstantNameAndTypeInfo;
import org.gjt.jclasslib.structures.constants.ConstantReference;

/**
 * Utility class for manipulating bytecodes.
 * 
 * @author suwimodh
 *
 */
public class BytecodeUtils {

	/**
	 * Returns the target label of the branch instruction inst.
	 * inst must be of type BranchInstruction.
	 * 
	 * @param methodName the method name.
	 * @param inst the abstract instruction.
	 * @return the target label.
	 */
	public static String branchTarget(String methodName, AbstractInstruction inst) {
		
		BranchInstruction branchInst = (BranchInstruction) inst;
		int target = branchInst.getOffset() + branchInst.getBranchOffset();
		
		return (target == 0) ? methodName :  methodName + String.valueOf(target);
	}
	
	public static String getInterfaceName(ClassFile cf, int index)
			throws InvalidByteCodeException {
		
		ConstantClassInfo info = (ConstantClassInfo) cf.getConstantPoolEntry(
				index, ConstantClassInfo.class);
		return info.getName();
	}
	
	/**
	 * Returns names of class and method and method's descriptor 
	 * in <code>invokevirtual</code>, <code>invokespecial</code>, 
	 * and <code>invokestatic</code> instruction.
	 *
	 * @param classFile the jclasslib class file.
	 * @param inst the <code>invokevirtual</code>, 
	 *				<code>invokespecial</code>, or <code>invokestatic</code>
	 *				instruction.
	 * @return array of string containing the class name at index 0,
	 *				 method name at index 1, and method descriptor at
	 *				 index 2.
	 */
	public static String[] getReferencedName(ClassFile cFile,
		AbstractInstruction aInst) throws InvalidByteCodeException {
			
		ImmediateShortInstruction isInst = (ImmediateShortInstruction) aInst;
		ConstantReference constRef = (ConstantReference) 
			cFile.getConstantPoolEntry(isInst.getImmediateShort(), 
					ConstantReference.class);
		ConstantClassInfo ccInfo = constRef.getClassInfo();
		ConstantNameAndTypeInfo cnatInfo = constRef.getNameAndTypeInfo();
		
		String[] invokedName = new String[3];
		invokedName[0] = ccInfo.getName();
		invokedName[1] = cnatInfo.getName();
		invokedName[2] = cnatInfo.getDescriptor();
		
		return invokedName;
	}
	
	public static ClassFile findClassFile(String className) throws IOException,
		InvalidByteCodeException
	{
		return findClassFile(className, new String[0]);
	}
	
	/**
	 * Searches all classpath directories for the file and all jars, otherwise
	 * looks into the rt.jar
	 * 
	 * @param className packages are expected to be separated by a '/' 
	 * or '.' , e.g. "java/util/StringTokenizer" or "java.util.StringTokenizer"
	 * @throws Exception if there was an error reading the class file or the
	 * file was not found
	 */
	public static ClassFile findClassFile(String className,
			String[] pathEntries) throws InvalidByteCodeException, IOException {
		
		// join given paths and class path entries
		String classPath = System.getProperty("java.class.path",".");
		String[] splitClassPath = 
			classPath.split(System.getProperty("path.separator"));
		String[] resultPaths = new String[pathEntries.length 
										  + splitClassPath.length];
		System.arraycopy(pathEntries, 0, resultPaths, 0, pathEntries.length);
		System.arraycopy(splitClassPath, 0, resultPaths, pathEntries.length, 
				splitClassPath.length);
		
		String subFilename = className.replace('.', File.separatorChar);
		subFilename = subFilename.replace('/', File.separatorChar) + ".class";
		
		for (int i = 0; i < resultPaths.length; i++) {
			File f = new File(resultPaths[i]);
			if (f.isFile()) {
				ClassFile classFile = findClassFile(new JarFile(f),
													subFilename);
				if (classFile != null) {
					return classFile;
				}
			}
			else {
				File possibleClassFile = new File(f, subFilename);
				if (possibleClassFile.isFile()) {
					return ClassFileReader.readFromFile(possibleClassFile);
				}
			}
		}
		// fall back on rt.jar.
		File rtFile = new File(PDSDefault.JAVA_HOME,  File.separator + "jre" 
				+ File.separator + "lib" + File.separator + "rt.jar");
		if (rtFile.isFile()) {
			ClassFile classFile = findClassFile(new JarFile(rtFile), subFilename);
			if (classFile != null) {
				return classFile;
			}
		}
		throw new FileNotFoundException("File for class " + className 
										+ " not found");
	}
	
	/**
	 * Converts the filesystem path to a jar path and looks it up in the
	 * jarFile. 
	 * <p>
	 * If successful a class file object is returned.
	 * @throws Exception
	 */
	public static ClassFile findClassFile(JarFile jarFile, String path)
			throws InvalidByteCodeException, IOException {
		
	    try {
	    	String jarPath = path.replace(File.separatorChar, '/');
	        JarEntry jarEntry = jarFile.getJarEntry(jarPath);
	        if (jarEntry != null) {
	        	return ClassFileReader.readFromInputStream
					(jarFile.getInputStream(jarEntry));
	        }
	    } 
	    finally {
	        jarFile.close();
	    }	
	    return null;
	}
	
	/**
	 * Returns true if the instruction gets the static "assertionsDisabled" field.
	 * @param instruction
	 * @return
	 * @throws InvalidByteCodeException 
	 */
	public static boolean getsAssertionsDisabledField(ClassFile classFile,
			AbstractInstruction instruction) throws InvalidByteCodeException
	{
		if (instruction.getOpcode() == Opcodes.OPCODE_GETSTATIC) {
			String[] names = getReferencedName(classFile, instruction);
			return names[1].contains("assertionsDisabled");
		}
		return false;
	}
	
	public static boolean isAssertionErrorConstructor(ClassFile classFile, AbstractInstruction inst)
		throws InvalidByteCodeException
	{
		if (inst.getOpcode() == Opcodes.OPCODE_INVOKESPECIAL) {
			String[] calledName = getReferencedName(classFile, inst);
			if (calledName[0].equals("java/lang/AssertionError")
					&& calledName[1].equals("<init>")) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Returns <code>true</code> if opcode is one of if-instructions,
	 * <code>false</code> otherwise.
	 * 
	 * @param opcode the opcode
	 * @return <code>true</code> if opcode is one of if-instructions.
	 */
	public static boolean isComparison(int opcode) {
		
		if (opcode >= Opcodes.OPCODE_IFEQ 
				&& opcode <= Opcodes.OPCODE_IF_ACMPNE) {
			return true;
		}
		if (opcode >= Opcodes.OPCODE_IFNULL 
				&& opcode <= Opcodes.OPCODE_IFNONNULL) {
			return true;
		}
		return false;
	}
}
