package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.log4j.Logger;
import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;
import org.gjt.jclasslib.io.ByteCodeReader;
import org.gjt.jclasslib.io.ClassFileReader;
import org.gjt.jclasslib.structures.AccessFlags;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.MethodInfo;
import org.gjt.jclasslib.structures.attributes.CodeAttribute;
import org.gjt.jclasslib.structures.constants.ConstantClassInfo;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import static de.uni_stuttgart.fmi.szs.jmoped.PDSDefault.IGNORED_CLASS;
import static de.uni_stuttgart.fmi.szs.jmoped.PDSDefault.IGNORED_METHOD;

public class InvokeAnalyzer {

	private File rootFile;
	private ClassFile rootClassFile;
	private MethodInfo rootMethodInfo;
	private String[] searchPaths;
	
	private Map<String, ClassInfo> MClassInfo = new HashMap<String, ClassInfo>();
	private Map<String, InterfaceNode> MInterfaceNode = new HashMap<String, InterfaceNode>();
	private Map<String, ClassNode> MClassNode = new HashMap<String, ClassNode>();
	
	/**
	 * The set of virtual method invocations. They will be traversed when a new
	 * class is instantiated, to see if it provides an implementation for a method
	 * invocation. For this to hold, the class has to be a subclass of the class of
	 * the method invocation and it must provide a non private implementation
	 * of the method.
	 */
	private Set<MethodInvocation> virtualInvocations 
		= new HashSet<MethodInvocation>();
	/**
	 * The set of interface invocations. When a new class is instantiated they will
	 * be traversed for each interface that the new class implements. And 
	 * {@link #addVirtualMethodForClass(String, MethodInvocation)} if the match.
	 */
	private Set<MethodInvocation> interfaceInvocations
		= new HashSet<MethodInvocation>();
	
	Logger logger = Logger.getLogger(InvokeAnalyzer.class);
	
	/**
	 * Constructor.
	 * 
	 * @param methodInfo
	 * @throws Exception
	 */
	public InvokeAnalyzer(File rootFile, String confName, String methodName, 
			String methodDesc) throws Exception {
		
		this.rootFile = rootFile;
		this.rootClassFile = ClassFileReader.readFromFile(rootFile);
		initialize(confName, methodName, methodDesc);
		logger.debug(includedMethodsInfo());
	}
	
	public InvokeAnalyzer(ClassFile classFile, String confName, 
			String methodName, String methodDesc, String[] searchPaths) 
			throws Exception {
		
		this.rootClassFile = classFile;
		this.searchPaths = searchPaths;
		initialize(confName, methodName, methodDesc);
		logger.debug(includedMethodsInfo());
	}
	
	/**
	 * Takes an array consisting of a class name, a method name and its 
	 * signature and checks if the method should be ignored in jmoped's
	 * output.
	 * @param names
	 * @return
	 */
	public boolean isIgnoredMethod(String[] names) {
		
		return isIgnoredMethod(names[0], names[1], names[2]);
	}
	
	public boolean isIgnoredMethod(String className, String methodName,
			String methodDescriptor) {
		
		ClassInfo classInfo = MClassInfo.get(className);
		if (classInfo == null) return false;
		
		return classInfo.isIgnored(methodName + methodDescriptor);
	}
	
	/**
	 * Takes an array of class name, method name and signature
	 * and returns true if the method is already due to be included.
	 * @param names
	 * @return
	 */
	public boolean isIncludedMethod(String[] names) {
		
		return isIncludedMethod(names[0], names[1], names[2]);
	}
	
	public boolean isIncludedMethod(String className, String methodName,
			String methodDescriptor) {
		
		ClassInfo classInfo = MClassInfo.get(className);
		if (classInfo == null) return false;
		
		return classInfo.isIncluded(methodName + methodDescriptor);
	}
	
	public boolean isIncludedMethod(String className, String methodName) {
		
		ClassInfo classInfo = MClassInfo.get(className);
		if (classInfo == null) return false;
		
		return classInfo.isMethodIncluded(methodName);
	}
	
	/**
	 * Returns true if the class should be ignored while generating
	 * Remopla output.
	 * 
	 * @param className
	 * @return
	 */
	public boolean isIgnoredClass(String className) {
		
		ClassInfo classInfo = MClassInfo.get(className);
		if (classInfo == null) return false;
		
		return classInfo.isIgnored();
	}
	
	/**
	 * Returns true if the class is already included.
	 * 
	 * @param className
	 * @return
	 */
	public boolean isIncludedClass(String className) {
		
		ClassInfo classInfo = MClassInfo.get(className);
		if (classInfo == null) return false;
		
		return classInfo.isIncluded();
	}
	
	/**
	 * Returns class file if the given className, or <code>null</code> if
	 * it does not exist.
	 * 
	 * @param className the class name
	 * @return the class file
	 */
	public ClassFile getClassFile(String className) {
		
		ClassInfo classInfo = MClassInfo.get(className);
		if (classInfo == null) return null;
		
		return classInfo.getClassFile();
	}
	
	public Set<ClassFile> getClassFiles() {
		
		Set<ClassFile> classFiles = new HashSet<ClassFile>();
		
		Collection<ClassInfo> classInfos = MClassInfo.values();
		for (ClassInfo classInfo : classInfos) {
			
			ClassFile cf = classInfo.getClassFile();
			if (cf != null) classFiles.add(cf);
		}
		
		return Collections.unmodifiableSet(classFiles);
	}
	
	public ClassFile getRootClassFile() {
		
		return rootClassFile;
	}
	
	/**
	 * Returns the class node of className, or <code>null</code> if not found.
	 * 
	 * @param className the name of class
	 * @return the class node
	 */
	public ClassNode getClassNode(String className) {
		
		logger.debug(String.format("className: %s", className));
		return  MClassNode.get(className);		
	}
	
	public MethodInfo getRootMethodInfo() {
		
		return rootMethodInfo;
	}
	
	/**
	 * Returns a collection of all class names that directly or 
	 * indirectly implement the interface <code>interfaceName</code>.
	 */
	public Collection<String> getImplementors(String interfaceName) {
		
		Set<String> directImplementers = getDirectImplementers(interfaceName);
		HashSet<String> implementers = new HashSet<String>();
		
		for (String className : directImplementers) {
			
			ClassNode classNode = getClassNode(className);
			Enumeration enumeration = classNode.breadthFirstEnumeration();
			while (enumeration.hasMoreElements()) {
				ClassNode node = (ClassNode) enumeration.nextElement();
				implementers.add(node.getClassName());
			}
		}
		return implementers;
	}
	
	/**
	 * Returns the methods of class <code>className</code> which
	 * are actually called on a path from the inital method, so they
	 * have to be translated to Remopla.
	 * 
	 * @param className
	 * @return
	 */
	public Set<String> getMethods(String className) {
		
		return Collections.unmodifiableSet(MClassInfo.get(className).getMethods());
	}
	
	public String includedMethodsInfo() throws InvalidByteCodeException {
		
		StringBuilder out = new StringBuilder();
		Set<String> classNames = MClassInfo.keySet();
		
		for (String className : classNames) {
			
			ClassInfo classInfo = MClassInfo.get(className);
			if (!classInfo.isIncluded()) continue;
			
			for (String methodName : classInfo.getMethods()) {
				
				out.append(className);
				out.append(".");
				out.append(methodName);
				out.append(PDSDefault.NEWLINE);
			}
			
		}
		
		return out.toString();
	}
	
	
	/**
	 * Returns the set of interface names that className implemented.
	 * 
	 * @param className the class name
	 * @return the interface names
	 */
	public Set<String> getImplementedInterfaces(String className) {
		
		ClassNode node = getClassNode(className);
		if (node == null) {
			return Collections.emptySet();
		}
		Set<String> interfaces = new HashSet<String>();
		
		// goes up in the extension tree
		while (node != null) {
			
			logger.debug("node.getClassName(): " + node.getClassName());
			
			// for each interface that className implemented
			for (InterfaceNode interfaceNode : node.getInterfaces()) {
				
				logger.debug("interfaceNode.getName(): " + interfaceNode.getName());
				interfaces.add(interfaceNode.getName());
				for (InterfaceNode superNode : interfaceNode.getAllSupers()) {
					
					logger.debug("superNode.getName(): " + superNode.getName());
					interfaces.add(superNode.getName());
					interfaces.addAll(getImplementedInterfaces(superNode.getName()));
				}
			}
			node = (ClassNode) node.getParent();
		}
		return interfaces;
	}

	/*******************************************************************/
	/**                       PRIVATE METHODS                         **/
	/*******************************************************************/
	
	private void initialize(String confName, String methodName, String methodDesc) 
			throws InvalidByteCodeException, IOException {
		
		for (String c : IGNORED_CLASS)
			MClassInfo.put(c, new ClassInfo());
		
		for (int i = 0; i < IGNORED_METHOD.length; i += 2) {
			
			ClassInfo ci = MClassInfo.get(IGNORED_METHOD[i]);
			if (ci == null) {
				MClassInfo.put(IGNORED_METHOD[i], 
						new ClassInfo(IGNORED_METHOD[i+1]));
			} else {
				ci.addIgnored(IGNORED_METHOD[i+1]);
			}
		}
		
		try {
			parseConfig(confName);
		} catch (Exception e) {
		}
		
		if (methodName != null) {
			
			rootMethodInfo = rootClassFile.getMethod(methodName, methodDesc);
			if (rootMethodInfo == null) {
				throw new IllegalArgumentException
					("Method " + methodName + methodDesc 
					 + " not found in input class file");
			}
			addMethod(rootMethodInfo);
		}
		else {
			
			rootMethodInfo = rootClassFile.getMethod("main", 
												"([Ljava/lang/String;)V");
			if (rootMethodInfo != null) {
				logger.debug("rootMethodInfo is main([Ljava/lang/String;)V");
				addMethod(rootMethodInfo);
			}
			// remove this branch eventually, still here for backwards
			// compatibility
			else {
				// add all methods of input class
				for (MethodInfo method : rootClassFile.getMethods()) {
					addMethod(method);
				}	
			}
		}
	}
	
	private void parseConfig(String confName) 
			throws ParserConfigurationException, SAXException, IOException {
		
		InputStream in = new BufferedInputStream(new FileInputStream(confName));
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();
		Document doc = builder.parse(in);
		
		NodeList nl1, nl2, nl3;
		Node n1, n2, n3, n4, n5, n6;
		
		nl1 = doc.getFirstChild().getChildNodes();
		for (int i = 0; i < nl1.getLength(); i++) {
			
			n1 = nl1.item(i);
			
			if (!n1.getNodeName().equals("ignore")) continue;
			
			// n1 is "ignore" node
			nl2 = n1.getChildNodes();
				
			/* For each child node of "ignore" node */
			for (int j = 0; j < nl2.getLength(); j++) {
				
				n2 = nl2.item(j);
				
				if (!n2.getNodeName().equals("class")) continue;
				
				// n2 is "class" node
				n4 = n2.getAttributes().item(0);	// n4 is "class name" node
				if (!n4.getNodeName().equals("name")) {
					logger.error("Invalid class attribute.");
					continue;
				}
				
				logger.info("Ignored class: " + n4.getNodeValue());
				// methodList = new ArrayList<String>();
				ClassInfo ci = MClassInfo.remove(n4.getNodeValue());
				if (ci == null) ci = new ClassInfo();
				
				nl3 = n2.getChildNodes();
				
				/* For each child node of "class" node */
				for (int k = 0; k < nl3.getLength(); k++) {
					n3 = nl3.item(k);
					if (n3.getNodeName().equals("method")) {// n3 is "method"
						
						// n5 is "method descriptor"
						n5 = n3.getAttributes().item(0);
						if (!n5.getNodeName().equals("descriptor")) {
							logger.error("Invalid method attribute.");
							continue;
						}
						
						// n6 is "method name"
						n6 = n3.getAttributes().item(1); 
						if (!n6.getNodeName().equals("name")) {
							logger.error("Invalid method attribute.");
							continue;
						}
						//methodList.add(n6.getNodeValue() + n5.getNodeValue());
						ci.addIgnored(n6.getNodeValue() + n5.getNodeValue());
						logger.info("\tIgnored method: " + n6.getNodeValue());
					}
				}
				
				MClassInfo.put(n4.getNodeValue(), ci);
				//ignoredTable.put(n4.getNodeValue(), methodList);
			}
			break;
		}
	}
	
	private boolean isIncludedMethod(MethodInfo method) 
			throws InvalidByteCodeException {
		
		return isIncludedMethod(method.getClassFile().getThisClassName(),
			method.getName(), method.getDescriptor());
	}
	
	private boolean isIgnoredMethod(MethodInfo method) 
			throws InvalidByteCodeException {
		
		return isIgnoredMethod(method.getClassFile().getThisClassName(),
				method.getName(), method.getDescriptor());
	}
	
	/**
	 * Searches through the set of included virtual methods -- 
	 * virtualInvocations. If one of the class in virtualInvocations is
	 * an ancestor of className, the same method is included for className.
	 * 
	 * @param className the class name
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void addVirtualMethodsForClass(String className)
			throws InvalidByteCodeException, IOException {
		
		ClassNode node = getClassNode(className);
		
		// for all included virtual invocations
		for (MethodInvocation invocation : virtualInvocations) {
			
			ClassNode candidate = getClassNode(invocation.className);
			
			// if candidate is an ancestor of node
			if (node.isNodeAncestor(candidate))
				addVirtualMethodForClass(className, invocation);
		}
		
//		// copy existing invocations, because new ones could be added in the
//		// process which will handle themselves
//		for (MethodInvocation invocation :
//			virtualInvocations.toArray(new MethodInvocation[0])) {
//			ClassNode candidate = getClassNode(invocation.className);
//			// we don't have to check || candidate.isAncestor(node), since
//			// super classes are there before a new virtual method is added, unless
//			// it's called on a null pointer, so that direction is covered in addVirtualMethod
//			if (node.isNodeAncestor(candidate)) {
//				addVirtualMethodForClass(className, invocation);
//			}
//		}
	}
	
	/**
	 * Searches through the set of included interface methods --
	 * intefaceInvocations. If className implements an interface that is in
	 * the set, the same method is included for className.
	 * 
	 * @param className
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void addInterfaceMethodsForClass(String className)
			throws InvalidByteCodeException, IOException {
		
		// gets all implemented interfaces
		Set<String> interfaces = getImplementedInterfaces(className);
		
		// for all included interface invocations
		for (MethodInvocation invocation : interfaceInvocations) {
			
			/* adds the virtual method, denoted by invocation, of className
			 * if the class of invocation is actually an interface which
			 * was already included
			 */  
			if (interfaces.contains(invocation.className))
				addVirtualMethodForClass(className, invocation);
		}
		
		// copy existing invocations, because new ones could be added in the
		// process which will handle themselves
//		for (MethodInvocation invocation :
//			interfaceInvocations.toArray(new MethodInvocation[0])) {
//			if (interfaces.contains(invocation.className)) {
//				addVirtualMethodForClass(className, invocation);
//			}
//		}
	}
	
	private void addMethod(MethodInfo methodInfo)
			throws InvalidByteCodeException, IOException {
		
		logger.debug("-- Entering --");
		logger.debug("methodInfo.getName(): " + methodInfo.getName());
		
		if (isIncludedMethod(methodInfo) || isIgnoredMethod(methodInfo)) {
			return;
		}
	
		ClassFile classFile = methodInfo.getClassFile();
		addClass(classFile);
		
		String className = classFile.getThisClassName();
		logger.debug("className: " + className);
		String methodName = methodInfo.getName();
		String methodDesc = methodInfo.getDescriptor();
		MClassInfo.get(className).addIncluded(methodName + methodDesc);
		
		if (methodInfo == rootMethodInfo || methodName.equals("<init>")) {
			
			// class is instantiated, inserts it into class hierarchy
			logger.debug("The method is root or <init>");
			insertIntoClassHierarchy(classFile);
			addVirtualMethodsForClass(className);
			addInterfaceMethodsForClass(className);
		}
		
		CodeAttribute codeAttr =
			(CodeAttribute) methodInfo.findAttribute(CodeAttribute.class);
		if (codeAttr != null) {
			List<AbstractInstruction> insts = 
				ByteCodeReader.readByteCode(codeAttr.getCode());
			parseInstructions(classFile, insts);
		}
		
		logger.debug("-- Leaving --");
	}
	
	/**
	 * Includes the class classFile. Also, includes the static initialization
	 * method, if any.
	 * 
	 * @param classFile the class file
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void addClass(ClassFile classFile) 
			throws InvalidByteCodeException, IOException {
		
		logger.debug("-- Entering --");
		logger.debug("classFile.getThisClassName(): " + classFile.getThisClassName());
		
		String className = classFile.getThisClassName();
		if (isIncludedClass(className) || isIgnoredClass(className)) return;
		
		MClassInfo.put(className, new ClassInfo(classFile));
		
		MethodInfo method = classFile.getMethod("<clinit>", "()V");
		if (method != null) addMethod(method);
		
		logger.debug("-- Leaving --");
	}
	
	/**
	 * Includes the class className. Also, includes the static initialization
	 * method, if any.
	 * 
	 * @param className the class name
	 * @return the class file
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 * @see findClassFile(String)
	 * @see addClass(ClassFile)
	 */
	private ClassFile addClass(String className) 
			throws InvalidByteCodeException, IOException {
		
		logger.debug("-- Entering --");
		logger.debug("className: " + className);
		
		if (isIgnoredClass(className)) {
			return null;
		}
		
		if (isIncludedClass(className)) {
			return getClassFile(className);
		}
		ClassFile classFile = findClassFile(className);
		addClass(classFile);
		
		logger.debug("-- Leaving --");
		return classFile;
	}
	
	/**
	 * Returns a possible file name of className. The method assumes that the
	 * class is in the same directory as the root file.
	 * 
	 * @param className the class name
	 * @return the file name
	 */
	private String formatFileName(String className) {
		
		StringBuilder out = new StringBuilder();
		
		if (rootFile != null && rootFile.getParent() != null) {
			logger.debug("rootFile.getName(): " + rootFile.getName());
			logger.debug("rootFile.getParent(): " + rootFile.getParent());
			out.append(rootFile.getParent());
			out.append(File.separator);
		}
		out.append(className.substring(className.lastIndexOf("/") + 1));
		out.append(".class");
		
		return out.toString();
	}
	
	/**
	 * Finds the class file of the given className. The method first tries
	 * at the same directory as the root file. If not found, it searches
	 * through the class paths.
	 * 
	 * @param className
	 * @return
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 * @see BytecodeUtils#findClassFile(String, String[])
	 */
	private ClassFile findClassFile(String className)
			throws InvalidByteCodeException, IOException {
		
		logger.debug("newClassFile.className: " + className);
		
		String calledFileName = formatFileName(className);
		logger.debug("newClassFile.calledFileName: " + calledFileName);
	
		File calledFile = new File(calledFileName);
		if (calledFile.exists()) {
			logger.debug("classFile found in current directory");
			return ClassFileReader.readFromFile(calledFile);
		} 
	
		logger.debug("classFile NOT found in current directory");
	
		// search for class in class path
		return BytecodeUtils.findClassFile(className, searchPaths);
	}
	
	/**
	 * Creates and adds a new interface node, named interfaceName, into 
	 * MInterfaceNode. If the interface has super interfaces, the method 
	 * recursively adds new nodes to MInterfaceNode. 
	 * 
	 * @param interfaceName the interface name
	 * @return the interface node
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private InterfaceNode createInterfaceNode(String interfaceName) 
			throws InvalidByteCodeException, IOException {
		
		InterfaceNode node = MInterfaceNode.get(interfaceName);
		if (node != null) return node;
		
		node = new InterfaceNode(interfaceName);
		MInterfaceNode.put(interfaceName, node);
		
		ClassFile interfaceClassFile = findClassFile(interfaceName);
		// TODO interfaceClassFile maybe null???
		int[] supInterfaces = interfaceClassFile.getInterfaces();
		for (int i : supInterfaces) {
			
			String supInterfaceName = BytecodeUtils.getInterfaceName(
					interfaceClassFile, i);
			InterfaceNode supInterfaceNode = createInterfaceNode(supInterfaceName);
			supInterfaceNode.addSub(node);
			node.addSuper(supInterfaceNode);
		}
		
		return node;
	}
	
	/**
	 * Creates a set of interface nodes which the class classFile implemented.
	 * 
	 * @param classFile
	 * @return
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private Set<InterfaceNode> createInterfaceNodes(ClassFile classFile) 
			throws InvalidByteCodeException, IOException {
		
		int[] interfaces = classFile.getInterfaces();
		String className = classFile.getThisClassName();
		Set<InterfaceNode> set = new HashSet<InterfaceNode>(interfaces.length);
		
		for (int i : interfaces) {
			
			String interfaceName = BytecodeUtils.getInterfaceName(classFile, i);
			InterfaceNode interfaceNode = createInterfaceNode(interfaceName);
			interfaceNode.addImplementer(className);
			set.add(interfaceNode);
		}
		return set;
	}
	
	/**
	 * Inserts the class classFile into the class extension hierarchy.
	 * The method also inserts all its parent classes into the hierarychy,
	 * if any.
	 * 
	 * @param classFile the class file
	 * @return the class node
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private ClassNode insertIntoClassHierarchy(ClassFile classFile)
			throws InvalidByteCodeException, IOException {
		
		String className = classFile.getThisClassName();
		ClassNode node = getClassNode(className);
		
		if (node != null) {
			return node;
		}
		
		node = new ClassNode(className, createInterfaceNodes(classFile));
		MClassNode.put(className, node);
		
		// adds this class as a child node of its parent, if any
		String superClassName = classFile.getSuperClassName();
		if (superClassName == null) {
			return node;
		}
		
		ClassFile superClassFile = addClass(superClassName);
		if (superClassFile != null) {
			ClassNode parent = insertIntoClassHierarchy(superClassFile);
			parent.add(node);
		}
		
		return node;
	}
	
	/*******************************************************************/
	/**                     PARSING INSTRUCTIONS                      **/
	/*******************************************************************/
	
	/**
	 * Parses the instruction insts of the class classFile.
	 * Includes new classes and methods. 
	 * 
	 * @param classFile the class file
	 * @param insts the list of abstract instructions
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void parseInstructions(ClassFile classFile,
			List<AbstractInstruction> insts) 
			throws InvalidByteCodeException, IOException {
		
		for (AbstractInstruction inst : insts) {
			
			switch (inst.getOpcode()) {
			
			case Opcodes.OPCODE_INVOKESTATIC:
			case Opcodes.OPCODE_INVOKESPECIAL:
				handleInvokestatic(classFile, inst);
				break;
			case Opcodes.OPCODE_PUTSTATIC:
			case Opcodes.OPCODE_GETSTATIC:
				handlePutstaticGetstatic(classFile, inst);
				break;
			case Opcodes.OPCODE_INVOKEVIRTUAL:
				handleInvokevirtual(classFile, inst);
				break;
			case Opcodes.OPCODE_INVOKEINTERFACE:
				handleInvokeinterface(classFile, inst);
				break;
			//case Opcodes.OPCODE_MULTIANEWARRAY:
			//	includeMultiNewArrayInitializerMethod = true;
			//	break;
			}
		}
	}
	
	/**
	 * Includes the method referenced by the invokestatic inst.
	 * 
	 * @param classFile the class file
	 * @param inst the instruction
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void handleInvokestatic(ClassFile classFile, AbstractInstruction inst) 
			throws InvalidByteCodeException, IOException {
		
		String[] names =  BytecodeUtils.getReferencedName(classFile, inst);
		
		if (isIgnoredClass(names[0])) {
			return;
		}
		
		ClassFile calledClassFile = getClassFile(names[0]);
		if (calledClassFile == null) {
			calledClassFile = findClassFile(names[0]);
		}
		
		MethodInfo calledMethodInfo = calledClassFile.getMethod(names[1], names[2]);
		if (calledMethodInfo != null) {
			addMethod(calledMethodInfo);
		}
	}
	
	/**
	 * Includes the class referenced by the putstatic or getstatic inst.
	 * 
	 * @param classFile
	 * @param inst
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void handlePutstaticGetstatic(ClassFile classFile,
			AbstractInstruction inst)
			throws InvalidByteCodeException, IOException {
		
		String[] names =  BytecodeUtils.getReferencedName(classFile, inst);
		addClass(names[0]);
	}
	
	/**
	 * Returns true if class <code>className</code> is possibly instantiated 
	 * somewhere on the static code path.
	 * 
	 * @param className the className
	 * @return true if its constructor was already included
	 */
	private boolean isInstantiatedClass(String className) {
		
		return isIncludedMethod(className, "<init>");
	}
	
	/**
	 * Adds the method denoted by {@link MethodInvocation} if the class
	 * <code>className</code> has been instantiated somewhere on the path
	 * from the initial method.
	 * 
	 * @param className
	 * @param invocation
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void addVirtualMethodForClass(String className, 
			MethodInvocation invocation) throws InvalidByteCodeException, 
			IOException {
		
		logger.debug("className: " + className);
		logger.debug("invocation: " + invocation);
		
		if (!isInstantiatedClass(className) 
				&& !rootClassFile.getThisClassName().equals(className)) 
			return;
			
		ClassFile classFile = getClassFile(className);
		MethodInfo methodInfo = classFile.getMethod(invocation.methodName,
				invocation.methodDescriptor);
		if (methodInfo == null)
			logger.debug("methodInfo is null");
		if (methodInfo != null
			&& (methodInfo.getAccessFlags() & AccessFlags.ACC_PRIVATE) == 0
			&& (methodInfo.getAccessFlags() & AccessFlags.ACC_ABSTRACT) == 0) {
			
			addMethod(methodInfo);
		}
	}
	
	/**
	 * Searches through the class hierarchy, and includes methods with the
	 * same name and descriptor for all predecessor and successor classes.
	 * 
	 * @param invocation
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void addVirtualMethod(MethodInvocation invocation)
			throws InvalidByteCodeException, IOException {
		
		logger.debug("invocation: " + invocation);
		
		// returns if the method was already included
		if (!virtualInvocations.add(invocation)) {
			logger.debug("invocation was already included in virtualInvocations");
			return;
		}
		
		logger.debug("virtualInvocations: " + virtualInvocations);
		
		ClassNode classNode = getClassNode(invocation.className);
		// class node can be null, if virtual method is called on a null pointer
		if (classNode == null) {
			logger.debug("classNode is null");
			return;
		}
		
		// includes all parent methods
		ClassNode node = (ClassNode) classNode.getParent();
		while (node != null) {
			addVirtualMethodForClass(node.getClassName(), invocation);
			node = (ClassNode)node.getParent();
		}

		// includes offspring methods, includes node itself as well
		Enumeration offspring = classNode.breadthFirstEnumeration();
		while (offspring.hasMoreElements()) {
			ClassNode next = (ClassNode) offspring.nextElement();
			addVirtualMethodForClass(next.getClassName(), invocation);
		}
	}
	
	/**
	 * Includes the method referenced by the invokevirtual inst.
	 * 
	 * @param classFile
	 * @param inst
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void handleInvokevirtual(ClassFile classFile,
			AbstractInstruction inst)
			throws InvalidByteCodeException, IOException {
		
		String[] names = BytecodeUtils.getReferencedName(classFile, inst);
		if (!isIgnoredClass(names[0])) {
			addVirtualMethod(new MethodInvocation(names[0], names[1], names[2]));
		}
	}
	
	/**
	 * Returns a set of class names that are direct implementers of 
	 * <code>interfaceName</code> or one of its sub interfaces.
	 * 
	 * @param interfaceName the interface name
	 * @return the set of implementers
	 */
	private Set<String> getDirectImplementers(String interfaceName) {
		
		InterfaceNode node = MInterfaceNode.get(interfaceName);
		if (node == null) return Collections.emptySet();
		
		Set<InterfaceNode> subs = node.getAllSubs();
		logger.debug("subs: " + subs);
		
		Set<String> implementers = new HashSet<String>(node.getImplementers());
		for (InterfaceNode sub : subs) {
			
			implementers.addAll(sub.getImplementers());
		}
		
		return implementers;
	}
	
	/**
	 * Searches through the hierarchy of all implementers of the interface, 
	 * and includes methods with the same name and signature for all 
	 * predecessor and successor classes.
	 * 
	 * @param invocation the method invocation
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void addInterfaceMethod(MethodInvocation invocation)
			throws InvalidByteCodeException, IOException {
		
		logger.debug("-- Entering --");
		logger.debug("invocation.className: " + invocation.className);
		logger.debug("invocation.methodName: " + invocation.methodName);
		logger.debug("invocation.methodDescriptor: " + invocation.methodDescriptor);
		
		// returns if the method was aleready included
		if (!interfaceInvocations.add(invocation)) {
			return;
		}
		
		Set<String> implementers = getDirectImplementers(invocation.className);
		logger.debug("implementers: " + implementers);
	
		// for all implementers of the interface
		for (String implementer : implementers) {
			
			// goes upwards, includes methods for all parent classes
			ClassNode implementerNode = getClassNode(implementer);
			ClassNode node = (ClassNode) implementerNode.getParent();
			while (node != null) {
				addVirtualMethodForClass(node.getClassName(), invocation);
				node = (ClassNode)node.getParent();
			}
			
			// go downwards, includes methods for all successor classes and itself
			Enumeration offspring = implementerNode.breadthFirstEnumeration();
			while (offspring.hasMoreElements()) {
				ClassNode next = (ClassNode) offspring.nextElement();
				addVirtualMethodForClass(next.getClassName(), invocation);
			}
		}
		logger.debug("-- Leaving --");
	}
	
	/**
	 * Includes the method referenced by the invokeinterface inst.
	 * 
	 * @param methodInfo
	 * @param inst
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void handleInvokeinterface
			(ClassFile classFile, AbstractInstruction inst)
			throws InvalidByteCodeException, IOException {
		
		String[] names = BytecodeUtils.getReferencedName(classFile, inst);
		
		if (isIgnoredClass(names[0])) return;
		
		addInterfaceMethod(new MethodInvocation(names[0], names[1], names[2]));
	}
	
	/*******************************************************************/
	/**                       PRIVATE CLASSES                         **/
	/*******************************************************************/
	
	private class ClassInfo {
		
		ClassFile classFile = null;
		Set<String> ignoredMethod = new HashSet<String>();
		Set<String> includedMethod = new HashSet<String>();
		
		ClassInfo() {
			
		}
		
		ClassInfo(String s) {
			
			ignoredMethod.add(s);
		}
		
		ClassInfo(ClassFile classFile) {
			
			this.classFile = classFile;
		}
		
		ClassFile getClassFile() {
			
			return classFile;
		}
		
		String getName() throws InvalidByteCodeException {
			
			return classFile.getThisClassName();
		}
		
		void addIgnored(String s) {
			
			ignoredMethod.add(s);
		}
		
		void addIncluded(String s) {
			
			includedMethod.add(s);
		}
		
		boolean isIgnored(String s) {
			
			return ignoredMethod.contains(s);
		}
		
		/**
		 * Returns true if the class is ignored. This is the case
		 * when its class file is null AND its set of ignored methods is empty.
		 * Note that if the set of ignored methods is not empty, it means that
		 * the class itself is not ignored, but only some of its methods. 
		 * 
		 * @return true, if the class is ignored
		 */
		boolean isIgnored() {
			
			return (classFile == null) && ignoredMethod.isEmpty();
		}
		
		boolean isIncluded() {
			
			return classFile != null;
		}
		
		boolean isIncluded(String s) {
			
			return includedMethod.contains(s); 
		}
		
		boolean isMethodIncluded(String methodName) {
			
			String methodNameWithLeftParenthesis = methodName + "(";
			
			for (String s : includedMethod)
				if (s.startsWith(methodNameWithLeftParenthesis))
					return true;
			
			return false;
		}
		
		public Set<String> getMethods() {
			
			return includedMethod;
		}
	}
	
	/**
	 * An interface node provides the following information:
	 * 
	 *  Its super interface
	 *  The class names of its direct implementers
	 *  Its sub inferfaces
	 */
	private static class InterfaceNode {
		
		private Set<InterfaceNode> supers = null;
		private Set<InterfaceNode> subs = null;
		private Set<String> implementers = null;
		private String name;
		
		private static Logger logger = Logger.getLogger(InterfaceNode.class);
		
		public InterfaceNode(String name) {
			
			this.name = name;
		}
		
		public String getName() {
			
			return name;
		}
		
		public void addSuper(InterfaceNode s) {
			
			if (supers == null) supers = new HashSet<InterfaceNode>();
			supers.add(s);
		}
		
		public void addSub(InterfaceNode s) {
			
			if (subs == null) subs = new HashSet<InterfaceNode>();
			subs.add(s);
		}
		
		public Set<InterfaceNode> getSupers() {
			
			if (supers != null) return Collections.unmodifiableSet(supers);
			else return Collections.emptySet();
		}
		
		public Set<InterfaceNode> getSubs() {
			
			if (subs != null) return Collections.unmodifiableSet(subs);
			else return Collections.emptySet();
		}
		
		public void addImplementer(String implementer) {
			
			if (implementers == null) implementers = new HashSet<String>();
			implementers.add(implementer);
		}
		
		public Set<String> getImplementers() {
			
			if (implementers != null) 
				return Collections.unmodifiableSet(implementers);
			else
				return Collections.emptySet();
		}
		
		public Set<InterfaceNode> getAllSupers() {
			
			logger.debug("name: " + name);
			return getAllSupers(supers);
		}
		
		public Set<InterfaceNode> getAllSubs() {
			
			return getAllSubs(subs);
		}
		
		public String toString() {
			
			return getName();
		}
		
		private static Set<InterfaceNode> getAllSupers(
				Set<InterfaceNode> interfaces) {
			
			logger.debug("-- Entering --");
			
			if (interfaces == null) return Collections.emptySet();
			
			Set<InterfaceNode> set = new HashSet<InterfaceNode>(interfaces);
			for (InterfaceNode i : interfaces) {
				
				logger.debug("i.getName(): " + i.getName());
				set.addAll(getAllSupers(i.getSupers()));
			}
			
			logger.debug("set.toString(): " + set.toString());
			logger.debug("-- Leaving --");
			return set;
		}
		
		private static Set<InterfaceNode> getAllSubs(
				Set<InterfaceNode> interfaces) {
			
			if (interfaces == null) return Collections.emptySet();
			
			Set<InterfaceNode> set = new HashSet<InterfaceNode>(interfaces);
			for (InterfaceNode i : interfaces) {
				
				set.addAll(getAllSubs(i.getSubs()));
			}
			
			return set;
		}
	}
	
	/**
	 * A class node provides the following information:
	 * 
	 *  It knows its parent in the class hierarchy
	 *  It knows its sub classes
	 *  It knows the interfaces it directly implements
	 */
	public static class ClassNode extends DefaultMutableTreeNode {
		
		private Set<InterfaceNode> interfaces;
		private String className;
		
		private ClassNode(String className, Set<InterfaceNode> interfaces) {
			
			super(className);
			this.className = className;
			this.interfaces = interfaces;
		}
		
		public String getClassName() {
			
			return className;
		}
		
		public Set<InterfaceNode> getInterfaces() {
			
			return interfaces;
		}
	}
	
	private static class MethodInvocation {
		
		public String className;
		public String methodName;
		public String methodDescriptor;

		public MethodInvocation(String className, String methodName, 
				String methodDescriptor) 
		{
			this.className = className;
			this.methodName = methodName;
			this.methodDescriptor = methodDescriptor;
		}
		
		@Override
		public boolean equals(Object obj) 
		{
			MethodInvocation other = (MethodInvocation)obj;
			return className.equals(other.className) 
				&& methodName.equals(other.methodName)
				&& methodDescriptor.equals(other.methodDescriptor);
		}
		
		@Override
		public int hashCode() 
		{
			return className.hashCode() + 31 * methodName.hashCode()
				+ 31 * 31 * methodDescriptor.hashCode();
		}
		
		public String toString() {
			
			return className + "." + methodName + methodDescriptor;
		}
	}
}
