/*
    Copyright (C) 2005 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;

import org.apache.log4j.Logger;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.MethodInfo;

/**
 * Represents a pushdown system in the form of Remopla language. 
 * Instatiate this class with a Java class.
 * Use {@link #translate() translate} method to translate the Java class into
 * Remopla. Afterwards, call {@link #toRemopla() toRemopla} to get the
 * Remopla code. 
 * 
 * @author suwimodh
 *
 */
public class PDS {
	
	static Logger logger = Logger.getLogger(PDS.class);
	
	private PDSInfo pdsInfo;
	private Set<PDSClass> pdsClasses = new HashSet<PDSClass>();

	/**
	 * Optional initial method which represents the entry point for the model
	 * checking analysis. This used to be the main method, but we want to
	 * allow other static and even non static methods as well.
	 */
	private PDSClass rootPDSClass;
	private PDSMethod rootPDSMethod;
	
	private PDSMethodWrapper pdsMethodWrapper = null;
	
	/**
	 * Creates an PDS instance for classFileName. The main method is assumed as
	 * the initial method.
	 * 
	 * @param classFileName path name to the file can be either absolute
	 * or relative to the current working directory.
	 * @throws Exception
	 */
	public PDS(String classFileName) throws Exception {
		
		this(new PDSInfo(new File(classFileName)));
	}
	
	/**
	 * Creates an PDS instance for classFileName. The initial method is
	 * suggested by methodName and methodDescriptor.
	 * 
	 * @param classFileName path name to the file can be either absolute
	 * or relative to the current working directory.
	 * @param methodName the initial method name.
	 * @param methodDescriptor the initial method descriptor.
	 * @throws Exception
	 */
	public PDS(String classFileName, String methodName, String methodDescriptor)
			throws Exception {
		
		this(new PDSInfo(new File(classFileName), methodName, 
						 methodDescriptor));
	}

	/**
	 * Creates an PDS instance for clazz. The main method is assumed as
	 * the initial method.
	 * 
	 * @param clazz the class.
	 * @throws Exception
	 */
	public PDS(Class clazz) throws Exception {
		
		this(new PDSInfo(clazz));
	}
	
	/**
	 * Creates an PDS instance for clazz. The initial method is
	 * suggested by methodName and methodDescriptor.
	 * 
	 * @param clazz the clazz.
	 * @param methodName the initial method name.
	 * @param methodDescriptor the initial method descriptor.
	 * @throws Exception
	 */
	public PDS(Class clazz, String methodName, String methodDescriptor) 
			throws Exception {
		
		this(new PDSInfo(clazz, methodName, methodDescriptor));
	}

	/**
	 * Creates an PDS instance for className. The array pathEntries suggest 
	 * where to find the class className. The main method is assumed as
	 * the initial method.
	 * 
	 * @param className the class name.
	 * @param pathEntries the paths.
	 * @throws Exception
	 */
	public PDS(String className, String[] pathEntries) throws Exception {
		
		this(new PDSInfo(className, pathEntries));
	}
	
	/**
	 * Creates an PDS instance for className. The array pathEntries suggest 
	 * where to find the class className. The initial method is
	 * suggested by methodName and methodDescriptor.
	 * 
	 * @param className the class name.
	 * @param pathEntries the paths.
	 * @param methodName the initial method name.
	 * @param methodDescriptor the initial method desciptor.
	 * @throws Exception
	 */
	public PDS(String className, String[] pathEntries, String methodName,
			String methodDescriptor) throws Exception {
		
		this(new PDSInfo(className, pathEntries, methodName, methodDescriptor));
	}
	
	/**
	 * Creates an PDS instance from info.
	 * 
	 * @param info the <code>PDSInfo</code> instance.
	 * @throws Exception
	 */
	public PDS(PDSInfo info) throws Exception {
		
		pdsInfo = info;
		logger.info("Entering PDS constructor.");
		initPDSClasses();
		initRootPDSMethod();
		logger.info("Leaving PDS constructor.");
	}
	
	/**
	 * Translates the Java class into Remopla.
	 * 
	 * @throws InvalidByteCodeException
	 */
	public void translate() throws InvalidByteCodeException {
		
		pdsInfo.setup(pdsClasses);
		if (pdsMethodWrapper != null) pdsMethodWrapper.wrap();
		for (PDSClass pdsClass : pdsClasses) {
			pdsClass.translate();
		}
	}
	
	/**
	 * Returns the configuration object responsible for this pds instance.
	 * 
	 * @return the <code>PDSInfo</code> instance.
	 */
	public PDSInfo getInfo() {
		
		return pdsInfo;
	}
	
	/**
	 * Returns <code>true</code> if the initial method is set.
	 * 
	 * @return <code>true</code> if the initial method is set.
	 */
	public boolean hasRootMethod() {
		
		return rootPDSMethod != null;
	}
	
	/**
	 * Initializes the initial method.
	 * 
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void initRootPDSMethod() throws InvalidByteCodeException, IOException {
		
		InvokeAnalyzer invokeAnalyzer = pdsInfo.getInvokeAnalyzer();
		
		ClassFile rootClassFile = invokeAnalyzer.getRootClassFile();
		rootPDSClass = getPDSClass(rootClassFile.getThisClassName());
		logger.debug("rootPDSClass: " + rootPDSClass.getName());
		
		MethodInfo rootMethodInfo = invokeAnalyzer.getRootMethodInfo();
		if (rootMethodInfo == null) return;
		logger.debug("rootMethodInfo: " + rootMethodInfo.getName());
		
		String methodName = rootMethodInfo.getName();
		String methodDesc = rootMethodInfo.getDescriptor();
		rootPDSMethod = rootPDSClass.getMethod(methodName, methodDesc);
		
		if (rootPDSMethod.isMainMethod()) return;
		logger.debug("rootPDSMethod: " + rootPDSMethod.getName());
		
		//if (initMethod.isStatic() && initMethod.getParamCount() == 0)
		//	return;
		
		pdsMethodWrapper = new PDSMethodWrapper(rootPDSClass, rootPDSMethod);
	}
	
	/**
	 * Returns the method that has been set to be the entry point for the 
	 * analysis or the main method of the class that is to be analysed.
	 * 
	 * @return the initial method, or<code>null</code> if no method has been 
	 * set and there is no main method in the initial class.
	 * @throws InvalidByteCodeException 
	 */
	public PDSMethod getRootPDSMethod() {
		
		return rootPDSMethod;
	}
	
	/**
	 * Returns the initial class. The initial class is the class that was
	 * given to the constructor.
	 * 
	 * @return the initial class.
	 */
	public PDSClass getRootPDSClass() {
		
		return rootPDSClass;
	}
	
	/**
	 * Returns the method wrapper, or null if there is none.
	 * 
	 * @return the method wrapper
	 */
	public PDSMethodWrapper getPDSMethodWrapper() {
		
		return pdsMethodWrapper;
	}
	
	/**
	 * Returns the Remopla code. The method {@link #translate() translate}
	 * must be called before calling this method.
	 * 
	 * @return the Remopla code.
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	public String toRemopla() throws InvalidByteCodeException, IOException {
		
		StringBuffer out = new StringBuffer();
		PDSMultiArrayInitializerMethod multiArrayMethod = 
			pdsInfo.includeMultiArrayInitializerMethod() ? 
					new PDSMultiArrayInitializerMethod(pdsInfo) : null;
		
		int heapSize = pdsInfo.getStoredHeapSize();
		int heapBit = PDSDefault.getMinimumBitsForHeapSize(heapSize);
		int intBit = pdsInfo.getBits();
		
		out.append("define DEFAULT_INT_BITS ");
		out.append(String.valueOf(intBit));
		out.append("\n\n");
		
		out.append("int ");
		out.append(pdsInfo.heapIndex(String.valueOf(pdsInfo.getHeapSize())));
		if (pdsInfo.getHeapOption() == PDSInfo.HeapOption.TWODIMS) {
			out.append("(1)");
		}
		out.append(";\n");
		
		out.append("int ");
		out.append(pdsInfo.getHeapPtrName());
		out.append("(");
		out.append(String.valueOf(heapBit));
		out.append(");\n\n");
		
		// collect declarations of static fields and declarations of
		// class methods
		
		if (pdsMethodWrapper != null) {
			out.append(pdsMethodWrapper.toRemoplaHead()).append(";\n");
		}
		if (multiArrayMethod != null 
				&& pdsInfo.getHeapOption() == PDSInfo.HeapOption.SIMPLE) {
			out.append(multiArrayMethod.toRemoplaHead()).append(";\n");
		}
		if (pdsInfo.isIncludeStringEqualsMethod()) {
			out.append(PDSString.remoplaHeadForEquals()).append(";\n");
		}
		
		for (PDSClass pdsClass : pdsClasses) {
			
			out.append(pdsClass.toRemoplaGlobal());
			out.append("\n");
		}
		
		out.append("init");
		
		if (PDSInfo.Stefan) {
			addStefansOutput(out);
		} 
		// We'll always use Stefan's option.
		/*else {

			for (int i = pdsClassList.size() - 1; i >= 0; i--) {
				
				pdsClass = (PDSClass) pdsClassList.get(i);
				pdsMethod = pdsClass.getMethodClinit();
				if (pdsMethod != null) {
					pdsMethod.addInitPtr();	//Dirty pointer initialization
					out.append(" ");
					out.append(pdsMethod.getFormattedName());
				}
			}
			pdsMethod = getInternalInitialMethod();
			if (pdsMethod != null) {
				out.append(" ");
				out.append(pdsMethod.getFormattedName());
			}
			out.append(";\n\n");
		}*/
		
		// collect all method definitions
		
		if (pdsMethodWrapper != null) {
			out.append(pdsMethodWrapper.toRemopla()).append("\n");
		}
		if (multiArrayMethod != null 
				&& pdsInfo.getHeapOption() == PDSInfo.HeapOption.SIMPLE) {
			out.append(multiArrayMethod.toRemopla()).append("\n");
		}
		if (pdsInfo.isIncludeStringEqualsMethod()) {
			out.append(PDSString.remoplaForEquals(pdsInfo)).append("\n");
		}
		
		for (PDSClass pdsClass : pdsClasses) {
			
			out.append(pdsClass.toRemoplaModule());
			out.append("\n");
		}

		// add assert error state
		// TODO maybe only add this if it might be reachable, i.e. if there are 
		// assert statements in the code
		out.append(PDSStmt.selfLoopStmt(pdsInfo.getLabelAssertError()));
		
		out.append(PDSStmt.selfLoopStmt(pdsInfo.getLabelGeneralException()));
		
		if (pdsInfo.checkForHeapOverflow()) {
			out.append(PDSStmt.selfLoopStmt(pdsInfo.getLabelHeapOverflow()));
		}
		
		if (pdsInfo.checkForIndexOutOfBounds()) {
			out.append(PDSStmt.selfLoopStmt(pdsInfo.getLabelIndexOutOfBounds()));
		}
		
		if (pdsInfo.checkForNullPointerExceptions()) {
			out.append(PDSStmt.selfLoopStmt(pdsInfo.getLabelNPE()));
		}
		
		if (pdsInfo.checkForStringBuilderOverflow()) {
			out.append(PDSStmt.selfLoopStmt(pdsInfo.getLabelStringBuilderOverflow()));
		}
		
		return out.toString();
	}
	
	/**
	 * Adds to <code>out</code> the initial symbol of the Remopla code, i.e.
	 * PDSDefault.JMOPED_INIT. The initial stack depth in this option is
	 * one. So, return statements of static initialization methods must be 
	 * replaced by goto statements.
	 * 
	 * The method loops from the end of class list to the beginning. If a 
	 * class has a static init method (clinit), the last statement (which is a 
	 * return statement) is replaced by a goto statement to the starting point 
	 * of the clinit method of the class that next class in the loop. The
	 * last clinit points to either the wrapper method if it is not null,
	 * or the initial method, otherwise. 
	 * 
	 * @param out
	 */
	public void addStefansOutput(StringBuffer out) {

		PDSClass nextClass;
		PDSMethod nextMethod;
		String initMethodName = null;
		PDSClass pdsClass;
		PDSMethod pdsMethod;
		String label;
		
		out.append(" ");
		out.append(PDSDefault.JMOPED_INIT);
		PDSClass[] classes = pdsClasses.toArray(new PDSClass[0]);
		for (int i = classes.length - 1; i >= 0; i--) {
			
			pdsClass = classes[i];
			pdsMethod = pdsClass.getMethodClinit();
			
			if (pdsMethod == null) continue;
				
			if (initMethodName == null) {
				initMethodName = pdsMethod.getFormattedName();
			}
			
			nextMethod = null;
			for (int j = i - 1; j >= 0; j--) {
				nextClass = classes[j];
				nextMethod = nextClass.getMethodClinit();
				if (nextMethod != null) {
					break;
				}
			}
			
			if (nextMethod != null) {
				
				label = nextMethod.getFormattedName();
			} else {
				
				if (pdsMethodWrapper != null) {
					label = pdsMethodWrapper.getName();
				} else if (rootPDSMethod != null) {
					label = rootPDSMethod.getFormattedName();
				} else {
					logger.error("Initialized method not found.");
					label = "";
				}
			}
			pdsMethod.replaceLastInst("goto " + label);
		}
		
		if (initMethodName == null) {
			
			if (pdsMethodWrapper != null) {
				initMethodName = pdsMethodWrapper.getName();
			} else if (rootPDSMethod != null) {
				initMethodName = rootPDSMethod.getFormattedName();
			} else {
				logger.error("Initialized method not found.");
			}
		}
		
		out.append(";\n\n");
		out.append(initModule(initMethodName));
	}
	
	/*******************************************************************/
	/**                       PDSInfo METHODS                         **/
	/*******************************************************************/
	
	public String fieldIndexTableInfo() {
		
		return pdsInfo.fieldIndexTableInfo();
	}
	
	public String getLabelAssertError() {
		
		return pdsInfo.getLabelAssertError();
	}
	
	public String getLabelHeapOverflow() {
		
		return pdsInfo.getLabelHeapOverflow();
	}

	public String getLabelNPE() {
	
	return pdsInfo.getLabelNPE();
	}
	
	public String getMopedPath() {
		
		return pdsInfo.getMopedPath();
	}
	
	public String includedTableInfo() throws InvalidByteCodeException {
		
		return pdsInfo.getInvokeAnalyzer().includedMethodsInfo();
	}
	
	public void putAllLineBitTable(Hashtable<Integer, Integer> lbTable) {
		
		pdsInfo.putAllLineBitTable(lbTable);
	}
	
	public void setBits(int bits) {
		
		pdsInfo.setBits(bits);
	}
	
	public void setHeapOption(PDSInfo.HeapOption heapOption) {
		
		pdsInfo.setHeapOption(heapOption);
	}
	
	public void setHeapSize(int heapSize) {
		
		pdsInfo.setHeapSize(heapSize);
	}
	
	/*******************************************************************/
	/**                       PRIVATE METHODS                         **/
	/*******************************************************************/
	
	/**
	 * Create a special module, used to initialize heap and its pointer.
	 * The module name is defined in PDSDefault.JMOPED_INIT.
	 * The method is called when Stefan option is on.
	 * After the initialization, the instruction "goto gotoName" is added.
	 * 
	 * @param gotoName the label of the last go statement.
	 * @return the Remopla code representing the init module.
	 */
	private String initModule(String gotoName) {
		
		StringBuffer out = new StringBuffer();
		
		out.append("module void ");
		out.append(PDSDefault.JMOPED_INIT);
		out.append("()\n");
		out.append("{\n");
		out.append(pdsInfo.getHeapPtrName());
		out.append("=1, ");
		out.append("A i (0,");
		out.append(String.valueOf(pdsInfo.getHeapSize() - 1));
		out.append(") ");
		out.append(pdsInfo.heapIndex("i"));
		out.append("=0;\n");
		out.append("goto ");
		out.append(gotoName);
		out.append(";\n");
		out.append("}\n\n");
		
		return out.toString();
	}
	

	/**
	 * Returns an unmodifiable set of all PDSClass objects that have been
	 * created.
	 *
	 * @return the set all <code>PDSClass</code> objects.
	 */
	public Set<PDSClass> getPDSClasses()
	{
		return Collections.unmodifiableSet(pdsClasses);
	}
	
	/**
	 * Gets the <code>PDSClass</code> object with className.
	 * If not found, null is returned.
	 * 
	 * @param className the class name.
	 * @return the <code>PDSClass</code> object, or null if not found.
	 */
	public PDSClass getPDSClass(String className) {
		
		for (PDSClass pdsClass : pdsClasses) {
			
			if (pdsClass.getName().equals(className))
				return pdsClass;
		}
		
		return null;
	}
	
	/**
	 * Gets the <code>PDSMethod</code> object with methodName.
	 * If not found, null is returned.
	 * 
	 * @param methodName the class name.
	 * @return the <code>PDSMethod</code> object, or null if not found.
	 */
	public PDSMethod getPDSMethod(String methodName) {
		
		for (PDSClass clazz : pdsClasses) {
			
			if (methodName.startsWith(clazz.getFormattedClassName())) {
				
				for (PDSMethod method : clazz.getMethodList()) {
					if (method.getFormattedName().equals(methodName)) {
						return method;
					}
				}
			}
		}
		return null;
	}
	
	/**
	 * Returns <code>true</code> if the method with methodName was included 
	 * in the pds.
	 * 
	 * @param methodName the method name.
	 * @return <code>true</code> there is a method name methodName.
	 */
	public boolean hasMethod(String methodName) {
		
		if (pdsMethodWrapper != null 
				&& pdsMethodWrapper.getName().equals(methodName))
			return true;
		
		if (getPDSMethod(methodName) != null) return true;
		
		return false;
	}
	
	/**
	 * Initializes pdsClasses.
	 * 
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	private void initPDSClasses() throws InvalidByteCodeException, IOException {
		logger.debug("Entering initPDSClassList");
		
		Set<ClassFile> classFiles = pdsInfo.getInvokeAnalyzer().getClassFiles();
		for (ClassFile classFile : classFiles) {
			pdsClasses.add(new PDSClass(classFile, pdsInfo));
		}
		
		pdsInfo.initClassTable(pdsClasses);
			
		logger.debug("Leaving initPDSClassList");
	}
	
}
