/*
    Copyright (C) 2005 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package de.uni_stuttgart.fmi.szs.jmoped;

import org.gjt.jclasslib.structures.*;
import org.gjt.jclasslib.structures.constants.ConstantClassInfo;
import org.apache.log4j.Logger;

import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo.HeapOption;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

/**
 * Class information for pushdown system.
 * 
 * @author suwimodh
 *
 */
public class PDSClass {
	
	static Logger logger = Logger.getLogger(PDSClass.class);
	
	private ClassFile classFile;
	PDSInfo pdsInfo;

	private String thisClassName;
	private String superClassName;
	private String[] interfaceName;
	private List<PDSField> pdsStaticFieldList = new ArrayList<PDSField>();
	private List<PDSField> pdsNonStaticFieldList = new ArrayList<PDSField>();
	private List<PDSMethod> pdsMethodList;
	
	private int index;
	
	/**
	 * The constructor.
	 * 
	 * @param classFile the class file.
	 * @param pdsInfo the pds info.
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	public PDSClass (ClassFile classFile, PDSInfo pdsInfo) 
			throws InvalidByteCodeException, IOException {
		
		logger.info("Entering PDSClass constructor");
		
		this.classFile = classFile;
		this.pdsInfo = pdsInfo;
		
		String className = classFile.getThisClassName();
		
		thisClassName = formatClassName(className);
		logger.info("\tthisClassName=" + thisClassName);
		
		superClassName = formatClassName(classFile.getSuperClassName());
		logger.info("\tsuperClassName=" + superClassName);
		
		initInterfaceName(classFile);
		
		initPDSFieldList(classFile, pdsInfo);
		
		initPDSMethodList(pdsInfo);
		
		logger.info("Leaving PDSClass constructor");
	}
	
	/**
	 * Returns the <code>ClassFile</code> object of this class.
	 * 
	 * @return the <code>ClassFile</code> object.
	 */
	public ClassFile getClassFile() {
		
		return classFile;
	}
	
	/**
	 * Sets the index of the class.
	 * 
	 * @param index the index.
	 */
	public void setIndex(int index) {
		
		this.index = index;
	}
	
	/**
	 * Gets the index of the class.
	 * 
	 * @return the index of the class.
	 */
	public int getIndex() {
		
		return index;
	}
	
	/**
	 * Translates this class to Remopla.
	 * 
	 * @throws InvalidByteCodeException
	 */
	public void translate() throws InvalidByteCodeException {
		
		for (PDSMethod pdsMethod : pdsMethodList) {
			pdsMethod.translate();
		}
	}

	/**
	 * Returns <code>true</code> if this class is a concrete class 
	 * which can be instantiated, i.e. not abstract nor interface.
	 * 
	 * @return <code>true</code> if this class can be instantiated.
	 */
	public boolean canBeInstantiated() {
		
		int flags = classFile.getAccessFlags();
		return (flags & AccessFlags.ACC_ABSTRACT) == 0 
			&& (flags & AccessFlags.ACC_INTERFACE) == 0
			&& (flags & AccessFlags.ACC_INTERFACE) == 0;
	}
		
	public static String formatClassName(String name) {
		// TODO review
//		return (name == null) ? null : name.replace('/', '_');
		return (name == null) ? null : PDSMethod.formatMethodName(name);
		// TODO refactor, it basically 
	}
	
	public String getFormattedClassName() {
		
		return thisClassName;
	}
	
	public String getFormattedSuperClassName() {
		
		return superClassName;
	}
	
	public String[] getInterfaceName() {
		
		return interfaceName;
	}
	
	/**
	 * Expects the full method signature including parameters of the method.
	 * <p>
	 * Examples:
	 * <pre>
	 * main([Ljava/lang/String;)V for main method
	 * </pre>
	 * <p>
	 * <code>name</code> can for example be constructed using 
	 * {@link ClassMember#getName()} + {@link ClassMember#getDescriptor()}.
	 * @return null if the method was not found or there was a bytecode
	 * exception
	 */	
	public PDSMethod getMethod(String name, String des) {
		
		PDSMethod pdsMethod;
		
		for (int i = 0; i < pdsMethodList.size(); i++) {
			
			pdsMethod = pdsMethodList.get(i);
			logger.debug("pdsMethod.getName(): " + pdsMethod.getName());
			logger.debug("pdsMethod.getDescriptor(): " + pdsMethod.getDescriptor());
			if (pdsMethod.isNameAndDescriptor(name, des)) {
				return pdsMethod;
			}
		}
		
		return null;
	}
	
	/**
	 * Added for testing purposes.
	 * @param index
	 * @return
	 */
	PDSMethod getMethod(int index)
	{
		return pdsMethodList.get(index);
	}
	
	/**
	 * Added for testing purposes.
	 * @return an unmodifiable list with elements of type {@link PDSMethod}.
	 */
	public List<PDSMethod> getMethodList() {
		return Collections.unmodifiableList(pdsMethodList);
	}
	
	PDSField getField(String fieldName) {
		
		for (PDSField pdsField : pdsStaticFieldList)
			if (pdsField.getName().equals(fieldName))
				return pdsField;
		
		for (PDSField pdsField : pdsNonStaticFieldList)
			if (pdsField.getName().equals(fieldName))
				return pdsField;
		
		return null;
	}
	
	/**
	 * 
	 * @return an unmodifiable list with elements of type {@link PDSField}.
	 */
	public List<PDSField> getStaticFieldList() 	{
		
		return Collections.unmodifiableList(pdsStaticFieldList);
	}
	
	/**
	 * Returns the static initializer method of this class if this class 
	 * has one, otherwise <code>null</code>.
	 */
	public PDSMethod getMethodClinit() {

		return getMethod("<clinit>", "()V");

	}
	
	/**
	 * Returns the static main method if this class has one otherwise 
	 * <code>null</code>.
	 */
	public PDSMethod getMethodMain() {
		
		return getMethod("main", "([Ljava/lang/String;)V");
	}
	
	public String getName() {
		
		try {
			return classFile.getThisClassName();
		} catch (InvalidByteCodeException e) {
			e.printStackTrace();
		}
		
		logger.error("Error in jclasslib");
		return null;
	}

	/**
	 * Returns the list of instance fields of this class.
	 * <p>
	 * The elements are of type {@link PDSField} and the list is unmodifiable. 
	 */
	public List<PDSField> getNonStaticFieldList() {
		return Collections.unmodifiableList(pdsNonStaticFieldList);
	}
	
	public String getSuperName() {
		
		try {
			return classFile.getSuperClassName();
		} catch (InvalidByteCodeException e) {
			e.printStackTrace();
		}
		
		logger.error("Error in jclasslib");
		return null;
	}
	
	public boolean hasField(String fieldName) {
	
		PDSField pdsField = getField(fieldName);
		return (pdsField != null) ? true : false;
	}
	
	public boolean isName(String name) {
		
		String className = null;
		
		try {
			className = classFile.getThisClassName();
		} catch (InvalidByteCodeException e) {
			e.printStackTrace();
		}
		
		return className.equals(name);
	}
	
	/**
	 * Returns the number of instance fields in this class.
	 */
	public int size() throws InvalidByteCodeException {
		
		int count = 0;
		int bits = pdsInfo.getBits();
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE)
			return pdsNonStaticFieldList.size();
		
		// HeapOption.TWODIMS
		for (PDSField pdsField : pdsNonStaticFieldList)
			count += pdsField.getBits(bits);
		
		return (int) Math.ceil(((float) count)/((float) bits));
	}
	
	/**
	 * Returns a string containing the declaration of all static fields and 
	 * all methods of this class in Remopla syntax.
	 * @return
	 * @throws InvalidByteCodeException
	 */
	public String toRemoplaGlobal() throws InvalidByteCodeException {
		
		StringBuilder out = new StringBuilder();
		
		for (PDSField pdsField : pdsStaticFieldList) {
			out.append(pdsField.toRemopla());
		}
		
		for (PDSMethod pdsMethod : pdsMethodList) {
			out.append(pdsMethod.toRemoplaHead());
			out.append(";\n");
		}
		
		return out.toString();
	}
	
	/**
	 * Added for testing purposes, returns {@link #getName()};
	 */
//	@Override
	public String toString()
	{
		return getName();
	}
	
	/**
	 * Two classes are considered equal when their classnames are the 
	 * same.
	 */
	//@Override
	public boolean equals(Object o) {
		if (o instanceof PDSClass) {
			return thisClassName.equals(((PDSClass)o).thisClassName);
		}
		return false;
	}
	
//	@Override
	public int hashCode() {
		return thisClassName.hashCode();
	}
	
	/**
	 * Returns a string that contains the definitions of all the methods in
	 * this class.
	 * @return
	 * @throws InvalidByteCodeException
	 */
	public String toRemoplaModule() throws InvalidByteCodeException {
		
		PDSMethod pdsMethod;
		StringBuffer out = new StringBuffer();
		
		for (int i = 0; i < pdsMethodList.size(); i++) {
			
			pdsMethod = pdsMethodList.get(i);
			out.append(pdsMethod.toRemopla());
			out.append("\n");
		}
		
		return out.toString();
	}
	
	private void initPDSMethodList(PDSInfo pdsInfo) 
			throws InvalidByteCodeException, IOException {
		
		logger.debug("thisClassName: " + thisClassName);
		
		Set<String> methods = pdsInfo.getInvokeAnalyzer().getMethods(
				classFile.getThisClassName());
		logger.debug("methods.toString: " + methods.toString());
		MethodInfo[] methodInfos = classFile.getMethods();
				
		pdsMethodList = new ArrayList<PDSMethod>();
		for (int i = 0; i < methodInfos.length; i++) {
			
			String nameAndType = methodInfos[i].getName() + methodInfos[i].getDescriptor();
			if (methods.contains(nameAndType)) {
				pdsMethodList.add(new PDSMethod(this, methodInfos[i], pdsInfo));
			}
		}
	}
	
	/*******************************************************************/
	/**                      PRIVATE METHODS                          **/
	/*******************************************************************/
	
	private void initInterfaceName(ClassFile cf) 
			throws InvalidByteCodeException {
		
		int[] name = cf.getInterfaces();
		ConstantClassInfo ccInfo;
		
		interfaceName = new String[name.length];
		for (int i = 0; i < name.length; i++) {
			
			logger.debug("\tinterface index = " + name[i]);
			ccInfo = (ConstantClassInfo) cf.getConstantPoolEntry(name[i], 
					ConstantClassInfo.class);
			interfaceName[i] = ccInfo.getName();
			logger.info("\timplements " + interfaceName[i]);
		}
	}
	
	private void initPDSFieldList(ClassFile cf, PDSInfo pdsInfo) 
		throws InvalidByteCodeException
	{	
		
		logger.debug("Entering initPDSFieldList");
		
		PDSField pdsField;
		FieldInfo[] fieldInfo;
		String fieldName;
		int bits = pdsInfo.getBits();
		
		while (cf != null) {
			
			logger.debug("initPDSFieldList.cf.getThisClassName(): " 
					+ cf.getThisClassName());
			fieldInfo = cf.getFields();
			for (int i = 0; i < fieldInfo.length; i++) {
				
				fieldName = fieldInfo[i].getName();
				logger.debug("initPDSFieldList.fieldName: " + fieldName);
				
				if (pdsInfo.isIgnoredField(fieldName)) {
					logger.debug(fieldName + " is ignored");
					continue;
				}
				
				if (!hasField(fieldName)) {
					
					pdsField = new PDSField(thisClassName, fieldInfo[i]);
					logger.debug(pdsField);
					
					if (pdsField.isStatic()) {
						pdsStaticFieldList.add(pdsField);
					} else {
						pdsNonStaticFieldList.add(pdsField);
					}
				}
			}
			
			cf = pdsInfo.getInvokeAnalyzer().getClassFile(cf.getSuperClassName());
		}
		
		logger.debug("Leaving initPDSFieldList");
	}
}
