/*
    Copyright (C) 2005 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package de.uni_stuttgart.fmi.szs.jmoped;

import org.gjt.jclasslib.structures.*;
import org.apache.log4j.Logger;

import de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSAnnotationUtils;

/**
 * Field information for pushdown system.
 */
public class PDSField {
	
	static Logger logger = Logger.getLogger(PDSField.class);

	private FieldInfo fieldInfo;
	private String fieldName;
	
	private int bits;	// zero means default
	private int index;
	
	/**
	 * The constructor.
	 * 
	 * @param className the class name.
	 * @param fieldInfo the field info.
	 * @throws InvalidByteCodeException
	 */
	public PDSField(String className, FieldInfo fieldInfo) 
			throws InvalidByteCodeException {
		
		this.fieldInfo = fieldInfo;
		fieldName = className + "_" + formatFieldName(fieldInfo.getName());
		logger.debug("PDSField.fieldName: " + fieldName);
		
		if (isBoolean()) bits = 1;
		else if (isReference()) bits = 0;
		else bits = PDSAnnotationUtils.getFieldBit(fieldInfo, 0);
	}
	
	/**
	 * Formats the field name to the one legal to Moped.
	 *
	 * @param name the field name.
	 * @return the formatted field name.
	 */
	public static String formatFieldName(String name) {
		
		return name.replaceAll("[./$]", "_");
	}
	
	/**
	 * Formats the field name to the one legal to Moped.
	 * The array must has at least two elements.
	 * The first element is the class name, while
	 * the seconde element is the field name.
	 *
	 * @param names the array of field name.
	 * @return the formatted field name.
	 */
	public static String formatFieldName(String[] name) {
		
		return formatFieldName(name[0], name[1]);
	}
	
	/**
	 * Formats the field name to the one legal to Moped.
	 *
	 * @param className the class name.
	 * @param fieldName the field name.
	 * @return the formatted field name.
	 */
	public static String formatFieldName(String className, String fieldName) {
		
		return PDSClass.formatClassName(className) + "_" 
				+ formatFieldName(fieldName);
	}
	
	/**
	 * Returns the Java name of this field.
	 * 
	 * @return the name of this field.
	 */
	public String getName() {
		
		try {
			return fieldInfo.getName();
		} catch (InvalidByteCodeException e) {
			e.printStackTrace();
		}
		
		logger.fatal("Error in jclasslib");
		return null;
	}
	
	/**
	 * Returns the Remopla name of this field.
	 * 
	 * @return the Remopla name of this field.
	 */
	public String getFormattedName() {
		
		return fieldName;
	}
	
	/**
	 * Returns the descriptor of this field.
	 * 
	 * @return the descriptor of this field.
	 * @throws InvalidByteCodeException
	 */
	public String getDescriptor() throws InvalidByteCodeException {
		
		return fieldInfo.getDescriptor();
	}
	
	/**
	 * Returns <code>true</code> if the field is of Boolean type.
	 * 
	 * @return <code>true</code> if the field s of Boolean type.
	 * @throws InvalidByteCodeException
	 */
	public boolean isBoolean() throws InvalidByteCodeException {
		
		return getDescriptor().equals("Z");
	}
	
	/**
	 * Returns <code>true</code> if the field is of reference type.
	 * 
	 * @return <code>true</code> if the field is of reference type.
	 * @throws InvalidByteCodeException
	 */
	public boolean isReference() throws InvalidByteCodeException {
		
		return isReference(getDescriptor());
	}
	
	/**
	 * Returns <code>true</code> if name is the name of this field.
	 * 
	 * @param name the name.
	 * @return <code>true</code> if name is the name of this field.
	 */
	public boolean isName(String name) {
		
		return getName().equals(name);
	}
	
	/**
	 * Returns the underlying field info object.
	 * 
	 * @return the field info
	 */
	public FieldInfo getFieldInfo() {
		
		return fieldInfo;
	}
	
	/**
	 * Returns the remopla code for a static field honoring the 
	 * {@link de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSFieldBits}.
	 * 
	 * @throws InvalidByteCodeException
	 * @throws IllegalStateException if this is called on a non-static field
	 */
	public String toRemopla() throws InvalidByteCodeException
	{
		if (!isStatic()) {
			throw new IllegalStateException("Should only be called on static methods");
		}
		
		StringBuilder builder = new StringBuilder("int ");
		builder.append(getFormattedName());
		
		if (bits > 0) {
			builder.append("(");
			builder.append(bits);
			builder.append(")");
		}
		
		builder.append(";\n");
		return builder.toString();
	}
	
	/**
	 * Sets the number of bits which should be used for this field.
	 * 
	 * @param bits the number of bits
	 */
	public void setBits(int bits) {
		
		this.bits = bits;
	}
	
	/**
	 * Gets the number of bits of this field.
	 * 
	 * @return the number of bits.
	 */
	public int getBits() {
		
		return bits;
	}
	
	/**
	 * Gets the number of bits of this field. If the stored bits if zero,
	 * defaultBits is returned.
	 * 
	 * @param defaultBits the default bits.
	 * @return the number of bits.
	 */
	public int getBits(int defaultBits) {
		
		return (bits != 0) ? bits : defaultBits;
	}
	
	/**
	 * Sets the index of this field.
	 * 
	 * @param index the index.
	 */
	public void setIndex(int index) {
		
		this.index = index;
	}
	
	/**
	 * Gets the index of this field.
	 * 
	 * @return the index.
	 */
	public int getIndex() {
		
		return index;
	}
	
	/**
	 * Returns <code>true</code> if this field is static.
	 * 
	 * @return <code>true</code> if this field is static.
	 */
	public boolean isStatic() {
		
		return isStatic(fieldInfo.getAccessFlags());
	}
	
	public String toString() {
		
		return String.format("name: %s, bits: %d, index: %d", 
				fieldName, bits, index);
	}
	
	/**
	 * Checks if the given access flag is static.
	 *
	 * @return <code>true</code> if static.
	 */
	public static boolean isStatic(int flag) {
		
		return (flag & AccessFlags.ACC_STATIC) != 0;
	}

	/**
	 * Returns <code>true</code> if the field descriptor denotes 
	 * a reference type, i.e. all objects and all arrays.
	 * 
	 * @param descriptor the field descriptor.
	 * @return <code>true</code> if the descriptor denotes a reference type.
	 */
	public static boolean isReference(String descriptor) {
		
		return descriptor.startsWith("L") || descriptor.startsWith("[");
	}
}
