/*
    Copyright (C) 2005 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.apache.log4j.Logger;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * This class manages all the configuration options that are currently used.
 * <p>
 * Its constructors are either given a class file or a file on the file system 
 * which is read as a {@link org.gjt.jclasslib.structures.ClassFile}. 
 * After PDSInfo's constructor is finished, PDSInfo contains all the 
 * {@link #getClassFileList() class files} that are 
 * needed when calling any method from the given class file.
 * <p>
 * Included are files which are either instantiated or otherwise invoked from 
 * the original class file. 
 */
public class PDSInfo {
	
	static Logger logger = Logger.getLogger(PDSInfo.class);
	
	InvokeAnalyzer invokeAnalyzer;

	private HashSet<String> ignoredFieldSet = new HashSet<String>();
	private Hashtable<String, PDSField> fieldTable;
	private Hashtable<String, PDSClass> classTable;
	private Hashtable<Integer, Integer> lineBitTable
		= new Hashtable<Integer, Integer>();
	
	private String labelAssertError = PDSDefault.LABEL_ASSERT_ERROR;
	private String labelHeapOverflow = PDSDefault.LABEL_HEAP_OVERFLOW;
	private String labelNPE = PDSDefault.LABEL_NULL_POINTER_EXCEPTION;
	private String labelGeneralException = PDSDefault.LABEL_GENERAL_EXCEPTION;
	private String labelIndexOutOfBounds = PDSDefault.LABEL_INDEX_OUT_OF_BOUNDS;
	private String labelStringBuilderOverflow = PDSDefault.LABEL_STRINGBUILDER_OVERFLOW;
	
	private int bits = PDSDefault.BITS;
	private String heapName = PDSDefault.HEAP_NAME;
	private String heapPtrName = PDSDefault.HEAP_PTR_NAME;
	private int heapSize = PDSDefault.HEAP_SIZE;
	private String mopedPath = PDSDefault.MOPED_PATH;
	private String varName = PDSDefault.LOCAL_VAR_NAME;
	private String stackName = PDSDefault.STACK_NAME;
	
	/**
	 * Two type of heap manipulation strategies:
	 * 
	 * SIMPLE is the same as before.
	 * 
	 * TWODIMS is similar to SIMPLE but represents a heap as a bit string. 
	 * For example if bits=4 and heapSize=15, the resulting Remopla will 
	 * declare: int heap[60](1); However, the heap pointer will have 4 bits 
	 * as usual. The trick is that we always access the heap as multiples 
	 * of 4, i.e. we always multiply heap pointer by 4 before accessing 
	 * the heap. This way we don't have to declare pointer with lots of 
	 * bits and are able to be flexible with elements in the heap, e.g.
	 * we can declare a field of an object having different bit size
	 * or declare an array of 1 bits. Nevertheless, we waste some space 
	 * in the heap anyway, since we always have to start with the offset
	 * of 4. 
	 */
	public static enum HeapOption { SIMPLE, TWODIMS }
	private HeapOption heapOption = HeapOption.SIMPLE;
	
	private boolean checkForHeapOverflow = PDSDefault.CHECK_FOR_HEAP_OVERFLOW;
	
	private boolean checkForIndexOutOfBounds = 
		PDSDefault.CHECK_FOR_INDEX_OUT_OF_BOUNDS;

	private boolean checkForNullPointerExceptions =
		PDSDefault.CHECK_FOR_NULL_POINTER_EXCEPTIONS;
	
	private boolean checkForStringBuilderOverflow = 
		PDSDefault.CHECK_FOR_STRINGBUILDER_OVERFLOW;
	
	private boolean abortAfterFailedAssertion = 
		PDSDefault.ABORT_AFTER_FAILED_ASSERTION;

	static boolean simplified = false;

	static boolean Stefan = true;
	
	private PDSString pdsString = new PDSString();
	
	/**
	 * Whether or not to include multiarray initializer method. It's final value
	 * is set when the included methods are determined.
	 */
	private boolean includeMultiNewArrayInitializerMethod = false;
	
	private boolean includeStringEqualsMethod = false;
	
	public PDSInfo(ClassFile classFile, String confName, String methodName,
			String methodDesc, String[] searchPaths) throws Exception {
		
		if (new File(confName).exists()) {
			parseConfig(confName);
		}
		
		invokeAnalyzer = new InvokeAnalyzer(classFile, confName, methodName, 
				methodDesc, searchPaths);
		pdsString.buildCharTable(invokeAnalyzer.getClassFiles());
	}
	
	/**
	 * Creates a pds info object for the class file at <code>rootFile</code>.
	 * @param rootFile
	 * @throws Exception
	 */
	public PDSInfo(File rootFile) throws Exception {
		this(rootFile, PDSDefault.CONFIG_NAME);
	}
	
	public PDSInfo(File rootFile, String confName, String methodName, 
			String methodDesc) throws Exception {

		if (new File(confName).exists()) {
			parseConfig(confName);
		}
		
		invokeAnalyzer = new InvokeAnalyzer(rootFile, confName, methodName, 
				methodDesc);
		pdsString.buildCharTable(invokeAnalyzer.getClassFiles());
	}
	
	public PDSInfo(File rootFile, String confName) throws Exception
	{
		this(rootFile, confName, null, null);
	}
	
	public PDSInfo(File rootFile, String methodName, String methodDescriptor)
		throws Exception
	{
		this(rootFile, PDSDefault.CONFIG_NAME, methodName, methodDescriptor);
	}
	
	public PDSInfo(ClassFile classFile) throws Exception {
		this(classFile, new String[0]);
	}
	
	public PDSInfo(ClassFile classFile, String[] searchPaths) throws Exception {
		this(classFile, PDSDefault.CONFIG_NAME, null, null, searchPaths);
	}
	
	public PDSInfo(ClassFile classFile, String methodName, 
				   String methodDescriptor, String[] searchPaths)
		throws Exception
	{
		this(classFile, PDSDefault.CONFIG_NAME, methodName, methodDescriptor, 
				searchPaths);
	}
	
	public PDSInfo(Class clazz) throws Exception {
		this(BytecodeUtils.findClassFile(clazz.getName()));
	}
	
	public PDSInfo(Class clazz,  String methodName, String methodDescriptor) 
		throws Exception 
	{
		this(BytecodeUtils.findClassFile(clazz.getName()), methodName, 
			 methodDescriptor, new String[0]);
	}
	
	public PDSInfo(String className, String[] pathEntries) throws Exception {
		this(BytecodeUtils.findClassFile(className, pathEntries), pathEntries);
	}
	
	public PDSInfo(String className, String[] pathEntries, String methodName, 
			String methodDescriptor) throws Exception 
	{
		this(BytecodeUtils.findClassFile(className, pathEntries), methodName, 
			 methodDescriptor, pathEntries);
	}

	public InvokeAnalyzer getInvokeAnalyzer() {
		
		return invokeAnalyzer;
	}
	
	public int getBits() {
		
		return bits;
	}
	
	public void putAllLineBitTable(Hashtable<Integer, Integer> lbTable) {
		
		lineBitTable.putAll(lbTable);
	}
	
	public void setBitsAtLine(int l, int b) {
		
		lineBitTable.put(l, b);
	}
	
	public int getBitsAtLine(int l) {
		
		logger.debug("getBitsAtLine(" + l + ")");
		Integer b = lineBitTable.get(l);
		return (b != null) ? b.intValue() : 0;
	}
	
	/**
	 * Sets the number of bits used for emulating integers, might also change
	 * the size of the heap.
	 * <p>
	 * The heap size can only be as large as the largest number that can be
	 * represented using <code>numOfBits</code> bits.
	 * @param numOfBits
	 */
	public void setBits(int numOfBits) {
		bits = numOfBits;
		int maximumHeapSize = PDSDefault.getMaxIntForBits(bits);
		if (maximumHeapSize < heapSize) {
			heapSize = maximumHeapSize; 
		}
	}

	/**
	 * 
	 * @param className
	 * @return -1 one if there is no index for this class, which might be the
	 * case for ignored, abstract or interface classes
	 */
	public int getClassIndex(String className) {
		
		PDSClass clazz = classTable.get(className);
		if (clazz != null) {
			return clazz.getIndex();
		} else {
			logger.error("Cannot find className=" + className);
			return -1;
		}
	}
	
	/**
	 * Returns the name of the pointer into the Remopla heap.
	 * <p>
	 * Java's heap is simulated in Remopla using a static array and
	 * a pointer pointing to the last used array element/ next free element.
	 * <p>
	 * Defaults to {@link PDSDefault#HEAP_PTR_NAME}.
	 * @return
	 */
	public int getClassSize(String className) throws InvalidByteCodeException {
		
		PDSClass clazz = classTable.get(className);
		if (clazz != null) {
			return clazz.size();
		} else {
			logger.error("Cannot find className=" + className);
			return -1;
		}
	}
	
	public HeapOption getHeapOption() {
		
		return heapOption;
	}
	
	public void setHeapOption(HeapOption heapOption) {
		
		this.heapOption = heapOption;
	}
	
	public String getHeapPtrName() {
		
		return heapPtrName;
	}
	
	/**
	 * Returns the size of the Remopla heap.
	 * <p>
	 * Defaults to {@link PDSDefault#HEAP_SIZE}.
	 * @return
	 */
	public int getHeapSize() {
		
		if (heapOption == HeapOption.SIMPLE)
			return heapSize;
		else // if (heapOption == HeapOption.TWODIMS)
			return bits * heapSize;
	}
	
	public int getStoredHeapSize() {
		
		return heapSize;
	}
	
	/**
	 * Sets the size of the Remopla heap and may change the number of bits.
	 * <p>
	 * The number of bits must be large enough to index all elements in the
	 * heap array, so it is adapted if necessary.
	 * 
	 * @throws IllegalArgumentException if <code>size</code> smaller than 1
	 */
	public void setHeapSize(int size) {
		if (size < 1) {
			throw new IllegalArgumentException("heap size must be >= 1");
		}
		heapSize = size;
		bits = Math.max(bits, PDSDefault.getMinimumBitsForHeapSize(heapSize));
	}
	
	/**
	 * Returns the name of the static array used in Remopla to simulate
	 * the heap.
	 * <p>
	 * Defaults to {@link PDSDefault#HEAP_NAME}.
	 * @return
	 */
	public String getHeapName() {
		return heapName;
	}
	
	public int getStringBuilderSize() {
		
		return pdsString.getStringBuilderSize();
	}
	
	public int getCharBits() {
		
		return pdsString.getCharBits();
	}
	
	public int getCharRep(char c) {
		
		return pdsString.getCharRep(c);
	}
	
	public void setCheckForIndexOutOfBounds(boolean newValue)
	{
		checkForIndexOutOfBounds = newValue;
	}
	
	public boolean checkForIndexOutOfBounds()
	{
		return checkForIndexOutOfBounds;
	}
	
	/**
	 * Returns true if the codepath should not be checked further, which is 
	 * possilble by just ignoring failed the assertion check altogether.
	 * @return
	 */
	public boolean abortAfterFailedAssertion()
	{
		return abortAfterFailedAssertion;
	}
	
	public void setAbortAfterFailedAssertion(boolean newValue)
	{
		abortAfterFailedAssertion = newValue;
	}
	
	public void setCheckForHeapOverflow(boolean newValue)
	{
		checkForHeapOverflow = newValue;
	}
	
	public boolean checkForHeapOverflow()
	{
		return checkForHeapOverflow;
	}
	
	public boolean checkForNullPointerExceptions()
	{
		return checkForNullPointerExceptions;
	}
	
	public boolean checkForStringBuilderOverflow() {
		
		return checkForStringBuilderOverflow;
	}
	
	public void setCheckForNullPointerExceptions(boolean newValue)
	{
		checkForNullPointerExceptions = newValue;
	}
	
	/**
	 * Returns the name of the variables used as stack in Remopla.
	 * <p>
	 * Defaults to {@link PDSDefault#STACK_NAME}.
	 * @return
	 */
	public String getStackName() {
		return stackName;
	}
	
	/**
	 * Returns the name of local variables created in jmoped's output.
	 * <p>
	 * Defaults to {@link PDSDefault#LOCAL_VAR_NAME}.
	 * @return
	 */
	public String getLocalVarName() {
		return varName;
	}

	/**
	 * Whether or not the initialization code for multidimensional arrays should be
	 * included in the Remopla output. 
	 * @return true if a multidimensional array is instantiated somewhere on the code
	 * path
	 */
	public boolean includeMultiArrayInitializerMethod()
	{
		return includeMultiNewArrayInitializerMethod;
	}
	
	public void setIncludeMultiArrayInitializerMethod() {
		
		includeMultiNewArrayInitializerMethod = true;
	}
	
	public boolean isIncludeStringEqualsMethod() {
		
		return includeStringEqualsMethod;
	}
	
	public void setIncludeStringEqualsMethod() {
		
		includeStringEqualsMethod = true;
	}
	
	public String getLabelAssertError() {
		
		return labelAssertError;
	}
	
	public String getLabelHeapOverflow() {
		
		return labelHeapOverflow;
	}
	
	public String getLabelNPE() {
		
		return labelNPE;
	}
	
	public String getLabelIndexOutOfBounds() {
		
		return labelIndexOutOfBounds;
	}
	
	public String getLabelGeneralException() {
		
		return labelGeneralException;
	}
	
	public String getLabelStringBuilderOverflow() {
		
		return labelStringBuilderOverflow;
	}

	/**
	 * Returns the name of the moped executable.
	 * <p>
	 * May be its absolute path name or just the filename if the executable
	 * is in PATH.
	 * @return
	 */	
	public String getMopedPath() {
		
		return mopedPath;
	}
	
	/**
	 * Returns the value of {@link #getHeapName()} subscripted
	 * by the given expression.
	 * @param expression
	 * @return
	 */
	public String heapIndex(String expression) {
		
		return heapIndex(expression, "");
	}
	
	public String heapIndex(String expression, String prime) {
		
		return stringIndex(heapName, expression, prime);
	}

	/**
	 * Returns the offset at which the given field is stored within its class.
	 * @param formattedFieldName
	 * @return -1 if field was not found
	 * TODO throw illegal argument exception?
	 */
	public int indexOfField(String formattedFieldName) {
		
		PDSField pdsField = fieldTable.get(formattedFieldName);
		if (pdsField != null) {
			return pdsField.getIndex();
		} else {
			logger.error("Cannot find formattedFieldName=" + formattedFieldName);
			logger.debug(fieldIndexTableInfo());
			return -1;
		}
	}
	
	public int getFieldBits(String formattedFieldName) 
			throws InvalidByteCodeException {
		
		PDSField pdsField = fieldTable.get(formattedFieldName);
		if (pdsField != null) {
			return pdsField.getBits(bits);
		} else {
			logger.error("Cannot find formattedFieldName=" + formattedFieldName);
			logger.debug(fieldIndexTableInfo());
			return -1;
		}
	}
	
	/**
	 * Returns true if the field should not be included in jmoped's output.
	 * @param fieldName
	 * @return
	 */
	public boolean isIgnoredField(String fieldName) {
		
		for (int i = 0; i < PDSDefault.IGNORED_FIELD.length; i++) {
			
			if (fieldName.startsWith(PDSDefault.IGNORED_FIELD[i])) {
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * Takes an array of class name and field name and checks if either
	 * the class or the field in the class should be ignored.
	 * @param fieldName
	 * @return
	 */
	public boolean isIgnoredField(String[] fieldName) {
		
		return invokeAnalyzer.isIgnoredClass(fieldName[0]) 
				|| isIgnoredField(fieldName[1]) 
				|| isIgnoredFormattedField(fieldName);
	}
	
	/**
	 * Formats the array consisting of a class name and a field to a fully
	 * qualified field name and checks if it should be ignored.
	 * @param fieldName
	 * @return
	 */
	private boolean isIgnoredFormattedField(String[] fieldName) {
		
		String name = PDSField.formatFieldName(fieldName);
		
		if (ignoredFieldSet.contains(name)) {
			return true;
		}
		return isIgnoredField(name);
	}

	/**
	 * Parses the xml configuration file.
	 * @param confName
	 * @throws Exception
	 */
	private void parseConfig(String confName) throws Exception {
		parseConfig(new BufferedInputStream(new FileInputStream(confName)));
	}
	
	
	private void parseConfig(InputStream in) throws Exception {
		
		DocumentBuilderFactory factory;
		DocumentBuilder builder;
		
		factory = DocumentBuilderFactory.newInstance();
		builder = factory.newDocumentBuilder();
		
		Document doc = builder.parse(in);
		
		NodeList nl1;
		Node n1, n2;
		NamedNodeMap nnm;
		String nodeName;
		
		nl1 = doc.getFirstChild().getChildNodes();
		for (int i = 0; i < nl1.getLength(); i++) {
			
			n1 = nl1.item(i);
			nodeName = n1.getNodeName();
			if (nodeName.equals("bits")) {
				
				nnm = n1.getAttributes();
				if (nnm == null || nnm.getLength() == 0) {
					bits = Integer.parseInt(n1.getFirstChild().getNodeValue());
					logger.debug("From jmoped.conf bits=" + bits);
				} else {
					n2 = nnm.item(0);
					if (n2.getNodeName().equals("line")) {
						int l = Integer.parseInt(n2.getNodeValue());
						int b = Integer.parseInt(n1.getFirstChild().getNodeValue());
						lineBitTable.put(l, b);
						logger.debug("From jmoped.conf bits=" + b + " at line=" +l);
					}
				}
			} else if (nodeName.equals("error")) {
				
				n2 = n1.getAttributes().item(0);
				if (n2.getNodeName().equals("name")) {
					labelAssertError = n2.getNodeValue();
				}
			} else if (nodeName.equals("heap")) {
				
				nnm = n1.getAttributes();
				for (int j = 0; j < nnm.getLength(); j++) {
					n2 = nnm.item(j);
					if (n2.getNodeName().equals("name")) {
						heapName = n2.getNodeValue();
					} else if (n2.getNodeName().equals("size")) {
						heapSize = Integer.parseInt(n2.getNodeValue());
					} else if (n2.getNodeName().equals("pointer")) {
						heapPtrName = n2.getNodeValue();
					}
				}
			} else if (nodeName.equals("moped")) {
				
				n2 = n1.getAttributes().item(0);
				if (n2.getNodeName().equals("path")) {
					mopedPath = n2.getNodeValue();
				}
			} else if (nodeName.equals("stack")) {
				
				n2 = n1.getAttributes().item(0);
				if (n2.getNodeName().equals("name")) {
					stackName = n2.getNodeValue();
				}
			} else if (nodeName.equals("stringbuilder")) {
				
				n2 = n1.getAttributes().item(0);
				if (n2.getNodeName().equals("size")) {
					int size = Integer.parseInt(n2.getNodeValue());
					pdsString.setStringBuilderSize(size);
					logger.debug("From jmoped.conf stringbuilder size=" + size);
				}
			} else if (nodeName.equals("var")) {
				
				n2 = n1.getAttributes().item(0);
				if (n2.getNodeName().equals("name")) {
					varName = n2.getNodeValue();
				}
			}
		}
		
		buildIgnoredTable(doc);
	}
	
	/**
	 * Returns the value of {@link #getStackName()} with the index
	 * appended to it. 
	 * @param expression
	 * @return
	 */
	public String stackIndex(int index) {
		
		return stackName + String.valueOf(index);
	}
	
	/**
	 * Returns a subscript string of the form
	 * <code>array[expression]</code>.
	 * @param array
	 * @param expression
	 * @return
	 */
	public static String stringIndex(String array, String expr, String prime) {
		
		StringBuilder out = new StringBuilder();
		
		out.append(array);
		out.append(prime);
		out.append("[");
		out.append(expr);
		out.append("]");
		
		return out.toString();
	}
	
	public static String stringIndex(String array, String expr) {
		
		return stringIndex(array, expr, "");
	}
	
	/**
	 * Returns the value of {@link #getLocalVarName() getLocalVarName} 
	 * with the index appended to it. 
	 * 
	 * @param index
	 * @return
	 */
	public String varIndex(int index) {
		
		return varName + String.valueOf(index);
	}
	
	public void setup(Set<PDSClass> pdsClasses) 
			throws InvalidByteCodeException {
		
		initFieldTable(pdsClasses);
	}
	
	/*******************************************************************/
	/**                          INFO METHODS                         **/
	/*******************************************************************/
	
	public String fieldIndexTableInfo() {
		
		return fieldTable.toString();
	}
	
	/*******************************************************************/
	/**               PUBLIC TABLES' BUILDING METHODS                 **/
	/*******************************************************************/
	
	/**
	 * Fills ignoredTable in such a way that its keys are class names,
	 * which map lists of method names to be ignored. A list of 
	 * size zero implies ignoring every method.
	 */
	private void buildIgnoredTable(Document confDoc) {
		
		NodeList nl1, nl2;
		Node n1, n2, n3, n4;

		ignoredFieldSet.clear();

		nl1 = confDoc.getFirstChild().getChildNodes();
		for (int i = 0; i < nl1.getLength(); i++) {
			
			n1 = nl1.item(i);
			
			if (!n1.getNodeName().equals("ignore")) continue;
			
			// n1 is "ignore" node
			nl2 = n1.getChildNodes();
			
			/* For each child node of "ignore" node */
			for (int j = 0; j < nl2.getLength(); j++) {
				
				n2 = nl2.item(j);
				
				if (!n2.getNodeName().equals("field")) continue;

				n3 = n2.getAttributes().item(0);
				if (!n3.getNodeName().equals("class")) {
					logger.error("Invalid field attribute.");
				}
				
				n4 = n2.getAttributes().item(1);
				if (!n4.getNodeName().equals("name")) {
					logger.error("Invalid field attribute.");
				}
				
				ignoredFieldSet.add(PDSField.formatFieldName(
						n3.getNodeValue(), n4.getNodeValue()));
				logger.info("\tIgnored field: " 
						+ n3.getNodeValue() + "." + n4.getNodeValue());
			}
			break;
		}
	}
	
	/*******************************************************************/
	/**                       PRIVATE METHODS                         **/
	/*******************************************************************/
	
	/**
	 * @param pdsClassList
	 */
	private void initFieldTable(Set<PDSClass> pdsClasses)
			throws InvalidByteCodeException {
		
		fieldTable = new Hashtable<String, PDSField>();
		
		for (PDSClass pdsClass : pdsClasses) {
			
			initFieldTable(pdsClass);
		}
		
		logger.info("buildFieldIndexTable completed");
		logger.debug(fieldTable);
	}
	
	/**
	 * Registers the class fields' indices by their formatted names in
	 * the field index table.
	 * @param pdsClass
	 */
	private void initFieldTable(PDSClass pdsClass) 
			throws InvalidByteCodeException {
		
		int fieldIndex = 0;
		
		for (PDSField pdsField : pdsClass.getNonStaticFieldList()) {
			
			pdsField.setIndex(fieldIndex);
			fieldTable.put(pdsField.getFormattedName(), pdsField);
			
			switch (heapOption) {
				case SIMPLE:
					fieldIndex++;
					break;
				case TWODIMS:
					fieldIndex += pdsField.getBits(bits);
					break;
			}
		}
	}
	
	/**
	 * @param pdsClasses
	 */
	public void initClassTable(Set<PDSClass> pdsClasses) {
		
		classTable = new Hashtable<String, PDSClass>();
		
		// start index at one, to avoid spurious excecutions for null pointer checks 
		int i = 1;
		for (PDSClass pdsClass : pdsClasses) {
			
			pdsClass.setIndex(i++);
			// TODO if i > 2^bits, program execution won't be guaranteed to be correct
			// for virtual calls and instanceof, so throw exception or produce warning
			if (i > Math.pow(2, bits)) {
				// produce a warning of some kind, since bits cannot be set in the constructor
		//		throw new InvalidByteCodeException("The number of included classes ("
		//				+ (i - 1) 
		//				+ ") cannot be handled with the current number of bits, please consider increasing the number of bits.");
			}
			classTable.put(pdsClass.getName(), pdsClass);
		}
		
		logger.info("buildClassSizeTable completed");
	}
}
