package de.uni_stuttgart.fmi.szs.jmoped;

import java.util.Arrays;
import java.util.Collection;
import java.util.Enumeration;
import java.util.List;
import java.util.ArrayList;

import javax.swing.tree.DefaultMutableTreeNode;

import org.apache.log4j.Logger;
import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.BranchInstruction;
import org.gjt.jclasslib.bytecode.ImmediateByteInstruction;
import org.gjt.jclasslib.bytecode.ImmediateShortInstruction;
import org.gjt.jclasslib.bytecode.IncrementInstruction;
import org.gjt.jclasslib.bytecode.InvokeInterfaceInstruction;
import org.gjt.jclasslib.bytecode.LookupSwitchInstruction;
import org.gjt.jclasslib.bytecode.MatchOffsetPair;
import org.gjt.jclasslib.bytecode.MultianewarrayInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;
import org.gjt.jclasslib.bytecode.TableSwitchInstruction;
import org.gjt.jclasslib.structures.AccessFlags;
import org.gjt.jclasslib.structures.CPInfo;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.constants.ConstantClassInfo;
import org.gjt.jclasslib.structures.constants.ConstantIntegerInfo;
import org.gjt.jclasslib.structures.constants.ConstantInterfaceMethodrefInfo;
import org.gjt.jclasslib.structures.constants.ConstantNameAndTypeInfo;
import org.gjt.jclasslib.structures.constants.ConstantReference;

import de.uni_stuttgart.fmi.szs.jmoped.InvokeAnalyzer.ClassNode;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo.HeapOption;
import de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSAnnotationUtils;

import static de.uni_stuttgart.fmi.szs.jmoped.PDSInfo.HeapOption;

public class PDSInst {
	
	/**
	 * Represents type of bytecode instructions that are not supported
	 * by jMoped
	 */
	public static enum Warning { BITWISE, FLOAT, DOUBLE, NEGATIVE, STRING }
	
	private AbstractInstruction absInst = null;
	private PDSMethod pdsMethod;
	private PDSInfo pdsInfo;
	private PDSStmt pdsStmt;
	private List<Warning> warningList = new ArrayList<Warning>();
	
	/**
	 * Marks if this instruction was replaced from the outside. 
	 * If this is the case, the Remopla output will not be rebuilt 
	 * in {@link #toRemopla()} and the actual 
	 * underlying bytecode instruction will be ignored.
	 */
	private boolean instReplaced = false;
	
	static Logger logger = Logger.getLogger(PDSInst.class);

	public PDSInst(
			PDSMethod pdsMethod, AbstractInstruction absInst, PDSInfo pdsInfo)
		throws InvalidByteCodeException {
		
		this.pdsMethod = pdsMethod;
		this.absInst = absInst;
		this.pdsInfo = pdsInfo;
	}
	
	public PDSInst(PDSMethod pdsMethod, PDSInfo pdsInfo) {
		
		this.pdsMethod = pdsMethod;
		this.pdsInfo = pdsInfo;
	}
	
	public void addInst(String stmt) {
		
		if (pdsStmt == null) pdsStmt = new PDSStmt();
		pdsStmt.addNext(stmt);
	}
	
	public void replaceInst(String st) {
		
		pdsStmt = new PDSStmt(st);
		instReplaced = true;
	}
	
	public String toRemopla() throws InvalidByteCodeException {
		
		StringBuilder out = new StringBuilder();

		out.append(getLabel());
		out.append(":\t");
		out.append(pdsStmt.toRemopla());
		
		return out.toString();
	}
	
	/**
	 * Returns the label for this instruction.
	 * <p>
	 * If the pc offset in the method is large than 0, it is appended,
	 * otherwise just the formatted method name is returned.
	 */
	private String getLabel() {
		
		StringBuilder builder = new StringBuilder(pdsMethod.getFormattedName());
		if ((absInst != null) && absInst.getOffset() > 0) {
			builder.append(absInst.getOffset());
		}
		return builder.toString();
	}
	
	public List<Warning> getWarningList() {
		
		return warningList;
	}
	
	public void translate() throws InvalidByteCodeException {
		
		if (absInst == null || instReplaced)
			return;
		
		pdsStmt = new PDSStmt();
		switch (absInst.getOpcode()) {
		
		case Opcodes.OPCODE_AALOAD: 
			aaloadInst();
			break;
		case Opcodes.OPCODE_AASTORE:
			aastoreInst();
			break;
		case Opcodes.OPCODE_ACONST_NULL: 
			pushWithInst(0);
			break;
		case Opcodes.OPCODE_ALOAD: 
			pushWithInst(pdsInfo.varIndex(immediateByte()));
			break;
		case Opcodes.OPCODE_ALOAD_0: 
		case Opcodes.OPCODE_ALOAD_1: 
		case Opcodes.OPCODE_ALOAD_2: 
		case Opcodes.OPCODE_ALOAD_3: 
			pushWithInst(pdsInfo.varIndex(
					absInst.getOpcode() - Opcodes.OPCODE_ALOAD_0));
			break;
		case Opcodes.OPCODE_ANEWARRAY:
			anewarrayInst();
			break;
		case Opcodes.OPCODE_ARETURN:
			pdsStmt.addReturn(pdsInfo.stackIndex(0));
			break;
		case Opcodes.OPCODE_ARRAYLENGTH: 
			arraylengthInst();
			break;
		case Opcodes.OPCODE_ASTORE:
			popToInst(pdsInfo.varIndex(immediateByte()));
			break;
		case Opcodes.OPCODE_ASTORE_0: 
		case Opcodes.OPCODE_ASTORE_1: 
		case Opcodes.OPCODE_ASTORE_2: 
		case Opcodes.OPCODE_ASTORE_3:
			popToInst(pdsInfo.varIndex(
					absInst.getOpcode() - Opcodes.OPCODE_ASTORE_0));
			break;
		case Opcodes.OPCODE_ATHROW: 
			// TODO 
			// actually handling exceptions would mean to go through the
			// call stack and look for possible exception handlers. To
			// avoid spurious executions we go to an extra general
			// exception state for now
			pdsStmt.addGoto(pdsInfo.getLabelGeneralException());
			break;
		case Opcodes.OPCODE_BALOAD: 
			ialoadInst();
			break;
		case Opcodes.OPCODE_BASTORE:
			iastoreInst();
			break;
		case Opcodes.OPCODE_BIPUSH: 
			pushWithInst(immediateByte());
			break;
		case Opcodes.OPCODE_CALOAD: 
			ialoadInst();
			break;
		case Opcodes.OPCODE_CASTORE:
			iastoreInst();
			break;
		case Opcodes.OPCODE_CHECKCAST: 
			//TODO
			pdsStmt.addSkip();
			break;
		case Opcodes.OPCODE_DADD:
			warningList.add(Warning.DOUBLE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_DALOAD:
			warningList.add(Warning.DOUBLE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_DASTORE:
			warningList.add(Warning.DOUBLE);
			popPushInst("undef", 3, 0);
			break;
		case Opcodes.OPCODE_DCMPG:
		case Opcodes.OPCODE_DCMPL:
			warningList.add(Warning.DOUBLE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_DCONST_0: 
		case Opcodes.OPCODE_DCONST_1:
			warningList.add(Warning.DOUBLE);
			pushWithInst(absInst.getOpcode() - Opcodes.OPCODE_DCONST_0);
			break;
		case Opcodes.OPCODE_DDIV:
			warningList.add(Warning.DOUBLE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_DLOAD: 
		case Opcodes.OPCODE_DLOAD_0: 
		case Opcodes.OPCODE_DLOAD_1: 
		case Opcodes.OPCODE_DLOAD_2: 
		case Opcodes.OPCODE_DLOAD_3:
			warningList.add(Warning.DOUBLE);
			pushWithInst("undef");
			break;
		case Opcodes.OPCODE_DMUL:
			warningList.add(Warning.DOUBLE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_DNEG:
			warningList.add(Warning.DOUBLE);
			warningList.add(Warning.NEGATIVE);
			popPushInst("undef", 1, 1);
			break;
		case Opcodes.OPCODE_DREM:
			warningList.add(Warning.DOUBLE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_DRETURN:
			warningList.add(Warning.DOUBLE);
			pdsStmt.addReturn(pdsInfo.stackIndex(0));
			break;
		case Opcodes.OPCODE_DSTORE: 
		case Opcodes.OPCODE_DSTORE_0: 
		case Opcodes.OPCODE_DSTORE_1: 
		case Opcodes.OPCODE_DSTORE_2: 
		case Opcodes.OPCODE_DSTORE_3:
			warningList.add(Warning.DOUBLE);
			popPushInst("undef", 1, 0);
			break;
		case Opcodes.OPCODE_DSUB:
			warningList.add(Warning.DOUBLE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_DUP: 
			pushWithInst(pdsInfo.stackIndex(0));
			break;
		case Opcodes.OPCODE_DUP_X1:
			pdsStmt.addNext(stackEqual(0, 0));
			pdsStmt.addPar(stackEqual(1, 1));
			pdsStmt.addPar(stackEqual(2, 0));
			pdsStmt.addPar(pushStmt(3, 1));
			break;
		case Opcodes.OPCODE_DUP_X2:
			pdsStmt.addNext(stackEqual(0, 0));
			pdsStmt.addPar(stackEqual(1, 1));
			pdsStmt.addPar(stackEqual(2, 2));
			pdsStmt.addPar(stackEqual(3, 0));
			pdsStmt.addPar(pushStmt(4, 1));
			break;
		case Opcodes.OPCODE_DUP2:
			pdsStmt.addNext(stackEqual(0, 0));
			pdsStmt.addPar(stackEqual(1, 1));
			pdsStmt.addPar(stackEqual(2, 0));
			pdsStmt.addPar(stackEqual(3, 1));
			pdsStmt.addPar(pushStmt(4, 1));
		case Opcodes.OPCODE_FADD:
			warningList.add(Warning.FLOAT);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_FALOAD:
			warningList.add(Warning.FLOAT);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_FASTORE:
			warningList.add(Warning.FLOAT);
			popPushInst("undef", 3, 0);
			break;
		case Opcodes.OPCODE_FCMPG: 
		case Opcodes.OPCODE_FCMPL:
			warningList.add(Warning.FLOAT);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_FCONST_0: 
		case Opcodes.OPCODE_FCONST_1: 
		case Opcodes.OPCODE_FCONST_2:
			warningList.add(Warning.FLOAT);
			pushWithInst(absInst.getOpcode() - Opcodes.OPCODE_FCONST_0);
			break;
		case Opcodes.OPCODE_FDIV:
			warningList.add(Warning.FLOAT);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_FLOAD: 
		case Opcodes.OPCODE_FLOAD_0: 
		case Opcodes.OPCODE_FLOAD_1: 
		case Opcodes.OPCODE_FLOAD_2: 
		case Opcodes.OPCODE_FLOAD_3: 
			pushWithInst("undef");
			break;
		case Opcodes.OPCODE_FMUL: 
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_FNEG:
			warningList.add(Warning.FLOAT);
			warningList.add(Warning.NEGATIVE);
			popPushInst("undef", 1, 1);
			break;
		case Opcodes.OPCODE_FREM: 
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_FRETURN:
			warningList.add(Warning.FLOAT);
			pdsStmt.addReturn(pdsInfo.stackIndex(0));
			break;
		case Opcodes.OPCODE_FSTORE: 
		case Opcodes.OPCODE_FSTORE_0: 
		case Opcodes.OPCODE_FSTORE_1: 
		case Opcodes.OPCODE_FSTORE_2: 
		case Opcodes.OPCODE_FSTORE_3:
			warningList.add(Warning.FLOAT);
			popPushInst("undef", 1, 0);
			break;
		case Opcodes.OPCODE_FSUB:
			warningList.add(Warning.FLOAT);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_GETFIELD: 
			getfieldInst();
			break;
		case Opcodes.OPCODE_GETSTATIC: 
			getstaticInst();
			break;
		case Opcodes.OPCODE_GOTO: 
			pdsStmt.addGoto(branchTarget());
			break;
		case Opcodes.OPCODE_IADD: 
			arithInst("+");
			break;
		case Opcodes.OPCODE_IALOAD: 
			ialoadInst();
			break;
		case Opcodes.OPCODE_IAND:
			warningList.add(Warning.BITWISE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_IASTORE:
			iastoreInst();
			break;
		case Opcodes.OPCODE_ICONST_0: 
		case Opcodes.OPCODE_ICONST_1: 
		case Opcodes.OPCODE_ICONST_2: 
		case Opcodes.OPCODE_ICONST_3: 
		case Opcodes.OPCODE_ICONST_4: 
		case Opcodes.OPCODE_ICONST_5: 
			pushWithInst(absInst.getOpcode() - Opcodes.OPCODE_ICONST_0);
			break;
		case Opcodes.OPCODE_ICONST_M1: 
			warningList.add(Warning.NEGATIVE);
			pushWithInst("undef");
			break;
		case Opcodes.OPCODE_IDIV: 
			arithInst("/");
			break;
		case Opcodes.OPCODE_IF_ACMPEQ:
			ifIcmpInst("==");
			break;
		case Opcodes.OPCODE_IF_ACMPNE:
			ifIcmpInst("!=");
			break;
		case Opcodes.OPCODE_IF_ICMPEQ:
			ifIcmpInst("==");
			break;
		case Opcodes.OPCODE_IF_ICMPGE:
			ifIcmpInst(">=");
			break;
		case Opcodes.OPCODE_IF_ICMPGT:
			ifIcmpInst(">");
			break;
		case Opcodes.OPCODE_IF_ICMPLE:
			ifIcmpInst("<=");
			break;
		case Opcodes.OPCODE_IF_ICMPLT:
			ifIcmpInst("<");
			break;
		case Opcodes.OPCODE_IF_ICMPNE:
			ifIcmpInst("!=");
			break;
		case Opcodes.OPCODE_IFEQ: 
			ifInst("==");
			break;
		case Opcodes.OPCODE_IFGE: 
			ifInst(">=");
			break;
		case Opcodes.OPCODE_IFGT: 
			ifInst(">");
			break;
		case Opcodes.OPCODE_IFLE: 
			ifInst("<=");
			break;
		case Opcodes.OPCODE_IFLT: 
			ifInst("<");
			break;
		case Opcodes.OPCODE_IFNE: 
			ifInst("!=");
			break;
		case Opcodes.OPCODE_IFNONNULL: 
			ifInst("!=");
			break;
		case Opcodes.OPCODE_IFNULL: 
			ifInst("==");
			break;
		case Opcodes.OPCODE_IINC: 
			iincInst();
			break;
		case Opcodes.OPCODE_ILOAD: 
			pushWithInst(pdsInfo.varIndex(immediateByte()));
			break;
		case Opcodes.OPCODE_ILOAD_0: 
		case Opcodes.OPCODE_ILOAD_1: 
		case Opcodes.OPCODE_ILOAD_2: 
		case Opcodes.OPCODE_ILOAD_3: 
			pushWithInst(pdsInfo.varIndex(
					absInst.getOpcode() - Opcodes.OPCODE_ILOAD_0));
			break;
		case Opcodes.OPCODE_IMUL: 
			arithInst("*");
			break;
		case Opcodes.OPCODE_INEG:
			warningList.add(Warning.NEGATIVE);
			popPushInst("undef", 1, 1);
			break;
		case Opcodes.OPCODE_INSTANCEOF: 
			instanceofInst();
			break;
		case Opcodes.OPCODE_INVOKEINTERFACE:
			invokeinterfaceInst();
			break;
		case Opcodes.OPCODE_INVOKESPECIAL: 
			invokespecialInst();
			break;
		case Opcodes.OPCODE_INVOKESTATIC:
			invokestaticInst();
			break;
		case Opcodes.OPCODE_INVOKEVIRTUAL: 
			invokevirtualInst();
			break;
		case Opcodes.OPCODE_IOR: 
			warningList.add(Warning.BITWISE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_IREM: 
			popPushInst(PDSStmt.mod(pdsInfo.stackIndex(1), pdsInfo.stackIndex(0)), 2, 1);
			break;
		case Opcodes.OPCODE_IRETURN: 
			pdsStmt.addReturn(pdsInfo.stackIndex(0));
			break;
		case Opcodes.OPCODE_ISHL: 
			arithInst("<<");
			break;
		case Opcodes.OPCODE_ISHR:
			warningList.add(Warning.BITWISE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_ISTORE:
			popToInst(pdsInfo.varIndex(immediateByte()));
			break;
		case Opcodes.OPCODE_ISTORE_0: 
		case Opcodes.OPCODE_ISTORE_1: 
		case Opcodes.OPCODE_ISTORE_2: 
		case Opcodes.OPCODE_ISTORE_3: 
			popToInst(pdsInfo.varIndex(
					absInst.getOpcode() - Opcodes.OPCODE_ISTORE_0));
			break;
		case Opcodes.OPCODE_ISUB: 
			arithInst("-");
			break;
		case Opcodes.OPCODE_IUSHR:
			warningList.add(Warning.BITWISE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_IXOR:
			warningList.add(Warning.BITWISE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_JSR: 
			pushWithInst(absInst.getOffset() + 2);
			pdsStmt.addGoto(branchTarget());
			break;
		case Opcodes.OPCODE_JSR_W:
			pushWithInst(absInst.getOffset() + 4);
			pdsStmt.addGoto(branchTarget());
			break;
		case Opcodes.OPCODE_LADD: 
			arithInst("+");
			break;
		case Opcodes.OPCODE_LALOAD: 
			ialoadInst();
			break;
		case Opcodes.OPCODE_LAND:
			warningList.add(Warning.BITWISE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_LASTORE:
			iastoreInst();
			break;
		case Opcodes.OPCODE_LCMP:
			lcmpInst();
			break;
		case Opcodes.OPCODE_LCONST_0: 
		case Opcodes.OPCODE_LCONST_1: 
			pushWithInst(absInst.getOpcode() - Opcodes.OPCODE_LCONST_0);
			break;
		case Opcodes.OPCODE_LDC: 
			ldcInst();
			break;
		case Opcodes.OPCODE_LDC_W: 
		case Opcodes.OPCODE_LDC2_W:
			//TODO
			pushWithInst("undef");
			break;
		case Opcodes.OPCODE_LDIV: 
			arithInst("/");
			break;
		case Opcodes.OPCODE_LLOAD: 
			pushWithInst(pdsInfo.varIndex(immediateByte()));
			break;
		case Opcodes.OPCODE_LLOAD_0: 
		case Opcodes.OPCODE_LLOAD_1: 
		case Opcodes.OPCODE_LLOAD_2: 
		case Opcodes.OPCODE_LLOAD_3: 
			pushWithInst(pdsInfo.varIndex(
					absInst.getOpcode() - Opcodes.OPCODE_LLOAD_0));
			break;
		case Opcodes.OPCODE_LMUL: 
			arithInst("*");
			break;
		case Opcodes.OPCODE_LNEG:
			warningList.add(Warning.NEGATIVE);
			popPushInst("undef", 1, 1);
			break;
		case Opcodes.OPCODE_LOOKUPSWITCH:
			lookupswitchInst();
			break;
		case Opcodes.OPCODE_LOR: 
			warningList.add(Warning.BITWISE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_LREM: 
			popPushInst(PDSStmt.mod(pdsInfo.stackIndex(1), pdsInfo.stackIndex(0)), 2, 1);
			break;
		case Opcodes.OPCODE_LRETURN: 
			pdsStmt.addReturn(pdsInfo.stackIndex(0));
			break;
		case Opcodes.OPCODE_LSHL: 
			arithInst("<<");
			break;
		case Opcodes.OPCODE_LSHR: 
			warningList.add(Warning.BITWISE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_LSTORE:
			popToInst(pdsInfo.varIndex(immediateByte()));
			break;
		case Opcodes.OPCODE_LSTORE_0: 
		case Opcodes.OPCODE_LSTORE_1: 
		case Opcodes.OPCODE_LSTORE_2: 
		case Opcodes.OPCODE_LSTORE_3: 
			popToInst(pdsInfo.varIndex(
					absInst.getOpcode() - Opcodes.OPCODE_LSTORE_0));
			break;
		case Opcodes.OPCODE_LSUB: 
			arithInst("-");
			break;
		case Opcodes.OPCODE_LUSHR: 
			warningList.add(Warning.BITWISE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_LXOR: 
			warningList.add(Warning.BITWISE);
			popPushInst("undef", 2, 1);
			break;
		case Opcodes.OPCODE_MULTIANEWARRAY: 
			multianewarrayInst();
			break;
		case Opcodes.OPCODE_NEW: 
			newInst();
			break;
		case Opcodes.OPCODE_NEWARRAY: 
			newarrayInst();
			break;
		case Opcodes.OPCODE_NOP: 
			pdsStmt.addSkip();
			break;
		case Opcodes.OPCODE_POP:
			pdsStmt.addNext(popStmt(1, 1));
			break;
		case Opcodes.OPCODE_POP2:
			pdsStmt.addNext(popStmt(2, 2));
			break;
		case Opcodes.OPCODE_PUTFIELD:
			putfieldInst();
			break;
		case Opcodes.OPCODE_PUTSTATIC: 
			putstaticInst();
			break;
		case Opcodes.OPCODE_RET: 
			retInst();
			break;
		case Opcodes.OPCODE_RETURN:
			pdsStmt.addReturn();
			break;
		case Opcodes.OPCODE_SALOAD: 
			ialoadInst();
			break;
		case Opcodes.OPCODE_SASTORE:
			iastoreInst();
			break;
		case Opcodes.OPCODE_SIPUSH: 
			pushWithInst(((ImmediateShortInstruction) absInst).getImmediateShort());
			break;
		case Opcodes.OPCODE_SWAP: 
			pdsStmt.addNext(stackEqual(0, 1));
			pdsStmt.addPar(stackEqual(1, 0));
			break;
		case Opcodes.OPCODE_TABLESWITCH: 
			tableswitchInst();
			break;
		}
	}
	
	/*******************************************************************/
	/**              FUNCTIONS FOR COMMON INSTRUCTIONS                **/
	/*******************************************************************/
	
	/**
	 * Allocates array of size s0+1 in the heap. The first block stores
	 * the length of array (s0).
	 * 
	 * For HeapOption.TWODIMS: unlike newarrayInst() the second block
	 * does not store the bits of element in the heap. Every object 
	 * inside the array always has the default bit size.
	 */
	private void anewarrayInst() {
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
			
			newarrayInst();
			return;
		}
		
		// if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
		
		int intBits = pdsInfo.getBits();
		String s0 = pdsInfo.stackIndex(0);
		String heapPtr = pdsInfo.getHeapPtrName();
		String newHeapPtr = heapPtr + "+" + s0	+ "+1";
		
		List<String> stmtList = new ArrayList<String>();
		stmtList.add(s0 + "=" + heapPtr);
		stmtList.addAll(PDSStmt.storeArray(pdsInfo.getHeapName(), 
				intBits + "*" + heapPtr, s0, intBits));
		stmtList.add(heapPtr + "=" + newHeapPtr);
		
		if (pdsInfo.checkForHeapOverflow()) {
			
			pdsStmt.addIf();
			pdsStmt.addCond(newHeapPtr + ">" + pdsInfo.getStoredHeapSize());
			pdsStmt.addNext(getOverflowCode());
			pdsStmt.addColonElse(stmtList);
			pdsStmt.addFi();
		} else {
			
			pdsStmt.addNext(stmtList);
		}
	}
	
	/**
	 * Pushes heap[s1+s0+1] into the stack.
	 *
	 */
	private void aaloadInst() {
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
			
			ialoadInst();
			return;
		}
		
		// if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
		String heapName = pdsInfo.getHeapName();
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		int bits = pdsInfo.getBits();
		String index = bits + "*("+ s1 + "+" + s0 + "+1)";
		String loadHeap = PDSStmt.loadArray(heapName, index, bits);
		List<String> stmt = popPushStmt(loadHeap, 2, 1);
		
		boolean chkIOOB = pdsInfo.checkForIndexOutOfBounds();
		boolean chkNPE = pdsInfo.checkForNullPointerExceptions();
		
		if (!chkIOOB && ! chkNPE) {
			pdsStmt.addNext(stmt);
			return;
		}
		
		// if (chkIOOB || chkNPE)
		String length = PDSStmt.loadArray(heapName, bits + "*" + s1, bits);
		pdsStmt.addIf();
		if (chkIOOB) pdsStmt.addColonStmt(s0 + ">=" + length, getIOOBCode());
		if (chkNPE) pdsStmt.addColonStmt(s1 + "==0", getNPECode());
		if (chkIOOB) pdsStmt.addCond(s0 + "<" + length);
		if (chkNPE) pdsStmt.addCond(s1 + "!=0");
		pdsStmt.addNext(stmt);
		pdsStmt.addFi();
	}
	
	/**
	 * Stores s0 into heap[s2+s1+1].
	 *
	 */
	private void aastoreInst() {
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
			
			iastoreInst();
			return;
		}
		
		// if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
		String heapName = pdsInfo.getHeapName();
		int bits = pdsInfo.getBits();
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		String s2 = pdsInfo.stackIndex(2);
		String index = bits + "*("+ s2 + "+" + s1 + "+1)";
		List<String> storeHeap = PDSStmt.storeArray(heapName, index, s0, bits);
		List<String> stmt = popPushStmt(storeHeap, 3, 0);
		
		boolean chkIOOB = pdsInfo.checkForIndexOutOfBounds();
		boolean chkNPE = pdsInfo.checkForNullPointerExceptions();
		
		if (!chkIOOB && ! chkNPE) {
			pdsStmt.addNext(stmt);
			return;
		}
		
		// if (chkIOOB || chkNPE)
		String length = PDSStmt.loadArray(heapName, bits + "*" + s2, bits);
		pdsStmt.addIf();
		if (chkIOOB) pdsStmt.addColonStmt(s1 + ">=" + length, getIOOBCode());
		if (chkNPE) pdsStmt.addColonStmt(s2 + "==0", getNPECode());
		if (chkIOOB) pdsStmt.addCond(s1 + "<" + length);
		if (chkNPE) pdsStmt.addCond(s2 + "!=0");
		pdsStmt.addNext(stmt);
		pdsStmt.addFi();
	}
	
	private void arithInst(String op) {
		
		pdsStmt.addNext(stackArith(op));
		pdsStmt.addPar(popStmt(2, 1));
	}
	
	private void arraylengthInst() {

		String s0 = pdsInfo.stackIndex(0);
		String arrLength; 
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE)
			arrLength = pdsInfo.heapIndex(s0);
		else	// if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
			arrLength = PDSStmt.loadArray(pdsInfo.getHeapName(), 
					pdsInfo.getBits() + "*" + s0, pdsInfo.getBits());
		
		if (pdsInfo.checkForNullPointerExceptions()) {
			
			pdsStmt.addIf();
			pdsStmt.addColonStmt(s0 + "==0", getNPECode());
			pdsStmt.addColonElse(popPushStmt(arrLength, 1, 1));
			pdsStmt.addFi();
		} else {
			
			popPushInst(arrLength, 1, 1);
		}
	}
	
	private void getfieldInst() throws InvalidByteCodeException {
		
		String[] fieldName = BytecodeUtils.getReferencedName(
				pdsMethod.getClassFile(), absInst);
		
		if (pdsInfo.isIgnoredField(fieldName)) {
			if (pdsInfo.checkForNullPointerExceptions()
					// Felix added this, I don't know why
					//&& PDSField.isReference(fieldName[2])
					) {
				popPushInst("!=0", 1,1);
			} else {
				popPushInst("undef", 1, 1);
			}
			return;
		}
		
		String s0 = pdsInfo.stackIndex(0);
		ClassNode thisNode = pdsInfo.getInvokeAnalyzer().getClassNode(fieldName[0]);
		if (thisNode == null) {
			popPushInst("undef", 1, 1);
			return; 
		}
		Enumeration e = thisNode.breadthFirstEnumeration();
		ClassNode subNode;
		String[] subName;
		String formattedSubName;
		int intBits = pdsInfo.getBits();
		String objType;
		int fieldIndex;
		int fieldBits;
		String h;
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
			objType = pdsInfo.heapIndex(s0);
		} else { // if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
			objType = PDSStmt.loadArray(pdsInfo.getHeapName(), intBits + "*" + s0, intBits);
		}
		
		pdsStmt.addIf();
		while (e.hasMoreElements()) {
			
			subNode = (ClassNode) e.nextElement();
			subName = replaceClassName(fieldName, subNode.getClassName());
			formattedSubName = PDSField.formatFieldName(subName);
			fieldIndex = pdsInfo.indexOfField(formattedSubName);
			if (fieldIndex >= 0) {
				
				if (pdsInfo.checkForNullPointerExceptions()) { 
					pdsStmt.addCond(s0 + "!=0");
				}
				pdsStmt.addCond(objType + "==" + pdsInfo.getClassIndex(subName[0]));
				
				if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
					
					h = pdsInfo.heapIndex(s0 + "+" + fieldIndex + "+" + 1);
				} else { // if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
					
					fieldBits = pdsInfo.getFieldBits(formattedSubName);
					h = PDSStmt.loadArray(pdsInfo.getHeapName(), intBits + "*" + s0 + "+" + intBits 
							+ "+" + fieldIndex, fieldBits);
				}
				
				pdsStmt.addNext(popPushStmt(h, 1, 1));
			}
		}
		
		if (pdsInfo.checkForNullPointerExceptions()) {
			pdsStmt.addColonStmt(s0 + "==0", getNPECode());
		}
		pdsStmt.addFi();
	}
	
	private void getstaticInst() throws InvalidByteCodeException {
		
		String[] fieldName = BytecodeUtils.getReferencedName(
				pdsMethod.getClassFile(), absInst);
		
		if (//pdsInfo.abortAfterFailedAssertion() && 
				fieldName[1].equals("$assertionsDisabled")) {
			
			pushWithInst("0");
			return;
		}
		
		if (pdsInfo.isIgnoredField(fieldName)) {
			pushWithInst("undef");
			return;
		} 
		
		pushWithInst(PDSField.formatFieldName(fieldName));
	}
	
	/**
	 * Push the value of array s1 at index s0.
	 * 
	 * For HeapOption.SIMPLE: s0 = heap[s1 + s0 + 1].
	 * 
	 * For HeapOption.TWODIMS: two local variables are added, elem and elemBit.
	 * elemBit represents how many bits each element of the array has, e.g.
	 * if the default bits are 3, then elemBit = heap[3*s1+3]<<2 
	 * + heap[3*s1+3+1]<<1 + heap[3*s1+3+2] (Note that the array length is
	 * heap[3*s1]<<2 + heap[3*s1+1] + heap[3*s1+2]). elem is the index 
	 * of the element in the heap, e.g. elem = 3*s1 + (s0*elemBit) + 2*3.
	 * The resulting Remopla statement is actually an if-statement with
	 * conditions on elemBits, e.g.
	 * if
	 *   :: elemBit == 1 -> s0 = heap[elem];
	 *   :: elemBit == 2 -> s0 = heap[elem]<<1 + heap[elem+1];
	 *   :: elemBit == 3 -> s0 = heap[elem]<<2 + heap[elem+1]<<1 + heap[elem+2];
	 * fi;
	 */
	private void ialoadInst() {
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		String length;
		String h;
		boolean chkIOOB = pdsInfo.checkForIndexOutOfBounds();
		boolean chkNPE = pdsInfo.checkForNullPointerExceptions();
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
			
			pdsStmt.addIf();
			if (chkIOOB) {
				length = pdsInfo.heapIndex(s1);
				pdsStmt.addColonStmt(s0 + ">=" + length, getIOOBCode());
			}
			if (chkNPE) {
				pdsStmt.addColonStmt(s1 + "==0", getNPECode());
			}
			
			h = pdsInfo.heapIndex(s1 + "+" + s0 + "+1");
			pdsStmt.addColonElse(popPushStmt(h, 2, 1));
			pdsStmt.addFi();
		} else { // if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
			
			String heapName = pdsInfo.getHeapName();
			int bits = pdsInfo.getBits();
			String arrPtr = String.valueOf(bits)+ "*" + s1;
			length = PDSStmt.loadArray(heapName, arrPtr, bits);
			String elemBit;
			String elemBitVar;
			
			int defaultArrayBits = PDSAnnotationUtils.getDefaultArrayBits(
					pdsMethod.getMethodInfo());
			if (defaultArrayBits != 0) {
				
				// elemBit = default array bits;
				elemBitVar = String.valueOf(defaultArrayBits);
			} else {
				
				// elemBit = load array at (bits*s1 + bits) for length bits;
				elemBit = PDSStmt.loadArray(heapName, arrPtr + "+" + bits, bits);
				elemBitVar = "elemBit";
				pdsMethod.addLocalVar(elemBitVar);
				pdsStmt.addNext(elemBitVar + "=" + elemBit);
			}
			
			// elem = bits*s1 + bits*2 + elemBit*s0;
			String elem = arrPtr + " + " + String.valueOf(bits*2) 
					+ " + " + elemBitVar + "*" + s0;
			String elemVar = "elem";
			pdsMethod.addLocalVar(elemVar, 
					PDSDefault.getMinimumBitsForHeapSize(pdsInfo.getHeapSize()));
			pdsStmt.addNext(elemVar + "=" + elem);
			
			pdsStmt.addIf();
			if (chkIOOB) {
				
				// :: s0 >= length -> goto IOOB;
				pdsStmt.addColonStmt(s0 + ">=" + length, getIOOBCode());
			}
			if (chkNPE) {
				
				// :: s1 == 0 -> goto NPE;
				pdsStmt.addColonStmt(s1 + "==0", getNPECode());
			}
			
			if (defaultArrayBits != 0) {
				
				// :: else -> load from heap at elemVar with defaultArrayBits;
				h = PDSStmt.loadArray(heapName, elemVar, defaultArrayBits);
				if (chkIOOB || chkNPE)
					pdsStmt.addColonElse(popPushStmt(h, 2, 1));
				else
					pdsStmt.addColonStmt("true", popPushStmt(h, 2, 1));
			} else {
				
				for (int i = 1; i <= bits; i++) {
					
					/* :: elemBit == i && [s0 < length]? && [s1 != 0]?
					 * 		-> load from heap at elemVar for bits i;
					 */
					pdsStmt.addCond(elemBitVar + "==" + String.valueOf(i));
					if (chkIOOB) {
						pdsStmt.addCond(s0 + "<" + length);
					}
					if (chkNPE) {
						pdsStmt.addCond(s1 + "!=0");
					}
					h = PDSStmt.loadArray(heapName, elemVar, i);
					pdsStmt.addNext(popPushStmt(h, 2, 1));
				}
				//pdsStmt.addColonElse(popPushStmt("undef", 2, 1));
			}
			
			pdsStmt.addFi();
		}
	}
	
	/**
	 * Stores s0 into array s2 at index s1.
	 * 
	 * For HeapOption.SIMPLE: heap[s2 + s1 + 1] = s0.
	 * 
	 * For HeapOption.TWODIMS: two local variables are added, elem and elemBit.
	 * elemBit represents how many bits each element of the array has, e.g.
	 * if the default bits are 3, then elemBit = heap[3*s2+3]<<2 
	 * + heap[3*s2+3+1]<<1 + heap[3*s2+3+2] (Note that the array length is
	 * heap[3*s2]<<2 + heap[3*s2+1] + heap[3*s2+2]). elem is the index 
	 * of the element in the heap, e.g. elem = 3*s2 + (s1*elemBit) + 2*3.
	 * The resulting Remopla statement is actually an if-statement with
	 * conditions on elemBits, e.g.
	 * if
	 *   :: elemBit == 1 -> heap[elem] = s0;
	 *   :: elemBit == 2 -> heap[elem] = s0>>1, 
	 *   					heap[elem+1]=s0 & 1;
	 *   :: elemBit == 3 -> heap[elem]=s0>>2, 
	 *   					heap[elem+1]=(s0>>1) & 1, 
	 *   					heap[elem+2]=s0 & 1;
	 * fi;
	 */
	private void iastoreInst() {
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		String s2 = pdsInfo.stackIndex(2);
		String length;
		boolean chkIOOB = pdsInfo.checkForIndexOutOfBounds();
		boolean chkNPE = pdsInfo.checkForNullPointerExceptions();
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
			
			pdsStmt.addIf();
			if (chkIOOB) {
				length = pdsInfo.heapIndex(s2);
				pdsStmt.addColonStmt(s1 + ">=" + length, getIOOBCode());
			}
			if (chkNPE) {
				pdsStmt.addColonStmt(s2 + "==0", getNPECode());
			}
			
			String h = pdsInfo.heapIndex(s2 + "+" + s1 + "+1");
			pdsStmt.addColonElse(popPushStmt(h + "=" + s0, 3, 0));
			pdsStmt.addFi();
		} else { // if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
		
			String heapName = pdsInfo.getHeapName();
			int bits = pdsInfo.getBits();
			String arrPtr = String.valueOf(bits)+ "*" + s2;
			length = PDSStmt.loadArray(heapName, arrPtr, bits);
			String elemBit;
			String elemBitVar;
			List<String> h;
			
			int defaultArrayBits = PDSAnnotationUtils.getDefaultArrayBits(
					pdsMethod.getMethodInfo());
			if (defaultArrayBits != 0) {
				
				// elemBit = default array bits;
				elemBitVar = String.valueOf(defaultArrayBits);
			} else {
				
				// elemBit = load array at (bits*s1 + bits) for length bits;
				elemBit = PDSStmt.loadArray(heapName, arrPtr + "+" + bits, bits);
				elemBitVar = "elemBit";
				pdsMethod.addLocalVar(elemBitVar);
				pdsStmt.addNext(elemBitVar + "=" + elemBit);
			}
			
			// elem = bits*s2 + bits*2 + elemBit*s1;
			String elem = arrPtr + " + " + String.valueOf(bits*2) 
					+ " + " + elemBitVar + "*" + s1;
			String elemVar = "elem";
			pdsMethod.addLocalVar(elemVar, 
					PDSDefault.getMinimumBitsForHeapSize(pdsInfo.getHeapSize()));
			pdsStmt.addNext(elemVar + "=" + elem);
			
			pdsStmt.addIf();
			if (chkIOOB) {
				
				// :: s0 >= length -> goto IOOB;
				pdsStmt.addColonStmt(s1 + ">=" + length, getIOOBCode());
			}
			if (chkNPE) {
				
				// :: s1 == 0 -> goto NPE;
				pdsStmt.addColonStmt(s2 + "==0", getNPECode());
			}
			
			if (defaultArrayBits != 0) {
				
				// :: else -> store s0 to heap at elemVar with defaultArrayBits;
				h = PDSStmt.storeArray(heapName, elemVar, s0, defaultArrayBits);
				if (chkIOOB || chkNPE)
					pdsStmt.addColonElse(popPushStmt(h, 3, 0));
				else
					pdsStmt.addColonStmt("true", popPushStmt(h, 3, 0));
			} else {
				
				for (int i = 1; i <= bits; i++) {
					
					/* :: elemBit == i && [s1 < length]? && [s2 != 0]?
					 * 		-> store s0 to heap at elemVar for bits i;
					 */
					pdsStmt.addCond(elemBitVar + "==" + String.valueOf(i));
					if (chkIOOB) {
						pdsStmt.addCond(s1 + "<" + length);
					}
					if (chkNPE) {
						pdsStmt.addCond(s2 + "!=0");
					}
					
					h = PDSStmt.storeArray(heapName, elemVar, s0, i);
					pdsStmt.addNext(popPushStmt(h, 3, 0));
				}
				//pdsStmt.addColonElse(popPushStmt("undef", 3, 0));
			}
			
			pdsStmt.addFi();
		}
	}
	
	private void ifIcmpInst(String op) {
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		
		pdsStmt.addIf();
		pdsStmt.addColonStmt(s1 + op + s0, 
				popPushStmt(PDSStmt.gotoStmt(branchTarget()), 2, 0));
		pdsStmt.addColonElse(popStmt(2, 2));
		pdsStmt.addFi();
	}
	
	private void ifInst(String op) {
		
		String s0 = pdsInfo.stackIndex(0);
		
		pdsStmt.addIf();
		pdsStmt.addColonStmt(s0 + op + "0", 
				popPushStmt(PDSStmt.gotoStmt(branchTarget()), 1, 0));
		pdsStmt.addColonElse(popStmt(1, 1));
		pdsStmt.addFi();
	}
	
	private void iincInst() {
		
		IncrementInstruction incInst = (IncrementInstruction) absInst;
		int index = incInst.getImmediateByte();
		int cnst = incInst.getIncrementConst();
		
		if (cnst <= 127) {
			pdsStmt.addNext(pdsInfo.varIndex(index) + "=" 
					+ pdsInfo.varIndex(index) + "+" + String.valueOf(cnst));
		} else {
			pdsStmt.addNext(pdsInfo.varIndex(index) + "=" 
					+ pdsInfo.varIndex(index) + "-" + String.valueOf(256 - cnst));
		}
	}
	
	private void invokeinterfaceInst() throws InvalidByteCodeException {
		
		String[] calledName = calleeInterfaceName();
		int paramCount = PDSMethod.countParam(calledName[2]) + 1;
		int retCount = PDSMethod.isVoid(calledName[2]) ? 0 : 1;
		//String heapPtr = pdsInfo.stackIndex(paramCount - 1);
		
		Collection<String> impList = pdsInfo.getInvokeAnalyzer().getImplementors(calledName[0]);
		
		for (String impName : impList) {
			String[] names = replaceClassName(calledName, impName);
			invokevirtualInst(pdsMethod.getClassFile(), names, paramCount, 
					retCount);
		}
		
		if (pdsStmt.isEmpty()) {
			pdsStmt.addSkip();
			logger.debug("No methods found for interface call "
					+ Arrays.toString(calledName));
		}
	}
	
	private void invokespecialInst() throws InvalidByteCodeException {
		
		logger.debug("invokespeical instruction");
		String[] calledName = BytecodeUtils.getReferencedName(pdsMethod.getClassFile(), absInst);
		int paramCount = PDSMethod.countParam(calledName[2]) + 1;
		int retCount = PDSMethod.isVoid(calledName[2]) ? 0 : 1;
		
		if (calledName[0].equals("java/lang/AssertionError")
				&& calledName[1].equals("<init>")) {
			
			pdsStmt.addGoto(pdsInfo.getLabelAssertError());
			return;
		}
		
		if (PDSString.isInvokespecialStringBuilder(calledName)) {
			
			pdsStmt = PDSString.invokespecialStringBuilder(
					pdsInfo, pdsMethod, this, calledName);
			if (!pdsStmt.isEmpty()) return;
		}
		
		if (!pdsInfo.getInvokeAnalyzer().isIncludedMethod(calledName)) {
			logger.debug("\tnot included method, paramCount=" + paramCount +
				", retCount=" + retCount);
			if (pdsInfo.checkForNullPointerExceptions()
					&& PDSMethod.returnsReference(calledName[2])) {
				popPushInst("!=0", paramCount, retCount);
			} else {
				popPushInst("undef", paramCount, retCount);
			}
			return;
		}
		
		String calledMethodName = PDSMethod.formatMethodName(calledName);
		String param = paramList(paramCount);
		if (pdsInfo.checkForNullPointerExceptions()) {
			pdsStmt.addIf();
			pdsStmt.addColonStmt(pdsInfo.stackIndex(paramCount - 1) + "==0", 
					getNPECode());
			pdsStmt.addColonElse(popPushStmt(calledMethodName + param, 
					paramCount, retCount));
			pdsStmt.addFi();
		} else {
			popPushInst(calledMethodName + param, paramCount, retCount);
		}
	}
	
	private void invokestaticInst() throws InvalidByteCodeException {
		
		String[] calledName = BytecodeUtils.getReferencedName(pdsMethod.getClassFile(), absInst);
		int paramCount = PDSMethod.countParam(calledName[2]);
		int retCount = PDSMethod.isVoid(calledName[2]) ? 0 : 1;
		
		if (!pdsInfo.getInvokeAnalyzer().isIncludedMethod(calledName)) {
			//if (pdsInfo.checkForNullPointerExceptions()) {
			//	popPushInst("!=0", paramCount, retCount);
			//} else {
				popPushInst("undef", paramCount, retCount);
			//}
			return;
		}
		
		String calledMethodName = PDSMethod.formatMethodName(calledName);
		String param = paramList(paramCount);
		
		popPushInst(calledMethodName + param, paramCount, retCount);
	}
	
	private void invokevirtualInst() throws InvalidByteCodeException {
		
		String[] calledName = BytecodeUtils.getReferencedName(
				pdsMethod.getClassFile(), absInst);
		
		if (PDSString.isInvokevirtualString(calledName)) {
			
			pdsStmt = PDSString.invokevirtualString(pdsInfo, pdsMethod, this, 
					calledName);
			if (!pdsStmt.isEmpty()) return;
		}
		
		int paramCount = PDSMethod.countParam(calledName[2]) + 1;
		int retCount = PDSMethod.isVoid(calledName[2]) ? 0 : 1;
		invokevirtualInst(pdsMethod.getClassFile(), calledName, paramCount, retCount);
		
		if (pdsStmt.isEmpty()) {
			pdsStmt.addSkip();
			logger.debug("No methods found for virtual call "
					+ Arrays.toString(calledName));
		}
	}
	
	private void invokevirtualInst(ClassFile cf, String[] calledName,
			int paramCount, int retCount) throws InvalidByteCodeException {
		
		String heapPtr = pdsInfo.stackIndex(paramCount - 1);
		String objType;
		String param = paramList(paramCount);
		InvokeAnalyzer invokeAnalyzer = pdsInfo.getInvokeAnalyzer();
		ClassNode callNode = invokeAnalyzer.getClassNode(calledName[0]);
		int intBits = pdsInfo.getBits();
		List<String> stmt;
		String cond;
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
			objType = pdsInfo.heapIndex(heapPtr);
		} else {
			objType = PDSStmt.loadArray(pdsInfo.getHeapName(), 
					intBits + "*" + heapPtr, intBits);
		}
		
		// If the called method is ignored
		if (invokeAnalyzer.isIgnoredMethod(calledName) || callNode == null) {
			
			if (callNode == null) logger.debug("callNode is null");
			logger.debug(getLabel() + ": invokevirtual " 
					+ Arrays.toString(calledName) + " ignored");
			if (pdsInfo.checkForNullPointerExceptions() 
					&& PDSMethod.returnsReference(calledName[2])) {
				
				pdsStmt.addIf();
				pdsStmt.addColonStmt(heapPtr + "==0", getNPECode());
				stmt = popPushStmt("undef", paramCount, retCount);
				//stmt = popPushStmt("!=0", paramCount, retCount);
				//if (pdsInfo.isIgnoredClass(calledName[0]) || callNode == null) {
					
					pdsStmt.addColonElse(stmt);
				//} else {
					
				//	cond = objType + "==" + pdsInfo.getClassIndex(calledName[0]);
				//	pdsStmt.addColonStmt(cond, stmt);
				//}
				pdsStmt.addFi();
			} 
			else if (retCount > 0) {
				
				stmt = popPushStmt("undef", paramCount, retCount);
				//if (pdsInfo.isIgnoredClass(calledName[0]) || callNode == null) {
					
					pdsStmt.addNext(stmt);
				//}
				//else {
					
				//	cond = objType + "==" + pdsInfo.getClassIndex(calledName[0]);
				//	pdsStmt.addIf(cond, stmt);
				//}
			}
			return;
		}
		
		// objects this method is called upon must extend the callNode, so
		// determine for each class that extends callNode the most specialized
		// method
		
		pdsStmt.addIf();
		Enumeration children = callNode.breadthFirstEnumeration();
		while (children.hasMoreElements()) {
			ClassNode classNode = (ClassNode)children.nextElement();
			String className = classNode.getClassName();

			// if class is abstract, ignore it
			ClassFile classFile = invokeAnalyzer.getClassFile(className);
			if ((classFile.getAccessFlags() & AccessFlags.ACC_ABSTRACT) != 0) {
				continue;
			}
			
			// go upwards to find most specialized method
			while (classNode != null) {
				String[] methodCall = replaceClassName(calledName,
						classNode.getClassName());
				if (invokeAnalyzer.isIncludedMethod(methodCall)) {
					
					String calledMethodName = PDSMethod.formatMethodName(methodCall);
					cond = objType + "==" + pdsInfo.getClassIndex(className);
					stmt = popPushStmt(calledMethodName + param, paramCount, retCount);
					pdsStmt.addColonStmt(cond, stmt);
					break;
				}
				classNode = (ClassNode)classNode.getParent();
			}
		}
		if (pdsInfo.checkForNullPointerExceptions()) {
			pdsStmt.addColonStmt(heapPtr + "==0", getNPECode());
		}
		pdsStmt.addFi();
		
		return;
	}
	
	private void instanceofInst() throws InvalidByteCodeException {

		ImmediateShortInstruction instruction = (ImmediateShortInstruction) absInst;
		ConstantClassInfo info = (ConstantClassInfo) 
			pdsMethod.getClassFile().getConstantPoolEntry(
					instruction.getImmediateShort(), 
					ConstantClassInfo.class);

		String typeName = info.getName();
		String objType  = pdsInfo.heapIndex(pdsInfo.stackIndex(0));
		
		InvokeAnalyzer invokeAnalyzer = pdsInfo.getInvokeAnalyzer();
		ClassNode classNode = invokeAnalyzer.getClassNode(typeName);
		if (classNode != null) {
			Enumeration children = classNode.breadthFirstEnumeration();
			pdsStmt.addIf();
			if (children.hasMoreElements()) {
				while (children.hasMoreElements()) {
					classNode = (ClassNode) children.nextElement();
					int classIndex = pdsInfo.getClassIndex(classNode.getClassName());
					pdsStmt.addColonStmt(objType + "==" + classIndex, popPushStmt("1", 1, 1));
				}
				pdsStmt.addColonElse(popPushStmt("0", 1, 1));
			}
			pdsStmt.addFi();
		}
		else {
			Collection<String> implementers = invokeAnalyzer.getImplementors(typeName);
			if (!implementers.isEmpty()) {
				pdsStmt.addIf();
				for (String className : implementers) {
					int classIndex = pdsInfo.getClassIndex(className);
					pdsStmt.addColonStmt(objType + "==" + classIndex, popPushStmt("1", 1, 1));
				}
				pdsStmt.addColonElse(popPushStmt("0", 1, 1));
				pdsStmt.addFi();
			}
			else {
				popPushInst("0", 1, 1);
			}
		}
	}
	
	private void lcmpInst() {
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		
		pdsStmt.addIf();
		pdsStmt.addColonStmt(s1 + ">" + s0, popPushStmt("1", 2, 1));
		pdsStmt.addColonStmt(s1 + "==" + s0, popPushStmt("0", 2, 1));
		pdsStmt.addColonElse(popPushStmt("undef", 2, 1));
		pdsStmt.addFi();
	}
	
	private void ldcInst() throws InvalidByteCodeException {
		
		ClassFile cf = pdsMethod.getClassFile();
		CPInfo[] cpInfo = cf.getConstantPool();
		int immByte = immediateByte();
		int cpTag = cpInfo[immByte].getTag();
		
		if (cpTag == CPInfo.CONSTANT_INTEGER) {
			
			ConstantIntegerInfo ciInfo = (ConstantIntegerInfo) cpInfo[immByte];
			int value = ciInfo.getInt();
			if (value == Integer.MAX_VALUE) {
				value = PDSDefault.getMaxIntForBits(pdsInfo.getBits());
			}
			else if (value == Integer.MIN_VALUE) {
				value = 0;
			}
			pushWithInst(value);
			return;
		} else if (cpTag == CPInfo.CONSTANT_STRING) {
			
			pdsStmt =  PDSString.ldcString(cpInfo, immByte, pdsInfo, this);
			return;
		} else { // if (cpTag == CPInfo.CONSTANT_FLOAT)
			
			warningList.add(Warning.FLOAT);
			pushWithInst("undef");
		}
	}
	
	private void lookupswitchInst() {
		
		LookupSwitchInstruction lsInst = (LookupSwitchInstruction) absInst;
		List moPairs = lsInst.getMatchOffsetPairs();
		int offset = lsInst.getOffset();
		int match;
		int target;
		String methodName = pdsMethod.getFormattedName();
		
		pdsStmt.addIf();
		for (int i =  0; i < moPairs.size(); i++) {
			
			match = ((MatchOffsetPair) moPairs.get(i)).getMatch();
			target = offset + ((MatchOffsetPair) moPairs.get(i)).getOffset();
			pdsStmt.addColonStmt(pdsInfo.stackIndex(0) + "==" + match, 
					popPushStmt(PDSStmt.gotoStmt(methodName + target), 1, 0));
		}
		target = offset + lsInst.getDefaultOffset();
		pdsStmt.addColonElse(popPushStmt(PDSStmt.gotoStmt(methodName + target), 1, 0));
		pdsStmt.addFi();
	}
	
	/**
	 * @throws InvalidByteCodeException if there are problems interpreting the 
	 * bytecode or if the instruction creates an array with more than 4 dimensions,
	 * which is not handled yet.
	 */
	private void multianewarrayInst() throws InvalidByteCodeException {
		
		int dim = ((MultianewarrayInstruction) absInst).getDimensions();
		
		if (dim > PDSMultiArrayInitializerMethod.MAX_DIMENSIONS) {
			throw new InvalidByteCodeException("Multidmensional arrays with more than "
					+ PDSMultiArrayInitializerMethod.MAX_DIMENSIONS 
					+ "  dimensions are not supported.");
		}
		
		String heapPtr = pdsInfo.getHeapPtrName();
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
			
			String newHeapPtr = heapPtr + "+" + createMultiArraySizeExpression(dim);
			String heapUpdate = heapPtr + "=" + newHeapPtr;
	
			// add heap initialization code here
			StringBuilder content = new StringBuilder();
			content.append(createMultiArrayInitializerCall(dim));
			content.append(";\n\t   ");
			List<String> push = popPushStmt(heapPtr, 0, 1);
			push.add(heapUpdate);
			content.append(PDSStmt.commaList(push));
		
			if (pdsInfo.checkForHeapOverflow()) {
				pdsStmt.addIf();
				pdsStmt.addColonStmt(newHeapPtr + ">" + pdsInfo.getStoredHeapSize(), 
						getOverflowCode());
				pdsStmt.addColonElse(content.toString());
				pdsStmt.addFi();
			} 
			else {
				pdsStmt.addNext(content.toString());
			}
			
			pdsInfo.setIncludeMultiArrayInitializerMethod();
			
			return;
		} 
		
		// if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
			
		int intBits = pdsInfo.getBits();
		String s0 = pdsInfo.stackIndex(0);
		int arrBits = pdsMethod.getBitsAtPc(absInst.getOffset());
		if (arrBits == 0) {
			arrBits = PDSAnnotationUtils.getDefaultArrayBits(
					pdsMethod.getMethodInfo());
			if (arrBits == 0) arrBits = pdsInfo.getBits();
		}
		String s0TimesArrBits = "(" + s0 + "*" + arrBits + ")";
		String arrLength1 = s0TimesArrBits + "/" + intBits + "+2";
		String s0ModIntBits = PDSStmt.mod(s0TimesArrBits, intBits);
		String cond1 = s0ModIntBits + "==0";
		
		/*
		 * if
		 *   :: (s0*arrBit) mod intBit == 0 -> 
		 *   			arrLength = (s0*arrBits)/intBit+2;
		 *   :: else -> arrLength = (s0*arrBits)/intBit+3;
		 * fi;
		 */
		String arrLengthVar = "arrLength";
		pdsMethod.addLocalVar(arrLengthVar);
		pdsStmt.addIf();
		pdsStmt.addColonStmt(cond1, arrLengthVar + "=" + arrLength1);
		pdsStmt.addColonElse(arrLengthVar + "=" + arrLength1 + "+1");
		pdsStmt.addFi();
		
		// Adds local variables
		String counterVar = "counter";
		pdsMethod.addLocalVar(counterVar);
		String counterVar2 = "counter2";
		pdsMethod.addLocalVar(counterVar2);
		String indexVar = "index";
		pdsMethod.addLocalVar(indexVar);
		String lastPtrVar = "lastPtr";
		pdsMethod.addLocalVar(lastPtrVar);
		
		// counter = 0, lastPtr = heap_ptr;
		pdsStmt.addNext(counterVar + "=0");
		pdsStmt.addPar(lastPtrVar + "=" + heapPtr);
		
		/*
		 * do
		 *   :: counter < s{dim-1}*...*s1 -> 
		 *   		stores s0 to heap at intBit*ptr,
		 *   		stores arrBit to heap at intBit*(ptr+1),
		 *   		ptr = ptr+arrLength,
		 *   		counter = counter + 1;
		 *   :: else -> skip;
		 * od;
		 */
		pdsStmt.addDo();
		StringBuilder s = new StringBuilder();
		for (int i = dim-1; i>0; i--) {
			s.append(pdsInfo.stackIndex(i));
			if (i > 1) s.append("*");
		}
		pdsStmt.addCond(counterVar + "<" + s);
		pdsStmt.addNext(PDSStmt.commaList(PDSStmt.storeArray(
				pdsInfo.getHeapName(), 
				intBits + "*" + heapPtr, s0, intBits)));
		pdsStmt.addPar(PDSStmt.commaList(PDSStmt.storeArray(
				pdsInfo.getHeapName(), 
				intBits + "*(" + heapPtr + "+1)", arrBits, intBits)));
		pdsStmt.addPar(heapPtr + "=" + heapPtr + "+" + arrLengthVar);
		pdsStmt.addPar(counterVar + "=" + counterVar + "+1");
		pdsStmt.addColonElseBreak();
		pdsStmt.addOd();
		
		// For i = 1 to dim-1
		for (int i = 1; i < dim; i++) {
			
			String si = pdsInfo.stackIndex(i);
			s = new StringBuilder();
			for (int j = dim-1; j>i; j--) {
				s.append(pdsInfo.stackIndex(j));
				if (j > i + 1) s.append("*");
			}
			
			// counter = 0, index = lastPtr, lastPtr = ptr;
			pdsStmt.addNext(counterVar + "=0");
			pdsStmt.addPar(indexVar + "=" + lastPtrVar);
			pdsStmt.addPar(lastPtrVar + "=" + heapPtr);
			
			/*
			 * do
			 *   :: counter < s{dim-1}*...*s{i+1} -> 
			 *     stores s{i} to heap at intBit*ptr, counter2 = 0;
			 *     do
			 *     :: counter2 < si -> 
			 *     			stores index to heap at intBit*(ptr+counter2+1),
			 *     			index = index + arrLength,
			 *     			counter2 = counter2 + 1;
			 *     :: else -> break;
			 *     od;
			 *     ptr = ptr+s{i}+1, counter = counter+1;
			 *   :: else ->break;
			 * od;
			 */
			pdsStmt.addDo();
			pdsStmt.addCond(counterVar + "<" + ((i < dim-1) ? s : "1"));
			pdsStmt.addNext(PDSStmt.commaList(
					PDSStmt.storeArray(pdsInfo.getHeapName(), 
							intBits + "*" + heapPtr, si, intBits)));
			pdsStmt.addPar(counterVar2 + "=0");
				pdsStmt.incNestedLevel();
				pdsStmt.addDo();
				pdsStmt.addCond(counterVar2 + "<" + si);
				pdsStmt.addNext(PDSStmt.commaList(
						PDSStmt.storeArray(pdsInfo.getHeapName(), 
								intBits + "*("+ heapPtr + "+" + counterVar2 + "+1)",
								indexVar, intBits)));
				pdsStmt.addPar(indexVar + "=" + indexVar + "+" + arrLengthVar);
				pdsStmt.addPar(counterVar2 + "=" + counterVar2 + "+1");
				pdsStmt.addColonElseBreak();
				pdsStmt.addOd();
				pdsStmt.decNestedLevel();
			pdsStmt.addNext(heapPtr + "=" + heapPtr + "+" + si + "+1");
			pdsStmt.addPar(counterVar + "=" + counterVar + "+1");
			pdsStmt.addColonElseBreak();
			pdsStmt.addOd();
			
			// arrLength = s{i} + 1;
			if (i < dim -1)
				pdsStmt.addNext(arrLengthVar + "=" + si + "+1");
		}
		
		pdsStmt.addNext(popPushStmt(lastPtrVar, dim, 1));
	}
	
	private String createMultiArraySizeExpression(int dim)
	{
		StringBuilder size = new StringBuilder();
		
		for (int i = dim - 1; i >= 0; i--) {
			for (int j = dim - 1; j > i; j--) {
				size.append(pdsInfo.stackIndex(j));
				size.append(" * ");
			}
			size.append("(").append(pdsInfo.stackIndex(i)).append(" + 1)");
			if (i > 0) {
				size.append(" + ");
			}
		}
		return size.toString();
	}
	
	private String createMultiArrayInitializerCall(int dim)
	{
		StringBuilder code = new StringBuilder(PDSMultiArrayInitializerMethod.NAME);
		code.append("(").append(pdsInfo.getHeapPtrName()).append(", ");
		for (int i = dim - 1; i >= 0; i--) {
			code.append(pdsInfo.stackIndex(i));
			if (i > 0) {
				code.append(", ");
			}
		}
		final int maxDim = PDSMultiArrayInitializerMethod.MAX_DIMENSIONS;
		if (dim < maxDim) {
			code.append(", ");
			for (int i = maxDim - dim - 1; i >= 0;  i--) {
				code.append("0");
				if (i > 0) {
					code.append(", ");
				}
			}
		}
		code.append(")");
		return code.toString();
	}
	
	private void newInst() throws InvalidByteCodeException {
		
		String className = getNewedClassName(pdsMethod.getClassFile(), absInst);
		
		if (className.equals("java/lang/StringBuilder")) {
			
			int sbSize = pdsInfo.getStringBuilderSize();
			String heapPtr = pdsInfo.getHeapPtrName();
			String s0 = pdsInfo.stackIndex(0);
			
			if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
				
				String newHeapPtr = heapPtr + "+" + sbSize + "+1";
				
				List<String> content = new ArrayList<String>();
				content.add(s0 + "=" + heapPtr);
				content.add(pdsInfo.heapIndex(heapPtr) + "=0");
				content.add(heapPtr + "=" + newHeapPtr);
				
				if (pdsInfo.checkForHeapOverflow()) {

					pdsStmt.addIf();
					pdsStmt.addColonStmt(newHeapPtr + ">" + pdsInfo.getHeapSize(),
							getOverflowCode());
					pdsStmt.addColonElse(content);
					pdsStmt.addFi();
				} else {
					
					pdsStmt.addNext(content);
				}
			} else { // if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
				
				int intBits = pdsInfo.getBits();
				int charBits = pdsInfo.getCharBits();
				int blocks = countBlocks(sbSize * charBits, intBits);
				String newHeapPtr = heapPtr + "+" + blocks + "+1";
				
				List<String> stmtList = new ArrayList<String>();
				stmtList.add(s0 + "=" + heapPtr);
				stmtList.addAll(PDSStmt.storeArray(pdsInfo.getHeapName(), 
						intBits + "*" + heapPtr, 0, intBits));
				stmtList.add(heapPtr + "=" + newHeapPtr);
				
				if (pdsInfo.checkForHeapOverflow()) {
					
					pdsStmt.addIf();
					String cond = newHeapPtr + ">" + pdsInfo.getStoredHeapSize();
					pdsStmt.addColonStmt(cond, getOverflowCode());
					pdsStmt.addColonElse(stmtList);
					pdsStmt.addFi();
				} else {
					
					pdsStmt.addNext(stmtList);
				}
			}
			return;
		}
		
		if (pdsInfo.getInvokeAnalyzer().isIncludedClass(className)) {
			
			String heapPtr = pdsInfo.getHeapPtrName();
			int classSize = pdsInfo.getClassSize(className);
			String newHeapPtr = heapPtr + "+" + classSize + "+1";
			String heapUpdate = heapPtr + "=" + newHeapPtr;
			int classIndex = pdsInfo.getClassIndex(className);
			
			List<String> content = popPushStmt(heapPtr, 0, 1);
			if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
				
				content.add(pdsInfo.heapIndex(heapPtr) + "=" + classIndex);
			} else { // if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
				
				int intBits = pdsInfo.getBits();
				content.addAll(PDSStmt.storeArray(pdsInfo.getHeapName(), 
						intBits + "*" + heapPtr, classIndex, intBits));
			}
			content.add(heapUpdate);
			
			if (pdsInfo.checkForHeapOverflow()) {
				
				pdsStmt.addIf();
				pdsStmt.addColonStmt(newHeapPtr + ">" + pdsInfo.getStoredHeapSize(), 
						getOverflowCode());
				pdsStmt.addColonElse(content);
				pdsStmt.addFi();
			} else {
				
				pdsStmt.addNext(content);
			}
		} else {
			
			pushWithInst("undef");
		}
	}
	
	/**
	 * For HeapOption.SIMPLE: Allocates (s0 + 1) blocks of memory in the heap.
	 * The first block stores the length of array.
	 * 
	 * For HeapOption.TWODIMS: Allocates (ceil((s0*arrBits)/intBits) + 2) 
	 * blocks of memory in the heap. The first block stores the length of array,
	 * whereas the second block tells how many bits each element of the array
	 * has. We simulate ceil function by using if-statement by the following
	 * way:
	 * <pre>
	 * if
	 *   :: (s0*arrBits)%intBits == 0 -> allocate (s0*arrBits)/intBits + 2
	 *   :: (s0*arrBits)%intBits != 0 -> allocate (s0*arrBits)/intBits + 3
	 * fi;
	 * </pre>
	 */
	private void newarrayInst() {
		
		String s0 = pdsInfo.stackIndex(0);
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
			
			String heapPtr = pdsInfo.getHeapPtrName();
			String newHeapPtr = heapPtr + "+" + s0 + "+1";
			
			List<String> content = new ArrayList<String>();
			content.add(s0 + "=" + heapPtr);
			content.add(pdsInfo.heapIndex(heapPtr) + "=" + s0);
			content.add(heapPtr + "=" + newHeapPtr);
			
			if (pdsInfo.checkForHeapOverflow()) {

				pdsStmt.addIf();
				pdsStmt.addColonStmt(newHeapPtr + ">" + pdsInfo.getHeapSize(),
						getOverflowCode());
				pdsStmt.addColonElse(content);
				pdsStmt.addFi();
			} else {
				
				pdsStmt.addNext(content);
			}
		} else { // if (pdsInfo.getHeapOption() == HeapOption.TWODIMS)
		
			int arrBits;
			if (absInst.getOpcode() == Opcodes.OPCODE_NEWARRAY
					&& immediateByte() == 4) {	// For boolean
				arrBits = 1;
			} else {
				arrBits = pdsMethod.getBitsAtPc(absInst.getOffset());
				if (arrBits == 0) {
					arrBits = PDSAnnotationUtils.getDefaultArrayBits(
							pdsMethod.getMethodInfo());
					if (arrBits == 0) arrBits = pdsInfo.getBits();
				}
			}
			
			PDSStmt.createArray(pdsStmt, pdsInfo, s0, arrBits, getOverflowCode());
		}
	}
	
	//private void oneInst(String content) {
		
	//	pdsStmt.addPar(content);
	//}
	
	private void popToInst(String var) {
		
		pdsStmt.addNext(popToStmt(var));
	}
	
	private void popPushInst(String content, int pop, int push) {
		
		pdsStmt.addNext(popPushStmt(content, pop, push));
	}
	
	private void pushWithInst(int content) {
		
		pushWithInst(String.valueOf(content));
	}
	
	private void pushWithInst(String content) {
		
		pdsStmt.addNext(popPushStmt(content, 0, 1));
	}
	
	private void putfieldInst() throws InvalidByteCodeException {
		
		String[] fieldName = BytecodeUtils.getReferencedName(pdsMethod.getClassFile(), absInst);
		
		if (pdsInfo.isIgnoredField(fieldName)) {
			popPushInst("undef", 2, 0);
			return;
		}
		
		String s0 = pdsInfo.stackIndex(0);
		String s1 = pdsInfo.stackIndex(1);
		ClassNode thisNode = pdsInfo.getInvokeAnalyzer().getClassNode(fieldName[0]);
		Enumeration e = thisNode.breadthFirstEnumeration();
		DefaultMutableTreeNode subNode;
		String[] subName;
		String formattedSubName;
		String objType;
		int fieldIndex;
		String h;
		List<String> cond, stmt;
		int intBits = pdsInfo.getBits();
		
		if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
			objType = pdsInfo.heapIndex(s1);
		} else {	// pdsInfo.getHeapOption() == HeapOption.TWODIMS
			objType = PDSStmt.loadArray(pdsInfo.getHeapName(), intBits + "*" + s1, intBits);
		}
		
		pdsStmt.addIf();
		while (e.hasMoreElements()) {
			
			subNode = (DefaultMutableTreeNode) e.nextElement();
			subName = replaceClassName(fieldName, 
					(String) subNode.getUserObject());
			formattedSubName = PDSField.formatFieldName(subName);
			fieldIndex = pdsInfo.indexOfField(formattedSubName);
			if (fieldIndex >= 0) {
				
				cond = new ArrayList<String>();
				cond.add(objType + "==" + pdsInfo.getClassIndex(subName[0]));
				if (pdsInfo.checkForNullPointerExceptions()) {
					cond.add(s1 + "!=0");
				}
				
				if (pdsInfo.getHeapOption() == HeapOption.SIMPLE) {
					
					h = pdsInfo.heapIndex(s1 + "+" + fieldIndex + "+" + 1);
					stmt = new ArrayList<String>();
					stmt.add(h + "=" + s0);
				} else {	// pdsInfo.getHeapOption() == HeapOption.TWODIMS
					
					int fieldBits = pdsInfo.getFieldBits(formattedSubName);
					stmt = PDSStmt.storeArray(pdsInfo.getHeapName(), 
							intBits + "*" + s1 + "+" + intBits 
							+ "+" + fieldIndex, s0, fieldBits);
				}
				
				pdsStmt.addColonStmt(cond, popPushStmt(stmt, 2, 0));
			}
		}
		
		if (pdsInfo.checkForNullPointerExceptions()) {
			pdsStmt.addColonStmt(s1 + "==0", getNPECode());
		}
		pdsStmt.addFi();
	}
	
	private void putstaticInst() throws InvalidByteCodeException {
		
		String[] fieldName = BytecodeUtils.getReferencedName(
				pdsMethod.getClassFile(), absInst);
		if (pdsInfo.isIgnoredField(fieldName)) {
			popPushInst("undef", 1, 0);
		} else {
			popToInst(PDSField.formatFieldName(fieldName));
		}
	}
	
	private void retInst() {
		
//		type = IF_INST;
		
		/*TODO*/
	}
	
	private void tableswitchInst() {
		
		TableSwitchInstruction tsInst = (TableSwitchInstruction) absInst;
		int offset = tsInst.getOffset();
		int[] jumpOffsets = tsInst.getJumpOffsets();
		int lowByte = tsInst.getLowByte();
		int highByte = tsInst.getHighByte();
		int target;
		String s0 = pdsInfo.stackIndex(0);
		String methodName = pdsMethod.getFormattedName();
		String cond;
		List<String> stmt;
		
		pdsStmt.addIf();
		for (int i = 0; i <= (highByte - lowByte); i++) {
			
			target = offset + jumpOffsets[i];
			cond = s0 + "==" + String.valueOf(lowByte + i);
			stmt = popPushStmt(PDSStmt.gotoStmt(methodName + target), 1, 0);
			pdsStmt.addColonStmt(cond, stmt);
		}
		target = offset + tsInst.getDefaultOffset();
		stmt = popPushStmt(PDSStmt.gotoStmt(methodName + target), 1, 0);
		pdsStmt.addColonElse(stmt);
		pdsStmt.addFi();
	}
	
	/*******************************************************************/
	/**                EXTERNAL UTILITIY FUNCTIONS                    **/
	/*******************************************************************/
	
	/**
	 * Returns name of class in <code>new</code> instruction.
	 *
	 * @param classFile the jclasslib class file.
	 * @param inst the <code>new</code> instruction.
	 * @return the class name.
	 */
	public static String getNewedClassName(ClassFile cFile, 
			AbstractInstruction aInst) throws InvalidByteCodeException {
		
		ImmediateShortInstruction isInst = (ImmediateShortInstruction) aInst;
		ConstantClassInfo ccInfo = 
			(ConstantClassInfo) cFile.getConstantPoolEntry(
				isInst.getImmediateShort(), ConstantClassInfo.class);
				
		return ccInfo.getName();
	}
	
	/*******************************************************************/
	/**                INTERNAL UTILITIY FUNCTIONS                    **/
	/*******************************************************************/
	
	private String branchTarget() {
		
		return BytecodeUtils.branchTarget(pdsMethod.getFormattedName(), absInst);
	}
	
	private String[] calleeInterfaceName() throws InvalidByteCodeException {
		
		InvokeInterfaceInstruction iiInst = (InvokeInterfaceInstruction) absInst;
		ConstantInterfaceMethodrefInfo cimInfo = 
				(ConstantInterfaceMethodrefInfo) 
				pdsMethod.getClassFile().getConstantPoolEntry(
				iiInst.getImmediateShort(), ConstantInterfaceMethodrefInfo.class);
		ConstantClassInfo ccInfo = cimInfo.getClassInfo();
		ConstantNameAndTypeInfo cnatInfo = cimInfo.getNameAndTypeInfo();
		
		String[] out = new String[3];
		out[0] = ccInfo.getName();
		out[1] = cnatInfo.getName();
		out[2] = cnatInfo.getDescriptor();
		
		return out;
	}
	
	public static int countBlocks(int totalBits, int intBits) {
		
		return (int) Math.ceil(((double) totalBits) / ((double) intBits));
	}
	
	/**
	 * Returns statements that handle the case of an exception. 
	 * @param targetLabel the label the code jumps to and which is used as a prefix
	 * @return
	 */
	public String getExceptionCode(String targetLabel)
	{
		//StringBuilder overflow = new StringBuilder("skip;\n\t   ");
		StringBuilder overflow = new StringBuilder();
		overflow.append(targetLabel).append(getLabel());
		overflow.append(": ");
		overflow.append(PDSStmt.gotoStmt(targetLabel));
		return overflow.toString();
	}
	
	/**
	 * Returns statements that handle the case of a Remopla StringBuilder overflow.
	 * @return
	 */
	public String getStringBuilderOverflowCode()
	{
		return getExceptionCode(pdsInfo.getLabelStringBuilderOverflow());
	}
	
	/**
	 * Returns statements that handle the case of a Remopla heap overflow.
	 * @return
	 */
	public String getOverflowCode()
	{
		return getExceptionCode(pdsInfo.getLabelHeapOverflow());
	}
	
	/**
	 * Returns statements that handle the case of referencing a null pointer. 
	 * @return
	 */
	public String getNPECode()
	{
		return getExceptionCode(pdsInfo.getLabelNPE());
	}
	
	/**
	 * Returns statements that handle the case of accessing an array beyond its size.
	 * @return
	 */
	public String getIOOBCode()
	{
		return getExceptionCode(pdsInfo.getLabelIndexOutOfBounds());
	}
	
	public int immediateByte() {
		
		return ((ImmediateByteInstruction) absInst).getImmediateByte();
	}
	
	private String paramList(int paramCount) {
		
		if (paramCount == 0) {
			return "()";
		}
		
		StringBuffer param = new StringBuffer();
		
		param.append("(");
		param.append(pdsInfo.stackIndex(paramCount - 1));
		for (int i = paramCount - 2; i >= 0; i--) {
			param.append(", ");
			param.append(pdsInfo.stackIndex(i));
		}
		param.append(")");
		
		return param.toString();
	}
	
	public List<String> popToStmt(String var) {
		
		return popPushStmt(var + "=" + pdsInfo.stackIndex(0), 1, 0);
	}
	
	public List<String> popStmt(int from, int step) {
		
		ArrayList<String> list = new ArrayList<String>();
		
		for (int i = from; i < pdsMethod.getMaxStack(); i++) {
			list.add(stackEqual(i - step, i));
		}
		
		return list;
	}
	
	public List<String> popPushStmt(String action, int pop, int push) {
		
		ArrayList<String> actions = new ArrayList<String>();
		actions.add(action);
		
		return popPushStmt(actions, pop, push);
	}
	
	public List<String> popPushStmt(List<String> actions, int pop, int push) {
		
		ArrayList<String> list = new ArrayList<String>();
		
		if ((pop == 0) && (push == 0)) {
			if (actions.get(0).equals("undef")
					|| actions.get(0).equals("!=0")) 
				list.add("skip");
			else
				list.addAll(actions);
		} else if ((pop == 0) && (push == 1)) {
			if (actions.get(0).equals("!=0"))
				list.add(pdsInfo.stackIndex(0) + "!=0");
			else
				list.add(pdsInfo.stackIndex(0) + "=" + actions.get(0));
			list.addAll(pushStmt(1, 1));
		} else if ((pop == 1) && (push == 1)) {
			if (actions.get(0).equals("!=0"))
				list.add(pdsInfo.stackIndex(0) + "!=0");
			else
				list.add(pdsInfo.stackIndex(0) + "=" + actions.get(0));
		} else if (push == 0) {
			if (!actions.get(0).equals("undef")
					&& !actions.get(0).equals("!=0"))
				list.addAll(actions);
			list.addAll(popStmt(pop, pop));
		} else {	// if (push == 1)
			if (actions.get(0).equals("!=0"))
				list.add(pdsInfo.stackIndex(0) + "!=0");
			else
				list.add(pdsInfo.stackIndex(0) + "=" + actions.get(0));
			list.addAll(popStmt(pop, pop - 1));
		}
		
		return list;
	}
	
	public List<String> pushStmt(int from, int step) {
		
		ArrayList<String> list = new ArrayList<String>();
		for (int i = from; i < pdsMethod.getMaxStack(); i++) {
			list.add(stackEqual(i, i - step));
		}
		
		return list;
	}
	
	/**
	 * Replaces the first element of array names with name
	 * 
	 * @param names
	 * @param name
	 * @return
	 */
	private static String[] replaceClassName(String[] names, String name) {
		
		String[] newNames = new String[names.length];
		newNames[0] = name;
		newNames[1] = names[1];
		if (names.length == 3) {
			newNames[2] = names[2];
		}
		
		return newNames;
	}
	
	/**
	 * Constructs a string representing:
	 * 		s0 = s1 op s0
	 * 
	 * @param op
	 * @return
	 */
	public String stackArith(String op) {
		
		StringBuilder out = new StringBuilder();
		
		out.append(pdsInfo.stackIndex(0));
		out.append("=");
		out.append(pdsInfo.stackIndex(1));
		out.append(op);
		out.append(pdsInfo.stackIndex(0));
		
		return out.toString();
	}
	
	/**
	 * Construct a string representing:
	 * 		pdsInfo.stackIndex(left) = pdsInfo.stackIndex(right)
	 * 
	 * @param left
	 * @param right
	 * @return
	 */
	public String stackEqual(int left, int right) {
		
		StringBuilder out = new StringBuilder();
		
		out.append(pdsInfo.stackIndex(left));
		out.append("=");
		out.append(pdsInfo.stackIndex(right));
		
		return out.toString();
	}
}
