package de.uni_stuttgart.fmi.szs.jmoped;

import java.util.List;
import java.util.Random;

import org.apache.log4j.Logger;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.attributes.LocalVariableTableEntry;

/**
 * Local variable information for pushdown system.
 * 
 * @author suwimodh
 *
 */
public class PDSLocalVar {
	
	LocalVariableTableEntry lvtEntry;
	String name;
	private int bits;			// zero means that it has the default value.
	private boolean external;	// true, if the variable is externally added.
	
	static Logger logger = Logger.getLogger(PDSLocalVar.class);
	
	public PDSLocalVar(String name) {
		
		this(name, 0);
	}
	
	public PDSLocalVar(String name, int bits) {
		
		this(null, name, bits, false);
	}
	
	public PDSLocalVar(String name, int bits, boolean external) {
		
		this(null, name, bits, external);
	}
	
	public PDSLocalVar(LocalVariableTableEntry lvtEntry, String name, int bits) {
		
		this(lvtEntry, name, bits, false);
	}
	
	public PDSLocalVar(LocalVariableTableEntry lvtEntry, String name, int bits, 
			boolean external) {
		
		this.lvtEntry = lvtEntry;
		this.name = name;
		this.bits = bits;
		this.external = external;
	}
	
	public void setLocalVariableTableEntry(LocalVariableTableEntry lvtEntry) {
		
		this.lvtEntry = lvtEntry;
	}
	
	/**
	 * Gets the name of this variable in the source if its
	 * <code>LocalVariableTableEntry</code> exists.
	 * Otherwise, {@link PDSInfo#varIndex(int) the default name} is returned.
	 * 
	 * @return the source name of this variable.
	 */
	public String getSourceName() {
		
		if (lvtEntry == null) return name;
		
		int index = lvtEntry.getNameIndex();
		ClassFile classFile = lvtEntry.getClassFile();
		try {
			return classFile.getConstantPoolUtf8Entry(index).getString();
		} catch (InvalidByteCodeException e) {
			return name;
		}
	}
	
	public void setName (String name) {
		
		this.name = name;
	}
	
	public String getName() {
		
		return name;
	}
	
	public void setBits(int bits) {
		
		this.bits = bits;
	}
	
	public int getBits() {
		
		return bits;
	}
	
	public int getBits(int defaultBits) {
		
		return (bits != 0) ? bits : defaultBits;
	}
	
	public int getDescriptorIndex() {
		
		return lvtEntry.getDescriptorIndex();
	}
	
	public int getIndex() {
		
		return lvtEntry.getIndex();
	}
	
	public int getLength() {
		
		return lvtEntry.getLength();
	}
	
	public int getStartPc() {
		
		return lvtEntry.getStartPc();
	}
	
	/**
	 * Returns <code>true</code> if the variable was added externally,
	 * i.e. not the one directly translated from bytecoe.
	 * 
	 * @return <code>true</code> if the variable was externally added.
	 */
	public boolean isExternal() {
		
		return external;
	}
	
	/**
	 * Returns the Remopla code declaring this local variable.
	 * 
	 * @return the Remopla code.
	 */
	public String toRemopla() {
		
		return toRemopla(getName(), getBits());
	}
	
	/**
	 * Returns the Remopla declaring the variable var with size bits.
	 * 
	 * @param var the variable name.
	 * @param bits the number of bits.
	 * @return the Remopla code.
	 */
	public static String toRemopla(String var, int bits) {
		
		StringBuilder out = new StringBuilder();
		
		out.append("int ");
		out.append(var);
		if (bits != 0) {
			out.append("(");
			out.append(bits);
			out.append(")");
		}
		
		return out.toString();
	}
	
	/**
	 * Returns candidate if it is different from all names in list, otherwise
	 * the method appends a random number to candidate until it is unique.
	 * 
	 * @param list the list of variable names.
	 * @param candidate the candidate.
	 * @return the unique name.
	 */
	public static String getUniqueVarName(List<String> list, String candidate) {
		
		if (list == null) return candidate;
		
		if (!list.contains(candidate)) return candidate;
		
		Random r = new Random();
		return getUniqueVarName(list, candidate + r.nextInt(Integer.MAX_VALUE));
	}
}
