/*
    Copyright (C) 2005 Dejvuth Suwimonteerabuth

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;
import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;
import org.gjt.jclasslib.io.ByteCodeReader;
import org.gjt.jclasslib.structures.AccessFlags;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.MethodInfo;
import org.gjt.jclasslib.structures.attributes.CodeAttribute;
import org.gjt.jclasslib.structures.attributes.LineNumberTableAttribute;
import org.gjt.jclasslib.structures.attributes.LineNumberTableEntry;
import org.gjt.jclasslib.structures.attributes.LocalVariableTableAttribute;
import org.gjt.jclasslib.structures.attributes.LocalVariableTableEntry;
import org.gjt.jclasslib.structures.constants.ConstantUtf8Info;
import org.gjt.jclasslib.structures.elementvalues.ElementValue;

import de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSAnnotationUtils;

/**
 * Method information for pushdown system.
 */
public class PDSMethod {
	
	static Logger logger = Logger.getLogger(PDSMethod.class);
	
	protected PDSClass pdsClass;
	protected ClassFile classFile;
	protected MethodInfo methodInfo;
	protected PDSInfo pdsInfo;
	private CodeAttribute codeAttr;
	private List<AbstractInstruction> instList = Collections.emptyList();
	protected String methodName;
	protected List<PDSLocalVar> pdsLocalVarList;
	protected List<PDSInst> pdsInstList;
	
	protected ElementValue[] annotatedBits;
	
	private int paramCount = -1;
	
	private String humanMethodName = null;
	
	/**
	 * The constructor.
	 * 
	 * @param pdsClass the class of this method.
	 * @param methodInfo the method info.
	 * @param pdsInfo the pds info.
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	public PDSMethod(PDSClass pdsClass, MethodInfo methodInfo, PDSInfo pdsInfo)
			throws InvalidByteCodeException, IOException {
		
		this.pdsClass = pdsClass;
		this.classFile = pdsClass.getClassFile();
		this.methodInfo = methodInfo;
		this.pdsInfo = pdsInfo;
		
		String[] name = new String[3];
		name[0] = classFile.getThisClassName();
		name[1] = methodInfo.getName();
		name[2] = methodInfo.getDescriptor();
		methodName = PDSMethod.formatMethodName(name);
		
		logger.debug("Initializing PDSMethod methodName=" + methodName);
		
		annotatedBits = PDSAnnotationUtils.getBits(methodInfo);
		if (annotatedBits == null) logger.debug("annotatedBits NULL");

		codeAttr = (CodeAttribute)methodInfo.findAttribute(CodeAttribute.class);

		if (codeAttr == null) {
			initPDSInstListForNative();
		} else {
			instList = (List<AbstractInstruction>) 
						ByteCodeReader.readByteCode(codeAttr.getCode());
			initPDSInstList();
		}
		
		initLocalVarList();
	}
	
	/**
	 * Translates this method to Remopla.
	 * 
	 * @throws InvalidByteCodeException
	 */
	public void translate() throws InvalidByteCodeException {
		
		for (PDSInst pdsInst : pdsInstList) {
			pdsInst.translate();
		}
	}
	
	/**
	 * Replaces the last Remopla instruction of the method with inst.
	 * 
	 * @param inst the Remopla instruction.
	 */
	public void replaceLastInst(String inst) {
		
		PDSInst pdsInst = pdsInstList.get(pdsInstList.size() - 1);
		pdsInst.replaceInst(inst);
	}
	
	public static String createHumanReadableName(String className, String methodName)
	{
		String humanName;
		int lastIndex = className.lastIndexOf('/');
		if (lastIndex > 0) {
			className = className.substring(lastIndex + 1, className.length());
		}
		humanName = className.replace('$', '.');
		if (methodName.equals("<init>")) {
			humanName += "." 
				+ humanName.substring(humanName.lastIndexOf('.') + 1);
		}
		else {
			humanName += "." + methodName;
		}
		return humanName;
	}
	
	public String getHumanReadableName()
	{
		if (humanMethodName == null) {
			ClassFile classFile = getMethodInfo().getClassFile();
			try {
				String className = classFile.getThisClassName();
				String methodName = getName();
				humanMethodName = createHumanReadableName(className, methodName);
			}
			catch (InvalidByteCodeException ive) {
				humanMethodName = getFormattedName();
			}
		}
		return humanMethodName;
	}
	
	/**
	 * Returns the <code>ClassFile</code> object for this method.
	 * 
	 * @return the <code>ClassFile</code> object.
	 */
	public ClassFile getClassFile() {
		
		return classFile;
	}
	
	/**
	 * Returns the <code>PDSClass</code> object for this method.
	 * 
	 * @return the <code>PDSClass</code> object.
	 */
	public PDSClass getPDSClass() {
		
		return pdsClass;
	}
	
	/**
	 * Returns the number of parameters given the method's descriptor in 
	 * the JVM standard.
	 *
	 * @param descriptor the method descriptor.
	 * @return the number of parameters.
	 */
	public static int countParam(String descriptor) {
		
		int count = 0;
		String param = descriptor.substring(1, descriptor.indexOf(")"));
		
		for (int i = 0; i < param.length(); i++) {
			if (param.charAt(i) == '[') {
				continue;
			}
			if (param.charAt(i) == 'L') {
				while (param.charAt(i) != ';') {
					i++;
				}
			}
			count++;
		}
		
		return count;
	}
	
	/**
	 * Returns true if the parameter at <code>index</code> is an array.
	 * <p>
	 * Indexes begin at zero and are smaller than {@link #getParamCount()}.
	 * @param descriptor
	 * @param index
	 * @return
	 */
	public static boolean isArray(String descriptor, int index)
	{
		int count = 0;
		char[] array = descriptor.substring(1, descriptor.indexOf(")")).toCharArray();
		for (int i = 0; i < array.length; i++) {
			if (array[i] == '[') {
				if (count == index) {
					return true;
				}
				else {
					while (i + 1 < array.length && array[i+1] == '[') {
						i++;
					}
					continue;
				}
			}
			if (array[i] == 'L') {
				while (array[i] != ';') {
					i++;
				}
			}
			count++;
			if (count > index) {
				break;
			}
		}
		return false;
	}
	
	/**
	 * Returns true if the descriptor returns a reference
	 * @param descriptor
	 * @return
	 */
	public static boolean returnsReference(String descriptor)
	{
		return descriptor.indexOf(")L") != -1 || descriptor.indexOf(")[") != -1;
	}
	
	/**
	 * Returns true if the parameter at <code>index</code> is an array.
	 * <p>
	 * Indexes begin at zero and are smaller than {@link #getParamCount()}.
	 */
	public boolean isParamArray(int index)
	{
		String desc = getDescriptor();
		if (desc != null) {
			return PDSMethod.isArray(desc, index);
		}
		return false;
	}
	
	public int getParamCount()
	{
		if (paramCount == -1) {
			String desc = getDescriptor();
			if (desc != null) {
				paramCount = PDSMethod.countParam(desc);
			}
		}
		return paramCount;
	}
	
	/**
	 * Returns the info object responsible for this method.
	 */
	public PDSInfo getInfo()
	{
		return pdsInfo;
	}
	
	public static String formatMethodName(String name) {
		
		String out = name;
		
		out = out.replaceAll("[()/;$]", "_");
		out = out.replace('[', 'A');
		out = out.replace('.', '_');
		out = out.replaceAll("_{2,}", "_");
		out = out.replaceAll("[<>]", "_");
		if (out.endsWith("_"))
			out = out.substring(0, out.length()-1);
		if (out.startsWith("_"))
			out = out.substring(1, out.length());
		
		return out;
	}
	
	public static String formatMethodName(String[] name) {
		
		StringBuffer out = new StringBuffer();
		
		if (PDSInfo.simplified) {
			if (name[1].equals("<init>")) {
				out.append(PDSClass.formatClassName(name[0]));
			} else {
				out.append(formatMethodName(name[1]));
			}
		} else {
			out.append(formatMethodName(PDSClass.formatClassName(name[0]) 
					+ "_" + name[1] + "_" + name[2]));
		}
		
		return out.toString();
	}
	
	public String getFormattedName() {
		
		return methodName;
	}
	
	public String getDescriptor() {
		
		String methodDes = null;
		
		try {
			methodDes = methodInfo.getDescriptor();
		} catch (InvalidByteCodeException e) {
			e.printStackTrace();
		}
		
		return methodDes;
	}
	
	/**
	 * Returns the name part of the method, this is not enough to uniquely
	 * identify it.
	 * 
	 * @return null if the name could not be retrieved from the underlying
	 * MethodInfo object
	 */
	public String getName() {
		
		String methodName = null;
		
		try {
			methodName = methodInfo.getName();
		} catch (InvalidByteCodeException e) {
			e.printStackTrace();
		}
		
		return methodName;
	}
	
	public int getMaxStack() {
		
		return codeAttr.getMaxStack();
	}
	
	/**
	 * Returns the method info object that represents this method.
	 * @return
	 */
	public MethodInfo getMethodInfo() {
		return methodInfo;
	}
	
	/**
	 * Returns the unmodifiable list of instructions of this method. 
	 * @return
	 */
	public List<AbstractInstruction> getInstructionList() {
		return Collections.unmodifiableList(instList);
	}
	
	public void addLocalVar(String name) {
		
		addLocalVar(name, 0);
	}
	
	/**
	 * Adds new variable name with size bits into pdsLocalVarList from external.
	 * The method does nothing if name is already exists in pdsLocalVarList.
	 * The variable is flagged as external.
	 * 
	 * @param name the name of the variable.
	 * @param bits the size of the variable
	 */
	public void addLocalVar(String name, int bits) {
		
		for (PDSLocalVar pdsLocalVar : pdsLocalVarList) {
			if (pdsLocalVar.getName().equals(name))
				return;
		}
		pdsLocalVarList.add(new PDSLocalVar(name, bits, true));
	}
	
	/**
	 * Returns a string which represents information inside pdsLocalVarList.
	 * 
	 * @return the info string.
	 */
	public String localVarInfo() {
		
		StringBuilder out = new StringBuilder();
		
		for (PDSLocalVar pdsLocalVar : pdsLocalVarList) {
			out.append("int ");
			out.append(pdsLocalVar.getName());
			out.append("(");
			out.append(pdsLocalVar.getBits());
			out.append(")");
			out.append(";\n");
		}
		
		return out.toString();
	}
	
	/**
	 * Find pdsLocalVarList's index of variable <code>var</code>
	 * at <code>pc</code>.
	 * 
	 * @param var the bytecode local variable
	 * @param pc the pc
	 * @return the index
	 */
	public int findLocalVarIndex(int var, int pc) {
		
		PDSLocalVar pdsLocalVar;
		
		for (int i = 0; i < pdsLocalVarList.size(); i++) {
			pdsLocalVar = pdsLocalVarList.get(i);
			if ((pdsLocalVar.getIndex() == var) 
					&& (pc > pdsLocalVar.getStartPc())
					&& (pc < pdsLocalVar.getLength())) {
				return i;
			}
		}
		
		return -1;
	}
	
	public boolean isLocalVarArray(int index) {
		
		int descIndex = pdsLocalVarList.get(index).getDescriptorIndex();
		String desc = null;
		try {
			desc = classFile.getConstantPoolUtf8Entry(descIndex).getString();
		} catch (InvalidByteCodeException e) {
			e.printStackTrace();
		}
		if (isArray(desc, 0))
			return true;
		else
			return false;
	}
	
	/**
	 * Returns the maximum number of bits of the variable name.
	 * 
	 * @param name the variable name
	 * @return the maximum number of bits
	 */
	public int getMaxLocalVarBits(String name) {
		
		int bits = pdsInfo.getBits();
		int max = 0;
		
		for (PDSLocalVar pdsLocalVar : pdsLocalVarList) {
			
			if (pdsLocalVar.getName().equals(name)) {
				if (pdsLocalVar.getBits(bits) > max) {
					max = pdsLocalVar.getBits();
				}
			}
		}
		
		return (max == 0) ? bits : max;
	}
	
	/**
	 * Returns the <code>PDSLocalVar</code> object at index.
	 * 
	 * @param index the index.
	 * @return the <code>PDSLocalVar</code> object.
	 * @throws ArrayIndexOutOfBoundsException if index is greater than the
	 * 		length of the list.
	 */
	public PDSLocalVar getPDSLocalVar(int index) {
		
		return pdsLocalVarList.get(index);
	}
	
	/**
	 * Returns the <code>PDSLocalVar</code> object with name in Java source.
	 * If none, <code>null</code> is returned.
	 * 
	 * @param name the variable name.
	 * @return the <code>PDSLocalVar</code> object, 
	 * 		or <code>null</code> if not found.
	 */
	public PDSLocalVar getPDSLocalVar(String name) {
		
		for (PDSLocalVar pdsLocalVar : pdsLocalVarList) {
			
			if (pdsLocalVar.getSourceName().equals(name))
				return pdsLocalVar;
		}
		
		return null;
	}
	
	/**
	 * Gets the annotated array bit of the variable <code>var</code>.
	 * 
	 * @param var the variable name
	 * @return the number of bits
	 * @throws InvalidByteCodeException
	 */
	public int getArrayBit(String var) throws InvalidByteCodeException {
		
		int bit = PDSAnnotationUtils.getArrayBits(annotatedBits, methodInfo, var);
		return (bit != 0) ? bit : pdsInfo.getBits();
	}
	
	/**
	 * @param name
	 * @return
	 */
	public boolean isNameAndDescriptor(String name, String des) {
		
		return getName().equals(name) && getDescriptor().equals(des);
	}
	
	public boolean isName(String name) {
		
		return getName().equals(name);
	}
	
	public boolean isStatic() {
		
		return isStatic(methodInfo);
	}
	
	public static boolean isStatic(MethodInfo methodInfo) {
		
		return (methodInfo.getAccessFlags() & AccessFlags.ACC_STATIC) != 0;
	}
	
	public boolean isMainMethod() {
		return isNameAndDescriptor("main", "([Ljava/lang/String;)V");
	}
	
	public int getBitsAtPc(int pc) {
		
		logger.debug("getBitsAtPc(" + pc + ")");
		LineNumberTableAttribute lntAttr = (LineNumberTableAttribute)
				codeAttr.findAttribute(LineNumberTableAttribute.class);
		
		// If for some reason LineNumberTable doesn't exist,
		// return 0.
		if (lntAttr == null) return 0;
		
		LineNumberTableEntry[] lineNumberTable = lntAttr.getLineNumberTable();
		
		if (lineNumberTable == null) return 0;
		
		int i = 0;
		while (i < lineNumberTable.length 
				&& lineNumberTable[i].getStartPc() < pc) {
			i++;
		}
		
		return pdsInfo.getBitsAtLine(lineNumberTable[i-1].getLineNumber());
	}
	
	/**
	 * Overriden for testing purposes, returns {@link #getName()}.
	 */
	@Override
	public String toString()
	{
		return getName();
	}
	
	/**
	 * Returns <code>true</code> if this method returns void.
	 * 
	 * @return <code>true</code> if this method returns void.
	 */
	public boolean isVoid() {
		
		return isVoid(getDescriptor());
	}
	
	/**
	 * Returns <code>true</code> if the method descriptor retruns void.
	 * 
	 * @param descriptor
	 * @return
	 */
	public static boolean isVoid(String descriptor) {
		
		return descriptor.endsWith("V");
	}
	
	/**
	 * Returns the Remopla code for this method.
	 * 
	 * @return the Remopla code.
	 * @throws InvalidByteCodeException
	 */
	public String toRemopla() throws InvalidByteCodeException {
	
		int paramCount = getParamCount();
		
		if (!isStatic()) {
			paramCount++;
		}
			
		PDSInst pdsInst;
		StringBuffer out = new StringBuffer();
		
		out.append(toRemoplaHead());
		out.append("\n{\n");
		if (codeAttr != null) {
			
			for (int i = paramCount; i < codeAttr.getMaxLocals(); i++) {
				
				String var = pdsInfo.varIndex(i);
				out.append(PDSLocalVar.toRemopla(var, getMaxLocalVarBits(var)));
				out.append(";\n");
			}
			
			for (PDSLocalVar pdsLocalVar : pdsLocalVarList) {
				
				if (pdsLocalVar.isExternal()) {
					
					out.append(pdsLocalVar.toRemopla());
					out.append(";\n");
				}
			}
			for (int i = 0; i < codeAttr.getMaxStack(); i++) {
				
				out.append("int ");
				out.append(pdsInfo.stackIndex(i));
				out.append(";\n");
			}
		}
		// The case where the method is native but return something.
		// A local variable is needed to return undef.
		else if (!isVoid()) {
			
			out.append(PDSLocalVar.toRemopla(pdsInfo.varIndex(paramCount), 0));
			out.append(";\n");
		}
		
		for (int i = 0; i < pdsInstList.size(); i++) {
			
			pdsInst = pdsInstList.get(i);
			out.append(pdsInst.toRemopla());
		}
		out.append("}\n");
		
		return out.toString();
	}
	
	/**
	 * Returns the Remopla module header for this method.
	 * 
	 * @return the Remopla module header.
	 * @throws InvalidByteCodeException
	 */
	public String toRemoplaHead() throws InvalidByteCodeException {
		
		int paramCount = getParamCount();
		String var;
		
		if (!isStatic()) {
			paramCount++;
		}
			
		StringBuffer out = new StringBuffer();
		
		out.append("module ");
		if (isVoid()) {
			out.append("void ");
		} else {
			out.append("int ");
		}
		out.append(methodName);
		logger.debug(methodName + ", paramCount=" + paramCount);
		out.append("(");
		if (paramCount > 0) {
			
			var = pdsInfo.varIndex(0);
			out.append(PDSLocalVar.toRemopla(var, getMaxLocalVarBits(var)));
			for (int i = 1; i < paramCount; i++) {
				
				var = pdsInfo.varIndex(i);
				out.append(", ");
				out.append(PDSLocalVar.toRemopla(var, getMaxLocalVarBits(var)));
			}
		}
		out.append(")");
		
		return out.toString();
	}
	
	/*******************************************************************/
	/**                       PRIVATE METHODS                         **/
	/*******************************************************************/
	
	private void initPDSInstList() throws InvalidByteCodeException {
		
		logger.debug("Entering initPDSInstList() of method: " + methodName);
		
		pdsInstList = new ArrayList<PDSInst>();
		
		for (Iterator<AbstractInstruction> i = instList.iterator(); i.hasNext();) {
			AbstractInstruction inst = i.next();
			
			pdsInstList.add(new PDSInst(this, inst, pdsInfo));
			
			if (isAssertionErrorConstructor(inst)) {
				// ignore the throw instruction after an assertion error constructo
				// since it is never reached
				if (i.hasNext()) {
					inst = i.next();
					if (inst.getOpcode() != Opcodes.OPCODE_ATHROW) {
						pdsInstList.add(new PDSInst(this, inst, pdsInfo));
					}
				}
			}
		}
	}
		
	
	
	boolean isAssertionErrorConstructor(AbstractInstruction inst)
	throws InvalidByteCodeException
	{
		if (inst.getOpcode() == Opcodes.OPCODE_INVOKESPECIAL) {
			String[] calledName = BytecodeUtils.getReferencedName(classFile, inst);
			if (calledName[0].equals("java/lang/AssertionError")
					&& calledName[1].equals("<init>")) {
				return true;
			}
		}
		return false;
	}
	
	private void initPDSInstListForNative() throws 
			InvalidByteCodeException {
		
		String descriptor = methodInfo.getDescriptor();
		PDSInst pdsInst;
		
		pdsInstList = new ArrayList<PDSInst>();
		if (isVoid(descriptor)) {
			
			pdsInst = new PDSInst(this, pdsInfo);
			pdsInst.addInst("return");
			pdsInstList.add(pdsInst);
		} else {
			
			String retVar = pdsInfo.varIndex(countParam(descriptor));
			pdsInst = new PDSInst(this, pdsInfo);
			pdsInst.addInst(retVar + "=undef");
			pdsInstList.add(pdsInst);
			
			pdsInst = new PDSInst(this, pdsInfo);
			pdsInst.addInst("return " + retVar);
			pdsInstList.add(pdsInst);
		}
	}
	
	/**
	 * Returns the list of parameters of this method. The list is created
	 * from the descriptor of the method. An element of the array is one
	 * of the following (see JVM specification, section 4.3.2 Field 
	 * Descriptor):
	 * 
	 * <table>
	 * <tr><td>BaseType Character</td>	<td>Type</td></tr>
	 * <tr><td>B</td>				<td>byte</td></tr>
	 * <tr><td>C</td>				<td>char</td></tr>
	 * <tr><td>D</td>				<td>double</td></tr>
	 * <tr><td>F</td>				<td>float</td></tr>
	 * <tr><td>I</td>				<td>int</td></tr>
	 * <tr><td>J</td>				<td>long</td></tr>
	 * <tr><td>L<classname>;</td>	<td>reference (a class instance)</td></tr>
	 * <tr><td>S</td>				<td>short</td></tr>
	 * <tr><td>Z</td>				<td>boolean</td></tr>
	 * </table>
	 * 
	 * Each element can be prefixed by (one or more) [, which denotes 
	 * (one or more) array dimension.
	 *
	 * @return the list
	 */
	public List<String> getParamList() throws InvalidByteCodeException {
		
		String descriptor = getDescriptor();
		String param = descriptor.substring(1, descriptor.indexOf(")"));
		StringBuilder tmp;
		
		List<String> paramList = new ArrayList<String>();
		for (int i = 0; i < param.length(); i++) {
			switch (param.charAt(i)) {		
				case '[': 
					tmp = new StringBuilder();
					while (param.charAt(i) == '[') {
						tmp.append("[");
						i++;
					}
					if (param.charAt(i) == 'L') {
						while (param.charAt(i) != ';') {
							tmp.append(param.charAt(i));
							i++;
						}
						tmp.append(";");
					} else {
						tmp.append(param.charAt(i));
					}
					paramList.add(tmp.toString());
					break;
					
				case 'L':
					tmp = new StringBuilder();
					while (param.charAt(i) != ';') {
						tmp.append(param.charAt(i));
						i++;
					}
					tmp.append(";");
					paramList.add(tmp.toString());
					break;
					
				default:
					paramList.add(new String(new char[] {param.charAt(i)}));
			}
		}
		
		return paramList;
	}

	/**
	 * Initializes pdsLocalVarList to a list of <code>PDSLocalVar</code>
	 * objects.
	 * 
	 * If the method contains local variable table, each object in 
	 * pdsLocalVarList corresponds to an entry in the table.
	 * Note that names of variables are determined by
	 * {@link PDSInfo#varIndex(int) PDSInfo.varIndex(index)}, where the index is 
	 * {@link org.gjt.jclasslib.structures.attributes.LocalVariableCommonEntry#getIndex()}.
	 * 
	 * If the table is <code>null</code>, the method 
	 * {@link CodeAttribute#getMaxLocals() CodeAttribute.getMaxLocals()}
	 * determines the number of objects in the list. 
	 * 
	 * @throws InvalidByteCodeException
	 */
	private void initLocalVarList() throws InvalidByteCodeException {
		
		pdsLocalVarList = new ArrayList<PDSLocalVar>();
		
		if (codeAttr == null) {
			
			addLocalVarListFromDescriptor();
			return;
		}
		
		PDSLocalVar pdsLocalVar;
		int varBits;
		LocalVariableTableAttribute lvtAttr = (LocalVariableTableAttribute)
				codeAttr.findAttribute(LocalVariableTableAttribute.class);
		if (lvtAttr != null) {
			
			LocalVariableTableEntry[] lvtEntry = lvtAttr.getLocalVariableTable();
			int index;
			ConstantUtf8Info info;
			
			for (int i = 0; i < lvtEntry.length; i++) {
				
				int varIndex = lvtEntry[i].getIndex();
				if (annotatedBits != null) {
					index = lvtEntry[i].getNameIndex();
					info = classFile.getConstantPoolUtf8Entry(index);
					varBits = PDSAnnotationUtils.getBit(
							annotatedBits, classFile, info.getString(), 
							isStatic() ? varIndex : varIndex - 1);
				} else {
					varBits = 0;
				}
				String varName = pdsInfo.varIndex(varIndex);
				pdsLocalVar = new PDSLocalVar(lvtEntry[i], varName, varBits);
				pdsLocalVarList.add(pdsLocalVar);
				logger.debug("Local Var: " + pdsLocalVar.getName() 
						+ "(" + pdsLocalVar.getBits() + ")");
			}
		} else {
			
			logger.debug("LocalVariableTableAttribute is null");
			for (int i = 0; i < codeAttr.getMaxLocals(); i++) {
				
				if (annotatedBits != null && i < annotatedBits.length) {
					varBits = PDSAnnotationUtils.getInt(annotatedBits[i], 
							classFile);
				} else {
					varBits = 0;
				}
				pdsLocalVar = new PDSLocalVar(pdsInfo.varIndex(i), varBits);
				pdsLocalVarList.add(pdsLocalVar);
			}	
		}
		
		logger.debug("initLocalVarTable completed");
	}
	
	private void addLocalVarListFromDescriptor() 
			throws InvalidByteCodeException {
		
		int paramCount = getParamCount();
		if (!isStatic()) {
			paramCount++;
		}
		PDSLocalVar pdsLocalVar;
		
		for (int i = 0; i < paramCount; i++) {
			
			pdsLocalVar = new PDSLocalVar(pdsInfo.varIndex(i));
			pdsLocalVarList.add(pdsLocalVar);
		}
	}
}
