package de.uni_stuttgart.fmi.szs.jmoped;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;

//import static de.uni_stuttgart.fmi.szs.jmoped.PDSInfo.HeapOption;
import static de.uni_stuttgart.fmi.szs.jmoped.PDSDefault.NEWLINE;

public class PDSMethodWrapper {

	PDSClass pdsClass;
	PDSMethod pdsMethod;
	PDSInfo pdsInfo;
	ClassFile classFile;
	private List<PDSLocalVar> pdsLocalVarList;
	private List<PDSStmt> pdsStmtList;
	private String counter;
	
	private final static String wrapperPrefix = "wrapperFor_";
	
	static Logger logger = Logger.getLogger(PDSMethodWrapper.class);
	
	/**
	 * Constructor.
	 * 
	 * @param pdsClass
	 * @param pdsMethod
	 * @throws InvalidByteCodeException
	 * @throws IOException
	 */
	public PDSMethodWrapper(PDSClass pdsClass, PDSMethod pdsMethod) 
			throws InvalidByteCodeException, IOException {
		
		this.pdsClass = pdsClass;
		this.pdsMethod = pdsMethod;
		this.pdsInfo = pdsMethod.getInfo();
		this.classFile = pdsMethod.getClassFile();
	}
	
	/**
	 * Gets the name of the method wrapper
	 * 
	 * @return the name of the method wrapper
	 */
	public String getName() {
		
		return wrapperPrefix + pdsMethod.getFormattedName();
	}
	
	/**
	 * Wraps the method. This is where Remopla statements are crated.
	 * The method must be called before toRemopla().
	 * 
	 * @throws InvalidByteCodeException
	 */
	public void wrap() throws InvalidByteCodeException {
		
		pdsLocalVarList = new ArrayList<PDSLocalVar>();
		pdsStmtList = new ArrayList<PDSStmt>();
		
		/* We don't wrap static fields,
		 * since it might have been initiliazed in <clinit>.
		 */
		//wrapStaticFields();
		if (pdsMethod.isStatic()) wrapStaticMethod();
		else wrapVirtualMethod();
		wrapCall();
	}
	
	/**
	 * Returns the String, which represents a Remopla module that
	 * wraps the method.
	 * 
	 * @return the Remopla module
	 */
	public String toRemopla() {
		
		StringBuilder out = new StringBuilder();
		
		out.append(toRemoplaHead());
		out.append(NEWLINE);
		out.append("{");
		out.append(NEWLINE);
		for (PDSLocalVar pdsLocalVar : pdsLocalVarList) {
			out.append(pdsLocalVar.toRemopla());
			out.append(";");
			out.append(NEWLINE);
		}
		for (PDSStmt pdsStmt : pdsStmtList) {
			out.append(pdsStmt.toRemopla());
		}
		out.append("}");
		out.append(NEWLINE);
		
		return out.toString();
	}
	
	/**
	 * Returns the String, which represents a Remopla module's header
	 * that wraps the method
	 * 
	 * @return the Remopla module header
	 */
	public String toRemoplaHead() {
		
		StringBuilder out = new StringBuilder();
		
		out.append("module void ");
		out.append(getName());
		out.append("()");
		
		return out.toString();
	}
	
	/*******************************************************************/
	/**                       PRIVATE METHODS                         **/
	/*******************************************************************/
	
	/**
	 * Wraps static fields.
	 * 
	 * @throws InvalidByteCodeException
	 */
	private void wrapStaticFields() throws InvalidByteCodeException {
		
		List<PDSField> pdsFieldList = pdsClass.getStaticFieldList();
		if (pdsFieldList.isEmpty()) return;
		
		int bits = pdsInfo.getBits();
		List<String> cond = new ArrayList<String>();
		PDSStmt stmt = new PDSStmt();
		
		for (PDSField pdsField : pdsFieldList) {
			String rhs = (pdsField.isReference()) 
				? "'==0" 
				: "'<=" + PDSDefault.getMaxIntForBits(pdsField.getBits(bits));
			cond.add(pdsField.getFormattedName() + rhs);
		}
		
		stmt.addSkip(cond);
		pdsStmtList.add(stmt);
	}
	
	/**
	 * Wraps the static method.
	 */
	private void wrapStaticMethod() throws InvalidByteCodeException {
		
		wrapParams(0);
	}
	
	/**
	 * Wraps the virtual method.
	 * 	this = heap_ptr, 
	 * 	heap_ptr = heap_ptr + classSize + 1,
	 * 	heap[heap_ptr] = classIndex,
	 * 	heap[heap_ptr+0+1] = undef, ..., heap[heap_ptr+classSize-1+1] = undef,
	 *  staticField = undef;
	 * 
	 * @throws InvalidByteCodeException
	 */
	private void wrapVirtualMethod() throws InvalidByteCodeException {
		
		String var = "this";
		int bits = pdsInfo.getBits();
		pdsLocalVarList.add(new PDSLocalVar(var, bits));
		
		String heapPtr = pdsInfo.getHeapPtrName();
		PDSStmt stmt = new PDSStmt();
		
		switch (pdsInfo.getHeapOption()) {
		case SIMPLE:
			
			List<String> cond = new ArrayList<String>();
			
			// this = heap_ptr,
			cond.add(var + "'==" + heapPtr);
			
			// heap_ptr = heap_ptr + classSize + 1,
			cond.add(heapPtr + "'==" + heapPtr + "+" + pdsClass.size() + "+1");
			
			// heap[heap_ptr] = classIndex,
			cond.add(pdsInfo.heapIndex(heapPtr, "'") + "==" + pdsClass.getIndex());
			
			/* Initializes elemental fields to undef, reference fields to null,
			 * i.e. heap[heap_ptr+fieldIndex+1] = undef, for elemental fields
			 *  and heap[heap_ptr+fieldIndex+1] = 0, for reference fields.
			 */
			for (PDSField pdsField : pdsClass.getNonStaticFieldList()) {
				String ptr = heapPtr + "+" + pdsField.getIndex() + "+1";
				String rhs;
				if (pdsField.isReference()) rhs = "==0";
				else rhs = "<=" + PDSDefault.getMaxIntForBits(pdsField.getBits(bits));
				cond.add(pdsInfo.heapIndex(ptr, "'") + rhs);
			}
			
			stmt.addSkip(cond);
			break;
			
		case TWODIMS:
			
			String heapName = pdsInfo.getHeapName();
			
			// this = heap_ptr,
			stmt.addPar(var + "=" + heapPtr);
			
			// heap_ptr = heap_ptr + classSize + 1,
			stmt.addPar(heapPtr + "=" + heapPtr + "+" + pdsClass.size() + "+1");
			
			// Stores classIndex into heap at bits*heap_ptr
			stmt.addPar(PDSStmt.storeArray(heapName, 
					bits + "*" + heapPtr, pdsClass.getIndex(), bits));
			
			/* Initializes elemental fields to undef, reference fields to null,
			 * i.e. stores undef (resp. 0) to heap at 
			 * bits*(heap_ptr+1)+fieldIndex, for elemental fields
			 */
			int count = 0;
			for (PDSField pdsField : pdsClass.getNonStaticFieldList()) {
				
				String ptr = bits + "*(" + heapPtr + "+1)+" + count;
				String rhs;
				if (pdsField.isReference()) rhs = "0";
				else rhs = "undef";
				logger.debug("pdsField.getFormattedName(): " + pdsField.getFormattedName());
				logger.debug(PDSStmt.commaList(
						PDSStmt.storeToEveryIndex(heapName, ptr, rhs, pdsField.getBits(bits))));
				stmt.addPar(PDSStmt.storeToEveryIndex(heapName, ptr, rhs,
						pdsField.getBits(bits)));
				count += pdsField.getBits(bits);
			}
			break;
		}
		pdsStmtList.add(stmt);
		
		wrapParams(1);
	}
	
	/**
	 * Wraps the parameters.
	 * 
	 * @param varIndex the index of the bytecode's local variable
	 * @throws InvalidByteCodeException
	 */
	private void wrapParams(int varIndex) throws InvalidByteCodeException {
		
		List<String> paramList = pdsMethod.getParamList();
		counter = PDSLocalVar.getUniqueVarName(paramList, "counter");
		
		for (String param : paramList) {
			
			PDSLocalVar pdsLocalVar = pdsMethod.getPDSLocalVar(varIndex);
			String var = pdsLocalVar.getSourceName();
			int bits = pdsLocalVar.getBits();
			if (bits == 0) bits = pdsInfo.getBits();
			
			switch (param.charAt(0)) {
			case 'L':
				unset(var, bits);
				break;
				
			case '[':
				undefArray(var, bits);
				break;
				
			default:
				undef(var, bits);
				break;
			}
			
			varIndex++;
		}
	}
	
	/**
	 * Adds local variable <code>var</code>, and creates the statement
	 * <code>var</code>=0.
	 * 
	 * @param var the variable name
	 * @param bits the number of bits
	 */
	private void unset(String var, int bits) {
		
		pdsLocalVarList.add(new PDSLocalVar(var, bits));
		pdsStmtList.add(new PDSStmt(var + "=0"));
	}
	
	/**
	 * Adds local variable <code>var</code>, and creates the statement
	 * <code>var</code>=undef.
	 * 
	 * @param var the variable name
	 * @param bits the number of bits
	 */
	private void undef(String var, int bits) {
		
		pdsLocalVarList.add(new PDSLocalVar(var, bits));
		pdsStmtList.add(new PDSStmt(var + "=undef"));
	}
	
	/**
	 * Adds into pdsStmtList a PDSStmt, which creates an array in the heap.
	 * <code>var</code> is the name of the array.
	 * 
	 * @param var the array's name
	 * @param bits the array's bits
	 */
	private void undefArray(String var, int bits) 
			throws InvalidByteCodeException {
		
		pdsLocalVarList.add(new PDSLocalVar(var, pdsInfo.getBits()));
		
		PDSStmt stmt;
		List<String> cond;
		String heapPtr = pdsInfo.getHeapPtrName();
		int intBit = pdsInfo.getBits();
		int arrBit = pdsMethod.getArrayBit(var);
		int maxArrInt = PDSDefault.getMaxIntForBits(arrBit);
		
		// Adds a counter variable
		addLocalVar(counter, bits);
		
		/* skip (var' <= 2^bits && counter' == 0); */
		cond = new ArrayList<String>();
		cond.add(var + "'<=" + PDSDefault.getMaxIntForBits(bits));
		cond.add(counter + "'==0");
		stmt = new PDSStmt();
		stmt.addSkip(cond);
		pdsStmtList.add(stmt);
		
		/* do
		 *   :: counter < var -> skip (heap'[counter+heap_ptr+1] <= 2^arrayBits
		 *                             && counter' == counter+1);
		 *   :: else -> break;
		 * od;
		 */
		cond = new ArrayList<String>();
		switch (pdsInfo.getHeapOption()) {
		case SIMPLE:
			cond.add(pdsInfo.heapIndex(heapPtr + "+" + counter + "+1", "'") 
					+ "<=" + maxArrInt);
			break;
		case TWODIMS:
			String base = intBit + "*(" + heapPtr + "+2) + " 
					+ arrBit + "*" + counter;
			for (int i = 0; i < arrBit; i++)
				cond.add(pdsInfo.heapIndex(base + " + " + i, "'") 
						+ "<=" + maxArrInt);
			break;
		}
		cond.add(counter + "'==" + counter + "+1");
		stmt = new PDSStmt();
		stmt.addDo();
		stmt.addColonStmt(counter + "<" + var, PDSStmt.skip(cond));
		stmt.addColonElseBreak();
		stmt.addOd();
		pdsStmtList.add(stmt);
		
		/* heap[heap_ptr] = var, 
		 * heap_ptr = heap_ptr + var + 1,
		 * var = heap_ptr; 
		 */
		String gotoOverflow = PDSStmt.gotoStmt(pdsInfo.getLabelHeapOverflow());
		stmt = new PDSStmt();
		switch (pdsInfo.getHeapOption()) {
		case SIMPLE:
			
			stmt.addPar(pdsInfo.heapIndex(heapPtr) + "=" + var);
			stmt.addPar(heapPtr + "=" + heapPtr + "+" + var + "+1");
			stmt.addPar(var + "=" + heapPtr);
			break;
		case TWODIMS:
			
			PDSStmt.createArray(stmt, pdsInfo, var, arrBit, gotoOverflow);
			break;
		}
		pdsStmtList.add(stmt);
	}

	/**
	 * Creates the statement for method call.
	 */
	private void wrapCall() {
		
		StringBuilder call = new StringBuilder(pdsMethod.getFormattedName());
		call.append("(");
		PDSLocalVar localVar = null;
		if (pdsLocalVarList.size() > 0) 
			localVar = pdsLocalVarList.get(0);
		if (localVar != null) call.append(localVar.getName());
		for (int i = 1; i < pdsLocalVarList.size(); i++) {
			
			localVar = pdsLocalVarList.get(i);
			if (localVar.isExternal()) continue;
			call.append(", ");
			call.append(localVar.getName());
		}
		call.append(")");
		
		pdsStmtList.add(new PDSStmt(call.toString()));
	}
	
	/**
	 * Adds a local variable with <code>name</code> and <code>bits</code>
	 * into pdsLocalVarList. If there is a variable with the same 
	 * <code>name</code> already in the list, no variable is added.
	 * Instead, the method updates <code>bits</code> of the local
	 * variable if it is greater than the stored <code>bits</code>.
	 * 
	 * @param name the name of the local variable
	 * @param bits the number of bits'
	 * @param flag true, if the variable is external
	 */
	private void addLocalVar(String name, int bits) {
		
		for (PDSLocalVar localVar : pdsLocalVarList) {
			
			if (localVar.getName().equals(name)) {
				
				int oldBits = localVar.getBits();
				if (oldBits == 0) {
					if (bits > pdsInfo.getBits()) localVar.setBits(bits);
				} else {
					if (bits > oldBits) localVar.setBits(bits);
				}
				return;
			}
		}
		
		pdsLocalVarList.add(new PDSLocalVar(name, bits, true));
	}
}
