package de.uni_stuttgart.fmi.szs.jmoped;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.log4j.Logger;

import static de.uni_stuttgart.fmi.szs.jmoped.PDSDefault.NEWLINE;

/**
 * The class represents a sequence of Remopla statements. A statement can be:
 * <ol>
 * <li>(parallel) assignment,</li>
 * <li>if statement, or</li>
 * <li>do statement.</li>
 * </ol>
 * 
 * The class is supposed to be used in the following ways:
 * <ol>
 * <li>Use addNext(stmt) to add stmt that will be executed after previously-
 * 		added statements.</li>
 * <li>For parallel assignment, use addPar(stmt). stmt will be executed in 
 * 		parallel with the previously-added statements. If there is no
 * 		previous statment, the method behaves the same as addNext(stmt).</li> 
 * <li>For if statement, always start with addIf() and end with addFi().
 * 		In between, for each colon statement, use addColonStmt(cond, stmt). 
 * 		Alternatively, use series of addCond(cond), and follow by addNext(stmt) 
 * 		or addPar(stmt).</li>
 * <li>For do statement, always start with addDo() and end with addOd().
 * 		In between, use the same methods as for if statments.</li>
 * </ol>
 * 
 *  For example, if you want to add the following statments:
 *  <pre>
 *  	stmt1, stmt2;
 *  	if
 *  		:: cond1 -> stmt3, stmt4;
 *  		:: cond2 -> stmt5;
 *  	fi;
 *  	stmt6;
 *  </pre>
 *  Call the methods as follows: 
 *  	addNext(stmt1); addPar(stmt2); addIf(); addCond(cond1); addNext(stmt3);
 *  	addPar(stmt4); addColonStmt(cond2, stmt5); addFi(); addNext(stmt6);
 * 
 * @author suwimodh
 *
 */
public class PDSStmt {

	// Label of the statement
	private String label = null;
	
	// Nested level of if/do statements
	int nested = 0;
	
	// Remopla statements
	private StringBuilder stmt = new StringBuilder();
	
	// Enumerated type of Remopla statements
	private static enum StmtType { NULL, NORMAL, COND, IFDO }
	
	// Type of the previously added Remopla statement
	private StmtType type = StmtType.NULL;
	
	// Logger
	static Logger logger = Logger.getLogger(PDSStmt.class);
	
	/**
	 * Constructor with empty Remopla statement.
	 *
	 */
	public PDSStmt() {
	}
	
	/**
	 * Constructor with single statement <code>next</code> added.
	 * 
	 * @param next the Remopla statement
	 */
	public PDSStmt(String next) {
		
		addNext(next);
	}
	
	/**
	 * Constructor with parallel statements <code>parList</code> added.
	 * 
	 * @param parList the list of parallel statements
	 */
	public PDSStmt(List<String> parList) {
		
		addPar(parList);
	}
	
	/**
	 * Adds a goto <code>to</code> statement.
	 * 
	 * @param to the label for goto statement
	 */
	public void addGoto(String to) {
		
		addNext("goto " + to);
	}
	
	/**
	 * Adds a skip statement.
	 *
	 */
	public void addSkip() {
		
		addNext(skip(null));
	}
	
	/**
	 * Adds a skip statement followed by a list of conditions 
	 * <code>condList<code>.
	 * 
	 * @param condList the list of conditions
	 */
	public void addSkip(List<String> condList) {
		
		addNext(skip(condList));
	}
	
	/**
	 * Adds a return statement.
	 *
	 */
	public void addReturn() {
		
		addNext("return");
	}
	
	/**
	 * Adds a return <code>ret</code> statement.
	 * 
	 * @param ret expression to be returned
	 */
	public void addReturn(String ret) {
		
		addNext("return " + ret);
	}
	
	/**
	 * Adds parallel statements <code>parList</code> that will be executed
	 * after the previously-added statement.
	 * 
	 * @param parList the list of parallel statements
	 */
	public void addNext(List<String> parList) {
		
		addNext(commaList(parList));
	}
	
	/**
	 * Adds a single statement <code>next</code> that will be executed
	 * after the previously-added statement. 
	 * 
	 * @param next the statement
	 */
	public void addNext(String next) {
		
		if (type == StmtType.COND)
			stmt.append(" -> ");
		else if (type == StmtType.NORMAL) 
			stmt.append(";").append(NEWLINE);
		stmt.append(next);
		type = StmtType.NORMAL;
	}
	
	/**
	 * Adds parallel statements <code>parList</code> that will be executed
	 * in parallel with the previously-added statement.
	 * 
	 * @param parList
	 */
	public void addPar(List<String> parList) {
		
		String par = commaList(parList, null);
		if (par != null) addPar(par);
	}
	
	/**
	 * Adds a statement <code>par</code> that will be executed
	 * in parallel with the previously-added statement.
	 * 
	 * @param parList
	 */
	public void addPar(String par) {
		
		if (type == StmtType.COND)
			stmt.append(" -> ");
		else if (type == StmtType.NORMAL) 
			stmt.append(", ");
		stmt.append(par);
		type = StmtType.NORMAL;
	}
	
	/**
	 * Adds a list of conditions <code>condList</code> for if/do statments.
	 * 
	 * @param condList the list of condition
	 */
	public void addCond(List<String> condList) {
		
		addCond(andList(condList));
	}
	
	/**
	 * Adds a condition <code>cond</code> for if/do statments.
	 * 
	 * @param cond the condition
	 */
	public void addCond(String cond) {
		
		switch (type) {
		case NORMAL: 
			stmt.append(";");
		case NULL:
		case IFDO:
			stmt.append(NEWLINE);
			for (int i = 0; i <= nested; i++)
				stmt.append("\t");
			stmt.append(":: ");
			break;
		case COND:
			stmt.append(" && ");
			break;
		}
		stmt.append(cond);
		type = StmtType.COND;
	}
	
	/**
	 * Adds a keyword <code>if</code>.
	 *
	 */
	public void addIf() {
		
		if (type != StmtType.NULL)
			stmt.append(";").append(NEWLINE);
		for (int i = 0; i < nested; i++)
			stmt.append("\t");
		stmt.append("if");
		type = StmtType.IFDO;
	}
	
	/**
	 * Adds a keyword <code>fi</code>.
	 *
	 */
	public void addFi() {
		
		for (int i = 0; i < nested; i++)
			stmt.append("\t");
		stmt.append(";").append(NEWLINE).append("fi");
	}
	
	/**
	 * Adds a keyword <code>do</code>.
	 *
	 */
	public void addDo() {
		
		if (type != StmtType.NULL)
			stmt.append(";").append(NEWLINE);
		for (int i = 0; i < nested; i++)
			stmt.append("\t");
		stmt.append("do");
		type = StmtType.IFDO;
	}
	
	/**
	 * Adds a keyword <code>od</code>.
	 *
	 */
	public void addOd() {
		
		for (int i = 0; i < nested; i++)
			stmt.append("\t");
		stmt.append(";").append(NEWLINE).append("od");
	}
	
	/**
	 * Adds the following Remopla code:
	 * <pre>
	 *     :: else -> break;
	 * </pre>
	 *
	 */
	public void addColonElseBreak() {
		
		addColonElse("break");
	}
	
	/**
	 * Adds the following Remopla code:
	 * <pre>
	 *     :: else -> <code>commaList(parList)</code>;
	 * </pre>
	 *
	 */
	public void addColonElse(List<String> parList) {
		
		addColonElse(commaList(parList));
	}
	
	/**
	 * Adds the following Remopla code:
	 * <pre>
	 *     :: else -> <code>stmt</code>;
	 * </pre>
	 *
	 */
	public void addColonElse(String stmt) {
		
		addColonStmt("else", stmt);
	}
	
	/**
	 * Adds the following Remopla code:
	 * <pre>
	 *     :: <code>cond1</code> && <code>cond2</code> -> <code>commaList(parList)</code>;
	 * </pre>
	 *
	 */
	public void addColonStmt(String cond1, String cond2, List<String> parList) {
		
		addColonStmt(cond1 + " && " + cond2, commaList(parList));
	}
	
	/**
	 * Adds the following Remopla code:
	 * <pre>
	 *     :: <code>cond1</code> && <code>cond2</code> -> <code>stmt</code>;
	 * </pre>
	 *
	 */
	public void addColonStmt(String cond1, String cond2, String stmt) {
		
		addColonStmt(cond1 + " && " + cond2, stmt);
	}
	
	/**
	 * Adds the following Remopla code:
	 * <pre>
	 *     :: <code>cond</code> -> <code>commaList(parList)</code>;
	 * </pre>
	 *
	 */
	public void addColonStmt(String cond, List<String> parList) {
		
		addColonStmt(cond, commaList(parList));
	}
	
	/**
	 * Adds the following Remopla code:
	 * <pre>
	 *     :: <code>andList(condList)</code> -> <code>commaList(parList)</code>;
	 * </pre>
	 *
	 */
	public void addColonStmt(List<String> condList, List<String> parList) {
		
		addColonStmt(andList(condList), commaList(parList));
	}
	
	/**
	 * Adds the following Remopla code:
	 * <pre>
	 *     :: <code>cond</code> -> <code>stmt</code>;
	 * </pre>
	 *
	 */
	public void addColonStmt(String cond, String stmt) {
		
		addCond(cond);
		addNext(stmt);
	}
	
	/**
	 * Sets the label of the statement.
	 * 
	 * @param label the label
	 */
	public void setLabel(String label) {
		
		this.label = label;
	}
	
	/**
	 * Gets the label of the statement
	 * 
	 * @return the label
	 */
	public String getLabel() {
		
		return label;
	}
	
	/**
	 * Return the Remopla code representing the statements.
	 * 
	 * @return the Remopla code
	 */
	public String toRemopla() {
		
		StringBuilder out = new StringBuilder();
		if (label != null) {
			out.append(label);
			out.append(": ");
		}
		out.append(stmt);
		if (type == StmtType.NORMAL)
			out.append(";").append(NEWLINE);
		
		return out.toString();
	}
	
	/**
	 * Return <code>true</code> if the statement is empty.
	 * 
	 * @return <code>true</code> if the statement is empty
	 */
	public boolean isEmpty() {
		
		return stmt.length() == 0;
	}
	
	/**
	 * Increases the nested level of the following statements.
	 *
	 */
	public void incNestedLevel() {
		
		nested++;
	}
	
	/**
	 * Decreases the nested level of the following statements.
	 *
	 */
	public void decNestedLevel() {
		
		nested--;
	}
	
	/*******************************************************************/
	/**                    PUBLIC STATIC METHODS                      **/
	/*******************************************************************/
	
	/**
	 * Returns the String which represents for all quantification in Remopla,
	 * i.e. A quantVar (from,to) expr
	 * 
	 * @param quantVar the quantifier
	 * @param from the starting index of the quantifiction
	 * @param to the end index of the quantification
	 * @param expr the expresstion
	 * @return the String
	 */
	public static String forAll(String quantVar, String from, String to, 
			String expr) {
		
		StringBuilder out = new StringBuilder();
		
		out.append("A ");
		out.append(quantVar);
		out.append(" (");
		out.append(from);
		out.append(",");
		out.append(to);
		out.append(") ");
		out.append(expr);
		
		return out.toString();
	}
	
	public static String andList(List<String> list) {
		
		return andList(list, list.size());
	}
	
	public static String andList(List<String> list, int to) {
		
		return createList(list, to, " && ");
	}
	
	public static String commaList(List<String> list) {
		
		return commaList(list, "skip");
	}
	
	public static String commaList(List<String> list, String defaultStmt) {
		
		String par = createList(list, list.size(), ", ");
		return (par != null) ? par : defaultStmt;
	}
	
//	public static String commaList(List<String> list, int to) {
//
//		return createList(list, to, ", ");
//	}
	
	public static String semicolonList(List<String> list) {
		
		return createList(list, list.size(), "; ");
	}
	
	/**
	 * Concatenates elements of the list to a symbol-separated list and 
	 * returns it as a string.
	 *
	 * @return <code>null</code> if there is an error
	 */
	public static String createList(List<String> list, int to, String symbol) {

		if (list.size() == 0) {
			logger.debug("createList.list.size() = 0");
			return null;
		}

		if (list.size() < to) {
			return null;
		}
		StringBuilder out = new StringBuilder();
		Iterator<String> iter = list.iterator();
		
		out.append(iter.next());
		for (int i = 1; i < to; i++) {
			out.append(symbol);
			out.append(iter.next());
		}
		
		return out.toString();
	}
	
	/**
	 * Creates a list of remopla statements which store <code>val</code> 
	 * into <code>array<code> at every index from <code>ptr</code> to
	 * <code>ptr</code> + <code>length</code> - 1. 
	 * 
	 * For example, if val = 1 and length = 3, then we get 
	 * array[ptr+0]=val, array[ptr+1]=val, array[ptr+2]=val. 
	 * 
	 * @param array the array
	 * @param ptr the array pointer
	 * @param val the value
	 * @param length the length
	 * @return the list of strings representing remopla statements
	 */
	public static List<String> storeToEveryIndex(String array, String ptr
			, String val, int length) {
		
		List<String> out = new ArrayList<String>();
		
		for (int i = 0; i < length; i++) {
			
			out.add(PDSInfo.stringIndex(array, ptr + "+" + i) + "=" + val);
		}
		
		return out;
	}
	
	/**
	 * For PDSInfo.HeapOption.TWODIMS: Stores <code>value</code> into 
	 * <code>array</code> starting at <code>ptr</code> with size 
	 * <code>bits</code>. 
	 * 
	 * For example, if bits = 4, the method will return a list of: 
	 * array[ptr] = value>>3,
	 * array[ptr+1] = (value>>2) & 1,
	 * array[ptr+2] = (value>>1) & 1,
	 * array[ptr+3] = value & 1;
	 * 
	 * @param array the array
	 * @param ptr the array pointer
	 * @param value the value to be stored
	 * @param bits the number of bits
	 * @return the array of string representing remopla code
	 */
	public static List<String> storeArray(String array, String ptr, 
			String value, int bits) {
		
		List<String> out = new ArrayList<String>();
		
		for (int i = 0; i < bits; i++) {
			out.add(PDSInfo.stringIndex(array, ptr + "+" + i) + "=(" + value 
					+ ">>" + (bits-i-1) + ")&1");
		}
		
		return out;
	}
	
	/**
	 * For PDSInfo.HeapOption.TWODIMS: Stores <code>value</code> into 
	 * <code>array</code> starting at <code>ptr</code> with size 
	 * <code>bits</code>. 
	 * 
	 * For example, if bits = 4 and value = 5, the method will return a list of: 
	 * heap[ptr] = 0, heap[ptr+1] = 1, heap[ptr+2] = 0, heap[ptr+2] = 1.
	 * 
	 * @param array the array
	 * @param ptr the array pointer
	 * @param value the value to be stored
	 * @param bits the number of bits
	 * @return the array of string representing remopla code
	 */
	public static List<String> storeArray(String array, String ptr, 
			int value, int bits) {
		
		String bin = Integer.toBinaryString(value);
		List<String> out = new ArrayList<String>();
		char b;
		
		for (int i = 0; i < bits; i++) {
			
			if (i < bits - bin.length()) {
				b = '0';
			} else {
				b = bin.charAt(i - bits + bin.length());
			}
			out.add(PDSInfo.stringIndex(array, ptr + "+" + i) + "=" + b);
		}
		
		return out;
	}
	
	/**
	 * For PDSInfo.HeapOption.TWODIMS: Creates array of <code>arrSize</code> 
	 * with <code>arrBit</code> into the heap.
	 * 
	 * @param stmt
	 * @param info
	 * @param arrSize
	 * @param arrBit
	 * @param overflowCode
	 */
	public static void createArray(PDSStmt stmt, PDSInfo info, String arrSize, 
			int arrBit, String overflowCode) {
		
		String heap = info.getHeapName();
		String ptr = info.getHeapPtrName();
		int intBit = info.getBits();
		
		String sizeXbit = "(" + arrSize + "*" + arrBit + ")";
		String newPtr1 = ptr + "+" + sizeXbit + "/" + intBit + "+2";
		String newPtr2 = newPtr1 + "+1";
		String modIntBits = mod(sizeXbit, intBit);
		String cond1 = modIntBits + "==0";
		String cond2 = modIntBits + "!=0";
		List<String> stmtList1 = new ArrayList<String>();
		List<String> stmtList2 = new ArrayList<String>();
		
		stmtList1.add(arrSize + "=" + ptr);
		stmtList1.addAll(storeArray(heap, intBit + "*" + ptr, arrSize, intBit));
		stmtList1.addAll(storeArray(heap, intBit + "*(" + ptr + "+1)", arrBit, 
				intBit));
		stmtList2.addAll(stmtList1);
		stmtList1.add(ptr + "=" + newPtr1);
		stmtList2.add(ptr + "=" + newPtr2);
		
		if (info.checkForHeapOverflow()) {
			
			int heapSize = info.getStoredHeapSize();
			stmt.addIf();
			stmt.addColonStmt(cond1, newPtr1 + ">" + heapSize, overflowCode);
			stmt.addColonStmt(cond1, newPtr1 + "<=" + heapSize, stmtList1);
			stmt.addColonStmt(cond2, newPtr2 + ">" + heapSize, overflowCode);
			stmt.addColonStmt(cond2, newPtr2 + "<=" + heapSize, stmtList2);
			stmt.addFi();
		} else {
			
			stmt.addColonStmt(cond1, stmtList1);
			stmt.addColonStmt(cond2, stmtList2);
		}
	}
	
	/**
	 * Returns a string representing integer value of <code>array</code> 
	 * at <code>ptr</code> with size <code>bits</code>. 
	 * 
	 * For example, if bits = 3, the method will return: 
	 * ((array[ptr+0]<<2) + (array[ptr+1]<<1) + (array[ptr+2])).
	 * 
	 * @param array the array
	 * @param ptr the array pointer
	 * @param bits the number of bits
	 * @return the remopla statement
	 */
	public static String loadArray(String array, String ptr, int bits) {
		
		StringBuilder b = new StringBuilder();
		String index;
		
		if (bits > 1) b.append("(");
		for (int i = bits-1; i > 0; i--) {
			b.append("(");
			index = ptr + "+" + String.valueOf(bits - i - 1);
			b.append(PDSInfo.stringIndex(array, index)
					+ "<<" + String.valueOf(i));
			b.append(")");
			b.append(" + ");
		}
		index = ptr + "+" + String.valueOf(bits - 1);
		b.append(PDSInfo.stringIndex(array, index));
		if (bits > 1) b. append(")");
		
		return b.toString();
	}
	
	public static String skip(List<String> condList) {
		
		if (condList == null || condList.isEmpty())
			return "skip";
		
		return "skip (" + andList(condList) + ")";
	}
	
	/**
	 * Returns a string representing a%b
	 * 
	 * @param a
	 * @param b
	 * @return a - (b*(a/b))
	 */
	public static String mod(String a, int b) {
		
		return mod(a, String.valueOf(b));
	}
	
	/**
	 * Returns a string representing a%b
	 * 
	 * @param a
	 * @param b
	 * @return a - (b*(a/b))
	 */
	public static String mod(String a, String b) {
		
		return a + "-(" + b + "*(" + a + "/" + b + "))";
	}
	
	/**
	 * Returns the Remopla code representing
	 * <pre>
	 * goto to;
	 * </pre>
	 * 
	 * @param to the to label.
	 * @return the Remopla code.
	 */
	public static String gotoStmt(String to) {
		
		return "goto " + to;
	}
	
	/**
	 * Returns the Remopla code representing 
	 * <pre>
	 * label: goto label;
	 * </pre>
	 * 
	 * @param label the label.
	 * @return the Remopla code.
	 */
	public static String selfLoopStmt(String label) {
		
		StringBuilder out = new StringBuilder();
		
		out.append(label);
		out.append(": ");
		out.append(gotoStmt(label));
		out.append(";\n");
		
		return out.toString();
	}
}

