package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumMap;
import java.util.List;

import org.gjt.jclasslib.bytecode.AbstractInstruction;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

/**
 * Wraps an {@link org.gjt.jclasslib.bytecode.AbstractInstruction} and manages 
 * coverage data and transitions for special states.
 */
public class AnnotatedInstruction implements Comparable
{
	private AbstractInstruction instruction;
	private CoverageMethod method;
	
	private EnumMap<State, List<Transition>> transitionsMap;
	
	/**
	 * Initializes the instruction.
	 * 
	 * @param inst the <code>AbstractInstruction</code>.
	 * @param method the <code>CoverageMethod</code>.
	 */
	public AnnotatedInstruction(AbstractInstruction inst, CoverageMethod method)
	{
		this.instruction = inst;
		this.method = method;
	}
	
	/**
	 * Returns the underlying <code>CoverageMethod</code>.
	 * 
	 * @return the <code>CoverageMethod</code>.
	 */
	public CoverageMethod getMethod()
	{
		return method;
	}
	
	/**
	 * Returns the offset of the bytecode instruction.
	 * 
	 * @return the offset.
	 */
	public int getOffset()
	{
		return instruction.getOffset();
	}
	
	/**
	 * Returns the underlying <code>AbstractInstruction</code>.
	 * 
	 * @return the <code>AbstractInstruction</code>
	 */
	public AbstractInstruction getInstruction()
	{
		return instruction;
	}
	
	/**
	 * Returns the list of automaton transitions which have the given state. 
	 * 
	 * @param state the state
	 * @return the list of transitions.
	 */
	public List<Transition> getTransitions(State state)
	{
		if (transitionsMap != null) {
			List<Transition> list = transitionsMap.get(state);
			if (list != null) {
				return Collections.unmodifiableList(list);
			}
		}
		List<Transition> emptyList = Collections.emptyList();
		return emptyList;
	}
	
	/**
	 * Adds the automaton transition with state. The state that is not one of
	 * the exception states is considered as COVERED.
	 * 
	 * @param transition the transtion.
	 * @param state the state.
	 */
	void addTransition(Transition transition, State state) 
	{
		if (!state.isException()) {
			state = State.COVERED;
		}
		
		if (transitionsMap == null) {
			transitionsMap = new EnumMap<State, List<Transition>>(State.class);
		}
		List<Transition> list = transitionsMap.get(state);
		if (list == null) {
			list = new ArrayList<Transition>(2);
			transitionsMap.put(state, list);
		}
		if (!list.contains(transition)) {
			list.add(transition);	
		}
	}
	
	/**
	 * Returns <code>true</code> if the instruction is covered.
	 * 
	 * @return <code>true</code> if the instruction is covered.
	 */
	public boolean isCovered()
	{
		return transitionsMap != null;
	}
	
	/**
	 * Returns the Remopla label.
	 * 
	 * @return the Remopla label.
	 */
	public String getLabel()
	{
		int offset = getOffset(); 
		if (offset == 0) {
			return method.getMethod().getFormattedName();
		}
		else {
			return method.getMethod().getFormattedName() + offset;
		}
	}

	public int compareTo(Object o) 
	{
		if (o instanceof Integer) {
			return  instruction.getOffset() - (Integer)o;
		}
		return 0;
	}
	
	public String toString() {
		
		return getLabel();
	}
}