package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

/**
 * Abstract class for all native moped analyses. 
 * <p>
 * All native calls must synchronize on {@link Coverage#class}.
 * <p>
 * The abstract class provides the support for 
 * {@link de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageListener CoverageListeners}
 * and manages the {@link de.uni_stuttgart.fmi.szs.jmoped.PDS} instance common
 * to all analyses.
 */
public abstract class Coverage 
{

	static {
		System.loadLibrary("mopedcoverage");
	}
	
	private Map<String, CoverageMethod> methodsByLabel =
		new HashMap<String, CoverageMethod>();
	
	private MulticastCoverageListener listenerSupport = new MulticastCoverageListener();
	
	private PDS pds;
	
	private boolean notifyOfReachedLabels = false;
	
	private String filterMethodName;
	
	private CoverageResult result;
	
	private String assertErrorLabel;
	
	private String generalExceptionLabel;
	
	private String heapOverflowLabel;
	
	private String indexOutOfBoundsLabel;
	
	private String npeLabel;
	
	private String[] ignoredPrefixes = new String[4];
	
	private static Logger logger = Logger.getLogger(Coverage.class);
	
	protected Coverage(PDS pds)
	{
		this.pds = pds;
		assertErrorLabel = pds.getLabelAssertError();
		
		generalExceptionLabel = pds.getInfo().getLabelGeneralException();
		ignoredPrefixes[0] = generalExceptionLabel;
		
		heapOverflowLabel = pds.getLabelHeapOverflow();
		ignoredPrefixes[1] = heapOverflowLabel;
		
		indexOutOfBoundsLabel = pds.getInfo().getLabelIndexOutOfBounds();
		ignoredPrefixes[2] = indexOutOfBoundsLabel;
		
		npeLabel = pds.getLabelNPE();
		ignoredPrefixes[3] = npeLabel;
	}
	
	protected Coverage(CoverageResult result)
	{
		this(result.getPDS());
		this.result = result;
	}
	
	/**
	 * Returns the PDS instance that is used inernally to create the intermediate 
	 * Remopla code.
	 * <p>
	 * You can configure the tester through PDS before you start the tests.
	 * @return
	 */
	public PDS getPDS()
	{
		return pds;
	}
	
	public static void setMaximumMemory(long memoryInBytes)
	{
		synchronized (Coverage.class) {
			setMaximumMemoryNative(memoryInBytes);
		}
	}
	
	private static native void setMaximumMemoryNative(long memoryInBytes);
	
	/**
	 * 
	 * @return null if there is none
	 */
	protected CoverageResult getResult()
	{
		return result;
	}
	
	/**
	 * Returns the PDSInfo instance that is used internally to create the 
	 * intermediate Remopla code.
	 * <p>
	 * You can configure the tester through PDS before you start the tests.
	 * @return
	 */
	public PDSInfo getInfo()
	{
		return pds.getInfo();
	}
	
	protected void clearMethodsMap()
	{
		methodsByLabel.clear();
	}
	
	protected String[] getIgnoredPrefixes()
	{
		return ignoredPrefixes;
	}
	
	/**
	 * Adds a coverage listener that is notified of the test progress.
	 * @param listener
	 */
	public void addCoverageListener(CoverageListener listener)
	{
		listenerSupport.addCoverageListener(listener);
	}
	
	/**
	 * Removes the coverage listener.
	 * @param listener
	 */
	public void removeCoverageListener(CoverageListener listener)
	{
		listenerSupport.removeCoverageListener(listener);
	}
	
	protected MulticastCoverageListener getListenerSupport()
	{
		return listenerSupport;
	}
	
	protected CoverageMethod[] getMethods(String[] labels)
	{
		for (String label : labels) {
			// register all methods
			getMethod(label);
		}
		return methodsByLabel.values().toArray(new CoverageMethod[0]);
	}
	
	/**
	 * Looks up the method for which the label is used in the remopla output.
	 * @param label
	 * @return <code>null</code> if the method could not be found
	 */
	protected CoverageMethod getMethod(String label) 
	{
		String lookupLabel = CoverageUtils.getMethodLabel(label);
		CoverageMethod method = methodsByLabel.get(lookupLabel);
		if (method == null) {
			method = CoverageUtils.createMethod(pds, label);
			if (method != null) {
				methodsByLabel.put(lookupLabel, method);
			}
		}
		return method;
	}
	
	/**
	 * May return null.
	 */
	public String getFilterMethodName()
	{
		return filterMethodName;
	}
	
	public void setFilterMethodName(String method)
	{
		filterMethodName = method;
	}
	
	public void setEnabledNotifyofReachedLabels(boolean newValue)
	{
		notifyOfReachedLabels = newValue;
	}
	
	public boolean notifyOfReachedLabels()
	{
		return notifyOfReachedLabels;
	}
	
	/**
	 * Call this when a label is reached???
	 * 
	 * @param label
	 * @param from
	 * @param name
	 * @param to
	 */
	protected void fireCoveredLabel(String label, int from, int name, int to) 
	{
		logger.debug("label: " + label);
		
		if (notifyOfReachedLabels) {
			getListenerSupport().labelReached(label);
		}
		
		if (label.startsWith(heapOverflowLabel)) {
			fireState(State.HEAP_OVERFLOW,
					CoverageUtils.stripLabelPrefix(label, heapOverflowLabel), from, name, to);
		}
		else if (label.startsWith(npeLabel)) {
			fireState(State.NPE, CoverageUtils.stripLabelPrefix(label, npeLabel),
					from, name, to);
		}
		else if (label.startsWith(indexOutOfBoundsLabel)) {
			fireState(State.INDEX_OUT_OF_BOUNDS, 
					CoverageUtils.stripLabelPrefix(label, indexOutOfBoundsLabel), from, name, to);
		}
		else {
			CoverageMethod method = getMethod(label);
			int pc = CoverageUtils.extractOffset(label);
			if (method != null && pc != -1) {
				if (method.isFailedAssertionInstruction(pc)) {
					fireState(State.ASSERTION_FAILED, label, from, name, to);
				}
				else {
					int lineOffset = method.getLineOffset(pc);
					Transition transition = new Transition(from, name, to);
					method.addTransition(pc, transition, State.PARTIALLY_COVERED);
					State state = method.getRawLineState(lineOffset);
					getListenerSupport().coveredMethodOffset(method, lineOffset, pc, state,
							transition);
				}
			}
		}
	}
	
	/**
	 * Call this when something special is reached???
	 * 
	 * @param state
	 * @param methodLabel
	 * @param from
	 * @param name
	 * @param to
	 */
	protected void fireState(final State state, final String methodLabel, 
			int from, int name, int to)
	{
		CoverageMethod method = getMethod(methodLabel);
		int pc = CoverageUtils.extractOffset(methodLabel);
		if (method != null && pc != -1) {
			int lineOffset = method.getLineOffset(pc);
			Transition transition = new Transition(from, name, to);
			method.addTransition(pc, transition, state);
			getListenerSupport().coveredMethodOffset(method, lineOffset, pc, 
					state, transition);
		}
	}

}
