package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.MethodInfo;
import org.gjt.jclasslib.structures.attributes.CodeAttribute;
import org.gjt.jclasslib.structures.attributes.LineNumberTableAttribute;
import org.gjt.jclasslib.structures.attributes.LineNumberTableEntry;

import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;
import de.uni_stuttgart.fmi.szs.jmoped.BytecodeUtils;

public class CoverageMethod 
{

	public static int ASSERTION_STATEMENTS_COUNT = 4;
	
	private PDSMethod method;
	private LineNumberTableEntry[] lineTable;
	/**
	 * Array that keeps track of how many instructions are already covered for a line.
	 */
	private int[] coveredInstructions;
	/**
	 * Keeps track of the number of failed assertions per line.
	 * This could be rather sparse, maybe use list or hash map instead.
	 */
	private int[] failedAssertions;

	/**
	 * The state of an instruction.
	 */
	public static enum State  {
		
		NOT_COVERED("Not Covered"),
		PARTIALLY_COVERED ("Partially Covered"),
		COVERED("Covered"),
		ASSERTION_FAILED("Assertion Failure", true),
		HEAP_OVERFLOW("Analysis Heap Overflow", true),
		NPE("Null Pointer Exception", true),
		INDEX_OUT_OF_BOUNDS("Array Index Out of Bounds", true);
		
		private final String name;
		private final boolean isException;
		
		private State(String name) {
			
			this(name, false);
		}
		
		private State(String name, boolean isException) {
			
			this.name = name;
			this.isException = isException;
		}
		
		@Override
		public String toString() {
			
			return name;
		}
		
		/**
		 * Return <code>true</code> if the state is an exception, i.e. one of 
		 * the ASSERTION_FAILED, HEAP_OVERFLOW, NPE, INDEX_OUT_OF_BOUTNDS.
		 * 
		 * @return <code>true</code> if the state is an exception.
		 */
		public boolean isException() {
			
			return isException;
		}
	
	}
	
	private AnnotatedInstruction[] instructions;
	
	/**
	 * Maps a line offset to index in the <code>instructions</code> array, 
	 * where the bytecode instructions of this line begin.
	 */
	private int[] instructionsOffsets;
	
	private Line[] lines = null;
	
	/**
	 * 
	 * @throws NullPointerException if the bytecode of <code>method</code> doesn't
	 * have a line number table
	 */
	public CoverageMethod(PDSMethod method) {
		
		this.method = method;
		CodeAttribute code = 
			(CodeAttribute)method.getMethodInfo().findAttribute(CodeAttribute.class);
		LineNumberTableAttribute table = (LineNumberTableAttribute)
			code.findAttribute(LineNumberTableAttribute.class);
		lineTable = table.getLineNumberTable();
		coveredInstructions = new int[lineTable.length];
		failedAssertions = new int[lineTable.length];
		initialize();
	}

	/**
	 * Initializes the field <code>instructions</code> and 
	 * <code>instructionOffsets</code>.
	 */
	private void initialize() {
		
		List<AbstractInstruction> pdsInstructions = method.getInstructionList();
		this.instructions = new AnnotatedInstruction[pdsInstructions.size()];
		this.instructionsOffsets = new int[lineTable.length];
		Iterator<AbstractInstruction> iter = pdsInstructions.iterator();
		int lineOffset = 0;
		
		for (int i = 0; i < instructions.length; i++) {
			
			AbstractInstruction inst = iter.next();
			instructions[i] = new AnnotatedInstruction(inst, this);
			if (lineTable.length > lineOffset 
					&& lineTable[lineOffset].getStartPc() == inst.getOffset()) {
				instructionsOffsets[lineOffset] = i;
				lineOffset++;
			}
		}
	}
	
	/**
	 * Gets its <code>ClassFile</code> instance.
	 * 
	 * @return the <code>ClassFile</code> instance.
	 */
	public ClassFile getClassFile() {
		
		return method.getMethodInfo().getClassFile();
	}
	
	/**
	 * Returns <code>true</code> if the method returns nothing (void).
	 * 
	 * @return <code>true</code> if the method returns nothing.
	 */
	public boolean isVoid() {
		
		return PDSMethod.isVoid(method.getFormattedName());
	}
	
	/**
	 * Returns <code>true</code> if this method is the static class 
	 * initializer method.
	 * 
	 * @return <code>true</code> if the method is an static initializer.
	 */
	public boolean isClinit() {
		
		MethodInfo info = method.getMethodInfo();
		try {
			return "<clinit>()V".equals(info.getName() + info.getDescriptor());
		}
		catch (InvalidByteCodeException ivbce) {
		}
		return false;
	}
	
	/**
	 * Returns <code>true</code> if this method is the static class initializer method and it only
	 * initializes the static assertion variables.
	 * @return
	 */
	public boolean clinitIsAssertionOnly()
	{
		if (!isClinit()) {
			return false;
		}
		List<AbstractInstruction> instructions =  method.getInstructionList();
		int staticCount = 0;
		for (AbstractInstruction instruction : instructions) {
			if (instruction.getOpcode() == Opcodes.OPCODE_PUTSTATIC) {
				staticCount++;
				try {
					String[] names = BytecodeUtils.getReferencedName(
							method.getMethodInfo().getClassFile(), 
							instruction);
					if (!names[1].contains("assertionsDisabled")) {
						staticCount++;
					}
				}
				catch (InvalidByteCodeException ivbce) {
				}
			}
		}
		return staticCount == 1;
	}
	
	/**
	 * Returns true if the line <code>offset</code> contains an assertion.
	 * @param offset
	 * @return
	 */
	public boolean containsAssertion(int offset)
	{
		return getAssertionCount(offset) > 0;
	}
	
	/**
	 * Returns the number of <code>assert</code> statements in this line.
	 * @param offset
	 * @return
	 */
	public int getAssertionCount(int offset)
	{
		return getAssertionConstructors(offset).size();
	}
	
	private List<AnnotatedInstruction> getAssertionConstructors(int offset)
	{
		List<AnnotatedInstruction> list = new ArrayList<AnnotatedInstruction>(2);
		List<AnnotatedInstruction> instructions = getInstructions(offset);
		ClassFile classFile = method.getMethodInfo().getClassFile();
		for (AnnotatedInstruction i : instructions) {
			AbstractInstruction inst = i.getInstruction();
			try {
				if (BytecodeUtils.getsAssertionsDisabledField(classFile, inst)) {
					list.add(i);
				}
			}
			catch (InvalidByteCodeException ibce) {
			}
		}
		return list;
	}

	
	/**
	 * Returns true if the method is a constructor.
	 * @return
	 */
	public boolean isConstructor()
	{
		try {
			return method.getMethodInfo().getName().equals("<init>");
		}
		catch (InvalidByteCodeException e) {
		}
		return method.getFormattedName().contains("_init_");
	}
	
	public PDSMethod getMethod()
	{
		return method;
	}
	

	/**
	 * Returns true if all lineTable of this method have been fully covered.
	 * @return
	 */
	public boolean isFullyCovered()
	{
		for (int i = 0; i < lineTable.length; i++) {
			if (getLineState(i) != State.COVERED) {
				return false;
			}
		}
		return true;
	}
	
	@Override
	public String toString() 
	{
		return method.getFormattedName();
	}
	
	public AnnotatedInstruction getInstruction(int pc)
	{
		int offset = Arrays.binarySearch(instructions, pc);
		if (offset < 0) {
			throw new IndexOutOfBoundsException("pc not found: " + pc);
		}
		return instructions[offset];
	}
	
	void addTransition(int pc, Transition transition, State state)
	{
		AnnotatedInstruction instruction = getInstruction(pc);
		if (state == State.ASSERTION_FAILED 
				&& !instruction.getTransitions(State.ASSERTION_FAILED).contains(transition)) {
			int line = getLineOffset(pc);
			failedAssertions[line]++;
		}
		instruction.addTransition(transition, state);
	}
	
	/**
	 * Returns the line offset this bytecode instruction offset belongs to.
	 * 
	 * @throws IndexOutOfBoundsException if <code>pc</code> can't be found in
	 * the line table of this method
	 */
	public int getLineOffset(int pc)
	{
		for (int i = 0; i < lineTable.length; i++) {
			if (pc >= lineTable[i].getStartPc()) {
				if (i == lineTable.length - 1 || lineTable[i + 1].getStartPc() > pc) {
					return i;
				}
			}
		}
		throw  new IndexOutOfBoundsException("pc wrong " + pc);
	}
	
	/**
	 * Returns the current coverage state for the line at <code>offset</code>.
	 * @param offset
	 * @return
	 */
	public State getLineState(int offset)
	{
		List<AnnotatedInstruction> instructions = getInstructions(offset);
		int coveredCount = 0;
		for (AnnotatedInstruction inst : instructions) {
			if (inst.isCovered()) {
				++coveredCount;
			}
		}
		int diff = coveredCount - instructions.size();
		int assertionCount = getAssertionCount(offset);
		// only subtract assertion statements if other statements were reached
		if (assertionCount > 0 && coveredCount > 0) {
			diff += (assertionCount - failedAssertions[offset]) * ASSERTION_STATEMENTS_COUNT;
			diff += failedAssertions[offset];
		}
		if (diff >= 0) {
			return State.COVERED;
		}
		else if (diff == -instructions.size()) {
			return State.NOT_COVERED;
		}
		return State.PARTIALLY_COVERED;
	}
	
	/**
	 * Does not treat assertion lines differently.
	 * @param offset
	 * @return
	 */
	public State getRawLineState(int offset)
	{
		List<AnnotatedInstruction> instructions = getInstructions(offset);
		int coveredCount = 0;
		for (AnnotatedInstruction inst : instructions) {
			if (inst.isCovered()) {
				++coveredCount;
			}
		}
		int diff = coveredCount - instructions.size();
		List<AnnotatedInstruction> constructors = getAssertionConstructors(offset);
		for (AnnotatedInstruction constructor : constructors) {
			if (constructor.isCovered()) {
				diff += ASSERTION_STATEMENTS_COUNT - 1;
			}
		}
		if (diff >= 0) {
			return State.COVERED;
		}
		else if (diff == -instructions.size()) {
			return State.NOT_COVERED;
		}
		return State.PARTIALLY_COVERED;
	}
	
	/**
	 * Returns the number of covered instructions for the given line offset.
	 * @param offset
	 * @return
	 */
	public int getCoveredInstructions(int offset)
	{
		return coveredInstructions[offset];
	}
	
	/**
	 * Returns true if the line at <code>offset</code> contains an assertion statement
	 * that has failed in one possible execution path.
	 */
	public boolean containsFailedAssertion(int offset)
	{
		return failedAssertions[offset] > 0;
	}
	
	/**
	 * Returns true if the bytecode instruction at offset <code>pc</code> calls the 
	 * constructor of {@link AssertionError}.
	 */
	public boolean isFailedAssertionInstruction(int pc)
	{
		for (Iterator<AbstractInstruction> i = method.getInstructionList().iterator(); i.hasNext();) {
			AbstractInstruction inst = i.next();
			if (inst.getOffset() == pc) {
				if (inst.getOpcode() == Opcodes.OPCODE_INVOKESPECIAL) {
					try {
						String[] calledName = BytecodeUtils.getReferencedName
						(method.getMethodInfo().getClassFile(), inst);
						if (calledName[0].equals("java/lang/AssertionError")
								&& calledName[1].equals("<init>")) {
							return true;
						}
					}
					catch (InvalidByteCodeException ibce) {
					}
				}
				return false;
			}
		}
		return false;
	}
	
	/**
	 * Returns the number of bytecode instructions that match to the source code
	 * on the line at <code>offset</code>
	 * @param offset the index into the {@link LineNumberTableEntry} array.
	 * 
	 * @throws IndexOutOfBoundsException if the index is negative or too large
	 */
	public int getInstructionCount(int offset)
	{
		int end = offset + 1 < instructionsOffsets.length ? instructionsOffsets[offset + 1] 
		                                                                        : instructions.length;
		return end - instructionsOffsets[offset];
	}
	
	/**
	 * Returns the list of instructions that belong to the line at <code>offset</code>.
	 * @param offset
	 * @return
	 */
	public List<AnnotatedInstruction> getInstructions(int offset)
	{
		int end = offset + 1 < instructionsOffsets.length ? 
				instructionsOffsets[offset + 1] : instructions.length;
		List<AnnotatedInstruction> list = new ArrayList<AnnotatedInstruction>();
		
		for (int i = instructionsOffsets[offset]; i < end; i++) { 
			list.add(instructions[i]);
		}
		return list;
	}

	public LineNumberTableEntry[] getLineTable()
	{
		return lineTable;
	}

	/**
	 * Returns the line number in the source code corresponding to the offset
	 * in the line number table.
	 * 
	 * @throws IndexOutOfBoundsException if offset is negative or too large
	 */
	public int getLineNumber(int offset)
	{
		return lineTable[offset].getLineNumber();
	}
	
	/**
	 * Returns true if the line at <code>offset</code> is an effective line, which maps
	 * to a line visible in the source code.
	 */
	public boolean isEffectiveLine(int offset)
	{
		if (isConstructor()) {
			if (offset == 0) {
				return false;
			}
			
		}
		if (isVoid()) {
			if (offset == lineTable.length - 1 && getInstructionCount(offset) <= 1) {
				return false;
			}
		}
		return true;
	}
	
	public List<String> getUncoveredLabels()
	{
		ArrayList<String> list = new ArrayList<String>();
		for (AnnotatedInstruction inst : instructions) {
			if (!inst.isCovered()) {
				list.add(inst.getLabel());
			}
		}
		return list;
	}
	
	public List<AnnotatedInstruction> getAnnotatedInstructions()
	{
		return Collections.unmodifiableList(Arrays.asList(instructions));
	}
	
	public String getHumanReadableName()
	{
		return method.getHumanReadableName();
	}
	
	public String getClassName()
	{
		return method.getPDSClass().getName();
	}
	
	/**
	 * Returns the number of effective source code lineTable for this method.
	 * @return
	 */
	public int getLineCount() 
	{
		int count;
		if (isConstructor()) {
			// brace is on same line as last instruction
			if (getInstructionCount(lineTable.length - 1) > 1) {
				count = lineTable.length - 1;
			}
			else {
				count = lineTable.length - 2;
			}
		}
		else if (isVoid()) {
			// brace is on same line as last instruction
			if (getInstructionCount(lineTable.length - 1) > 1) {
				count = lineTable.length;
			}
			else {
				count = lineTable.length - 1;
			}
		}
		else {
			count = lineTable.length;
		}
		return Math.max(count, 0);
	}
	
	public Line[] getLines()
	{
		if (lines == null) {
			Line[] lines = new Line[lineTable.length];
			for (int i = 0; i < lines.length; i++) {
				lines[i] = new Line(this, i);
			}
			this.lines = lines;
		}
		return lines;
	}
	
}
