package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.io.File;
import java.io.IOException;

import org.gjt.jclasslib.structures.InvalidByteCodeException;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;
import de.uni_stuttgart.fmi.szs.moped.IntPds;

public class CoverageTester extends Coverage
{

	private String className;
	private String fileName;
	private File remoplaFile;
	
	/**
	 * Convenience construtor for {@link #CoverageTester(String, String[], String)
	 * CoverageTester(clazz.getName(), searchPaths, methodName)}.
	 */
	public CoverageTester(Class clazz, String[] searchPaths, String methodName, 
			String methodDescriptor, String fileName) 
		throws Exception
	{
		this(clazz.getName(), searchPaths, methodName, methodDescriptor, fileName);
	}
	
	/**
	 * Constructs a CoverageTester for class <code>className</code>.
	 * <p>
	 * 
	 * @param className should be the fully qualified name of the class as it is returned
	 * by {@link Class#getName()}.
	 * @param searchPaths array of names of directories and jar files that are searched
	 * for the class file of <code>className</code>
	 * @param methodName the method that should be tested for coverage. The format
	 * should match {@link PDSMethod#getFormattedName()}, can be <code>null</code>
	 * @throws Exception
	 */
	public CoverageTester(String className, String[] searchPaths, String methodName,
			String methodDescriptor, String fileName) 
		throws Exception
	{
		this(new PDS(className, searchPaths, methodName, methodDescriptor),
				className, methodName, methodDescriptor, fileName);
	}
	
	public CoverageTester(String className, String[] searchPaths, String fileName) throws Exception
	{
		this(className, searchPaths, null, null, fileName);
	}
	
	public CoverageTester(PDS pds, String className, String methodName, 
			String methodDescriptor, String fileName)
	{
		super(pds);
		this.className = className;
		this.fileName = fileName;
	}
	
	public File getRemoplaFile() {
		
		return remoplaFile;
	}
	
	/**
	 * Convenience constructor for {@link #CoverageTester(Class, String[], String),
	 * CoverageTester(clazz, new String[0], null)}.
	 * @param clazz
	 * @throws Exception
	 */
	public CoverageTester(Class clazz) throws Exception
	{
		this(clazz, null, null);
	}
	
	public CoverageTester(Class clazz, String methodName, String methodDescriptor)
		throws Exception
	{
		this(clazz, new String[0], methodName, methodDescriptor, null);
	}
	
	public CoverageTester(CoverageResult result)
	{
		super(result);
	}
	
	public CoverageMethod[] test()
	{
		return test(new NullProgressMonitor());
	}
	
	
	/**
	 * Tests the class' main method for coverage and returns the labels
	 * that have not been reached.
	 * <p>
	 * Listeners are notified of the progress of the test.
	 * 
	 * @param clazz the class must have a correct main method for now
	 * @throws  
	 */
	public CoverageMethod[] test(ProgressMonitor monitor)
	{
		CoverageListener listenerSupport = getListenerSupport();
		clearMethodsMap();
		
		listenerSupport.coverageStarted(getPDS());
		monitor.beginTask("Coverage Analysis", 10000);
		
		CoverageResult result = null;
		if (getResult() == null) {
			result = createRemoplaFileAndRun(new SubProgressMonitor(10000, monitor));
		}
		else {
			synchronized (Coverage.class) {
				result = analyseCoverage(getResult().getIntPds(), getFilterMethodName(), 
						new SubProgressMonitor(10000, monitor));
			}
		}
		
		if  (result == null) {
			listenerSupport.coverageFailed("Coverage was cancelled", 
					new InterruptedException());
		}
		else {
			result.setMethods(getMethods(new String[0]));
			result.setPDS(getPDS());
			listenerSupport.coverageEnded(result);
		}
		
		monitor.done();
		return getMethods(new String[0]);
	}
	
	private void unreachedLabels(String[] labels)
	{
		getMethods(labels == null ? new String[0] : labels);
	}
	
	private CoverageResult createRemoplaFileAndRun(ProgressMonitor monitor)
	{
		CoverageListener listenerSupport = getListenerSupport();
		remoplaFile = null;
		try {
			remoplaFile = CoverageUtils.writeRemopla(getPDS(), className, fileName);
		}
		catch (InvalidByteCodeException ibce) {
			listenerSupport.coverageFailed("Invalid byte code", ibce);
			monitor.done();
			return null;
		}
		catch (IOException ie) {
			listenerSupport.coverageFailed("Could not write temporary file", ie);
			monitor.done();
			return null;
		}
		CoverageResult result = null;
		synchronized (Coverage.class) {
			if (monitor.isCanceled()) {
				monitor.done();
				return null;
			}
			result = analyseCoverage(remoplaFile.getAbsolutePath(), getFilterMethodName(), 
					monitor);
		}
		return result;
	}

	private native CoverageResult analyseCoverage(String fileName, String method, 
			ProgressMonitor monitor);
	
	private native CoverageResult analyseCoverage(IntPds pds,
			String filterMethodName, ProgressMonitor monitor);
}
