package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import org.gjt.jclasslib.structures.InvalidByteCodeException;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;

/**
 * Class that provides static helper methods for handling covered labels of the
 * pushdown system.
  */
public class CoverageUtils 
{

	/**
	 * Creates a coverage method by looking up the {@link PDSMethod} which
	 * contains <code>label</code>
	 * @param pds
	 * @param label
	 * @return null if the pds method could not be found
	 */
	public static CoverageMethod createMethod(PDS pds, String label) 
	{
		label = getMethodLabel(label);
		PDSClass clazz = null;
		for (PDSClass c : pds.getPDSClasses()) {
			if (label.startsWith(c.getFormattedClassName())) {
				clazz = c;
				break;
			}
		}
		if (clazz == null) {
			return null;
		}
		for (PDSMethod m : clazz.getMethodList()) {
			if (label.startsWith(m.getFormattedName())) {
				return new CoverageMethod(m);
			}
		}
		return null;
	}

	/**
	 * Strips pc offset from the labe if there is one and returns the stripped label 
	 * which equals the formatted method name of {@link PDSMethod}.
	 * @param label
	 * @return
	 */
	public static String getMethodLabel(String label) 
	{
		int offset = getOffsetPosition(label);
		if (offset >= 0) {
			return label.substring(0, offset);
		} else {
			return label;
		}
	}

	/**
	 * Returns the position of the first digit of the offset number.
	 * @return -1 if there is none
	 */
	public static int getOffsetPosition(String label) {
		int offset = label.length() - 1;
		while (offset >= 0 && Character.isDigit(label.charAt(offset))) {
			offset--;
		}
		// at least one digit
		if (offset < label.length() - 1) {
			return offset + 1;
		}
		return -1;
	}
	
	/**
	 * Extracts the offset number at the end of the label string.
	 * 
	 * @param label
	 * @return o if there is no number
	 */
	public static int extractOffset(String label)
	{
		int offset = getOffsetPosition(label);
		if (offset >= 0) {
			return Integer.parseInt(label.substring(offset));
		}
		return 0;
	}
	
	/**
	 * Strips the first <code>prefix.length()</code> characters from <code>label</code>.
	 * 
	 * @param label
	 * @param prefix
	 * @return
	 */
	public static String stripLabelPrefix(String label, String prefix)
	{
		return label.substring(prefix.length());
	}
	
	public static File writeRemopla(PDS pds, String fileName)
			throws InvalidByteCodeException, IOException
	{
		return writeRemopla(pds, fileName, null);
	}
	
	/**
	 * Writes the Remopla code of <code>pds</code> to a unique temporary file having 
	 * the prefix name <code>fileName</code>.
	 *
	 * @return the temporary file on success
	 * @throws InvalidByteCodeException if the Remopla code could not be generated
	 * @throws IOException if there was an error writing to the temporary file
	 */
	public static File writeRemopla(PDS pds, String className, String fileName)
			throws InvalidByteCodeException, IOException
	{
		pds.translate();
		String mopedInput = pds.toRemopla();
		BufferedWriter out = null;
		File tmpFile;
		
		try {
			if (className.length() == 0) {
				throw new IllegalArgumentException("empty filename");
			}
			while (className.length() < 3) {
				className += className;
			}
			
			if (fileName == null) 
				tmpFile = File.createTempFile(className, ".rem");
			else tmpFile = new File(fileName);
			
//			tmpFile.deleteOnExit();
			out = new BufferedWriter(new FileWriter(tmpFile));
			out.write(mopedInput);
			return tmpFile;
		} 
		finally {
			if (out != null) {
				try {
					out.close();
				} 
				catch (IOException ie) {
				}
			}
		}
	}
	
	public static CoverageMethod getMethod(CoverageMethod[] methods, 
			String stackSymbol)
	{
		for (CoverageMethod method: methods) {
			if (stackSymbol.startsWith(method.getMethod().getFormattedName())) {
				return method;
			}
		}
		return null;
	}
}
