package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.Arrays;

import de.uni_stuttgart.fmi.szs.moped.IntPds;

/**
 * Peer class that keeps a handle to the native <code>intConfig</code> struct, 
 * and caches all values retrieved from the native structure.
 * 
 * TODO move this class to moped package.
 */
public class IntConfig 
{
	
	static {
		System.loadLibrary("mopedcoverage");
	}
	
	private final int handle;
	
	private String stackSymbol = null;
	
	/**
	 * IntConfigs created from a previous point to it, so it is not deleted on the
	 * native side as long as references to the children exist.
	 */
	private final IntConfig previous;

	private IntConfig next = null;
	
	private boolean hasNext = true;
	
	private int stackLength = -1;
	
	private IntVariable[] stackVars = null;
	
	private IntVariable[] globalVars = null;
	
	private final IntPds pds;
	
	 IntConfig(int handle, IntPds pds)
	{
		this.handle = handle;
		this.previous = null;
		this.pds = pds;
	}
	
	IntConfig(int handle, IntConfig previous)
	{
		this.handle = handle;
		this.previous = previous;
		this.pds = null;
	}
	
	/**
	 * Gets all global variables of this configuration.
	 * 
	 * @return the global variables.
	 */
	public IntVariable[] getGlobalVariables()
	{
			synchronized (Coverage.class) {
				if (globalVars == null) {
					globalVars = getGlobalVariablesNative();
					if (globalVars == null) {
						globalVars = new IntVariable[0];
					}
				}
			}
			return globalVars;
	}
	
	private native IntVariable[] getGlobalVariablesNative();
	
	/**
	 * Gets all local variables at the top-of-stack of this configuration.
	 * 
	 * @return the local variables at the top-of-stack.
	 */
	public IntVariable[] getStackVariables()
	{
		synchronized (Coverage.class) {
			if (stackVars == null) {
				stackVars = getStackVariablesNative();
				if (stackVars == null) {
					stackVars = new IntVariable[0];
				}
			}
		}
		return stackVars;
	}
	
	private native IntVariable[] getStackVariablesNative();
	
	/**
	 * Calls {@link #getNextConfig() getNextConfig()} until label is
	 * the top-of-stack symbol.
	 * 
	 * @param label the label.
	 * @return the config with label as the top-of-stack symbol.
	 */
	public IntConfig getConfigAt(String label)
	{
		if (getStackSymbol().equals(label)) {
			return this;
		}
		
		IntConfig nextConfig = getNextConfig();
		if (nextConfig != null) {
			return nextConfig.getConfigAt(label);
		}
		
		return null;
	}
	
	private native IntConfig getConfigAtNative(String label);

	/**
	 * Gets the top-of-stack symbol of this config.
	 * 
	 * @return the top-of-stack symbol.
	 */
	public String getStackSymbol()
	{
		synchronized (Coverage.class) {
			if (stackSymbol == null) {
				stackSymbol = getStackSymbolNative();
			}
		}
		return stackSymbol;
	}
	
	private native String getStackSymbolNative();
	
	/**
	 * Gets next config in the counterexample.
	 * 
	 * @return the next config, or <code>null</code> if there is none.
	 */
	public IntConfig getNextConfig()
	{
		synchronized (Coverage.class) {
			if (next == null && hasNext) {
				next = getNextConfigNative();
				hasNext = false;
			}
		}
		return next;
	}
	
	/**
	 * Gets the previous config in the counterexample.
	 * 
	 * @return the previous config, or <code>null</code> 
	 * 		if this the start configuration.
	 */
	public IntConfig getPreviousConfig()
	{
		return previous;
	}
	
	private native IntConfig getNextConfigNative();
	
	/**
	 * Returns the stack length of this config.
	 * 
	 * @return the stack length.
	 */
	public int getStackLength()
	{
		synchronized (Coverage.class) {
			if (stackLength == -1) {
				stackLength = getStackLengthNative();
			}
		}
		return stackLength;
	}
	
	private native int getStackLengthNative();
	
	@Override
	public boolean equals(Object obj) 
	{
		if (obj instanceof IntConfig) {
			IntConfig other = (IntConfig)obj;
			String symbol = getStackSymbol();
			String otherSymbol = other.getStackSymbol();
			// TODO is hard coded here, see coverage.c as well
			if (!symbol.equals(otherSymbol) 
					&& !(symbol.startsWith("DEFAULT") && otherSymbol.startsWith("DEFAULT"))) {
				return false;
			}
			if (getStackLength() != other.getStackLength()) {
				return false;
			}
			if (!Arrays.equals(getStackVariables(), other.getStackVariables())
				|| !Arrays.equals(getGlobalVariables(), other.getGlobalVariables())) {
				return false;
			}
			return true;
		}
		return false;
	}
	
	@Override
	public String toString() 
	{
		StringBuilder builder = new StringBuilder();
		builder.append("Label: ").append(getStackSymbol());
		builder.append(", Depth: ").append(getStackLength()).append("\n");
		builder.append("Global Variables: \n");
		builder.append(Arrays.toString(getGlobalVariables()));
		builder.append("\nStack Variables: \n");
		builder.append(Arrays.toString(getStackVariables()));
		builder.append("\n");
		return builder.toString();
	}
}
