package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.BranchInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;

import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo;
import de.uni_stuttgart.fmi.szs.jmoped.BytecodeUtils;

public class JavaAssertionExpressionResolver 
{
	public static String getFailedAssertionExpression(CoverageResult result,
			Transition transition) throws InvalidByteCodeException
	{
		IntConfig config = getLastConfig(result, transition);
		CoverageMethod method = CoverageUtils.getMethod(result.getMethods(),
				config.getStackSymbol());
		
		InstructionConfigPair[] comparisons = getComparisons(config, method);
		BranchInstruction[] instructions = getBranchInstructions(config, method);
		
		PDSInfo info = result.getPDS().getInfo();
		
		if (comparisons.length == 0) {
			return Boolean.FALSE.toString();
		}
		else if (comparisons.length == 1) {
			if (isOr(instructions)) {
				return getComparison(comparisons[0], false, info);
			}
			else if (isAnd(instructions)) {
				return getComparison(comparisons[0], true, info);
			}
			else {
				return createMixedExpression(comparisons, instructions, info);
			}
		}
		else if (isAnd(instructions)) {
			StringBuilder builder = new StringBuilder();
			for (int i = 0; i < comparisons.length - 1; i++) {
				builder.append(getComparison(comparisons[i], true, info));
				builder.append(" && ");
			}
			builder.append(getComparison(comparisons[comparisons.length - 1], false, info));
			return builder.toString();
		}
		else if (isOr(instructions)) {
			StringBuilder builder = new StringBuilder();
			for (int i = 0; i < comparisons.length - 1; i++) {
				builder.append(getComparison(comparisons[i], false, info));
				builder.append(" || ");
			}
			builder.append(getComparison(comparisons[comparisons.length - 1], false, info));
			return builder.toString();
		}
		else {
			// mixed case
			return createMixedExpression(comparisons, instructions, info);
		}
	}
	
	private static String createMixedExpression(InstructionConfigPair[] comparisons,
			BranchInstruction[] instructions, PDSInfo info) 
	{
		StringBuilder builder = new StringBuilder();
		int maxOffset = getMaxJumpTarget(instructions);
	
		for (int i = 0; i < comparisons.length; i++) {
			BranchInstruction branchInst = comparisons[i].instruction;
			int address = branchInst.getOffset() + branchInst.getBranchOffset(); 
			if (address < maxOffset) {
				builder.append(getComparison(comparisons[i], true, info));
				if (i < comparisons.length - 1) {
					if (address < comparisons[i + 1].instruction.getOffset()) {
						builder.append(" || ");
					}
					else {
						builder.append(" && ");
					}
				}
			}
			else {
				builder.append(getComparison(comparisons[i], false, info));
				if (i < comparisons.length - 1) {
					builder.append(" || ");
				}
			}
		}
		return builder.toString();
	}
	
	private static int getMaxJumpTarget(BranchInstruction[] instructions)
	{
		int maxOffset = -1;
		for (BranchInstruction inst: instructions) {
			maxOffset = Math.max(maxOffset, inst.getOffset() + inst.getBranchOffset());
		}
		return maxOffset;
	}

	private static String getComparison(InstructionConfigPair comparison, boolean inverse, 
			PDSInfo info)
	{
		IntConfig config = comparison.config;
		BranchInstruction instruction = comparison.instruction;
		
		IntVariable arg0 = JavaMethodArgumentsResolver.findVariable(config
				.getStackVariables(), info.getStackName() + 0);
		IntVariable arg1 = JavaMethodArgumentsResolver.findVariable(config
				.getStackVariables(), info.getStackName() + 1);
		
		int value0;
		int value1;
		if (isZeroComparison(instruction)) {
			value0 =  0 ;
			value1 = arg0.intValue();
		}
		else {
			value0 = arg0.intValue();
			value1 = arg1.intValue();
		}

		StringBuilder builder = new StringBuilder();
		
		builder.append(value1);
		builder.append(' ');
		builder.append(inverse ? getInverseComparator(instruction) : getComparator(instruction));
		builder.append(' ');
		builder.append(value0);
		
		return builder.toString();
	}
	
	private static boolean isAnd(BranchInstruction[] comparisons) 
	{
		BranchInstruction last = comparisons[comparisons.length - 1];
		int lastOffset = last.getOffset() + last.getBranchOffset();
		
		int maxOffset = 0;
		for (int i = 0; i < comparisons.length - 1; i++) {
			BranchInstruction inst = comparisons[i];
			maxOffset = Math.max(maxOffset, inst.getOffset() + inst.getBranchOffset());
		}
		return maxOffset < lastOffset;
	}
	
	private static boolean isOr(BranchInstruction[] comparisons)
	{
		int offset = -1;
		for (int i = 0; i < comparisons.length; i++) {
			BranchInstruction inst = comparisons[i];
			if (offset == -1) {
				offset = inst.getOffset() + inst.getBranchOffset();
			}
			else {
				if (offset != inst.getOffset() + inst.getBranchOffset()) {
					return false;
				}
			}
		}
		return true;
	}
	
	public static IntConfig getLastConfig(CoverageResult result, Transition transition)
	{
		IntConfig config = TraceGenerator.getTraceConfig(result, transition);

		// move to last config
		while (config.getNextConfig() != null) {
			config = config.getNextConfig();
		}
		return config;
	}
	
	private static BranchInstruction[] getBranchInstructions(IntConfig config,
			CoverageMethod method) throws InvalidByteCodeException
	{
		AnnotatedInstruction lastInst = getInstruction(method, config);
		ClassFile classFile = method.getClassFile();
		int lineOffset = method.getLineOffset(lastInst.getOffset());
		List<AnnotatedInstruction> instructions = method.getInstructions(lineOffset);
		List<BranchInstruction> result = new ArrayList<BranchInstruction>();
		boolean skipNextBranch = true;
		
		for (AnnotatedInstruction inst : instructions) {
			if (inst == lastInst) {
				break;
			}
			if (BytecodeUtils.getsAssertionsDisabledField(classFile, inst.getInstruction())) {
				result.clear();
				skipNextBranch = true;
			}
			if (isComparison(inst.getInstruction())) {
				if (skipNextBranch) {
					skipNextBranch = false;
				}
				else {
					result.add((BranchInstruction)inst.getInstruction());
				}
			}
		}
		return result.toArray(new BranchInstruction[result.size()]);
	}
	
	private static InstructionConfigPair[] getComparisons(IntConfig config,
			CoverageMethod method) throws InvalidByteCodeException
	{
		ClassFile classFile = method.getClassFile();
		String methodName = method.getMethod().getFormattedName();
		int stackLength = config.getStackLength();
		LinkedList<InstructionConfigPair> list = new LinkedList<InstructionConfigPair>();

		
		// collect all comparisons
		for (; config != null; config = config.getPreviousConfig()) {
			// ignore methods that are called in the expression
			if (config.getStackLength() != stackLength 
					|| !config.getStackSymbol().startsWith(methodName)) {
				continue;
			}
			AnnotatedInstruction instruction = getInstruction(method, config);
			if (instruction == null) {
				break;
			}
			if (isComparison(instruction.getInstruction())) {
				IntConfig previous = config.getPreviousConfig();
				if (previous.getStackLength() != stackLength 
						|| !previous.getStackSymbol().startsWith(methodName)) {
					list.add(0, 
							new InstructionConfigPair((BranchInstruction) instruction.getInstruction(), 
									config));
				}
				else {
					AnnotatedInstruction prevInstruction = getInstruction(method, previous);
					if (prevInstruction != null) {
						// check if it's not the comparison that checks if assertions are enabled
						if (!BytecodeUtils.getsAssertionsDisabledField(classFile, 
								prevInstruction.getInstruction())) {
							// push to front, since we're going backwards
							list.add(0, 
									new InstructionConfigPair((BranchInstruction) instruction.getInstruction(), 
											config));
						}
						else {
							break;
						}
					}
				}
			}
		}
		return list.toArray(new InstructionConfigPair[list.size()]);		
	}
	
	private static AnnotatedInstruction getInstruction(CoverageMethod method, 
			IntConfig config)
	{
		int pc = CoverageUtils.extractOffset(config.getStackSymbol());
		return method.getInstruction(pc);
	}
	
	public static boolean isZeroComparison(AbstractInstruction instruction)
	{
		return isZeroComparison(instruction.getOpcode());
	}
	
	public static boolean isZeroComparison(int opcode) 
	{
		return opcode >= Opcodes.OPCODE_IFEQ && opcode <= Opcodes.OPCODE_IFLE;
	}
	
	/**
	 * Returns true if the instruction is a comparison.
	 */
	public static boolean isComparison(AbstractInstruction instruction)
	{
		return BytecodeUtils.isComparison(instruction.getOpcode());
	}

	public static String getComparator(AbstractInstruction inst)
	{
		return getComparator(inst.getOpcode());
	}
	
	public static String getComparator(int opcode) 
	{
		switch (opcode) {
		case Opcodes.OPCODE_IF_ACMPEQ:
			return "==";
		case Opcodes.OPCODE_IF_ACMPNE:
			return "!=";
		case Opcodes.OPCODE_IF_ICMPEQ:
			return "==";
		case Opcodes.OPCODE_IF_ICMPGE:
			return ">=";
		case Opcodes.OPCODE_IF_ICMPGT:
			return ">";
		case Opcodes.OPCODE_IF_ICMPLE:
			return "<=";
		case Opcodes.OPCODE_IF_ICMPLT:
			return "<";
		case Opcodes.OPCODE_IF_ICMPNE:
			return "!=";
		case Opcodes.OPCODE_IFEQ:
			return "==";
		case Opcodes.OPCODE_IFGE:
			return ">=";
		case Opcodes.OPCODE_IFGT:
			return ">";
		case Opcodes.OPCODE_IFLE:
			return "<=";
		case Opcodes.OPCODE_IFLT:
			return "<";
		case Opcodes.OPCODE_IFNE:
			return "!=";
		case Opcodes.OPCODE_IFNONNULL:
			return "!=";
		case Opcodes.OPCODE_IFNULL:
			return "==";
		}
		return " ";
	}
	

	public static String getInverseComparator(AbstractInstruction inst)
	{
		return getInverseComparator(inst.getOpcode());
	}
	
	public static String getInverseComparator(int opcode) 
	{
		switch (opcode) {
		case Opcodes.OPCODE_IF_ACMPEQ:
			return "!=";
		case Opcodes.OPCODE_IF_ACMPNE:
			return "==";
		case Opcodes.OPCODE_IF_ICMPEQ:
			return "!=";
		case Opcodes.OPCODE_IF_ICMPGE:
			return "<";
		case Opcodes.OPCODE_IF_ICMPGT:
			return "<=";
		case Opcodes.OPCODE_IF_ICMPLE:
			return ">";
		case Opcodes.OPCODE_IF_ICMPLT:
			return ">=";
		case Opcodes.OPCODE_IF_ICMPNE:
			return "==";
		case Opcodes.OPCODE_IFEQ:
			return "!=";
		case Opcodes.OPCODE_IFGE:
			return "<";
		case Opcodes.OPCODE_IFGT:
			return "<=";
		case Opcodes.OPCODE_IFLE:
			return ">";
		case Opcodes.OPCODE_IFLT:
			return ">=";
		case Opcodes.OPCODE_IFNE:
			return "==";
		case Opcodes.OPCODE_IFNONNULL:
			return "==";
		case Opcodes.OPCODE_IFNULL:
			return "!=";
		}
		return " ";
	}
	
	private static class InstructionConfigPair
	{
		BranchInstruction instruction;
		IntConfig config;
		
		public InstructionConfigPair(BranchInstruction instruction, IntConfig config) 
		{
			this.instruction = instruction;
			this.config = config;
		}
	}
}
