package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import org.apache.log4j.Logger;
import org.gjt.jclasslib.bytecode.AbstractInstruction;
import org.gjt.jclasslib.bytecode.ImmediateByteInstruction;
import org.gjt.jclasslib.bytecode.MultianewarrayInstruction;
import org.gjt.jclasslib.bytecode.Opcodes;
import org.gjt.jclasslib.structures.CPInfo;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;
import org.gjt.jclasslib.structures.constants.ConstantStringInfo;
import org.gjt.jclasslib.structures.constants.ConstantUtf8Info;

import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInst;

/**
 * This class provides means to determine the number of missing heap fields for 
 * an overflow. This can give a clue to the user how much to enlarge the heap.
 */
public class JavaHeapOverflowResolver 
{

	private static Logger logger = Logger.getLogger(JavaHeapOverflowResolver.class);
	
	/**
	 * Returns the number of heap fields that were missing for instantiating the class or
	 * array at the point the transition points to.
	 * 
	 * @param result
	 * @param transition the transition with state 
	 * {@link de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State#HEAP_OVERFLOW}
	 * @return
	 * @throws InvalidByteCodeException 
	 */
	public static int getMissingHeapFields(CoverageResult result, Transition transition) 
		throws InvalidByteCodeException 
	{
		IntConfig config = TraceGenerator.getTraceConfig(result, transition);
		PDSInfo info = result.getPDS().getInfo();

		// move to last config
		while (config.getNextConfig() != null) {
			config = config.getNextConfig();
		}
		
		CoverageMethod method = CoverageUtils.getMethod(result.getMethods(), 
				CoverageUtils.stripLabelPrefix(config.getStackSymbol(), info.getLabelHeapOverflow()));
		
		int pcOffset = CoverageUtils.extractOffset(config.getStackSymbol());
		
		AnnotatedInstruction instruction =	method.getInstruction(pcOffset);;
		AbstractInstruction inst = instruction.getInstruction();

		IntVariable heapPtr =
			JavaMethodArgumentsResolver.findVariable(config.getGlobalVariables(), 
					info.getHeapPtrName());
		
		IntVariable heap = 
			JavaMethodArgumentsResolver.findVariable(config.getGlobalVariables(),
					info.getHeapName());
		
		int size = 
			getSize(method.getMethod().getMethodInfo().getClassFile(),	inst, config, info);
		
		int index = heapPtr.intValue();
		int diff = size - (heap.arrayValue().length -  index);
		return diff;
	}
	
	private static int getSize(ClassFile classFile, AbstractInstruction inst, 
			IntConfig config, PDSInfo info) throws InvalidByteCodeException
	{
		if (inst.getOpcode() == Opcodes.OPCODE_NEW) {
			String className = PDSInst.getNewedClassName(classFile, inst);
			// +1 for class type
			return info.getClassSize(className) + 1;	
		}
		else if (inst.getOpcode() == Opcodes.OPCODE_NEWARRAY ||
				inst.getOpcode() == Opcodes.OPCODE_ANEWARRAY) {
			IntVariable stackTop = 
				JavaMethodArgumentsResolver.findVariable(config.getStackVariables(),
						info.stackIndex(0));
			// +1 for array length
			return stackTop.intValue() + 1;
		}
		else if (inst.getOpcode() == Opcodes.OPCODE_MULTIANEWARRAY) {
			final int dim = ((MultianewarrayInstruction)inst).getDimensions();
			int[] sizes = new int[dim];
			for (int i = 0; i < dim; i++) {
				IntVariable s_i = 
					JavaMethodArgumentsResolver.findVariable(config.getStackVariables(),
							info.stackIndex(i));
				sizes[i] = s_i.intValue();
			}
			return getMultiArraySize(sizes);
		} else if (inst.getOpcode() == Opcodes.OPCODE_LDC) { // String
			int immByte = ((ImmediateByteInstruction) inst).getImmediateByte();
			CPInfo[] cpInfo = classFile.getConstantPool();
			ConstantStringInfo csInfo = (ConstantStringInfo) cpInfo[immByte];
			ConstantUtf8Info cuInfo = (ConstantUtf8Info) cpInfo[csInfo.getStringIndex()];
			logger.debug("cuInfo.getString(): " + cuInfo.getString());
			int length = cuInfo.getString().length();
			logger.debug("cuInfo.getString().length(): " + cuInfo.getString().length());
			
			return length + 1;
		}
		throw new IllegalStateException("No size could be determinted for instruction of type "
				+ inst.getOpcodeVerbose());
	}

	// XXX code duplication to PDSInst.createMultiArraySizeExpression
	static int getMultiArraySize(int sizes[])
	{
		final int dim = sizes.length;
		int size = 0;
		for (int i = dim - 1; i >= 0; i--) {
			int product = 1;
			for (int j = dim - 1; j > i; j--) {
				product *= sizes[j];
			}
			size += product * (sizes[i] + 1); 
		}
		return size;
	}

}
