package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;
import org.gjt.jclasslib.structures.ClassFile;
import org.gjt.jclasslib.structures.InvalidByteCodeException;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethodWrapper;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo.HeapOption;

/**
 * This class resolves the Java method call trace from IntConfig counter examples. 
 * <p>
 * It maps the
 * {@link de.uni_stuttgart.fmi.szs.jmoped.coverage.IntConfig#getStackSybmol() labels}
 * of the intconfigs back to the methods in {@link de.uni_stuttgart.fmi.szs.jmoped.PDS}.
 * <p>
 * 
 * TODO 
 * The curent implementation uses instance methods but doesn't manage any state.
 * The pds parameter could be given in the constructor.
 */
public class JavaMethodArgumentsResolver 
{
	
	static Logger logger = Logger.getLogger(JavaMethodArgumentsResolver.class);

	/**
	 * Returns the wood of trace trees.
	 * <p>
	 * This should usually be called on the first IntConfig of a counterexample, 
	 * the one returned by {@link PostTraceGenerator}.
	 * 
	 * @param config 
	 * @param pds the push down system to look up the java methods.
	 * @return
	 */
	public TraceNode[] resolveMethodCallTraceTree(IntConfig config, PDS pds)
	{
		LinkedList<TraceNode> children = new LinkedList<TraceNode>();
		config = resolveChildCalls(config, pds, null, children);
		for (TraceNode child : children)
			logger.debug("child: " + child.getMethodCall().getMethod());
		return children.toArray(new TraceNode[children.size()]);
	}
	
	
	private IntConfig resolveChildCalls(IntConfig config, PDS pds, 
			TraceNode parent, List<TraceNode> children)
	{
		if (config == null) {
			return null;
		}
		
		logger.debug("entering, config.getStackSymbol(): " + config.getStackSymbol());
		int stackOffset = config.getStackLength();
		config = getAnyNextMethodCallConfig(config, pds);
		
		while (config != null) {
			if (config.getStackLength() == stackOffset) {
				
				String symbol = config.getStackSymbol();
				logger.debug("symbol: " + symbol);
				
				PDSMethodWrapper wrapper = pds.getPDSMethodWrapper();
				if (wrapper != null && wrapper.getName().equals(symbol)) {
					// short circuit so the method that is called non-deterministically
					// shows up on the top level
					logger.debug("wrapper found");
					return resolveChildCalls(getAnyNextMethodCallConfig
							(config.getNextConfig(), pds), pds, parent, children);
				}
				
				PDSMethod method = pds.getPDSMethod(symbol);
				logger.debug("method: " + method);
				if (method != null) {
					
					MethodCall call = resolveMethodCall(config, method, pds);
					
					config = getAnyNextMethodCallConfig(config.getNextConfig(), pds);
					logger.debug(String.format("length: %d, stackOffset: %d",
							config != null ? config.getStackLength() : -1, 
							stackOffset));
					if (config == null ||config.getStackLength() < stackOffset) {
						
						logger.debug("config is null, return");
						children.add(new DefaultTraceNode(call, parent, 
								new TraceNode[0]));
						return config;
					}
					else if (config.getStackLength() == stackOffset) {
						
						children.add(new DefaultTraceNode(call, parent, 
								new TraceNode[0]));
					}
					else /*if (config.getStackLength() > stackOffset) */{
						
						DefaultTraceNode node = new DefaultTraceNode(call, 
								parent, null);
						LinkedList<TraceNode> kids = 
							new LinkedList<TraceNode>();
						config = resolveChildCalls(config, pds, node, kids);
						node.setChildren(kids.toArray(new TraceNode[0]));
						children.add(node);
					}
				}
				else {
					config = 
						getAnyNextMethodCallConfig(config.getNextConfig(), pds);
				}
			}
			else {
				if (config.getStackLength() > stackOffset) {
					throw new IllegalStateException(config.toString());
				}
				return config;
			}
		}
		return null;
	}
	
	/**
	 * Returns a flat array of method calls on the whole config path.
	 *
	 * @param config
	 * @param pds
	 * @return
	 */
	public MethodCall[] resolveMethodCallTrace(IntConfig config, PDS pds)
	{
		ArrayList<MethodCall> calls = new ArrayList<MethodCall>();

		while (config != null) {
			
			PDSMethod method = pds.getPDSMethod(config.getStackSymbol());
			if (method != null/* && method != pds.getInitialMethodWrapper()*/) {
				calls.add(resolveMethodCall(config, method, pds));
			}
			config = config.getNextConfig();
		}
		return calls.toArray(new MethodCall[calls.size()]);
	}
	
	/**
	 * Returns the first occurrence of a call to <code>method</code> 
	 * on the config path.
	 * 
	 * @param config
	 * @param method
	 * @param pds
	 * @return
	 */
	public MethodCall resolveMethodCall(IntConfig config, PDSMethod method, PDS pds)
	{
		if (method.isStatic()) {
			return resolveStaticMethodCall(config, method, pds);
		}
		else {
			return resolveInstanceMethodCall(config, method, pds);
		}
	}
	
	/**
	 * Returns all method calls on the path from config.
	 * 
	 * @return empty array if method is not called on the config path.
	 */
	public MethodCall[] resolveMethodCalls(IntConfig config, 
			PDSMethod method, PDS pds) 
	{
		ArrayList<MethodCall> calls = new ArrayList<MethodCall>(5);
		boolean isStatic = method.isStatic();
		IntConfig callConfig = getNextMethodCallConfig(config, method);
		
		while (callConfig != null) {
			if (isStatic) {
				calls.add(resolveStaticMethodCall(callConfig, method, pds));
			}
			else {
				calls.add(resolveInstanceMethodCall(callConfig, method, pds));
			}
			
			callConfig = callConfig.getNextConfig();
			
			if (callConfig != null) {
				callConfig = getNextMethodCallConfig(callConfig, method);
			}
		}
		return calls.toArray(new MethodCall[calls.size()]);
	}
	
	private MethodCall resolveStaticMethodCall(IntConfig callConfig, 
			PDSMethod method, PDS pds) 
	{
		PDSInfo info = pds.getInfo();
		IntVariable[] stackVars = callConfig.getStackVariables();
		IntVariable[] globalVars = callConfig.getGlobalVariables();
		Parameter[] params = DescriptorParser.parse(method.getDescriptor());
		
		ArrayList<MethodArgument> arguments = new ArrayList<MethodArgument>();
		
		for (int i = 0; i < params.length; i++) {
			Parameter param = params[i];
			if (param.isValue()) {
				arguments.add(new IntVariableMethodArgument
						(param, findLocalVariable(stackVars, i, info)));
			}
			else if (param.isValueArray()) {
				arguments.add(new ValueArrayMethodArgument
							  (param, getArray(globalVars, 
											   findLocalVariable(stackVars, i, 
																 info), info)));
			}
			else if (param.isReference()) {
				arguments.add(new ReferenceMethodArgument
							  (param, findLocalVariable(stackVars, i, info), 
							   callConfig, pds));
			}
		}
		return new MethodCallImpl(callConfig, 
				arguments.toArray(new MethodArgument[0]), method, null);
	}
	
	private MethodCall resolveInstanceMethodCall(IntConfig callConfig,
			PDSMethod method, PDS pds) 
	{
		PDSInfo info = pds.getInfo();
		IntVariable[] stackVars = callConfig.getStackVariables();
		IntVariable[] globalVars = callConfig.getGlobalVariables();
		Parameter[] params = DescriptorParser.parse(method.getDescriptor());
		
		ArrayList<MethodArgument> arguments = new ArrayList<MethodArgument>(3);
		
		for (int i = 0; i < params.length; i++) {
			Parameter param = params[i];
			if (param.isValue()) {
				arguments.add(new IntVariableMethodArgument
						(param, findLocalVariable(stackVars, i + 1, info)));
			}
			else if (param.isValueArray()) {
				arguments.add(new ValueArrayMethodArgument
						(param, getArray(globalVars, 
										 findLocalVariable(stackVars, i + 1, 
														   info), info)));
			}
			else if (param.isReference()) {
				arguments.add(new ReferenceMethodArgument(param, 
						findLocalVariable(stackVars, i + 1, info), callConfig, 
														  pds));
			}
		}
		ReferenceMethodArgument ref = new ReferenceMethodArgument(
				getInstanceTypeParameter(method),
				findLocalVariable(stackVars, 0, info), 
				callConfig, pds);
		return new MethodCallImpl(callConfig, 
				arguments.toArray(new MethodArgument[0]), method, ref);
	}
	
	private Parameter getInstanceTypeParameter(PDSMethod method)
	{
		ClassFile classFile = method.getMethodInfo().getClassFile();
		try {
			return new MethodReferenceParameter(classFile.getThisClassName());
		}
		catch (InvalidByteCodeException ive) {
			return new MethodReferenceParameter("");
		}
	}

	/**
	 * Returns the next config where <code>method</code> is called.
	 * 
	 * @param config the config.
	 * @param method the method.
	 * @return the config where method is called.
	 */
	public IntConfig getNextMethodCallConfig(IntConfig config, PDSMethod method) {
		
		return config.getConfigAt(method.getFormattedName());
	}
	
	/**
	 * Returns the next config where a method from <code>pds</code> is called.
	 */
	public IntConfig getAnyNextMethodCallConfig(IntConfig config, PDS pds)
	{
		while (config != null 
				&& !pds.hasMethod(config.getStackSymbol())) {
			config = config.getNextConfig();
		}
		return config;
	}
	
	/**
	 * Retrieves the array indexed by indexVar from the heap.
	 * 
	 * @param globalVars
	 * @param indexVar
	 * @param info
	 * @return
	 */
	private int[] getArray(IntVariable[] globalVars, IntVariable indexVar, 
			PDSInfo info) {
		
		logger.debug("indexVar: " + indexVar);
		int offset = indexVar.intValue();
		logger.debug("offset: " + offset);
		
		IntVariable heapVar = findVariable(globalVars, info.getHeapName());
		int[] heap = heapVar.arrayValue();
		
		int size = 0;
		int[] array;
		
		if (info.getHeapOption() == HeapOption.SIMPLE) {
			
			size = heap[offset];
			array = new int[size];
			System.arraycopy(heap, offset + 1, array, 0, size);
			return array;
		}
		
		// HeapOption.TWODIMS
		int bits = info.getBits();
		for (int i = 0; i < bits; i++)
			size += heap[bits*offset + i] << (bits - i - 1);
		array = new int[size];
		logger.debug("size: " + size);
		
		int elemBits = 0;
		for (int i = 0; i < bits; i++)
			elemBits += heap[bits*offset + bits + i] << (bits - i - 1);
		logger.debug("elemBits: " + elemBits);
		
		for (int i = 0; i < size; i++) {
			for (int j = 0; j < elemBits; j++)
				array[i] += heap[bits*offset + 2*bits + i*elemBits + j] 
				                 << (elemBits - j - 1);
			logger.debug("array[" + i + "]: " + array[i]);
		}
		
		return array;
	}
	
	/**
	 * Looks up the local variable with index in the array of <code>vars</code>.
	 * 
	 * @param index the index of the local variable
	 * @param info needed to determine the name of the local variable
	 * @throws IllegalArgumentException if the variable could not be found
	 */
	public static IntVariable findLocalVariable(IntVariable[] vars, int index, 
			PDSInfo info) {
		
		String name = info.varIndex(index);
		return findVariable(vars, name);
	}
	
	/**
	 * Looks up the variable <code>name</code> in the array of <code>vars</code>.
	 * 
	 * @throws IllegalArgumentException if the variable could not be found
	 */
	public static IntVariable findVariable(IntVariable[] vars, String name)
	{
		for (IntVariable var : vars) {
			if (var.getName().equals(name)) {
				return var;
			}
		}
		throw new IllegalArgumentException("Not found var " + name);
	}
	
	private static class IntVariableMethodArgument implements MethodArgument
	{
		
		private Parameter param;
		
		private IntVariable variable;

		public IntVariableMethodArgument(Parameter param, IntVariable var)
		{
			this.param = param;
			this.variable = var;
		}
		
		public boolean isValue()
		{
			return true;
		}

		public IntVariable getVariable() 
		{
			return variable;
		}

		public boolean isArray() 
		{
			return false;
		}

		public int[] getArray() 
		{
			return null;
		}

		public boolean isReference() 
		{
			return false;
		}

		public MethodCall getConstructorCall() 
		{
			return null;
		}

		public Parameter getFormalParameter() 
		{
			return param;
		}
		
		@Override
		public String toString() 
		{
			StringBuilder builder = new StringBuilder();
			if (variable.getType() == IntVariable.Type.BOOLEAN) {
				builder.append(variable.booleanValue());
			}
			else {
				builder.append(variable.intValue());
			}
			builder.append(" : ");
			builder.append(param);
			return builder.toString();
		}
		
	}
	
	private static class ValueArrayMethodArgument implements MethodArgument
	{

		private int[] array;
		
		private Parameter param;
		
		public ValueArrayMethodArgument(Parameter param, int[] array)
		{
			this.param = param;
			this.array = array;
		}
		
		public boolean isValue() 
		{
			return false;
		}

		public IntVariable getVariable() 
		{
			return null;
		}

		public boolean isArray() 
		{
			return true;
		}

		public int[] getArray() 
		{
			return array;
		}

		public boolean isReference() 
		{
			return false;
		}

		public MethodCall getConstructorCall() 
		{
			return null;
		}

		public Parameter getFormalParameter() 
		{
			return param;
		}
		
		@Override
		public String toString() 
		{
			StringBuilder builder = new StringBuilder();
			builder.append(Arrays.toString(array));
			builder.append(" : ");
			builder.append(param);
			return builder.toString();
		}
		
	}
	
	private static class MethodCallImpl implements MethodCall
	{
		private IntConfig config;
		
		private MethodArgument[] args;
		
		private PDSMethod method;
		
		private MethodArgument instance;
		
		public MethodCallImpl(IntConfig config, MethodArgument[] args, 
							  PDSMethod method, MethodArgument instance) 
		{
			this.config = config;
			this.args = args;
			this.method = method;
			this.instance = instance;
		}

		public MethodArgument[] getArguments() 
		{
			return args;
		}

		public IntConfig getConfig() 
		{
			return config;
		}
		
		public String toString()
		{
			StringBuilder builder = new StringBuilder(method.getHumanReadableName());
			builder.append('(');
			for (int i = 0; i < args.length; i++) {
				builder.append(args[i]);
				if (i != args.length - 1) {
					builder.append(", ");
				}
			}
			builder.append(')');
			return builder.toString();
		}

		public PDSMethod getMethod() 
		{
			return method;
		}

		public MethodArgument getInstance() 
		{
			return instance;
		}
	
	}
	
	private static class MethodReferenceParameter implements Parameter
	{

		private String type;
		
		public MethodReferenceParameter(String type)
		{
			this.type = type;
		}
		
		public boolean isValue() 
		{
			return false;
		}

		public boolean isValueArray() 
		{
			return false;
		}

		public boolean isReference() 
		{
			return true;
		}

		public boolean isReferenceArray() 
		{
			return false;
		}

		public String getType() 
		{
			return type;
		}

		@Override
		public String toString() 
		{
			return "reference to " + type;
		}
		
	}
}
