package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

/**
 * Represents a line in Java program plus coverage information.
 * 
 * @author suwimodh
 *
 */
public class Line
{

	private final CoverageMethod method;
	
	private final int lineOffset;
	
	/**
	 *  The annotated bytecode instructions of this line.
	 */
	private List<AnnotatedInstruction> instructions = null;
	
	/**
	 * The constructor.
	 * 
	 * @param method the <code>CoverageMethod</code>
	 * @param lineOffset the line offset.
	 */
	public Line(CoverageMethod method, int lineOffset)
	{
		this.method = method;
		this.lineOffset = lineOffset;
	}
	
	/**
	 * Gets the underlying list of annotated bytecode instructions.
	 * 
	 * @return the list of annotated instructions.
	 */
	public List<AnnotatedInstruction> getInstructions()
	{
		if (instructions == null) {
			instructions = method.getInstructions(lineOffset);
		}
		return Collections.unmodifiableList(instructions);
	}
	
	/**
	 * Returns the list of automaton transitions having state in this line.
	 * 
	 * @param state the state.
	 * @return the list of transitions.
	 */
	public List<Transition> getTransitions(State state)
	{
		List<Transition> emptyList = Collections.emptyList();
		if (state != State.COVERED && !state.isException()) {
			return emptyList;
		}
		List<Transition> list = null;
		for (AnnotatedInstruction instruction : getInstructions()) {
			List<Transition> trans = instruction.getTransitions(state);
			if (!trans.isEmpty()) {
				if (list == null) {
					list = new ArrayList<Transition>(state == State.COVERED ? 4 : 2);
				}
				list.addAll(trans);
			}
		}
		return list != null ? list: emptyList;
	}
	
	public boolean isEffectiveLine()
	{
		return method.isEffectiveLine(lineOffset);
	}
	
	public State getState()
	{
		return method.getLineState(lineOffset);
	}
	
	public CoverageMethod getMethod()
	{
		return  method;
	}
	
	public int getLineNumber()
	{
		return method.getLineNumber(lineOffset);
	}
	
	@Override
	public String toString() 
	{
		State state = getState();
		if (state == State.PARTIALLY_COVERED) {
			List<AnnotatedInstruction> list = getInstructions();
			StringBuilder builder = new StringBuilder("Not covered: ");
			for (Iterator<AnnotatedInstruction> iter = list.iterator(); iter.hasNext();) {
				AnnotatedInstruction i = iter.next();
				if (!i.isCovered()) {
					builder.append(i.getInstruction().getOpcodeVerbose());
					if (iter.hasNext()) {
						builder.append(", ");
					}
				}
			}
			return builder.toString();
		}
		return "";
	}
	
}
