package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.ArrayList;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

/**
 * This class provides support for classes that want to notify 
 * {@link CoverageListener CoverageListeners} by implmenting the interface
 * itself and forwarding and multicasting calls to added listeners.
 * <p>
 * The class is thread safe.  
 */
public class MulticastCoverageListener implements CoverageListener
{

	private ArrayList<CoverageListener> coverageListeners = 
		new ArrayList<CoverageListener>();
	
	/**
	 * Adds a coverage listener that is notified of the test progress.
	 * @param listener
	 */
	public void addCoverageListener(CoverageListener listener)
	{
		synchronized (coverageListeners) {
			coverageListeners.add(listener);
		}
	}
	
	/**
	 * Removes the coverage listener.
	 * @param listener
	 */
	public void removeCoverageListener(CoverageListener listener)
	{
		synchronized (coverageListeners) {
			coverageListeners.remove(listener);
		}
	}
	
	public void coverageStarted(PDS pds)
	{
		CoverageListener[] listeners = getListeners();
		for (int i = listeners.length - 1; i >= 0; i--) {
			listeners[i].coverageStarted(pds);
		}
	}
	
	public void coverageEnded(CoverageResult result)
	{
		CoverageListener[] listeners = getListeners();
		for (int i = listeners.length - 1; i >= 0; i--) {
			listeners[i].coverageEnded(result);
		}
	}
	
	public void coverageFailed(final String message, final Throwable t)
	{
		CoverageListener[] listeners = getListeners();
		for (int i = listeners.length - 1; i >= 0; i--) {
			listeners[i].coverageFailed(message, t);
		}
	}
	
	public CoverageListener[] getListeners()
	{
		synchronized (coverageListeners) {
			return coverageListeners.toArray(new CoverageListener[coverageListeners.size()]);
		}
	}

	public void labelReached(String label) 
	{
		CoverageListener[] listeners = getListeners();
		for (int i = listeners.length - 1; i >= 0; i--) {
			listeners[i].labelReached(label);
		}
	}

	public void coveredMethodOffset(CoverageMethod method, int lineOffset,
			int pcOffset, State state, Transition transition) 
	{
		CoverageListener[] listeners = getListeners();
		for (int i = listeners.length - 1; i >= 0; i--) {
			listeners[i].coveredMethodOffset(method, lineOffset, pcOffset, state, transition);
		}
	}
	
}
