/*
    Copyright (C) 2005 Felix Berger <bflat1@gmx.net>

    This file is part of jMoped.

    jMoped is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    jMoped is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with jMoped; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package de.uni_stuttgart.fmi.szs.jmoped;

import java.awt.event.ComponentAdapter;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.Enumeration;
import java.util.Set;
import java.util.concurrent.Delayed;
import java.util.concurrent.TimeUnit;

import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;

import org.gjt.jclasslib.io.ClassFileReader;
import org.gjt.jclasslib.structures.ClassFile;

public class PDSInfoTest extends PDSTestCase 
{

	public void testFindClassFile() throws Exception
	{
		assertNotNull(BytecodeUtils.findClassFile("de/uni_stuttgart/fmi/szs/jmoped/PDSInfoTest"));
		assertNotNull(BytecodeUtils.findClassFile("java/util/StringTokenizer"));
		assertNotNull(BytecodeUtils.findClassFile("junit/framework/TestCase"));
		
		try {
			// if the class loader can find the file, we should too
			Class emmaClass = Class.forName("com.vladium.emma.rt.RT");
			assertNotNull(BytecodeUtils.findClassFile("com/vladium/emma/rt/RT"));
		}
		catch (ClassNotFoundException cnfe) {
			try {
				assertNull(BytecodeUtils.findClassFile("com/vladium/emma/rt/RT"));
				fail("Classloader didn't find it, neither should we");
			}
			catch (FileNotFoundException fnfe) {
			}
		}
	}
	
	public void testSetBits() throws Exception
	{
		PDS pds = getPDS(TestClass.class);
		PDSInfo info = pds.getInfo();
		
		info.setBits(1);
		assertEquals(1, info.getBits());
		assertEquals(1, info.getHeapSize());
		
		info.setBits(4);
		assertEquals(4, info.getBits());
		assertEquals(1, info.getHeapSize());
		info.setHeapSize(15);
		assertEquals(15, info.getHeapSize());
		assertEquals(4, info.getBits());
		
		info.setHeapSize(16);
		assertEquals(16, info.getHeapSize());
		assertEquals(5, info.getBits());
	}
	
	public static class TestClass
	{
		static int i = 1;
		private int k = 1;
		
		public void me()
		{
		}
		
		private void priv()
		{
			
		}
		
		public void onlyInTestClass()
		{
			
		}
		
	}
	
	public void testSuperMethodInclusion() throws Exception
	{
		PDS pds = getPDS(SubTestClass.class);
		PDSClass clazz = pds.getPDSClass("de/uni_stuttgart/fmi/szs/jmoped/PDSInfoTest$TestClass");
		assertNotNull(clazz.getMethod("me", "()V"));
	}
	
	public void testDoesNotIncludePrivateSuperMethod() throws Exception
	{
		PDS pds = getPDS(SubSubTestClass.class);
		
		// TestClass does not have it
		PDSClass clazz = pds.getPDSClass("de/uni_stuttgart/fmi/szs/jmoped/PDSInfoTest$TestClass");
		assertNull(clazz.getMethod("priv", "()V"));
		
		// but SubTestClass has it
		clazz = pds.getPDSClass("de/uni_stuttgart/fmi/szs/jmoped/PDSInfoTest$SubTestClass");
		assertNotNull(clazz.getMethod("priv", "()V"));
		
		// and SubSubTestClass has it
		clazz = pds.getPDSClass("de/uni_stuttgart/fmi/szs/jmoped/PDSInfoTest$SubSubTestClass");
		assertNotNull(clazz.getMethod("priv", "()V"));
	}
	
	public static class SubTestClass extends TestClass
	{
		
		public void me()
		{
			super.me();
		}
		
		public static void main(String[] args)
		{
			new SubTestClass().me();
		}
	
		public void priv()
		{
			
		}
		
	}
	
	public static class SubSubTestClass extends SubTestClass
	{
		
		public void me()
		{
			super.me();
		}
		
		public static void main(String[] args)
		{
			SubSubTestClass subsub = new SubSubTestClass();
			subsub.me();
			SubTestClass sub = subsub;
			sub.me();
			sub.priv();
			TestClass test = sub;
			test.me();
		}
		
		public void priv()
		{
			InstantiatedInVirtualMethodClass i = new InstantiatedInVirtualMethodClass();
		}
	}
	
	public static class InstantiatedInVirtualMethodClass extends SubSubTestClass
	{
		public void me()
		{
			
		}
		
		public void priv()
		{
			
		}
	}
	
	public void testParallelHierarchy() throws Exception
	{
		PDS pds = getPDS(HierarchyNeutralClass.class);

		PDSClass clazz = pds.getPDSClass("de/uni_stuttgart/fmi/szs/jmoped/PDSInfoTest$SubClass");
		assertNotNull(clazz.getMethod("up", "()V"));
		assertNotNull(clazz.getMethod("down", "()V"));
		
		clazz = pds.getPDSClass("de/uni_stuttgart/fmi/szs/jmoped/PDSInfoTest$SuperClass");
		assertNotNull(clazz.getMethod("up", "()V"));
		assertNotNull(clazz.getMethod("down", "()V"));
	}
	
	public static class HierarchyNeutralClass
	{
		public static void main(String[] args)
		{
			SuperClass sc = new SubClass();
			sc.up();
		}
	}
	
	public void testOnlySuperClassIncluded() throws Exception
	{
		PDS pds = getPDS(OnlyInstantiateSuperClass.class);

		assertEquals(2, pds.getPDSClasses().size());
		
		PDSClass clazz = pds.getPDSClass("de/uni_stuttgart/fmi/szs/jmoped/PDSInfoTest$SuperClass");
		assertNotNull(clazz.getMethod("up", "()V"));
		assertNotNull(clazz.getMethod("down", "()V"));
	}
	
	public static class OnlyInstantiateSuperClass
	{
		public static void main(String[] args)
		{
			SuperClass sc = new SuperClass();
			sc.up();
		}
	}
	
	public void testNullPointerCall() throws Exception
	{
		PDS pds = getPDS(NullPointerCallClass.class);

		assertEquals(1, pds.getPDSClasses().size());
	}
	
	public static class NullPointerCallClass
	{
		public static void main(String[] args)
		{
			SuperClass sc = null;
			sc.up();
		}
	}
	
	public static class SuperClass
	{
		public void up()
		{
			down();
		}
		
		public void down()
		{
			up();
		}
	}
	
	public static class SubClass extends SuperClass
	{
		public void up()
		{
			super.up();
		}
		
		public void down()
		{
			up();
		}
	}
	
	public void testGetImplementedInterfaces() throws Exception
	{
		PDS pds = getPDS(InstantiateInterfaceImplementerClass.class);
		InvokeAnalyzer analyzer = pds.getInfo().getInvokeAnalyzer();
		Set<String> interfaces = analyzer.getImplementedInterfaces(
				"de/uni_stuttgart/fmi/szs/jmoped/PDSInfoTest$InterfaceImplementerClass");
		assertTrue(interfaces.contains("java/awt/event/ComponentListener"));
		assertTrue(interfaces.contains("java/util/concurrent/Delayed"));
		assertTrue(interfaces.contains("javax/swing/tree/MutableTreeNode"));
		assertTrue(interfaces.contains("javax/swing/tree/TreeNode"));
		assertTrue(interfaces.contains("java/lang/Comparable"));
		assertTrue(interfaces.contains("java/util/EventListener"));
		assertEquals(6, interfaces.size());
	}
	
	public void testInterfaceInvocation() throws Exception
	{
		PDS pds = getPDS(InstantiateInterfaceImplementerClass.class);
		System.out.println(pds.getInfo().getInvokeAnalyzer().includedMethodsInfo());
		PDSClass clazz = pds.getPDSClass("de/uni_stuttgart/fmi/szs/jmoped/PDSInfoTest$InterfaceImplementerClass");
		assertNotNull(clazz.getMethod("getChildCount", "()I"));
		assertNotNull(clazz.getMethod("compareTo", "(Ljava/lang/Object;)I"));
		assertNotNull(clazz.getMethod("insert", "(Ljavax/swing/tree/MutableTreeNode;I)V"));
	}
	
	public static class InstantiateInterfaceImplementerClass
	{
		public static void main(String[] args)
		{
			InterfaceImplementerClass i = new InterfaceImplementerClass();
			TreeNode tn = i;
			tn.getChildCount();
			Comparable c = i;
			c.compareTo(i);
			MutableTreeNode mtn = i;
			mtn.insert(i, 0);
		}
	}
	
	public void testNullpoinerInterfaceInvocation() throws Exception
	{
		PDS pds = getPDS(NullPoinerInterfaceClass.class);
		// should not crash
	}
	
	public static class NullPoinerInterfaceClass
	{
		public static void main(String[] args)
		{
			TreeNode n = null;
			n.getChildCount();
		}
	}
	
	public static class InterfaceImplementerClass extends ComponentAdapter 
		implements MutableTreeNode, Delayed
	{

		public void insert(MutableTreeNode child, int index) {
		}

		public void remove(int index) {
		}

		public void remove(MutableTreeNode node) {
		}

		public void setUserObject(Object object) {
		}

		public void removeFromParent() {
		}

		public void setParent(MutableTreeNode newParent) {
		}

		public TreeNode getChildAt(int childIndex) {
			return null;
		}

		public int getChildCount() {
			return 0;
		}

		public TreeNode getParent() {
			return null;
		}

		public int getIndex(TreeNode node) {
			return 0;
		}

		public boolean getAllowsChildren() {
			return false;
		}

		public boolean isLeaf() {
			return false;
		}

		public Enumeration children() {
			return null;
		}

		public long getDelay(TimeUnit unit) {
			// TODO Auto-generated method stub
			return 0;
		}

		public int compareTo(Delayed o) {
			// TODO Auto-generated method stub
			return 0;
		}
		
	}
	
	/**
	 * Added because fields were used although their class was explicitly ignored.
	 */
	public void testIsIgnoredField() throws Exception
	{
		PDS pds = getPDS(TestClass.class);
		PDSInfo info = pds.getInfo();
		
		assertTrue(info.isIgnoredField(new String[] { "java/lang/System", "out", }));
	}
	
	public void testMiddleClassWithNoImplementation() throws Exception
	{
		PDS pds = getPDS(ReimplmentsMethodClass.class);
		assertTrue(pds.getInfo().getInvokeAnalyzer().isIncludedMethod(
				SuperMethodClass.class.getName().replace('.', '/'),
				"method", "()V"));
		
		pds = new PDS(ReimplmentsMethodClass.class, "test", "()V");
		// we approximate here, that's why it's in as well
		assertTrue(pds.getInfo().getInvokeAnalyzer().isIncludedMethod(
				SuperMethodClass.class.getName().replace('.', '/'), 
				"method", "()V"));
	}
	
	public static class SuperMethodClass
	{
		public void method()
		{
			
		}
	}
	
	public static class MiddleNoMethodClass extends SuperMethodClass
	{
		
	}
	
	public static class ReimplmentsMethodClass extends MiddleNoMethodClass
	{

		public void method() 
		{
			
		}
		
		public static void main(String[] args) 
		{
			ReimplmentsMethodClass instance = new ReimplmentsMethodClass();
			MiddleNoMethodClass test = new MiddleNoMethodClass();
			test.method();
		}
		
		public static void test()
		{
			ReimplmentsMethodClass instance = new ReimplmentsMethodClass();
			MiddleNoMethodClass test = instance;
			test.method();
		}
	}

	public void testSuperClassImplementsInterface() throws Exception
	{
		PDS pds = getPDS(ClassImplementsInterfaceThroughSuperClass.class);
		assertTrue(pds.getInfo().getInvokeAnalyzer().isIncludedMethod(
				SuperClassImplementsInterfaceClass.class.getName().replace('.', '/'),
				"call", "()V"));
	}
	
	public static class SuperClassImplementsInterfaceClass
	{
		public void call()
		{
			
		}
	}
	
	public static interface CallInterfaceClass
	{
		void call();
	}
	
	public static class ClassImplementsInterfaceThroughSuperClass 
		extends SuperClassImplementsInterfaceClass implements CallInterfaceClass
	{
		public static void main(String[] args) 
		{
			CallInterfaceClass call = new ClassImplementsInterfaceThroughSuperClass();
			call.call();
		}
	}
	
}
