package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.Arrays;
import java.util.List;

import org.gjt.jclasslib.bytecode.AbstractInstruction;

import de.uni_stuttgart.fmi.szs.jmoped.PDSClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;
import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

public class CoverageMethodTest extends PDSTestCase 
{

	public void testIsVoid() throws Exception
	{
		PDSClass clazz = getPDSClass(VoidTestClass.class);
		for (PDSMethod m : clazz.getMethodList()) {
			CoverageMethod method = new CoverageMethod(m);
			assertTrue(method.isVoid());
		}
		clazz = getPDSClass(NonVoidTestClass.class);
		for (PDSMethod m : clazz.getMethodList()) {
			CoverageMethod method = new CoverageMethod(m);
			if (!method.getMethod().getFormattedName().contains("init")) {
				assertFalse(method + " void", method.isVoid());
			}
		}
	}
	
	public static class VoidTestClass
	{
		public VoidTestClass()
		{
			
		}
		
		public static void main(String[] args)
		{
			
		}
		
		private void me()
		{
			
		}
		
		private void test(int i)
		{
			
		}
		
	}
	
	public static class NonVoidTestClass
	{
		public static int main(String[] args)
		{
			return 9;
		}
		
		public String me()
		{
			return null;
		}
		
		public Object test(int i)
		{
			return null;
		}
		
	}
	
	public void testGetInstructionCountForLine() throws Exception
	{
		PDSClass clazz = getPDSClass(LineCountClass.class);
		CoverageMethod method = new CoverageMethod(clazz.getMethod("<init>", "()V"));
		// implicit super call => 2 instructions
		assertEquals(2, method.getInstructionCount(0));
		assertEquals(1, method.getInstructionCount(1));
		
		method = new CoverageMethod(clazz.getMethod("<init>", "(I)V"));
		assertEquals(2, method.getInstructionCount(0));
		assertEquals(3, method.getInstructionCount(1));
		assertEquals(1, method.getInstructionCount(2));
		
		method = new CoverageMethod(clazz.getMethod("twolines", "()V"));
		assertEquals(3, method.getInstructionCount(0));
		assertEquals(6, method.getInstructionCount(1));
		assertEquals(1, method.getInstructionCount(2));
		
		method = new CoverageMethod(clazz.getMethod("braceTwolines", "()V"));
		assertEquals(3, method.getInstructionCount(0));
		assertEquals(7, method.getInstructionCount(1));
		
		method = new CoverageMethod(clazz.getMethod("empty", "()V"));
		assertEquals(1, method.getInstructionCount(0));
	}
	
	public void testGetLines() throws Exception
	{
		PDSClass clazz = getPDSClass(LineCountClass.class);
		CoverageMethod method = new CoverageMethod(clazz.getMethod("<init>", "()V"));
		assertEquals(0, method.getLineCount());
		
		method = new CoverageMethod(clazz.getMethod("<init>", "(I)V"));
		assertEquals(1, method.getLineCount());
		
		method = new CoverageMethod(clazz.getMethod("twolines", "()V"));
		assertEquals(2, method.getLineCount());
		
		method = new CoverageMethod(clazz.getMethod("braceTwolines", "()V"));
		assertEquals(2, method.getLineCount());
		
		method = new CoverageMethod(clazz.getMethod("nonvoid", "()I"));
		assertEquals(1, method.getLineCount());
		
		method = new CoverageMethod(clazz.getMethod("braceNonvoid", "()I"));
		assertEquals(1, method.getLineCount());
		
		method = new CoverageMethod(clazz.getMethod("nonvoidtwolines",
				"()Ljava/lang/Object;"));
		assertEquals(2, method.getLineCount());
		
		method = new CoverageMethod(clazz.getMethod("empty", "()V"));
		assertEquals(0, method.getLineCount());
		
		method = new CoverageMethod(clazz.getMethod("braceEmpty", "()V"));
		assertEquals(0, method.getLineCount());
	}
	
	public void testIsEffectiveLine() throws Exception
	{
		PDSClass clazz = getPDSClass(LineCountClass.class);
		CoverageMethod method = new CoverageMethod(clazz.getMethod("<init>", "()V"));
		assertFalse(method.isEffectiveLine(0));
		
		method = new CoverageMethod(clazz.getMethod("<init>", "(I)V"));
		assertFalse(method.isEffectiveLine(0));
		assertTrue(method.isEffectiveLine(1));
		assertFalse(method.isEffectiveLine(2));
		
		method = new CoverageMethod(clazz.getMethod("twolines", "()V"));
		assertTrue(method.isEffectiveLine(0));
		assertTrue(method.isEffectiveLine(1));
		assertFalse(method.isEffectiveLine(2));
		
		method = new CoverageMethod(clazz.getMethod("braceTwolines", "()V"));
		assertTrue(method.isEffectiveLine(0));
		assertTrue(method.isEffectiveLine(1));
		
		method = new CoverageMethod(clazz.getMethod("nonvoid", "()I"));
		assertTrue(method.isEffectiveLine(0));
		
		method = new CoverageMethod(clazz.getMethod("braceNonvoid", "()I"));
		assertTrue(method.isEffectiveLine(0));
		
		method = new CoverageMethod(clazz.getMethod("nonvoidtwolines",
				"()Ljava/lang/Object;"));
		assertTrue(method.isEffectiveLine(0));
		assertTrue(method.isEffectiveLine(1));
		
		method = new CoverageMethod(clazz.getMethod("empty", "()V"));
		assertFalse(method.isEffectiveLine(0));
		
		method = new CoverageMethod(clazz.getMethod("braceEmpty", "()V"));
		assertFalse(method.isEffectiveLine(0));
	}
	
	public void testIsClinit() throws Exception
	{
		PDSClass clazz = getPDSClass(AssertionClinitClass.class);
		CoverageMethod method = new CoverageMethod(clazz.getMethodClinit());
		assertTrue(method.isClinit());
		method = new CoverageMethod(clazz.getMethodMain());
		assertFalse(method.isClinit());
		
		clazz = getPDSClass(ClinitClass.class);
		method = new CoverageMethod(clazz.getMethodClinit());
		assertTrue(method.isClinit());
	}
	
	public void testAssertionClinitOnly() throws Exception
	{
		PDSClass clazz = getPDSClass(AssertionClinitClass.class);
		
		CoverageMethod method = new CoverageMethod(clazz.getMethodClinit());
		assertTrue(method.clinitIsAssertionOnly());
		method = new CoverageMethod(clazz.getMethodMain());
		assertFalse(method.clinitIsAssertionOnly());
		
		clazz = getPDSClass(ClinitClass.class);
		method = new CoverageMethod(clazz.getMethodClinit());
		assertFalse(method.clinitIsAssertionOnly());
		
		clazz = getPDSClass(ClinitPlusAssertionClinitClass.class);
		method = new CoverageMethod(clazz.getMethodClinit());
		assertFalse(method.clinitIsAssertionOnly());
	}
	
	public static class AssertionClinitClass
	{
		public static void main(String[] args)
		{
			assert(args.length > 0);
		}
	}
	
	public static class ClinitClass
	{
		private static int i = 0;
	}
	
	public static class ClinitPlusAssertionClinitClass
	{
		private static int k = 0;
		
		public static void main(String[] args)
		{
			assert(args.length > 0);
		}
	}
	
	public void testContainsAssertion() throws Exception
	{
		PDSClass clazz = getPDSClass(ClinitPlusAssertionClinitClass.class);
		CoverageMethod method = new CoverageMethod(clazz.getMethodMain());
		assertTrue(method.containsAssertion(0));
		
		clazz = getPDSClass(ContainsAssertionsClass.class);
		method = new CoverageMethod(clazz.getMethodMain());
		
		for (int i = 0; i < method.getLineTable().length; i++) {
			if (!method.isEffectiveLine(i)) {
				continue;
			}
			if (i % 2 == 0) {
				assertTrue(i + "", method.containsAssertion(i));
			}
			else {
				assertFalse(i + "", method.containsAssertion(i));
			}
		}
	}
	
	public static class ContainsAssertionsClass
	{
		public static void main(String[] args)
		{
			assert(args.length > 0);
			int i = 1;
			assert(i == 1 && i != 5);
			i = 5;
		}
	}
	
	public void testAssertionLinesCovered() throws Exception
	{
		CoverageTester tester = new CoverageTester(AssertionLinesCoveredClass.class);
		CoverageMethod method = tester.test()[0];
		
		assertFalse(method.isFullyCovered());
		
		assertEquals(State.COVERED, method.getLineState(0));
		assertEquals(State.COVERED, method.getLineState(1));
		assertEquals(State.COVERED, method.getLineState(2));
		assertEquals(State.COVERED, method.getLineState(3));
		assertEquals(State.PARTIALLY_COVERED, method.getLineState(4));
		assertTrue(method.containsFailedAssertion(4));
		assertEquals(State.NOT_COVERED, method.getLineState(5));
		assertFalse(method.containsFailedAssertion(5));
		assertEquals(State.NOT_COVERED, method.getLineState(6));
		assertFalse(method.containsFailedAssertion(6));
		assertEquals(State.NOT_COVERED, method.getLineState(7));
		assertFalse(method.containsFailedAssertion(7));
	}
	
	public static class AssertionLinesCoveredClass
	{
		public static void main(String[] args)
		{
			/*0*/int i = 1;
			/*1*/assert(i == 1);
			/*2*/assert(i == 1 && i != 2);
			/*3*/assert(i == 1); assert(i != 2);
			/*4*/assert(i != 1 && i == 2);
			/*5*/assert(i == 2);
			/*6*/assert(i != 1); assert(i == 2);
			/*7*/assert(i == 1 && i == 2);
		}
	}
	
	public void testIsFailedAssertionInstruction() throws Exception
	{
		PDSClass clazz = getPDSClass(AssertionLinesCoveredClass.class);
		CoverageMethod method = new CoverageMethod(clazz.getMethodMain());
		List<Integer> failedAssertionOffsets = Arrays.asList(17, 41, 60, 79, 103, 122, 141, 146, 160, 184);
		for (AbstractInstruction inst : method.getMethod().getInstructionList()) {
			if (failedAssertionOffsets.contains(inst.getOffset())) {
				assertTrue(method.isFailedAssertionInstruction(inst.getOffset()));
			}
			else {
				assertFalse("failed for " + inst.getOffset(), 
						method.isFailedAssertionInstruction(inst.getOffset()));
			}
		}
	}
	
}
