package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;

import de.uni_stuttgart.fmi.szs.jmoped.PDS;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo;
import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSArrayBits;
import de.uni_stuttgart.fmi.szs.jmoped.annotation.PDSParameterBits;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

public class CoverageTesterTest extends PDSTestCase
{
	
	public void testTest() throws Exception
	{
		CoverageTester tester = new CoverageTester(TestClass.class);
		tester.getPDS().getInfo().setBits(1);
		assertTrue(tester.test().length > 0);
		
		tester = new CoverageTester(CoveredClass.class);
		tester.setFilterMethodName("de_uni_stuttgart_fmi_szs_jmoped_coverage_" +
				"CoverageTesterTest_CoveredClass_main_" +
				"ALjava_lang_String_V");
		// one bit is not enough for full state coverage
		tester.getPDS().getInfo().setBits(1);
		assertFalse(tester.test()[0].isFullyCovered());
		// two bits already suffice
		tester.getPDS().getInfo().setHeapSize(4);
		CoverageMethod main = getMethod(tester.test(), "main"); 
		assertTrue(main.isFullyCovered());
	}
	
	public void testBubbleSortCoverage() throws Exception
	{
		CoverageTester tester = new CoverageTester(BubbleSort.class);
		tester.setFilterMethodName("de_uni_stuttgart_fmi_szs_jmoped_coverage_" +
		"CoverageTesterTest_BubbleSort_sort_AI_V");
		// one bit is not enough for full state coverage
		tester.getPDS().getInfo().setBits(1);
		
		CoverageMethod[] ms = tester.test();
		assertFalse(getMethod(ms, "sort").isFullyCovered());

		// 3 bits suffice
		tester.getInfo().setHeapSize(7);
		ms = tester.test();
		assertTrue(getMethod(ms, "sort").isFullyCovered());
	}
	
	private void log(CoverageMethod[] methods)
	{
		for (CoverageMethod m : methods) {
			System.out.println(m + " " + m.isFullyCovered());
			if (!m.isFullyCovered()) {
				for (int i = 0; i < m.getLineTable().length; i++) {
					System.out.println(m.getLineTable()[i].getStartPc() + " " + m.getLineState(i));
				}
			}
		}
	}
	
	public void testFindMethod() throws Exception
	{
		CoverageTester tester = new CoverageTester(TestClass.class);
		String label = "de_uni_stuttgart_fmi_szs_jmoped_coverage_" +
		"CoverageTesterTest_TestClass_main_" +
		"ALjava_lang_String_V";
		CoverageMethod method = tester.getMethod(label);
		assertNotNull(method);
		assertEquals(label, method.getMethod().getFormattedName());
	}
	
	public static class TestClass
	{
		
		public static void main(String[] args)
		{
			int i = 0;
			args = new String[2];
			if (args.length > 2) {
				i = 3;
			}
			else {
				i = 5;
			}
		}
	}
	
	public static class CoveredClass
	{
		public static void main(String[] args)
		{
			int i = Integer.parseInt("");
			if (i > 1) {
				i = 3;
			}
			else {
				i = 5;
			}
		}
	}
	
	public static class BubbleSort
	{
		public static void main(String[] args)
		{
			int [] array  = new int[2];
			for (int i = 0; i < array.length; i++) {
				array[i] = array.length - 1 - i;
			}
			new BubbleSort().sort(array);
			assert(new BubbleSort().isSorted(array));
		}
		
		public void sort(int[] array)
		{
			for (int i = 0; i < array.length; i++) {
				for (int j = 0; j < array.length - 1; j++) {
					if (array[j] > array[j+1]) {
						int tmp = array[j];
						array[j] = array[j+1];
						array[j+1] = tmp;
					}
				}
			}
		}
		
		boolean isSorted(int[] array)
		{
			for (int i = 0; i < array.length - 1; i++) {
				if (array[i] > array[i+1]) {
					return false;
				}
			}
			return true;
		}
	}
	
	public void testListener() throws Exception
	{
		CoverageTester tester = new CoverageTester(ListenerTestClass.class);
		tester.setFilterMethodName( "de_uni_stuttgart_fmi_szs_jmoped_coverage_" +
		"ListenerTestClass_main_ALjava_lang_String_V");
		TestListener listener = new TestListener();
		tester.addCoverageListener(listener);
		CoverageMethod[] methods = tester.test();
		log(methods);
		// main + clinit
		assertEquals(2, methods.length);
		assertTrue(listener.started);
		assertTrue(listener.endedOrFailed);
		assertNotNull(listener.coveredMethod);
		assertFalse(listener.offsets.isEmpty());
		assertEquals(1, listener.assertCount);
		
		
		// same tests without stack prefix
		tester = new CoverageTester(ListenerTestClass.class);
		listener = new TestListener();
		tester.addCoverageListener(listener);
		methods = tester.test();
		// 2 methods, main and clinit because of asserts, init is not called 
		assertEquals(2, methods.length);
		assertTrue(listener.started);
		assertTrue(listener.endedOrFailed);
		assertNotNull(listener.coveredMethod);
		assertFalse(listener.offsets.isEmpty());
		assertEquals(1, listener.assertCount);
	}
	
	private void log(String[] strings)
	{
		for (String s: strings) {
			System.out.println(s);
		}
	}
	
	public static class TestListener extends CoverageAdapter
	{

		boolean started;
		boolean endedOrFailed;
		CoverageMethod coveredMethod;
		ArrayList<Integer> offsets = new ArrayList<Integer>();
		int assertCount = 0;
		
		@Override
		public void coverageStarted(PDS pds) 
		{
			started = true;
		}

		@Override
		public void coveredMethodOffset(CoverageMethod method, int lineOffset,
				int pcOffset, State state, Transition transition) 
		{
			if (coveredMethod == null) {
				coveredMethod = method;
			}
			offsets.add(lineOffset);
			if (state == State.ASSERTION_FAILED) {
				++assertCount;
			}
		}

		@Override
		public void coverageEnded(CoverageResult result) 
		{
			endedOrFailed = true;
		}

		@Override
		public void coverageFailed(String message, Throwable t) 
		{
			endedOrFailed = true;
		}

		

	}
	
	public void testProgressMonitor() throws Exception
	{
		CoverageTester tester = new CoverageTester(BubbleSort.class);
		tester.setFilterMethodName("de_uni_stuttgart_fmi_szs_jmoped_coverage_" +
			"CoverageTesterTest_BubbleSort_sort_AI_V");
		TestProgressMonitor monitor = new TestProgressMonitor();
		tester.getInfo().setBits(3);
		CoverageMethod[] methods = tester.test(monitor);
		assertTrue(monitor.beginCalled);
		assertTrue(monitor.doneCalled);
		assertTrue(methods[0].isFullyCovered());
		assertEquals(10000, monitor.worked);
		assertEquals(monitor.workAmount, monitor.worked);
		
		monitor = new TestProgressMonitor();
		monitor.setCanceled(true);
		
		methods = tester.test(monitor);
//		assertTrue(monitor.beginCalled);
		assertTrue(monitor.doneCalled);
		assertEquals(0, monitor.worked);
		assertEquals(0, methods.length);
	}
	
	public static class TestProgressMonitor implements ProgressMonitor
	{

		public boolean canceled = false;
		public boolean beginCalled = false;
		public boolean doneCalled = false;
		public int worked = 0;
		public int workAmount = 0;
		
		public void beginTask(String description, int workAmount) 
		{
			this.workAmount = workAmount; 
			beginCalled = true;
		}

		public void done() 
		{
			doneCalled = true;		
		}

		public boolean isCanceled() 
		{
			return canceled;
		}

		public void setCanceled(boolean cancel) 
		{
			canceled = cancel;
		}

		public void subTask(String description) 
		{
			
		}

		public void worked(int amount) 
		{
			worked += amount;
		}
		
	}
	
	public void testNPEThrown() throws Exception
	{
		CoverageTester tester = new CoverageTester(NPEClass.class);
		tester.getInfo().setCheckForNullPointerExceptions(true);
		
		NPEListener listener = new NPEListener();
		tester.addCoverageListener(listener);
		
		tester.test();
		
		assertEquals(2, listener.lines.size());
		assertTrue(listener.lines.contains(4));
		assertTrue(listener.lines.contains(7));
	}
	
	public static class NPEClass
	{
		public boolean foo()
		{
			return true;
		}
		
		public static void main(String[] args)
		{
			int i = Integer.parseInt("");
			if (i == 0) {
				NPEClass n = new NPEClass();
				n = null;
				boolean t = n.foo();
			}
			else {
				NPEClass[] array = new NPEClass[1];
				NPEClass o = array[0];
				o.foo();
			}
		}
	}
	
	public static class NPEListener extends CoverageAdapter
	{
		public HashSet<Integer> lines = new HashSet<Integer>();
		
		@Override
		public void coveredMethodOffset(CoverageMethod method, int lineOffset,
				int pcOffset, State state, Transition transition) 
		{
			System.out.println(lineOffset + " " + state);
			if (state == State.NPE) {
				nullPointerExceptionThrown(method, lineOffset);
			}
		}
		
		public void nullPointerExceptionThrown(CoverageMethod method, int lineOffset) 
		{
			lines.add(lineOffset);
		}
	}
	
	public void testResultConstructor() throws Exception
	{
		CoverageTester tester = new CoverageTester(CoveredClass.class);
		GetResultListener listener = new GetResultListener();
		tester.addCoverageListener(listener);
		getMethod(tester.test(), "main").isFullyCovered();
		assertNotNull(listener.result);
		tester = new CoverageTester(listener.result);
		getMethod(tester.test(), "main").isFullyCovered();
	}
	
	/**
	 * This test belongs to a fix in {@link PDSInfo#buildClassIndexTable(List)}.
	 * @throws Exception
	 */
	public void testAvoidSpuriousExcecutionAfterNPECheck() throws Exception
	{
		CoverageTester tester = new CoverageTester(NullPointerTestClass.class);
		CoverageMethod method = getMethod(tester.test(), "main");
		assertFalse(method.isFullyCovered());
	}
	
	public static class NullPointerTestClass
	{

		public void npe(int i)
		{
			
		}
		
		public static void main(String[] args)
		{
			int i = 0;
			NullPointerTestClass npe = new NullPointerTestClass();
			npe = null;
			npe.npe(i);
			NullPointerTestClass[] array = new NullPointerTestClass[1];
			array[0].npe(i);
			assert(npe instanceof NullPointerTestClass);
		}
		
	}

	private static class GetResultListener extends CoverageAdapter
	{
		public CoverageResult result;
		@Override
		public void coverageEnded(CoverageResult result) 
		{
			this.result = result;
		}
	}
	
	public void testQuickPerf() throws Exception
	{
		CoverageTester tester = getTester(QuickSort.class, "sort", "([I)V");
		tester.test();
	}
	
	public void testQuickSort() throws Exception
	{
		CoverageTester tester = new CoverageTester(QuickSort.class);
		tester.getInfo().setHeapSize(6);
		CoverageMethod[] methods = tester.test();
		CoverageMethod main = getMethod(methods, "main");

		assertTrue(main.isFullyCovered());
		assertTrue(getMethod(methods, "sort", "([I)V").isFullyCovered());
//		print(getMethod(methods, "sort", "([III)V"));
		assertTrue(getMethod(methods, "sort", "([III)V").isFullyCovered());
	}
	
	public static class QuickSort
	{
		void sort(int a[], int lo0, int hi0) throws Exception {
	        int lo = lo0;
	        int hi = hi0;
	        //pause(lo, hi);
	        if (lo >= hi) {
	            return;
	        }
	        else if( lo == hi - 1 ) {
	            /*
	             *  sort a two element list by swapping if necessary 
	             */
	            if (a[lo] > a[hi]) {
	                int T = a[lo];
	                a[lo] = a[hi];
	                a[hi] = T;
	            }
	            return;
	        }


	        /*
	         *  Pick a pivot and move it out of the way
	         */
	        int pivot = a[(lo + hi) / 2];
	        a[(lo + hi) / 2] = a[hi];
	        a[hi] = pivot;

	        while( lo < hi ) {
	            /*
	             *  Search forward from a[lo] until an element is found that
	             *  is greater than the pivot or lo >= hi 
	             */
	            while (a[lo] <= pivot && lo < hi) {
	                lo++;
	            }

	            /*
	             *  Search backward from a[hi] until element is found that
	             *  is less than the pivot, or lo >= hi
	             */
	            while (pivot <= a[hi] && lo < hi ) {
	                hi--;
	            }

	            /*
	             *  Swap elements a[lo] and a[hi]
	             */
	            if( lo < hi ) {
	                int T = a[lo];
	                a[lo] = a[hi];
	                a[hi] = T;
	                assert(lo > hi);
	                //pause();
	            }

	            //if (stopRequested) {
	              //  return;
	           // }
	        }

	        /*
	         *  Put the median in the "center" of the list
	         */
	        a[hi0] = a[hi];
	        a[hi] = pivot;

	        /*
	         *  Recursive calls, elements a[lo0] to a[lo-1] are less than or
	         *  equal to pivot, elements a[hi+1] to a[hi0] are greater than
	         *  pivot.
	         */
	        sort(a, lo0, lo-1);
	        sort(a, hi+1, hi0);
	    }

		@PDSParameterBits(2)
		@PDSArrayBits(2)
	    public void sort(int a[]) throws Exception {
	        sort(a, 0, a.length-1);
	    }
	    
	    public static void main(String[] args) throws Exception {
	    	
	    	Random r = new Random();
	    	int[] array = new int[r.nextInt()];
	    	for (int i = 0; i < array.length; i++) {
	    		array[i] = r.nextInt();
	    	}
	    	new QuickSort().sort(array);
	    }
	}
}
