package de.uni_stuttgart.fmi.szs.jmoped.coverage;

import java.util.List;

import de.uni_stuttgart.fmi.szs.jmoped.PDSClass;
import de.uni_stuttgart.fmi.szs.jmoped.PDSInfo;
import de.uni_stuttgart.fmi.szs.jmoped.PDSMethod;
import de.uni_stuttgart.fmi.szs.jmoped.PDSTestCase;
import de.uni_stuttgart.fmi.szs.jmoped.coverage.CoverageMethod.State;

public class HeapOverflowTest extends PDSTestCase 
{

	public void testHeapOverflow() throws Exception
	{
		CoverageTester tester = new CoverageTester(AllocatorClass.class);
		CountHeapOverflowListener listener = new CountHeapOverflowListener();
		tester.addCoverageListener(listener);
		
		tester.getInfo().setHeapSize(19);
		assertEquals(5, tester.getInfo().getBits());
		
		tester.test();
		
		assertEquals(0, listener.overflows);
		assertEquals(0, listener.assertions);
		
		tester.getInfo().setBits(1);
		tester.getInfo().setHeapSize(8);
		assertEquals(4, tester.getInfo().getBits());
		
		tester.test();
		
		assertEquals(1, listener.overflows);
		assertEquals(3, listener.lineOffset);
		assertEquals(0, listener.assertions);
	}
	
	public void testHeapOverflowReachable() throws Exception
	{
		PostStarReachabilityTester tester = 
			new PostStarReachabilityTester(AllocatorClass.class);
		tester.setEnabledNotifyofReachedLabels(true);
		tester.getInfo().setHeapSize(3);
		CountHeapOverflowListener listener = new CountHeapOverflowListener();
		tester.addCoverageListener(listener);
		
		tester.test("HeapOverflowde_uni_stuttgart_fmi_szs_jmoped_coverage_HeapOverflowTest_AllocatorClass_main_ALjava_lang_String_V");
		assertEquals(0, listener.labelsReached);
		
		tester.getInfo().setHeapSize(2);
		tester.test("HeapOverflowde_uni_stuttgart_fmi_szs_jmoped_coverage_HeapOverflowTest_AllocatorClass_main_ALjava_lang_String_V");
		assertEquals(1, listener.labelsReached);
	}

	public static class AllocatorClass
	{
		public int i;
		
		public AllocatorClass(int i)
		{
			this.i = i;
		}
		
		public static void main(String[] args)
		{
			AllocatorClass a1 = new AllocatorClass(1);
			AllocatorClass a2 = new AllocatorClass(2);
			AllocatorClass a3 = new AllocatorClass(3);
			AllocatorClass a4 = new AllocatorClass(4); // 4

			AllocatorClass a5 = new AllocatorClass(5);
			AllocatorClass a6 = new AllocatorClass(6);
			AllocatorClass a7 = new AllocatorClass(7);
			AllocatorClass a8 = new AllocatorClass(3); // 8

			AllocatorClass a9 = new AllocatorClass(2);
			
			assert(a1.i == 1);
			assert(a2.i == 2);
			assert(a3.i == 3);
			assert(a4.i == 4);
			
			assert(a5.i == 5);
			assert(a6.i == 6);
			assert(a7.i == 7);
			assert(a8.i == 3);
			
			assert(a9.i == 2);
		}
	}
	
	/**
	 * Original implementation of 
	 * {@link de.uni_stuttgart.fmi.szs.jmoped.MethodWrapper} causes
	 * HeapOverflow which should be ignored.
	 * @throws Exception
	 */
	public void testNondeterministicArrayMethod() throws Exception
	{
		CoverageTester tester = getTester(BubbleSort.class, "sort", "([I)V");
		tester.setFilterMethodName(tester.getPDS().getRootPDSMethod().getFormattedName());
		CountHeapOverflowListener listener = new CountHeapOverflowListener();
		tester.addCoverageListener(listener);
		PDSInfo info = tester.getInfo();
		tester.test();
		assertEquals(0, listener.overflows);
	}
	
	public void testNonDeterministicArrayMethod() throws Exception
	{
		CoverageTester tester = getTester(BubbleSort.class, "sort", "([I)V");
		PostStarReachabilityTester psTester = new PostStarReachabilityTester(tester.getPDS());
		psTester.setEnabledNotifyofReachedLabels(true);
		psTester.getInfo().setBits(2);
		CountHeapOverflowListener listener = new CountHeapOverflowListener();
		psTester.addCoverageListener(listener);
		psTester.test(psTester.getInfo().getLabelHeapOverflow());
		assertEquals(1, listener.labelsReached);
		assertEquals(0, listener.overflows);
	}
	
	public void testOverflowsWithAndWithoutFilterPrefix() throws Exception
	{
		CoverageTester tester = new CoverageTester(AllocatorClass.class);
		CountHeapOverflowListener listener = new CountHeapOverflowListener();
		tester.addCoverageListener(listener);
		
		for (int heapSize : new int[] { 1, 2, 5, 8, 19 }) {
			tester.setFilterMethodName(null);
			listener.overflows = 0;
			
			tester.test();
			assertTrue(listener.overflows > 0);
			
			int overflows = listener.overflows;
			listener.overflows = 0;
			
			PDSClass clazz = tester.getPDS().getRootPDSClass();
			PDSMethod method = clazz.getMethodMain();
			
			tester.setFilterMethodName(method.getFormattedName());
			
			tester.test();
			assertEquals(overflows, listener.overflows);
		}
	}
	
	public static class CountHeapOverflowListener extends CoverageAdapter
	{
		
		public int overflows = 0;
		public int lineOffset = -1;
		public int assertions = 0;
		public int covered = 0;
		public int labelsReached = 0;

		
		@Override
		public void coveredMethodOffset(CoverageMethod method, int lineOffset,
				int pcOffset, State state, Transition transition) 
		{
			System.out.println(lineOffset + " " + state);
			if (state == State.COVERED) {
				++covered;
			}
			else if (state == State.ASSERTION_FAILED) {
				assertions++;
			}
			else if (state == State.HEAP_OVERFLOW) {
				overflows++;
				this.lineOffset = lineOffset;
			}
		}
				
				
		public void labelReached(String label) 
		{
			++labelsReached;
		}
		
	}
	
	/**
	 * In the current implementation only the overflow error is only reached once
	 */
	public void testOverFlowReachedMultipleTimes() throws Exception
	{
		CoverageTester tester = getTester(MultiAllocClass.class, "test", "(I)V");
		tester.getInfo().setHeapSize(2);
		CoverageMethod method = getMethod(tester.test(), "test");
		List<Transition> overflows = method.getLines()[0].getTransitions(State.HEAP_OVERFLOW);
		assertEquals(1 , overflows.size());
	}
	
	public static class MultiAllocClass
	{
		public static void test(int i)
		{
			int[] array = new int[i];
		}
	}
	
	public static class BubbleSort
	{
		public static void main(String[] args)
		{
			// TODO undef works?
//			int[] array = new int[args.length];
			int [] array  = new int[2];
			for (int i = 0; i < array.length; i++) {
				array[i] = Integer.parseInt(args[0]);
			}
			new BubbleSort().sort(array);
			assert(new BubbleSort().isSorted(array));
			assert(!(new BubbleSort().isSorted(array)));
		}
		
		public static void driver()
		{
			int[] array = new int[2];
			for (int i = 0; i < array.length; i++) {
				array[i] = Integer.parseInt("");
			}
			new BubbleSort().sort(array);
			new BubbleSort().nop();
			// crashes unless there is another line here
		}
		
		public void nop()
		{
			
		}
		
		public void sort(int[] array)
		{
			for (int i = 0; i < array.length; i++) {
				for (int j = 0; j < array.length - 1; j++) {
					if (array[j] > array[j+1]) {
						int tmp = array[j];
						array[j] = array[j+1];
						array[j+1] = tmp;
					}
				}
			}
		}
		
		boolean isSorted(int[] array)
		{
			for (int i = 0; i < array.length - 1; i++) {
				if (array[i] > array[i+1]) {
					return false;
				}
			}
			return true;
		}
	}
	
}
