#include "CoverageTester.h"
#include "common.h"

/**
 * The callback method that is registered with our instance of the intProblem.
 *
 * It checks if the current state is an assert error state or if it's just a
 * new state that has been covered.
 *
 * @return true if all states we're interested are covered
 */
int callbackIsCoverageSolved(wTrans *transition, bddProblem *problem) {

	callbackContext *callback = (callbackContext*)problem->user;
	coverageContext *context = callback->context;
  
	if (IS_CANCELED(callback->monitor)) return 1;

	if (transition->from == problem->state 
			&& !bddEmpty(transition->label)
			&& !isMarkedAsReached(context, transition->name)) {
    
		if (matchPrefix(transition->name, context->stackPrefix)) {
			//printf("worked %s\n", wIdentString(transition->name));
			// notify progress monitor
			WORKED(1, callback->monitor);
			markAsReached(context, transition->name);
		} else {
			//printf("notified %s\n", wIdentString(transition->name));
		}
    
		JNIEnv *env = callback->env;
		const char *label = wIdentString(transition->name);
		fireCovered(env, callback->tester, label, 
			transition->from, transition->name,
			transition->to, callback->fireCovered);
	}
	return context->unvisitedStackElements == 0;
}

extern wIdent default_state_name;

/**
 * Creates a callback problem for post* analysis.
 */
intProblem* coverageProblem(JNIEnv *env, jobject analyzer, intPds *ipds,
		const char *stackPrefix, progressMonitor *monitor) {

	intProblem *problem = Malloc(sizeof(intProblem));
	problem->type = PROBLEM_REACHABILITY;
	problem->state = default_state_name;
	problem->isSolved = &callbackIsCoverageSolved;
	problem->user = createCallbackContext(env, analyzer, ipds, stackPrefix, monitor);
	//problem->queue = NULL; 
	//problem->queue = createPreferringNewTransitionsQueueStrategy();
	//problem->queue = createFIFOPreferringNewTransitionsQueueStrategy();
	//problem->queue = createFIFOQueueStratey();
	problem->queue = createLIFOQueueStratey();
	return problem;
}

jobject analyse(JNIEnv *env, jobject analyzer, intPds *ipds, 
		jstring sPrefix, progressMonitor *javaMonitor, jobject jpds) {

	collectGarbage(env);

	progressMonitor *analyseMonitor = CREATE_SUB_MONITOR(7000, javaMonitor);

	const char *stackPrefix = getJavaString(env, sPrefix);
	intProblem *iproblem = coverageProblem(env, analyzer, ipds, stackPrefix, analyseMonitor);

	jobject result = jniAnalyse(env, analyzer, ipds, iproblem, sPrefix, analyseMonitor, jpds);
  
	freeIntProblem(iproblem); 
	freeSubProgressMonitor(analyseMonitor);
	if (!stackPrefix) Free(stackPrefix);

	return result;
}

/**
 * Implements the native method analyseCoverage of CoverageTester.
 *
 * Creates a pushdown system by parsing the given file, a callback problem and
 * returns an instance of CoverageResult.
 */
JNIEXPORT jobject JNICALL
Java_de_uni_1stuttgart_fmi_szs_jmoped_coverage_CoverageTester_analyseCoverage__Ljava_lang_String_2Ljava_lang_String_2Lde_uni_1stuttgart_fmi_szs_jmoped_coverage_ProgressMonitor_2
(JNIEnv *env, jobject analyzer, jstring fName, jstring sPrefix, jobject monitor) {

	progressMonitor *javaMonitor = createJavaProgressMonitor(env, monitor);
	BEGIN_TASK("Starting coverage analysis", 10000, javaMonitor);

	progressMonitor *readMonitor = CREATE_SUB_MONITOR(3000, javaMonitor);
	const char *fileName = (*env)->GetStringUTFChars(env, fName, 0);
	intPds *ipds = readPds(fileName, readMonitor);
	(*env)->ReleaseStringUTFChars(env, fName, fileName);
	freeSubProgressMonitor(readMonitor);

	if (!ipds) {
		freeJavaProgressMonitor(javaMonitor);
		return NULL;
	}

	jobject result = analyse(env, analyzer, ipds, sPrefix, javaMonitor, NULL);

	freeJavaProgressMonitor(javaMonitor);

	return result;
}

/**
 * Implements the native method analyseCoverage of CoverageTester.
 * Use this method if a pds was already created.
 */
JNIEXPORT jobject JNICALL
Java_de_uni_1stuttgart_fmi_szs_jmoped_coverage_CoverageTester_analyseCoverage__Lde_uni_1stuttgart_fmi_szs_moped_IntPds_2Ljava_lang_String_2Lde_uni_1stuttgart_fmi_szs_jmoped_coverage_ProgressMonitor_2
(JNIEnv *env, jobject analyzer, jobject jpds, jstring sPrefix, jobject monitor) {

	progressMonitor *javaMonitor = createJavaProgressMonitor(env, monitor);
	BEGIN_TASK("Starting coverage analysis", 10000, javaMonitor);

	progressMonitor *readMonitor = CREATE_SUB_MONITOR(3000, javaMonitor);
	intPds *ipds = (intPds*) getHandlePointer(env, jpds, "handle");
	assert(ipds != NULL);
	freeSubProgressMonitor(readMonitor);

	jobject result = analyse(env, analyzer, ipds, sPrefix, javaMonitor, jpds);
	
	freeJavaProgressMonitor(javaMonitor);

	return result;
}
