#include "common.h"
#include <ir.h>

/*****************************************/
/** FUNCTIONS FOR HANDLING JAVA OBJECTS **/
/*****************************************/

/**
 * The data structure for pairs of Java objects and their
 * native data structure.
 */
typedef struct peerReference {

	jobject object;
	void *handle;
	deconstructor delete;
	struct peerReference *next;
} peerReference;

/**
 * The global linked list keeping references to pairs of Java objects
 * and their native data structure.
 */
static peerReference *references = NULL;

/**
 * Creates a new peerReferece (object, handle) and puts it into
 * the list references.
 */
void registerReference(JNIEnv *env, jobject object, void *handle, 
		       deconstructor dec) {

	peerReference *reference = Malloc(sizeof(peerReference));
	reference->object = (*env)->NewWeakGlobalRef(env, object);
	reference->handle = handle;
	reference->delete = dec;
	reference->next = references;
	references = reference;
}

/**
 * Frees the native data structure if its corresponding Java object
 * was freed by the garbage collector.
 */
void collectGarbage(JNIEnv *env) {

	peerReference *precursor = NULL;
	peerReference *current = references;

	while (current) {
		// frees the handle if the object was freed by the garbage collector
    		if ((*env)->IsSameObject(env, current->object, NULL)) {
			if (precursor) {
				precursor->next = current->next;
			} else {
				references = current->next;
			}
			current->delete(current->handle);
			Free(current);
			if (precursor) {
				current = precursor->next;
			} else {
				current = references;
			}
    		} else {
			precursor = current;
			current = current->next;
		}
	}
}

/**
 * General constructor.
 */
jobject initPeerObject(JNIEnv *env, const char *className, void *handle,
			 deconstructor dec) {

	jclass clazz = (*env)->FindClass(env, className);
	jmethodID constructor = (*env)->GetMethodID(env, clazz, "<init>", "(I)V");
	assert(constructor != NULL);
	jobject object = (*env)->NewObject(env, clazz, constructor, (int)handle);
	assert(object != NULL);
	registerReference(env, object, handle, dec);

	(*env)->DeleteLocalRef(env, clazz);

	return object;
}

/**
 * Destructor for IntPds.
 */
static void deleteIntPds(void *handle) {

	intPdsDelete((intPds*) handle);
}

/**
 * Constructor for IntPds.
 */
jobject initIntPds(JNIEnv *env, intPds *pds) {

	return initPeerObject(env, "de/uni_stuttgart/fmi/szs/moped/IntPds",
		pds, &deleteIntPds);
}

/**
 * Destructor for WFa.
 */
static void deleteWFa(void *handle) {

	wFADelete((wFA*) handle);
}

/**
 * Constructor for WFa.
 */
jobject initWFa(JNIEnv *env, wFA *fa) {

	jclass clazz = (*env)->FindClass(env, "de/uni_stuttgart/fmi/szs/moped/WFa");
	jmethodID constructor = (*env)->GetMethodID(env, clazz, "<init>", "(II)V");
	assert(constructor != NULL);
	jobject jfa = (*env)->NewObject(env, clazz, constructor, fa, fa->dir == SAT_PRE ? 0 : 1);
	assert(jfa != NULL);
	registerReference(env, jfa, fa, &deleteWFa);

	(*env)->DeleteLocalRef(env, clazz);

	return jfa;
}

/**
 * Constructor for ConverageResult.
 */
jobject initCoverageResult(JNIEnv *env, jobject pds, jobject fa) {

	jclass clazz = (*env)->FindClass(env, 
		"de/uni_stuttgart/fmi/szs/jmoped/coverage/CoverageResult");
	jmethodID constructor = (*env)->GetMethodID(env, clazz, "<init>", 
			"(Lde/uni_stuttgart/fmi/szs/moped/IntPds;"
			"Lde/uni_stuttgart/fmi/szs/moped/WFa;)"
			"V");
	assert(constructor != NULL);
	jobject result= (*env)->NewObject(env, clazz, constructor, pds, fa);
	assert(result != NULL);

	(*env)->DeleteLocalRef(env, clazz);

	return result;
}

static void deleteIntConfig(void *handle) {

	intDeleteConfig((intConfig*) handle);
}

jobject createIntConfig(JNIEnv *env, intConfig *config, jobject ipds) {

	jclass clazz = (*env)->FindClass(env, 
		"de/uni_stuttgart/fmi/szs/jmoped/coverage/IntConfig");
	jmethodID constructor = (*env)->GetMethodID(env, clazz, "<init>", 
		"(ILde/uni_stuttgart/fmi/szs/moped/IntPds;)V");
  	assert(constructor != NULL);
	jobject object = (*env)->NewObject(env, clazz, constructor, (int)config, ipds);
	assert(object != NULL);
	registerReference(env, object, config, &deleteIntConfig);

	(*env)->DeleteLocalRef(env, clazz);

	return object;
}

/****************************/
/** MAIN ANALYSIS FUNCTION **/
/****************************/

jobject jniAnalyse(JNIEnv *env, jobject analyzer, intPds *ipds, intProblem *problem,
		jstring sPrefix, progressMonitor *monitor, jobject jpds) {

	coverageContext *context = ((callbackContext*) problem->user)->context;

	//excludeIgnoredPrefixes(env, tester, ipds, context);

	BEGIN_TASK("Analysing", context->unvisitedStackElements, monitor);
	setProgressMonitor(ipds, monitor);

	wFA *fa;
	wTrans *found;
	switch (problem->type) {
	case PROBLEM_REACHABILITY:
		fa = bddPoststar(ipds->pds, problem, &found);
		break;
	case PROBLEM_REACHABILITY_PRE:
		fa = bddPrestar(ipds->pds, problem, &found);
		break;
	case PROBLEM_REACHABILITY_2:
		fa = twoPoststar(ipds, problem, &found);
		break;
	case PROBLEM_REACHABILITY_PRE_2:
		fa = twoPrestar(ipds, problem, &found);
		break;
	}

	resetProgressMonitor(ipds);
	DONE(monitor);

	// convert unreached states to Java String array 
	wIdentList *states = getUnreachableStates(ipds, context);
	if (states) {
		jobjectArray labels = createStringArray(env, states);
		jclass clazz = (*env)->GetObjectClass(env, analyzer);
		jmethodID method = (*env)->GetMethodID(env, clazz, 
			"unreachedLabels", "([Ljava/lang/String;)V");
		assert(method != NULL);
		(*env)->CallVoidMethod(env, analyzer, method, labels);
		(*env)->DeleteLocalRef(env, clazz);
		(*env)->DeleteLocalRef(env, labels);
		freeWIdentList(states);
	}

	if (jpds == NULL) jpds = initIntPds(env, ipds);
	jobject jfa = initWFa(env, fa);
	return initCoverageResult(env, jpds, jfa);
}

/*jobject analyseCoverageJNI(JNIEnv *env, jobject tester, intPds *ipds, intProblem *problem,
			   jstring sPrefix, progressMonitor *analyseMonitor, 
			   jobject jpds)
{
  wIdentList *states;
  //intResult *result;

  // commented out for now
   //excludeIgnoredPrefixes(env, tester, ipds,  
 //			 ((callbackContext*)problem->user)->context); 

  // notify monitor of work amount
  BEGIN_TASK("Analysing", ((callbackContext*)problem->user)->
	     context->unvisitedStackElements, analyseMonitor);  

  // set progress monitor for cudd
  setProgressMonitor(ipds, analyseMonitor);

  // run coverage analysis and retrieve unreached states
  //result = intAnalyse(ipds, problem);
  wFA *fa = intCoverage(ipds, problem);

  // reset progress monitor
  resetProgressMonitor(ipds);

  DONE(analyseMonitor);

  states = getUnreachableStates(ipds,
				((callbackContext*)problem->user)->context);

  // convert unreached states to Java String array 
  if (states) {
    jobjectArray result = createStringArray(env, states);
    jclass clazz = (*env)->GetObjectClass(env, tester);
    jmethodID method = (*env)->GetMethodID(env, clazz, "unreachedLabels",
					   "([Ljava/lang/String;)V");
    assert(method != NULL);
    (*env)->CallVoidMethod(env, tester, method, result);
    (*env)->DeleteLocalRef(env, clazz);
    (*env)->DeleteLocalRef(env, result);
    freeWIdentList(states);
  }
  
  //jpds = (jpds != NULL) ? jpds : createIntPds(env, ipds);
  //int isPost;
  //if (problem->type == PROBLEM_REACHABILITY 
	//	|| problem->type == PROBLEM_REACHABILITY_2)
	//isPost = 1;
//  else isPost = 0;
  //jobject jresult = createIntResult(env, result, jpds, isPost);
  //return createCoverageResult(env, jpds, jresult);
  if (jpds == NULL) jpds = initIntPds(env, ipds);
  jobject jfa = initWFa(env, fa);
  return initCoverageResult(env, jpds, jfa);
}*/

/****************************/
/** JAVA UTILITY FUNCTIONS **/
/****************************/

/**
 * Gets char* from a jstring 
 */
char* getJavaString(JNIEnv *env, jstring string) {

	if (string == NULL) return NULL;

	const char *native = (*env)->GetStringUTFChars(env, string, 0);
	char *copy = Strdup(native);
	(*env)->ReleaseStringUTFChars(env, string, native);
	return copy;
}

/**
 * Gets wIdent from a jstring 
 */
wIdent getIdent(JNIEnv *env, jstring string) {

	const char *native = (*env)->GetStringUTFChars(env, string, 0);
	wIdent ident = wIdentCreate(native);
	(*env)->ReleaseStringUTFChars(env, string, native);
	return ident;
}

/**
 * Gets wIdent from a Java method.
 */
wIdent getIdentFromMethod(JNIEnv *env, jobject object, const char *method) {

	jclass clazz = (*env)->GetObjectClass(env, object);
	jmethodID mId = (*env)->GetMethodID(env, clazz, method,
				"()Ljava/lang/String;");
	assert(mId != NULL);
	jstring string = (*env)->CallObjectMethod(env, object, mId);
	wIdent ident = getIdent(env, string);
	(*env)->DeleteLocalRef(env, string);
	return ident;
}

/**
 * Gets C pointer from the field of object.
 */
void* getHandlePointer(JNIEnv *env, jobject object, const char *fieldname) {

	jclass clazz = (*env)->GetObjectClass(env, object);
	jfieldID field = (*env)->GetFieldID(env, clazz, fieldname, "I");
	assert(field != NULL);
	(*env)->DeleteLocalRef(env, clazz);

	return (void*)(*env)->GetIntField(env, object, field);
}

/**
 * Gets boolean value from the field of object.
 */
jboolean getBooleanField(JNIEnv *env, jobject object, const char *fieldname) {

	jclass clazz = (*env)->GetObjectClass(env, object);
	jfieldID field = (*env)->GetFieldID(env, clazz, fieldname, "Z");
	assert(field != NULL);
	(*env)->DeleteLocalRef(env, clazz);

	return (*env)->GetBooleanField(env, object, field);
}

/**
 * Creates a wIdentList list and fills it with the labels from the string
 * array <code>jLabels</code>. Order of labels is inverted in the process.
 */
wIdentList* extractLabels(JNIEnv* env, jobjectArray jLabels) {

	wIdentList* list = NULL;
	jsize length = (*env)->GetArrayLength(env, jLabels);
	jsize i;
  
	for (i = 0; i < length; i++) {

		jstring jLabel = (*env)->GetObjectArrayElement(env, jLabels, i);
		const char *label = (*env)->GetStringUTFChars(env, jLabel, 0);
		insert(wIdentCreate(label), &list);
		(*env)->ReleaseStringUTFChars(env, jLabel, label);
		(*env)->DeleteLocalRef(env, jLabel);
	}

	return list;
}

/**
 * Creates a callback context object.
 *
 * @param stackPrefix if not NULL, only stack symbols starting with
 * <code>startPrefix</code> are checked for coverage.
 */
callbackContext* createCallbackContext(JNIEnv *env, jobject tester, 
		intPds *ipds, const char *stackPrefix, progressMonitor *monitor) {

	callbackContext *context = Malloc(sizeof(callbackContext));
	wIdentList *excluded = NULL;
	context->env = env;
	context->tester = tester;
	context->monitor = monitor;

	jclass clazz = (*env)->GetObjectClass(env, tester);
	context->fireCovered = (*env)->GetMethodID(env, clazz, 
		"fireCoveredLabel", "(Ljava/lang/String;III)V");
	assert(context->fireCovered != NULL);
	(*env)->DeleteLocalRef(env, clazz);

	// TODO filter exception states, get them from the tester
  
	context->context = createCoverageContext(ipds, stackPrefix, excluded);
	return context;
}

/**
 * Frees the heap allocated resources of <code>context</code>.
 */
void freeCallbackContext(callbackContext *context) {

	freeCoverageContext(context->context);
	Free(context);
}

/**
 * Frees the heap allocated resources of a callback <code>problem</code>.
 */
void freeIntProblem(intProblem *problem)
{
  freeCallbackContext((callbackContext*)problem->user);
  if (problem->queue) {
    freeQueueStrategy(problem->queue);
  }
  Free(problem);
}

reachabilityContext* createReachabilityContext(JNIEnv *env, 
					       jobject tester,
					       progressMonitor *monitor,
					       wIdentList **labels) 
{ 
  reachabilityContext *context = Malloc(sizeof(reachabilityContext)); 
  context->env = env;
  context->tester = tester;
  context->monitor = monitor;
  context->labels = labels;
  // TODO
  /*
  context->heapOverflowLabel = getIdentFromMethod(env, tester, 
						  "getHeapOverflowLabel");
  context->assertErrorLabel = getIdentFromMethod(env, tester,
						 "getAssertErrorLabel");
  */
  return context;
} 

void freeReachabilityContext(reachabilityContext *context)
{
  Free(context);
}

extern int irVerbose;

JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *jvm, void *reserved)
{
  // turn off verbosity
  irVerbose = 0;
  wInit();
  bddInit();
  intInit();
  return JNI_VERSION_1_4;
}

JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *jvm, void *reserved)
{
  bddFinish();
  wFinish();
}

void fireCoveredLabel(JNIEnv *env, jobject tester, const char *label)
{
  fireCoveredLabelWithTransition(env, tester, label, -1, -1, -1);
}

void fireCoveredLabelWithTransition(JNIEnv *env, jobject tester,
				    const char *label, long from, long name,
				    long to)
{
  jclass clazz = (*env)->GetObjectClass(env, tester);
  jmethodID mId = (*env)->GetMethodID(env, clazz, "fireCoveredLabel", 
				      "(Ljava/lang/String;III)V");
  assert(mId != NULL);
  fireCovered(env, tester, label, from, name, to, mId);
  (*env)->DeleteLocalRef(env, clazz);
}

void fireCovered(JNIEnv *env, jobject tester, const char *label, long from,
		 long name, long to, jmethodID method)
{
/*   printf("%s - %s - %s\n", wIdentString(from), wIdentString(name),   */
/*  	 wIdentString(to));  */
/*   printf("%d - %d - %d\n", from, name, to);  */
  jstring jLabel = (*env)->NewStringUTF(env, label);
  (*env)->CallVoidMethod(env, tester, method, jLabel, from, name, to);
  (*env)->DeleteLocalRef(env, jLabel);
}

intPds* readPds(const char *filename, progressMonitor *monitor)
{
  intPds *pds = irReadPdsMonitored(filename, 0, monitor);
  if (pds && IS_CANCELED(monitor)) {
    intPdsDelete(pds);
    pds = NULL;
  }
  return pds;
}

#include <sys/time.h>
#include <time.h>

typedef struct javaProgressMonitorData
{
  JNIEnv *env;
  jobject monitor;
  jmethodID worked;
  jmethodID isCanceled;
  
  int lastWorkUpdate;
  suseconds_t lastCancelCheck;

  int accumlatedWork;

  suseconds_t timeUsed;

} javaProgressMonitorData;

#define CANCEL_DELAY 2500
#define WORK_DELAY 2500

static int isCanceled(progressMonitor *this)
{
  javaProgressMonitorData *data = (javaProgressMonitorData*)this->user;
  JNIEnv *env = data->env;
  if (data->lastCancelCheck % 100 == 0) {
    jboolean canceled = 
      (*env)->CallBooleanMethod(env, data->monitor, data->isCanceled);
    if (!canceled) {
      data->lastCancelCheck++;
    }
    return canceled;
  }
  else {
      data->lastCancelCheck++;
  }
  return 0;
}

static void worked(int amount, progressMonitor *this)
{
  javaProgressMonitorData *data = (javaProgressMonitorData*)this->user;
  JNIEnv *env = data->env;
  data->lastWorkUpdate++;
  if (data->lastWorkUpdate % 1 == 0) {
    (*env)->CallVoidMethod(env, data->monitor, data->worked, 
			   amount + data->accumlatedWork);
  }
}

static void beginTask(const char *name, int workAmount,
		      progressMonitor *this)
{
  javaProgressMonitorData *data = (javaProgressMonitorData*)this->user;
  JNIEnv *env = data->env;

  // flush work amount
  if (data->accumlatedWork > 0) {
    data->lastWorkUpdate = 0;
    worked(0, this);
  }

  // TODO cache method id and class
  jclass clazz = (*env)->GetObjectClass(env, data->monitor);
  jmethodID mId = (*env)->GetMethodID(env, clazz, "beginTask", 
				      "(Ljava/lang/String;I)V");
  jstring jName = (*env)->NewStringUTF(env, name);
  assert(mId != NULL);
  (*env)->CallVoidMethod(env, data->monitor, mId, jName, workAmount);
  (*env)->DeleteLocalRef(env, jName);
  (*env)->DeleteLocalRef(env, clazz);
}


static void subTask(const char *name, progressMonitor *this)
{
  javaProgressMonitorData *data = (javaProgressMonitorData*)this->user;
  JNIEnv *env = data->env;
  // TODO cache method id and class
  jclass clazz = (*env)->GetObjectClass(env, data->monitor);
  jmethodID mId = (*env)->GetMethodID(env, clazz, "subTask", 
				      "(Ljava/lang/String;)V");
  jstring jName = (*env)->NewStringUTF(env, name);
  assert(mId != NULL);
  (*env)->CallVoidMethod(env, data->monitor, mId, jName);
  (*env)->DeleteLocalRef(env, jName);
  (*env)->DeleteLocalRef(env, clazz);
}

static void done(progressMonitor *this)
{
 javaProgressMonitorData *data = (javaProgressMonitorData*)this->user;
  JNIEnv *env = data->env;
  // TODO cache method id and class
  jclass clazz = (*env)->GetObjectClass(env, data->monitor);
  jmethodID mId = (*env)->GetMethodID(env, clazz, "done", "()V");
  (*env)->CallVoidMethod(env, data->monitor, mId);
  (*env)->DeleteLocalRef(env, clazz);
}

progressMonitor* createJavaProgressMonitor(JNIEnv *env, jobject jMonitor)
{
  javaProgressMonitorData *data = Malloc(sizeof(javaProgressMonitorData));
  jclass clazz;

  data->env = env;
  data->monitor = jMonitor;

  progressMonitor *monitor = Malloc(sizeof(progressMonitor));
  monitor->user = data;

  monitor->isCanceled = &isCanceled;
  monitor->beginTask = &beginTask;
  monitor->worked = &worked;
  monitor->subTask = &subTask;
  monitor->done = &done;

  clazz = (*env)->GetObjectClass(env, data->monitor);
  data->worked = (*env)->GetMethodID(env, clazz, "worked", "(I)V");
  assert(data->worked != NULL);

  data->isCanceled = (*env)->GetMethodID(env, clazz, "isCanceled", "()Z");
  assert(data->isCanceled != NULL);

  data->lastWorkUpdate = 0;
  data->lastCancelCheck = 0;
  data->accumlatedWork = 0;
  data->timeUsed = 0;

  return monitor;
}

void freeJavaProgressMonitor(progressMonitor *monitor)
{
  if (monitor) {
    monitor->done(monitor);
    Free(monitor->user);
    Free(monitor);
  }
}

jobjectArray createStringArray(JNIEnv *env, wIdentList *list)
{
  if (!list) {
    return NULL;
  }
  int length = lengthSymList(list);
  jsize index;
  wIdentList* state = list;
  jclass stringClass = (*env)->FindClass(env, "java/lang/String");
  jobjectArray result = (*env)->NewObjectArray(env, length, stringClass, 
					       NULL);

  for (index = 0; index < length; index++, state = state->next) {
    const char *label = wIdentString(state->wId);
    jstring jLabel = (*env)->NewStringUTF(env, label);
    (*env)->SetObjectArrayElement(env, result, index, jLabel);
    (*env)->DeleteLocalRef(env, jLabel);
  }
  return result;
}

//typedef struct coverageResultRef
//{
  /* Weak reference to java object */
//  jobject coverageResult;
  /* Reference counting reference to intPds */
//  intPds *pds;
  /* Reference to result */
//  intResult *result;
  /* Pointer to next coverageResultRef */
//  struct coverageResultRef *next;
//} coverageResultRef;

/* Pointer to coverageResultRef list of coverage results still in use */
//static coverageResultRef *resultReferences = NULL;

/* Callback to create a CoverageResult object */
/*jobject createCoverageResult(JNIEnv *env, jobject pds, jobject result)
{
  jclass clazz = (*env)->FindClass(env, "de/uni_stuttgart/fmi/szs/jmoped"
				     "/coverage/CoverageResult");
  jmethodID constructor = 
    (*env)->GetMethodID(env, clazz, "<init>", 
			"(Lde/uni_stuttgart/fmi/szs/jmoped/coverage/IntPds;"
			"Lde/uni_stuttgart/fmi/szs/jmoped/coverage/IntResult;)"
			"V");
  jobject coverageResult;

  assert(constructor != NULL);
  coverageResult = (*env)->NewObject(env, clazz, constructor, pds, result);
  assert(result != NULL);

  (*env)->DeleteLocalRef(env, clazz);

  return coverageResult;
}*/

/*void collectGarbage(JNIEnv *env)
{
  coverageResultRef *precursor = NULL;
  coverageResultRef *current = resultReferences;
  while (current) {
    if ((*env)->IsSameObject(env, current->coverageResult, NULL)) {

      if (precursor) {
	precursor->next = current->next;
      }
      else {
	resultReferences = current->next;
      }

      (*env)->DeleteWeakGlobalRef(env, current->coverageResult);
      intDeleteResult(current->result);
      intPdsDelete(current->pds);
      Free(current);
      
      if (precursor) {
	current = precursor->next;
      }
      else {
	current = resultReferences;
      }
    }
    else {
      precursor = current;
      current = current->next;
    }
  }
  garbageCollectReferences(env);
}*/

/*void* getHandlePointer(JNIEnv *env, jobject object, const char *fieldname)
{
  jclass clazz = (*env)->GetObjectClass(env, object);
  jfieldID field = (*env)->GetFieldID(env, clazz, fieldname, "I");

  assert(field != NULL);
  (*env)->DeleteLocalRef(env, clazz);
  return (void*)(*env)->GetIntField(env, object, field);
}

jboolean getBooleanField(JNIEnv *env, jobject object, const char *fieldname)
{
  jclass clazz = (*env)->GetObjectClass(env, object);
  jfieldID field = (*env)->GetFieldID(env, clazz, fieldname, "Z");

  assert(field != NULL);
  (*env)->DeleteLocalRef(env, clazz);
  return (*env)->GetBooleanField(env, object, field);
}*/

/*typedef struct peerReference
{
  jobject object;
  void *handle;
  deconstructor delete;
  struct peerReference *next;
} peerReference;

static peerReference *references = NULL;*/

/*void garbageCollectReferences(JNIEnv *env)
{
  peerReference *precursor = NULL;
  peerReference *current = references;*/
  //peerReference *pre = NULL;
  //peerReference *next = NULL;

	/*while (current) {
		next = current->next;
		if ((*env)->IsSameObject(env, current->object, NULL)) {
			current->delete(current->handle);
			Free(current);
			if (pre) pre->next = next;
		} else {
			pre = current;
		}
		current = next;
	}*/


/*  while (current) {
    // frees the handle if the object was freed by the garbage collector
    if ((*env)->IsSameObject(env, current->object, NULL)) {
      if (precursor) {
	precursor->next = current->next;
      }
      else {
	references = current->next;
      }
      current->delete(current->handle);
      Free(current);
      if (precursor) {
	current = precursor->next;
      }
      else {
	current = references;
      }
    }
    else {
      precursor = current;
      current = current->next;
    }
  }
}*/

/*void registerReference(JNIEnv *env, jobject object, void *handle, 
		       deconstructor dec)
{
  peerReference *reference = Malloc(sizeof(peerReference));
  reference->object = (*env)->NewWeakGlobalRef(env, object);
  reference->handle = handle;
  reference->delete = dec;*/
  
  /* put wrapper into list */
/*  reference->next = references;
  references = reference;
}*/

/*jobject createPeerObject(JNIEnv *env, const char *className, void *handle,
			 deconstructor dec)
{
  jclass clazz = (*env)->FindClass(env, className);
  jmethodID constructor = (*env)->GetMethodID(env, clazz, "<init>", "(I)V");
  jobject object;

  assert(constructor != NULL);
  object = (*env)->NewObject(env, clazz, constructor, (int)handle);
  assert(object != NULL);

  registerReference(env, object, handle, dec);

  (*env)->DeleteLocalRef(env, clazz);

  return object;
}*/

/*static void deleteIntConfig(void *handle)
{
  intDeleteConfig((intConfig*)handle);
}

jobject createIntConfig(JNIEnv *env, intConfig *config, jobject ipds)
{
  jclass clazz = (*env)->FindClass(env, 
				   "de/uni_stuttgart/fmi/szs/jmoped/coverage/"
				   "IntConfig");
  jmethodID constructor = 
    (*env)->GetMethodID(env, clazz, "<init>", "(I"
			"Lde/uni_stuttgart/fmi/szs/jmoped/coverage/"
			"IntPds;)V");
  jobject object;

  assert(constructor != NULL);
  object = (*env)->NewObject(env, clazz, constructor, (int)config, ipds);
  assert(object != NULL);

  registerReference(env, object, config, &deleteIntConfig);

  (*env)->DeleteLocalRef(env, clazz);

  return object;
}*/

/*static void deleteIntPds(void *handle)
{
  intPdsDelete((intPds*)handle);
}*/

/* Callback to create an IntPds object. */
/*jobject createIntPds(JNIEnv *env, intPds *ipds)
{
  return createPeerObject(env, 
			  "de/uni_stuttgart/fmi/szs/jmoped/coverage/IntPds", 
			  ipds, &deleteIntPds);
}*/

/*static void deleteIntResult(void *handle)
{
  intResult *iresult = (intResult*) handle;
  intDeleteResult((intResult*)handle);
}*/

/* Callback to create an IntResult object. */
/*jobject createIntResult(JNIEnv *env, intResult *result, jobject ipds, 
			int isPostTrace)
{
  jclass clazz = (*env)->FindClass(env, 
				   "de/uni_stuttgart/fmi/szs/jmoped/coverage/"
				   "IntResult");
  jmethodID constructor = 
    (*env)->GetMethodID(env, clazz, "<init>", "(I"
			"Lde/uni_stuttgart/fmi/szs/jmoped/coverage/"
			"IntPds;Z)V");
  jobject object;
  
  assert(constructor != NULL);
  object = (*env)->NewObject(env, clazz, constructor, (int)result, ipds, 
			     isPostTrace ? JNI_TRUE : JNI_FALSE);
  assert(object != NULL);

  registerReference(env, object, result, &deleteIntResult);

  (*env)->DeleteLocalRef(env, clazz);

  return object;
}*/

void excludeIgnoredPrefixes(JNIEnv *env, jobject tester, intPds *ipds, 
			    coverageContext *context)
{
  jclass clazz = (*env)->FindClass(env, 
				   "de/uni_stuttgart/fmi/szs/jmoped/coverage/"
				   "Coverage");
  jmethodID method = (*env)->GetMethodID(env, clazz, "getIgnoredPrefixes",
					 "()[Ljava/lang/String;");
  jobjectArray array;
  wIdentList *ignoredPrefixes;

  assert(method != NULL);

  array = (*env)->CallObjectMethod(env, tester, method);
  assert(array != NULL);

  ignoredPrefixes = extractLabels(env, array);
  
  excludeStackPrefixes(ipds, context, ignoredPrefixes);
  freeWIdentList(ignoredPrefixes);
}
