#ifndef COMMON_H
#define COMMON_H

#include <coverage.h>
#include <absRef.h>
#include <int.h>
#include <two.h>
#include <ir.h>

#include <jni.h>

typedef struct reachabilityContext
{ 
  JNIEnv *env;
  jobject tester;
  progressMonitor *monitor;
  wIdentList **labels;
  
  wIdent state;
  wIdent stack;
  wIdent finalstate;
  
} reachabilityContext;

/**
 * Callback context struct, which is used as the private user data in 
 * <code>intProblem.user</code>.
 * <p>
 * Holds the Java environment and instance from the calling context, the
 * coverageContext data and the name of the assert error state.
 */
typedef struct callbackContext
{
  coverageContext *context;
  JNIEnv *env;
  jobject tester;
  progressMonitor *monitor;
  jmethodID fireCovered;
} callbackContext;

jobject jniAnalyse(JNIEnv *env, jobject analyzer, intPds *ipds, intProblem *problem, 
			   jstring sPrefix, progressMonitor *analyseMonitor, 
			   jobject jpds);

callbackContext* createCallbackContext(JNIEnv *env, jobject tester, 
				       intPds *ipds, const char *stackPrefix,
				       progressMonitor *monitor);

void freeCallbackContext(callbackContext *context);

void freeIntProblem(intProblem *problem);

reachabilityContext* createReachabilityContext(JNIEnv *env, 
					       jobject tester,
					       progressMonitor *monitor,
					       wIdentList **labels);

void freeReachabilityContext(reachabilityContext *context);

wIdentList* extractLabels(JNIEnv *env, jobjectArray jLabels);

void fireCoveredLabel(JNIEnv *env, jobject tester, const char *label);

void fireCoveredLabelWithTransition(JNIEnv *env, jobject tester, 
				    const char *label, long from, long name,
				    long to);

void fireCovered(JNIEnv *env, jobject tester, const char *label, long from,
		 long name, long to, jmethodID method);

/**
 * Returns a copy of the java string, so memory management is the caller's
 * responsibility.
 */
char* getJavaString(JNIEnv *env, jstring string);

wIdent getIdentFrom(JNIEnv *env, jstring string);

/**
 * Calls <code>method</code> on <code>object</code> and creates and
 * identifier from the string that is returned.
 */
wIdent getIdentFromMethod(JNIEnv *env, jobject object, const char *method);
	
/**
 * Called when the library is loaded, initializes wpds and moped.
 */
JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *jvm, void *reserved);

/**
 * Called when library is unloaded.
 */
JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *jvm, void *reserved);

/**
 * Calls irReadPdsMonitored using the java progress monitor.
 * 
 * @return NULL if monitor was canceled.
 */
intPds* readPds(const char *filename, progressMonitor *monitor);

/**
 * Creates a progress monitor that forwards all calls to it java progress
 * monitor.
 */
progressMonitor* createJavaProgressMonitor(JNIEnv *env, jobject jMonitor);

/**
 * Frees the java progress monitor and also ensures that it is done.
 */
void freeJavaProgressMonitor(progressMonitor *monitor);

/**
 * Creates a string array from a wIdent list. Returns NULL if list is NULL.
 */
jobjectArray createStringArray(JNIEnv *env, wIdentList *list);

/**
 * Creates a CoverageResult, sets its fields to pds and result, and stores a
 * weak reference of it in a struct that also manages pds and result, so they
 * can be deleted if the weak reference points to NULL.
 * 
 * Side effect: the ref count of <code>pds</code> is increased.
 */
jobject createCoverageResult(JNIEnv *env, jobject ipds, jobject result);

/**
 * Delete push down systems and results that are referenced by no longer
 * exiting CoverageResults.
 *
 * Decreases the reference counter of the respective pds and deletes it if
 * applicable.
 */
void collectGarbage(JNIEnv *env);

/**
 * Returns the pointer from the handle field
 */
void* getHandlePointer(JNIEnv *env, jobject object, const char *fieldname);

jboolean getBooleanField(JNIEnv *env, jobject object, const char *fieldname);

typedef void (*deconstructor)(void *handle);

/**
 * Creates a Java peer class of type <code>className</code> for
 * <code>handle</code>.
 * <p>
 * <code>dec is called when the native handle can be garbage collected.
 */
jobject createPeerObject(JNIEnv *env, const char *className, void *handle,
			 deconstructor dec);

void garbageCollectReferences(JNIEnv *env);

jobject createIntConfig(JNIEnv *env, intConfig *config, jobject ipds);

jobject initIntPds(JNIEnv *env, intPds *ipds);

jobject createIntResult(JNIEnv *env, intResult *result, jobject ipds, 
			int isPostTrace);

void excludeIgnoredPrefixes(JNIEnv *env, jobject tester, intPds *ipds, 
			    coverageContext *context);

#endif
