/****************************************************************************/

#include "bdd.h"
#include "absRef.h"
#include "dpn.h"

#include <cudd.h>

#include <time.h>

#include "../jni/common.h"

/****************************************************************************/

int ABSTRACTION = 0;
int NEWTRACES = 0;
/* from main.c */

/****************************************************************************/

static wFA* bddComputePrestar (bddPds *pds, wFA *fa,
			       bddProblem *problem, wTrans **found)
{
        wSat *prcs = wSatCreate(pds->pds->sr,SAT_PRE,
			NEWTRACES? TRACE_REF : TRACE_COPY);
	wIdent state = pds->start->state;
	wIdent stack = *(pds->start->stack);
	wIdent finalstate = bddFinalState();
        wRule *pt;
	wTrans *t;

	/* Remy: for callbacks (debugging purpose) */
	callbackContext *context = (callbackContext*) problem->user;
	progressMonitor *monitor = NULL;
	if (context) monitor = context->monitor;

	SUB_TASK("problem->state: ", monitor);
	SUB_TASK(wIdentString(problem->state), monitor);
	SUB_TASK("problem->stack: ", monitor);
	SUB_TASK(wIdentString(problem->stack), monitor);

	/* Add all the rules */
        for (pt = pds->pds->rules; pt; pt = pt->next)
	{
		if (pt->to_stack1)
		{
			wSatAugRule(prcs,pt->from_state,pt->from_stack,
				pt->to_state,pt->to_stack1,pt->to_stack2,
				pt->label,pt);
			continue;
		}

		/* pop rules need a different signature */
		bddDomain *d = bddDomConvert(pt->label);
		wSatAugRule(prcs,pt->from_state,pt->from_stack,
			pt->to_state,pt->to_stack1,pt->to_stack2,d,pt);
		bddDomDeref(d);
	}

	/* Add all initial transitions (to the queue) */
        for (t = fa->transitions; t; t = t->next)
                 wSatAddTrans(prcs,t->from,t->name,t->to,t->label);

	if (problem->queue) {
	  while (1) {
	    queueStrategy *queue = problem->queue;
	    while (t = wSatPopTrans(prcs)) {
	      queue->pushTransition(t, queue);
	    }
	    t = queue->popTransition(queue);
	    if (!t 
		|| (problem->isSolved != NULL && problem->isSolved(t, problem))
		|| (t->from == state && t->name == stack && t->to == finalstate 
		    && !bddEmpty(t->label)))
	      break;

	    /* fixme: we should prevent propagation of 'false' values
	       at this point (for efficiency) */
	    wSatAugTrans(prcs, t);
	  }
	}
	else {
	  /* Process all the transitions */
	  while ((t = wSatPopTrans(prcs)))
	  {
	    /* Remy: Callback */
	    if (monitor) {
		monitor->subTask(wIdentString(t->name), monitor);
	    }
	    int verbose = 1;
	    if (verbose) {
	    printf(wIdentString(t->from)); printf(" -");
	    printf(wIdentString(t->name)); printf("-> ");
	    printf(wIdentString(t->to));
	    printf(", "); 
	    if (bddEmpty(t->label))
		printf("bdd empty");
	    else
		printf("bdd NOT empty");
	    printf("\n");
	    }

	    if ((problem->isSolved != NULL && problem->isSolved(t, problem))
		|| (t->from == state && t->name == stack && t->to == finalstate 
		    && !bddEmpty(t->label)))
	      break;
	    
	    /* fixme: we should prevent propagation of 'false' values
	       at this point (for efficiency) */
	    wSatAugTrans(prcs,t);
	  }
	}
	    
	if (problem->queue && problem->queue->delete) {
	  /* Call optional destructor */
	  problem->queue->delete(problem->queue);
	}

	/* we could have left the loop because isSolved is true which is
	   abused to interrupt the computation */
	if (!t || 
	    (t->from == state && t->name == stack &&  t->to == finalstate 
	     && !bddEmpty(t->label)))
	  *found = t;
	
        return wSatDelete(prcs);
}

wFA *bddPrestar(bddPds *pds, bddProblem *problem, wTrans **found)
{
	wFA *fa = bddTargetAutomaton(pds, problem->state, problem->stack);
	wFA *fa_pre = bddComputePrestar(pds, fa, problem, found);

	bddFaDelete(fa);

	return fa_pre;
}


static bddResult* bddPreReachability (bddPds *pds, bddProblem *problem)
{
	bddResult *result = Malloc(sizeof(struct bddResult));

	DdManager *mgr = pds->mgr;
        wTrans *t;
	wFA *fa_pre = bddPrestar(pds,problem,&t);

	/* pre* is complete; now compute the result and an error trace. */

	result->yesno = t? RESULT_YES : RESULT_NO;
	result->prefix = NULL;
	result->loop = NULL;
	//result->sat = sat;

 	if (t && !Cudd_IsCancelled(mgr)) {
	  bddGetPreTrace(pds, fa_pre,t,result);
	}

 	bddFaDelete(fa_pre);
	
	return result;
}

/****************************************************************************/


/* This version may substitute the functionality of bddComputePostar,
   see below. It is useful for the version with abstraction and
   performs a complete Poststar search. */
static wFA* bddComputePoststarCompl (bddPds *pds, wFA *fa,
				     wIdent state, wIdent stack, wTrans **found)
{
	/* Note: I don't evaluate NEWTRACES here because I didn't implement
	   TRACE_REF for wsUpdateSpecial. - Stefan S. */
        wSat *prcs = wSatCreate(pds->pds->sr,SAT_POST,TRACE_YES);
 	prcs->special = 1;
	wRule *pt;
	wTrans *t;

	/* Add all the rules */
        for (pt = pds->pds->rules; pt; pt = pt->next)
                wSatAugRule(prcs,pt->from_state,pt->from_stack,
                        pt->to_state,pt->to_stack1,pt->to_stack2,pt->label,pt);

	/* Add all initial transitions (to the queue) */
        for (t = fa->transitions; t; t = t->next)
                 wSatAddTrans(prcs,t->from,t->name,t->to,t->label);

	*found = NULL; /* this is new */
	/* Process all the transitions */
        while ((t = wSatPopTrans(prcs)))
                /* for a breadth-first search do something like that:
		   do {
		     while (t=pop()) list.add(t);
		     if (list.empty()) break;
		     t  = list.remove();
		     aug(t);
		   } while(1);
		*/
	{
 		/* found target with non-empty BDD? */
 		if (t->from == state && t->name == stack && !bddEmpty(t->label))
			/* old: break; */
			if (*found == NULL) *found = t;
		/* fixme: we should prevent propagation of 'false' values
			  at this point (for efficiency) */
                if (!bddEmpty(t->label))
			wSatAugTrans(prcs,t);
	}

	/* old : *found = t; */
	/* bddPrintwNodeGraph(prcs); */
        return wSatDelete(prcs);
}


int containsCount = 0;

wFA* bddComputePoststar (bddPds *pds, wFA *fa, 
			 bddProblem *problem, wTrans **found)
{
	if (ABSTRACTION) 
		return bddComputePoststarCompl(pds,fa,problem->state,
					       problem->stack,found);

        wSat *prcs = wSatCreate(pds->pds->sr,SAT_POST,
			NEWTRACES? TRACE_REF : TRACE_COPY);
	prcs->special = 0;
        wRule *pt;
	wTrans *t = NULL;
	int count = 0;

	/* Add all the rules */
        for (pt = pds->pds->rules; pt; pt = pt->next)
                wSatAugRule(prcs,pt->from_state,pt->from_stack,
                        pt->to_state,pt->to_stack1,pt->to_stack2,pt->label,pt);

	/* Add all initial transitions (to the queue) */
        for (t = fa->transitions; t; t = t->next)
                 wSatAddTrans(prcs,t->from,t->name,t->to,t->label);

	if (problem->queue) {
	  while (1) {
	    queueStrategy *queue = problem->queue;
	    while (t = wSatPopTrans(prcs)) {
/*  	      printf("drain %s\n", wIdentString(t->name)); */
	      if (problem->isSolved(t, problem)) {
		goto exit;
	      }
	      queue->pushTransition(t, queue);
	    }
/*  	    printf("\n\n"); */
	    
	    t = queue->popTransition(queue);
	    if (t) {
/*  	      printf("work %s\n",  wIdentString(t->name)); */
	      count++;
	    }
	    if (!t || problem->isSolved(t, problem)) {
	      break;
	    }
	    /* fixme: we should prevent propagation of 'false' values
	       at this point (for efficiency) */
/*  	    printf("Augmenting %s\n", wIdentString(t->name)); */
/* 	    fflush(stdout); */
	    wSatAugTrans(prcs, t);
	  }
	}
	else {
	  /* Process all the transitions */
	  while ((t = wSatPopTrans(prcs)))
	    {
	      if (t) {
//printf("%s -%s-> %s\n", wIdentString(t->from), wIdentString(t->name), wIdentString(t->to)); fflush(stdout);
		count++;
	      }
	      if (problem->isSolved(t, problem))
		break;
//else printf("problem is not yet solved\n");
	      
	      /* fixme: we should prevent propagation of 'false' values
		 at this point (for efficiency) */
	      wSatAugTrans(prcs,t);
	    }
	}
	
exit:	if (problem->queue && problem->queue->delete) {
	  /* Call optional destructor */
	  problem->queue->delete(problem->queue);
	}

	*found = t;

/* 	printf("transitions looked at %d\n", count); */
/* 	printf("contains count %d\n", containsCount); */

        return wSatDelete(prcs);
}

wFA *bddPoststar(bddPds *pds, bddProblem *problem, wTrans **found)
{
	wFA *fa = bddInitialAutomaton(pds);
	wFA *fa_post = bddComputePoststar(pds, fa, problem, found);
	
	bddFaDelete(fa);

	return fa_post;
}

bddResult* bddPostReachability (bddPds *pds, bddProblem *problem)
{
	bddResult *result = Malloc(sizeof(struct bddResult));

	DdManager *mgr = pds->mgr;
        wTrans *t;

	wFA *fa_post = bddPoststar(pds,problem,&t);
	//wFA *fa_post = bddComputePoststar(pds, fa, state, stack, &t);

	/* post* is complete; now compute the result and an error trace. */

	result->yesno = t? RESULT_YES : RESULT_NO;
	result->prefix = NULL;
	result->loop = NULL;
	//result->sat = sat;

	/* only compute post trace if we weren't cancelled */
 	if (t && !bddEmpty(t->label) && !Cudd_IsCancelled(mgr)) {
	  bddGetPostTrace(pds, fa_post,t,result);
	}

 	bddFaDelete(fa_post);
	
	return result;
}

/***************************************************************************/

bddResult* bddAnalyse (bddPds *pds, bddProblem *problem)
{
	if (problem->type == PROBLEM_REACHABILITY) {
		if (ABSTRACTION)
			return absRef(pds,problem->state,problem->stack);
		else
			return bddPostReachability(pds, problem);
	}
	else if (problem->type == PROBLEM_REACHABILITY_PRE)
		return bddPreReachability(pds, problem);
	else if (problem->type == PROBLEM_REACHABILITY_DPN)
		return dpnPreReachability(pds,problem->state,problem->stack);
	else if (problem->type == PROBLEM_REACHABILITY_ADPN)
		return adpnPreReachability(pds,problem->state,problem->stack);
	else if (problem->type == PROBLEM_LTL)
		fprintf(stderr,"not implemented yet\n");
	else
		fprintf(stderr,"unknown problem\n");
	return NULL;
}

/***************************************************************************/

static void bddPrintCube (int *cube, DdNode **vars, int num_vars, char *mask)
{
	int i;

	for (i = 0; i < num_vars; i++)
	{
		if (cube[vars[i]->index] == 2) continue;
		printf("\t"); printf(mask,i);
		printf("=%d",cube[vars[i]->index]);
		if (i%9 == 8) printf("\n");
	}
}

static void bddPrintConfig (bddPds *pds, bddConfig *config)
{
	bddPath *path;
	int *cube;
	int i = 0;

	printf("<%s,",wIdentString(config->state));
	for (path = config->stack; path; path = path->rest_path)
		printf(" %s",wIdentString(path->stack));
	printf(">\n");

	printf("Global variables:\n");
	bddInitCube(pds->mgr,&cube);
	Cudd_BddToCubeArray(pds->mgr,config->bdd,cube);
	bddPrintCube(cube,pds->g0,pds->num_globals,"g%d");
	Free(cube);
	printf("\n");

	for (path = config->stack; path; path = path->rest_path)
	{
		printf("Local variables at stack position %d:\n",++i);
		bddInitCube(pds->mgr,&cube);
		Cudd_BddToCubeArray(pds->mgr,path->bdd,cube);
		bddPrintCube(cube,pds->l0,pds->num_locals,"l%d");
		Free(cube);
		printf("\n");
	}
}

static void bddPrintVars (bddPds *pds, DdNode *bdd, int *i)
{
	int *cube;

	printf("Local variables at stack position %d:\n",++(*i));
	bddInitCube(pds->mgr,&cube);
	Cudd_BddToCubeArray(pds->mgr,bdd,cube);
	bddPrintCube(cube,pds->l0,pds->num_locals,"l%d");
	Free(cube);
	printf("\n");
}

static void bddPrintReverse (bddPds *pds, bddPath *path, int *i)
{
	if (!path) return;
	bddPrintReverse(pds,path->rest_path,i);
	if (i)
		bddPrintVars(pds,path->bdd,i);
	else
		printf("%s ",wIdentString(path->stack));
}

static void bddPrintForward (bddPds *pds, bddPath *path, int *i)
{
	while (path)
	{
		if (i)
			bddPrintVars(pds,path->bdd,i);
		else
			printf("%s ",wIdentString(path->stack));
		path = path->rest_path;
	}
}


void bddPrintConfigDPN (bddPds *pds, bddConfig *config)
{
	int i = 0;

	bddPrintReverse(pds,config->prefix,NULL);
	bddPrintForward(pds,config->stack,NULL);
	printf("\n");

	bddPrintReverse(pds,config->prefix,&i);
	bddPrintForward(pds,config->prefix,&i);
}

extern void bddPrintResult (bddPds *pds, bddResult *result)
{
	bddConfig *config;

	if (result->yesno == RESULT_NO)
	{
		printf("Not reachable.\n");
		return;
	}

	printf("Reachable. Trace:\n");
	printf("\n-------------------\n\n");
	/*return;*/

	for (config = result->prefix; config; config = config->next_config)
	{
		bddPrintConfig(pds,config);
		printf("\n-------------------\n\n");
		if (config->rule)
		{
			wRule *r = config->rule;
			printf("rule used: <%s,%s> -> <%s",
				wIdentString(r->from_state),
				wIdentString(r->from_stack),
				wIdentString(r->to_state));
			if (r->to_stack1)
				printf(",%s",wIdentString(r->to_stack1));
			if (r->to_stack2)
				printf(" %s",wIdentString(r->to_stack2));
			printf(">\n\n-------------------\n\n");
		}
	}
}

/***************************************************************************/

static void bddDeletePath (bddPds *pds, bddPath *path)
{
	while (path)
	{
		bddPath *next = path->rest_path;
		if (--path->refcount) break;
		Cudd_RecursiveDeref(pds->mgr,path->bdd);
		Free(path);
		path = next;
	}
}

static void bddDeleteConfig (bddPds *pds, bddConfig *config)
{
	while (config)
	{
		bddConfig *next = config->next_config;
		Cudd_RecursiveDeref(pds->mgr,config->bdd);
		bddDeletePath(pds,config->stack);
		Free(config);
		config = next;
	}
}

extern void bddDeleteResult (bddPds *pds, bddResult *result)
{
	bddDeleteConfig(pds,result->prefix);
	bddDeleteConfig(pds,result->loop);
	Free(result);
}

int bddStateIsReached(wTrans *t, bddProblem *problem)
{
  if (t->from == problem->state && t->name == problem->stack
      && !bddEmpty(t->label))
    return 1;
  return 0;
}
