#include "coverage.h"

#include <stdlib.h>
#include <stdio.h>

#include "bdd.h"
#include "ir.h"
#include "bp.h"
#include "absRef.h"

#include <time.h>

/**
 * Compare method for hashtable.
 */
static int compare(void *obj1, void *obj2)
{
  return wIdentCompare(((stackElement*)obj1)->stack,
		       ((stackElement*)obj2)->stack);
}

int isMarkedAsReached(coverageContext *context, wIdent label)
{
  stackElement stack = { .hash = label, .stack = label };
  return wHashFind(context->visitedStackElements, &stack) != NULL;
}

int markAsReached(coverageContext *context, wIdent label)
{
  stackElement *stackToInsert = Malloc(sizeof(stackElement)); 
  stackToInsert->hash = label;
  stackToInsert->stack = label;
  assert(wHashInsert(context->visitedStackElements, stackToInsert) 
	 == stackToInsert);
  return --(context->unvisitedStackElements);
}

/**
 * Returns value != 0 if all stack elements have been visited at least once.
 *
 * Makes use of context struct which is stored in problem->user.
 */
int isCoverageSolved(wTrans *transition, bddProblem *problem)
{
  coverageContext *c = (coverageContext*)(problem->user);

  if (problem->state == transition->from && !bddEmpty(transition->label)
      && !isMarkedAsReached(c, transition->name)
      && matchPrefix(transition->name, c->stackPrefix)) {

    int remainingLabels = markAsReached(c, transition->name);
    return remainingLabels == 0;
  }
  return c->unvisitedStackElements == 0;
}

int startsWith(const char *string, const char *prefix)
{
  while (*prefix && *prefix == *string) {
    ++prefix;
    ++string;
  }
  return !*prefix;
}
  
int matchPrefix(wIdent id, const char *stackPrefix)
{
  if (stackPrefix == NULL) {
    // TODO these values should not be hardcoded here
    return id != 0 && !startsWith(wIdentString(id), "DEFAULT") 
      && !startsWith(wIdentString(id), "initialWrapperFor_");
  }
  else {
    return startsWith(wIdentString(id), stackPrefix);
  }
}

/**
 * Wraps Free macro for hash table cleanup.
 */
void freeWrapper(void *mem)
{
  Free(mem);
}

wHash* createStackElementHash()
{
  return wHashCreate(compare);
}

void freeStackElementHash(wHash *table)
{
  wHashDelete(table, freeWrapper);
}

/**
 * Counts how many different stack elements there are in pushdown system.
 */
int countStackElements(intPds *ipds, const char *stackPrefix)
{ 
  wRule* rule;
  wHash* seenStates = createStackElementHash();
  stackElement stack;
  int count = 0;

  for (rule = ipds->pds->pds->rules; rule; rule = rule->next) 
  {
      stack.hash = rule->from_stack;
      stack.stack = rule->from_stack;

      // add and count element if it's not already in the table and
      // it starts with the prefix if the prefix is not null
      if (matchPrefix(stack.hash, stackPrefix)
	  && !wHashFind(seenStates, &stack)) 
      {
	  stackElement *stackToInsert = Malloc(sizeof(stackElement));
	  stackToInsert->hash = rule->from_stack;
	  stackToInsert->stack = rule->from_stack;
	  wHashInsert(seenStates, stackToInsert);
/*   	  printf("%s\n", wIdentString(rule->from_stack)); */
/* 	  fflush(stdout); */
	  ++count;
      }
  }
  freeStackElementHash(seenStates); 
  return count;
}

void excludeStackPrefixes(intPds *ipds, coverageContext *context, 
			   struct wIdentList *prefixes)
{
  wRule* rule;
  wHash* table = context->visitedStackElements;
  wIdentList *prefix;
  wIdentList *finalList = NULL;

  // remove prefixes that wouldn't be matched anyway
  if (context->stackPrefix != NULL) {
    for (prefix = prefixes; prefix; prefix = prefix->next) {
      if (startsWith(context->stackPrefix, wIdentString(prefix->wId))) {
	insert(prefix->wId, &finalList);
      }
    }
  }
  else {
    finalList = prefixes;
  }
  
  for (rule = ipds->pds->pds->rules; rule; rule = rule->next) {
    for (prefix = finalList; prefix; prefix = prefix->next) {
      if (matchPrefix(rule->from_stack, context->stackPrefix)
	  && matchPrefix(rule->from_stack, wIdentString(prefix->wId))) {

	wHashElement *found;
	stackElement *elem;

	elem = Malloc(sizeof(stackElement));
	elem->hash = elem->stack = rule->from_stack;
	found = wHashInsert(table, elem);
	if (found == elem) {
	  --(context->unvisitedStackElements);
	}
	else {
	  // already excluded
	  Free(elem);
	}
      }
    }
  }
}

void excludeStacks(coverageContext *context, wIdentList *excludedStacks)
{
  wIdentList *listPtr;
  wHash* table = context->visitedStackElements;
  stackElement* elem;
  wHashElement* found;

  for (listPtr = excludedStacks; listPtr; listPtr = listPtr->next)
  {
    /* only exclude if it also matches the prefix */
    if (!matchPrefix(listPtr->wId, context->stackPrefix)) {
      continue;
    }

    elem = Malloc(sizeof(stackElement));
    elem->hash = elem->stack = listPtr->wId;
    found = wHashInsert(table, elem);
    if (found == elem) {
      --(context->unvisitedStackElements);
    }
    else {
      // already excluded
      Free(elem);
    }
  }
}

/**
 *
 * @param stackPrefix if not NULL, only states starting with
 * <code>startPrefix</code> are checked for coverage
 */
coverageContext* createCoverageContext(intPds *ipds, const char *stackPrefix,
				       wIdentList *excludedStacks)
{
  coverageContext *context = Malloc(sizeof(coverageContext));
  context->visitedStackElements = createStackElementHash();
  context->unvisitedStackElements = countStackElements(ipds, stackPrefix);
  context->stackPrefix = stackPrefix != NULL ? strdup(stackPrefix) : NULL;
  excludeStacks(context, excludedStacks);
  return context;
}

extern wIdent default_state_name;

intProblem* createCoverageProblem(intPds *ipds, const char *stackPrefix,
				  wIdentList *excludedStacks)
{
  intProblem *problem = Malloc(sizeof(intProblem));
  problem->type = PROBLEM_REACHABILITY;
  problem->state = default_state_name;
  problem->user = createCoverageContext(ipds, stackPrefix, excludedStacks);
  problem->isSolved = &isCoverageSolved;
  problem->queue = NULL;
  return problem;
}

void freeCoverageProblem(intProblem *problem)
{
  freeCoverageContext((coverageContext*)problem->user);
  Free(problem);
}

void freeCoverageContext(coverageContext *context)
{
  freeStackElementHash(context->visitedStackElements);
  Free(context->stackPrefix);
  Free(context);
}

intProblem* analyseCoverage(intPds *ipds, const char *stackPrefix,
			    wIdentList *excludedStacks)
{
  intProblem *problem = createCoverageProblem(ipds, stackPrefix, 
					      excludedStacks);
  intResult *result = intAnalyse(ipds, problem);
  problem->queue = createLIFOQueueStratey();
  intDeleteResult(result);
  return problem;
}

wIdentList* getUnreachableStates(intPds *ipds, coverageContext *context)
{
  wHash *table = context->visitedStackElements;
  int count = context->unvisitedStackElements;
  const char *stackPrefix = context->stackPrefix;
  int prefixLength = stackPrefix == NULL ? 0 : strlen(stackPrefix);
  wIdentList *list = NULL;
  wRule *rule;
  stackElement elem;

  for (rule = ipds->pds->pds->rules; rule; rule = rule->next) {
    elem.hash = elem.stack = rule->from_stack;
    // consider stack element not reached, if it's not in the table and
    // it starts with stackPrefix if stackPrefix is not NULL
    if (!wHashFind(table, &elem) 
	&& matchPrefix(rule->from_stack, context->stackPrefix)) {
      count -= insert(rule->from_stack, &list);
    }
  }
  assert(count == 0);
  return list;
}

void printLabels(intPds *ipds)
{
  wRule* rule;

  for (rule = ipds->pds->pds->rules; rule; rule = rule->next) {
    printf("%s -> %s, %s\n", wIdentString(rule->from_stack),
	   wIdentString(rule->to_stack1), wIdentString(rule->to_stack2));
  }
}

/**
 * Internal data structure, need not be known outside of coverage.c
 */
typedef struct newTransitionsQueue
{
  void *newTransitions;
  void *oldTransitions;
  wTrans *lastNewTransition;
  wTrans *lastOldTransition;
} newTransitionsQueue;

newTransitionsQueue* createNewTransitionQueue()
{
  newTransitionsQueue *queue = Malloc(sizeof(newTransitionsQueue));
  /* the memory addresses of the pointers are used as a markers */
  queue->newTransitions = &(queue->oldTransitions);
  queue->oldTransitions = &(queue->oldTransitions);
  queue->lastNewTransition = queue->lastOldTransition = NULL;
  return queue;
}

void pushPreferringNewTransitions(wTrans *transition, queueStrategy *this)
{
  if (this->data == NULL) {
    /* construct lazily */
    this->data = createNewTransitionQueue();
  }
  newTransitionsQueue *queue = (newTransitionsQueue*)this->data;
  /* it's a new transition */
  if (transition->user == NULL) {
    transition->user = queue->newTransitions;
    queue->newTransitions = transition;
  }
  else if (transition->user == queue) {
    transition->user = queue->oldTransitions;
    queue->oldTransitions = transition;
  }
  /* else: transition is currently in one of the queues and will be looked at
     eventually */
}

void pushFIFOPreferringNewTransitions(wTrans *transition, queueStrategy *this)
{
  if (this->data == NULL) {
    /* construct lazily */
    this->data = createNewTransitionQueue();
  }
  newTransitionsQueue *queue = (newTransitionsQueue*)this->data;
  /* new transitions */
  if (transition->user == NULL) {
    if (queue->lastNewTransition) {
      queue->lastNewTransition->user = transition;
    }
    else {
      queue->newTransitions = transition;
    }
    queue->lastNewTransition = transition;
    /* mark transition as being in one of the queues */
    transition->user = &(queue->oldTransitions);
  }
  /* old transitions */
  else if (transition->user == queue) {
    if (queue->lastOldTransition) {
      queue->lastOldTransition->user = transition;
    }
    else {
      queue->oldTransitions = transition;
    }
    queue->lastOldTransition = transition;
    /* mark transition as being in one of the queues */
    transition->user = &(queue->oldTransitions);
  }
}

wTrans* popPreferringNewTransitions(queueStrategy *this)
{
  newTransitionsQueue *queue = (newTransitionsQueue*)this->data;
  wTrans *transition;

  if (queue == NULL) {
    return NULL;
  }

  /* queue pointer is used as marker for end of new transitions list */
  if (queue->newTransitions != &queue->oldTransitions) {
    transition = (wTrans*)queue->newTransitions;
    queue->newTransitions = transition->user;
  }
  /* queue pointer is used as marker for end of old transitions list */
  else if (queue->oldTransitions != &queue->oldTransitions) {
    transition = (wTrans*)queue->oldTransitions;
    queue->oldTransitions = transition->user;
  }
  else {
    return NULL;
  }
  if (bddEmpty(transition->label)) {
    /* mark epsilon transitions as new again */
    transition->user = NULL;
  }
  else {
    /* mark transition as not in any queue */
    transition->user = queue;
  }
  return transition;
}

wTrans* popFIFOPreferringNewTransitions(queueStrategy *this)
{
  newTransitionsQueue *queue = (newTransitionsQueue*)this->data;
  wTrans *transition;

  if (queue == NULL) {
    return NULL;
  }

  /* compare against empty marker */
  if (queue->newTransitions != &(queue->oldTransitions)) {
    transition = (wTrans*)queue->newTransitions;
    if (transition == queue->lastNewTransition) {
      queue->lastNewTransition = NULL;
      /* set empty marker */
      queue->newTransitions = &(queue->oldTransitions);
    }
    else {
      queue->newTransitions = transition->user;
    
    }
    wTrans *cycle = (wTrans*)transition;
    while (cycle != queue->lastNewTransition) {
      //      printf("new queue label empty %s, %s\n", bddEmpty(cycle->label) ? "true" : "false ", wIdentString(cycle->name));
      if (queue->lastNewTransition == NULL) {
	break;
      }
      cycle = (wTrans*)cycle->user;
    }
    //    printf("\n\n");
  }
  else if (queue->oldTransitions != &(queue->oldTransitions)) {
    transition = (wTrans*)queue->oldTransitions;
    if (transition == queue->lastOldTransition) {
      queue->lastOldTransition = NULL;
      queue->oldTransitions = &(queue->oldTransitions);
    }
    else {
      queue->oldTransitions = transition->user;
    }
    wTrans *cycle = (wTrans*)transition;
    while (cycle != queue->lastOldTransition) {
      //      printf("old queue label empty %s, %s\n", bddEmpty(cycle->label) ? "true" : "false ", wIdentString(cycle->name));
      if (queue->lastOldTransition == NULL) {
	break;
      }
      cycle = (wTrans*)cycle->user;
    }
    //    printf("\n\n");
  }
  else {
    return NULL;
  }

  if (bddEmpty(transition->label)) {
    transition->user = NULL;
  }
  else {
    transition->user = queue;
  }
  return transition;
}

void deleteNewTransitionsQueue(queueStrategy *this)
{
  Free((newTransitionsQueue*)this->data);
  this->data = NULL;
}

queueStrategy* createPreferringNewTransitionsQueueStrategy()
{
  queueStrategy *queue = Malloc(sizeof(queueStrategy));
  queue->pushTransition = &pushPreferringNewTransitions;
  queue->popTransition = &popPreferringNewTransitions;
  queue->delete = &deleteNewTransitionsQueue;
  queue->data = NULL;
  return queue;
}

queueStrategy* createFIFOPreferringNewTransitionsQueueStrategy()
{
  queueStrategy *queue = Malloc(sizeof(queueStrategy));
  queue->pushTransition = &pushFIFOPreferringNewTransitions;
  queue->popTransition = &popFIFOPreferringNewTransitions;
  queue->delete = &deleteNewTransitionsQueue;
  queue->data = NULL;
  return queue;
}

void freeQueueStrategy(queueStrategy *queue)
{
  if (queue && queue->delete) {
    queue->delete(queue);
  }
  Free(queue);
}

extern int containsCount;

static void pushLIFOTransition(wTrans *transition, queueStrategy *this)
{
  /* (ab)use wTrans' user pointer for a single linked list impl */
  if (this->data) {
    wTrans *predecessor = (wTrans*)this->data;
    /* look if it's already in the list */
    while (predecessor) {
      if (transition == predecessor) {
	containsCount++;
	return;
      }
      predecessor = (wTrans*)predecessor->user;
    }
    predecessor = (wTrans*)this->data;
    this->data = transition;
    transition->user = predecessor;
  }
  else {
    this->data = transition;
    transition->user = NULL;
  }
}

wTrans* popLIFOTransition(queueStrategy *this)
{
  if (this->data) {
    wTrans *next = (wTrans*)this->data;
    this->data = next->user;
    next->user = NULL;
    return next;
  }
  else {
    return NULL;
  }
}

void pushFIFOTransition(wTrans *transition, queueStrategy *this)
{
  if (transition->user) {
    // in queue
    containsCount++;
    return;
  }
  if (this->data) {
    wTrans* predecessor;
    wTrans* current = (wTrans*)this->data;
    while (current && current != transition) {
      predecessor = current;
      current = (wTrans*)current->user;
    }
    assert(current == NULL || current == transition);
    assert(predecessor != NULL);
    if (!current) {
      predecessor->user = transition;
    }
    else {
      // already in queue
    }
  }
  else {
    this->data = transition;
  }
}

wTrans* popFIFOTransition(queueStrategy *this)
{
  wTrans *next = (wTrans*)this->data;
  if (next) {
    this->data = next->user;
    next->user = NULL;
  }
  return next;
}

queueStrategy* createFIFOQueueStratey()
{
  queueStrategy *queue = Malloc(sizeof(queueStrategy));
  queue->pushTransition = &pushFIFOTransition;
  queue->popTransition = &popFIFOTransition;
  queue->delete = NULL;
  queue->data = NULL;
  return queue;
}

queueStrategy* createLIFOQueueStratey()
{
  queueStrategy *queue = Malloc(sizeof(queueStrategy));
  queue->pushTransition = &pushLIFOTransition;
  queue->popTransition = &popLIFOTransition;
  queue->delete = NULL;
  queue->data = NULL;
  return queue;
}
