#ifndef COVERAGE_H
#define COVERAGE_H

#include <wpds.h>
#include <wpdsInt.h>
#include "int.h"

struct wIdentList;

/**
 * Hashtable entry for stack elements, value of stack is used as hash.
 */
typedef struct stackElement
{
  struct stackElement *next;
  unsigned long hash;
  
  wIdent stack;
} stackElement;

/**
 * Application specific intproblem/bddProblem user context, stores the 
 * hash table of already visited stack elements and the count of how many
 * have not been visited at all yet.
 */
typedef struct coverageContext
{
  wHash* visitedStackElements;
  wHash* culprit;
  int unvisitedStackElements;
  /**
   * If not NULL only stack elements starting with this prefix are tested
   * for reachability.
   */
  char* stackPrefix;
} coverageContext;

int countStackElements(intPds *ipds, const char *stackPrefix);

/**
 * Counts how many different stack elements there are in the pushdown system.
 */
int countStackElementsC(intPds *ipds, const char *stackPrefix, 
		       coverageContext *context);

wHash* createStackElementHash();

void freeStackElementHash(wHash*);

/**
 * Removes the excluded stacks in the list excludeStacks from the hash table
 * of the coverageContext and also decrements the number which have not been
 * visited yet. List may be NULL.
 */
void excludeStacks(coverageContext *context, struct wIdentList *excludedStacks);

void excludeStackPrefixes(intPds * ipds, coverageContext *context, 
			   struct wIdentList *prefixes);

coverageContext* createCoverageContext(intPds *ipds, const char *stackPrefix,
				       struct wIdentList *excludedStacks);

void freeCoverageContext(coverageContext *context);

/**
 * Creates a new coverage problem on the heap.
 * 
 * <code>stackPrefix</code> if not NULL only states which start with 
 * stackPrefix are considered in the coverage search
 * <code>excludedStacks</code> may also be NULL.
 */
intProblem* createCoverageProblem(intPds *ipds,
				  const char *stackPrefix,
				  struct wIdentList *excludedStacks);

/**
 * Deletes a coverage problem that was created on the heap, e.g. by calling
 * createCoverageProblem().
 */
void freeCoverageProblem(intProblem *problem);

/**
 * Computes coverage for all states in the push down system except for the
 * states in the zero terminated array excludedStacks. Array can be NULL.
 * Returns the problem used for analysing so you can check if all states have
 * been visited.
 */
intProblem* analyseCoverage(intPds *ipds, const char *stackPrefix,
			    struct wIdentList *excludedStacks);

/**
 * Returns the list of stack symbols that have not been
 * reached. <code>problem</code> must be a coverage problem.
 * Returns NULL if all states were reached.
 */
struct wIdentList* getUnreachableStates(intPds *ipds, coverageContext *context);

/**
 * Returns value != 0 if <code>string</code> starts with <code>prefix</code>.
 */
int startsWith(const char *string, const char *prefix);

/**
 * Returns value != 0 if the id's string starts with the stackPrefix
 * in context or if the stackPrefix is NULL.
 */
int matchPrefix(wIdent id, const char *stackPrefix);

/**
 * Pushes new transitions which are seen for the first time into the queue
 * that is emptied first and transitons we have visited already into a second
 * queue.
 * <p>
 * <code>wTrans.user</code> is used for building the queues and discerning if
 * a transition is either new, in a queue, or old.
 */
void pushPreferringNewTransitions(wTrans *transition, queueStrategy *this);

/**
 * Returns new transitions from the new queue until it is empty then falls
 * back to the old queue, or returns NULL if both queues are empty.
 */
wTrans* popPreferringNewTransitions(queueStrategy *this);

/**
 * Deletes the internal data structure that holds the two queues.
 */
void deleteNewTransitionsQueue(queueStrategy *this);

/**
 * Creates a queueStrategy struct that prefers new transitions over old ones.
 * <p> 
 * <code>wTrans.user</code> is used for this purpose.
 */
queueStrategy* createPreferringNewTransitionsQueueStrategy();

/**
 * Creates a queueStrategy struct that prefers new transitions over old ones
 * and queues them FIFO wise.
 * <p>
 * <code>wTrans.user</code> is used for this purpose.
 */
queueStrategy* createFIFOPreferringNewTransitionsQueueStrategy();

queueStrategy* createFIFOQueueStratey();

queueStrategy* createLIFOQueueStratey();

/**
 * Calls destructor <code>queue->delete(queue)</code> if there is one and
 * frees the queueStrategy struct.
 */
void freeQueueStrategy(queueStrategy *queue);

void printLabels(intPds *ipds);

/**
 * Marks a label as reached.
 * <p>
 * Precondition: label is not marked yet.
 *
 * Returns the number of labels which have not been covered yet.
 */
int markAsReached(coverageContext *context, wIdent label);

/**
 * Returns value != 0 if label has been marked as reached.
 */
int isMarkedAsReached(coverageContext *context, wIdent label);

#endif
