#ifndef IR_H
#define IR_H

/* External definitions of the IR module */

#include "int.h"

extern int DPN;
extern int irVerbose;

/**
 * Progress monitor struct. At least the isCanceled function pointer must be
 * not null. All other functions are optional.
 *
 * The progress monitor contract:
 *
 * If a method is given a progress monitor, it's a new one, and it calls
 * <code>beginTask</code> exactly once. At the end of its operations
 * <code>done</code> must be called.
 *
 * If you want to give the progress monitor to a method, create a sub progress
 * monitor using <code>createSubProgressMonitor</code> giving it a part of the
 * work amount of the current progress monitor.
 */
typedef struct progressMonitor
{
  int (*isCanceled)(struct progressMonitor *this);
  void (*beginTask)(const char *name, int workAmount, 
		    struct progressMonitor *this);
  void (*worked)(int amount, struct progressMonitor *this);
  void (*subTask)(const char *name, struct progressMonitor *this);
  void (*done)(struct progressMonitor *this);
  /* private user data */
  void *user;
} progressMonitor;

/**
 * @param monitor may be NULL
 * @return NULL if monitor was canceled
 */
intPds* irReadPdsMonitored (const char*, wIdent, progressMonitor *monitor);

intPds* irReadPds (const char*, wIdent);

progressMonitor* createSubProgressMonitor(int ticks, progressMonitor *parent);

/**
 * Frees the sub progress monitor and also ensures that it is done.
 */
void freeSubProgressMonitor(progressMonitor *monitor);

/**
 * Sets the progress monitor for cudd. Just uses the isCancelled pointer.
 */
void setProgressMonitor(intPds *pds, progressMonitor *monitor);

/**
 * Resets the progress monitor for cudd.
 */
void resetProgressMonitor(intPds *pds);

/**
 * Progress monitor macros that do null checks.
 */
#define IS_CANCELED(monitor) \
 (monitor && monitor->isCanceled(monitor))

#define BEGIN_TASK(name, totalAmount, monitor) \
do { if (monitor && monitor->beginTask) { monitor->beginTask(name, totalAmount, monitor); } } while(0)

#define SUB_TASK(name, monitor) \
do { if (monitor && monitor->subTask) { monitor->subTask(name, monitor); } } while (0)

#define WORKED(amount, monitor) \
do { if (monitor && monitor->worked) { monitor->worked(amount, monitor); } } while(0);

#define DONE(monitor) \
do { if (monitor && monitor->done) { monitor->done(monitor); } } while(0);

#define CREATE_SUB_MONITOR(ticks, monitor) \
(monitor ? createSubProgressMonitor(ticks, monitor) : NULL)

#endif
