#include "test.h"

#include <coverage.h>
#include <int.h>
#include <ir.h>
#include <absRef.h>


static int countRules(intPds *pds)
{
  int count = 0;
  wRule *rule;
  for (rule = pds->pds->pds->rules; rule; rule = rule->next, ++count);
  return count;
}

static void testNumberOfRules()
{
  intPds *pds = irReadPds("err.rem", 0);
  assert(4 == countRules(pds));
  intPdsDelete(pds);
}

static void testCountStackElements()
{
  intPds *pds = irReadPds("err.rem", 0);
/*   assertEquals(4, countStackElements(pds, NULL)); */
  assertEquals(3, countStackElements(pds, NULL)); /* main, ok, error */
  assertEquals(1, countStackElements(pds, "error"));
  assertEquals(1, countStackElements(pds, "ok"));
  intPdsDelete(pds);

  pds = irReadPds("jmopedoutput.rem", 0);
  assertEquals(7, countStackElements(pds, "de_uni_stuttgart_fmi_szs_jmoped_PDSTest_TestClass_test_I_V"));
  assertEquals(3, countStackElements(pds, "de_uni_stuttgart_fmi_szs_jmoped_PDSTest_TestClass__init__V"));
  assertEquals(6, countStackElements(pds, "de_uni_stuttgart_fmi_szs_jmoped_PDSTest_TestClass_main_Ljava_lang_String_V"));
  intPdsDelete(pds);

  pds = irReadPds("bubble.rem", 0);

  assertEquals(23, countStackElements(pds, "BubbleSort_isSorted_I_Z"));
  
  assertEquals(38,
	       countStackElements(pds, "BubbleSort_main_Ljava_lang_String_V"));

  assertEquals(46,
	       countStackElements(pds, "BubbleSort_sort_I_V"));

  assertEquals(8, countStackElements(pds, "BubbleSort__clinit__V"));

  assertEquals(3, countStackElements(pds, "BubbleSort__init__V"));

  assertEquals(118, countStackElements(pds, "BubbleSort"));

  assertEquals(119, countStackElements(pds, NULL));

  intPdsDelete(pds);
}

static int hasLabel(intPds *pds, const char *label)
{
  wRule *rule;
  for (rule = pds->pds->pds->rules; rule; rule = rule->next)
  {
    const char *currentLabel = wIdentString(rule->from_stack);
    if (strcmp(currentLabel, label) == 0)
    {
      return 1;
    }
  }
  return 0;
}

static void testLabelsExist()
{
  intPds *pds = irReadPds("err.rem", 0);
  assert(hasLabel(pds, "error"));
  assert(hasLabel(pds, "ok"));
  intPdsDelete(pds);
}

static void testExcludeStacks()
{
  coverageContext context = { createStackElementHash(), 0 };
  wIdent ok = wIdentCreate("ok");
  wIdent error = wIdentCreate("error");
  wIdentList toExclude = { ok, NULL };
  context.unvisitedStackElements = 2;

  assert(0 == context.visitedStackElements->num_entries);
  
  // NULL argument
  excludeStacks(&context, NULL);
  assertEquals(0, context.visitedStackElements->num_entries);
  assertEquals(2, context.unvisitedStackElements);

  // exclude ok
  excludeStacks(&context, &toExclude);
  assertEquals(1, context.visitedStackElements->num_entries);
  assertEquals(1, context.unvisitedStackElements);

  // exclude error
  toExclude.wId = error;
  excludeStacks(&context, &toExclude);
  assertEquals(2, context.visitedStackElements->num_entries);
  assertEquals(0, context.unvisitedStackElements);

  freeStackElementHash(context.visitedStackElements);
}

static void testCreateCoverageProblem()
{
  wIdentList list = { wIdentCreate("ok"), NULL };
  intPds *pds = irReadPds("err.rem", 0);
  
  intProblem *problem = createCoverageProblem(pds, NULL, NULL);
  assert(problem->isSolved);
  coverageContext *context = (coverageContext*)problem->user;
  assert(context);
  assertEquals(0, context->visitedStackElements->num_entries);
  assertEquals(countStackElements(pds, NULL), context->unvisitedStackElements);

  freeCoverageProblem(problem);

  problem = createCoverageProblem(pds, NULL, &list);
  assert(problem->isSolved);
  context = (coverageContext*)problem->user;
  assert(context);
  assertEquals(1, context->visitedStackElements->num_entries);
  assertEquals(context->unvisitedStackElements,
	       countStackElements(pds, NULL) - 1);

  freeCoverageProblem(problem);

  intPdsDelete(pds);
}

static wIdentList* createStateList()
{
  wIdentList *states = NULL;
  insert(wIdentCreate("BubbleSort__clinit__V"), &states);
  insert(wIdentCreate("BubbleSort__clinit__V12"), &states);
  insert(wIdentCreate("BubbleSort__clinit__V13"), &states);
  insert(wIdentCreate("BubbleSort__clinit__V16"), &states);
  insert(wIdentCreate("BubbleSort__clinit__V2"), &states);
  insert(wIdentCreate("BubbleSort__clinit__V5"), &states);
  insert(wIdentCreate("BubbleSort__clinit__V8"), &states);
  insert(wIdentCreate("BubbleSort__clinit__V9"), &states);
  insert(wIdentCreate("BubbleSort__init__V"), &states);
  insert(wIdentCreate("BubbleSort__init__V1"), &states);
  insert(wIdentCreate("BubbleSort__init__V4"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z1"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z10"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z11"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z12"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z13"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z16"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z17"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z18"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z2"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z21"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z22"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z23"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z24"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z25"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z26"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z29"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z30"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z5"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z6"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z7"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z8"), &states);
  insert(wIdentCreate("BubbleSort_isSorted_I_Z9"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V10"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V11"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V12"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V13"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V14"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V15"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V16"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V17"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V18"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V19"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V2"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V22"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V23"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V24"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V25"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V28"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V31"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V32"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V35"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V36"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V39"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V4"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V42"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V45"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V48"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V49"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V5"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V52"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V53"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V56"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V59"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V6"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V62"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V63"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V66"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V67"), &states);
  insert(wIdentCreate("BubbleSort_main_Ljava_lang_String_V7"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V1"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V10"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V11"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V12"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V13"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V14"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V15"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V16"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V17"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V18"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V2"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V21"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V22"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V23"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V24"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V26"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V27"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V28"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V29"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V30"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V31"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V32"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V33"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V34"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V35"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V36"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V37"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V38"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V40"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V41"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V44"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V45"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V46"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V47"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V48"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V49"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V5"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V52"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V55"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V56"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V57"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V58"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V6"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V61"), &states);
  insert(wIdentCreate("BubbleSort_sort_I_V7"), &states);
  insert(wIdentCreate("assertError"), &states);

  return states;
}

static void testIsCoverageSolved()
{
  intPds *pds = irReadPds("bubble.rem", 0); 
  intProblem *problem = createCoverageProblem(pds, NULL, NULL); 
  coverageContext *context = (coverageContext*)problem->user;
  wIdentList *states = createStateList();
  wIdentList *state;
  bddPds bpds = { .zero = (struct DdNode*)1 };
  bddDomain bdd = { .bdd = NULL, .pds = &bpds };
  wTrans trans = { .label = &bdd, .from = wIdentCreate("q") };
  int i;

  assertEquals(119, context->unvisitedStackElements);
  assertEquals(119, lengthSymList(states));

  for (state = states, i = 0; state; i++) {
    trans.name = state->wId;
    if (state->next) {
      assert(!isCoverageSolved(&trans, problem));
      assert(context->unvisitedStackElements > 0);
    }
    else {
      assert(isCoverageSolved(&trans, problem));
      assertEquals(0, context->unvisitedStackElements);
    }
    /* only advance every second time, so every state is tested twice */
    if (i % 2) {
      state = state->next;
    }
  }

  {
    wIdentList *unreached = getUnreachableStates(pds, context);
    assertEquals(0, lengthSymList(unreached));
    freeWIdentList(unreached);
  }

  intPdsDelete(pds); 
  freeCoverageProblem(problem); 
  freeWIdentList(states);
}

static void testAanalyseCoverage()
{
  intPds *pds = irReadPds("err.rem", 0);
  wIdentList list = { wIdentCreate("error"), NULL };
  intProblem *problem = analyseCoverage(pds, NULL, NULL);
  coverageContext *context = (coverageContext*)problem->user;
  wIdentList *resultList;

  // without any exclusions: error state is not reachable
  assertEquals(1, context->unvisitedStackElements);

  resultList = getUnreachableStates(pds, context);
  assertEquals(1, lengthSymList(resultList));
  // error must be in list of unreachable states
  assert(in(list.wId, resultList));

  freeWIdentList(resultList);
  freeCoverageProblem(problem);

  // exclude unreachable error state
  problem = analyseCoverage(pds, NULL, &list);
  context = (coverageContext*)problem->user;
  // all states reached
  assertEquals(0, context->unvisitedStackElements);

  resultList = getUnreachableStates(pds, context);
  assertEquals(0, lengthSymList(resultList));
  // error-state must not be in list of unreachable states
  assert(!in(list.wId, resultList));

  freeWIdentList(resultList);
  freeCoverageProblem(problem);

  intPdsDelete(pds);
}

static void testBubbleCoverage()
{
  intPds *pds = irReadPds("bubble.rem", 0);
  intProblem *problem = analyseCoverage(pds, NULL, NULL);
  wIdentList *unreachedStates = getUnreachableStates
    (pds, (coverageContext*)problem->user);

  /* 4 for never reached assertion error code, 2 for "return false" in
     isSorted and assertError state */
  assertEquals(7, lengthSymList(unreachedStates));

  intPdsDelete(pds);
  freeCoverageProblem(problem);
  freeWIdentList(unreachedStates);
}

static void testStackPrefix()
{
  intPds *pds = irReadPds("bubble.rem", 0);

  intProblem *problem = analyseCoverage(pds, NULL, NULL);
  wIdentList *unreachedStates = getUnreachableStates
    (pds, (coverageContext*)problem->user);

  intProblem *problem2 = analyseCoverage(pds, "BubbleSort", NULL);
  wIdentList* unreachedStates2 = getUnreachableStates
    (pds, (coverageContext*)problem2->user);

  // +1 for the assertionError which is not counted in the second run
  assertEquals(lengthSymList(unreachedStates), 
	       lengthSymList(unreachedStates2) + 1);

  freeCoverageProblem(problem2);
  freeWIdentList(unreachedStates2);

  problem2 = analyseCoverage(pds, "BubbleSort_sort_I_V", NULL);
  unreachedStates2 = getUnreachableStates(pds, 
					  (coverageContext*)problem2->user);
  // sort is completely covered
  assertEquals(0, lengthSymList(unreachedStates2));

  freeCoverageProblem(problem2);
  freeWIdentList(unreachedStates2);

  problem2 = analyseCoverage(pds, "BubbleSort_isSorted_I_Z", NULL);
  unreachedStates2 = getUnreachableStates
    (pds, (coverageContext*)problem2->user);
  
  // 2 states are not reached, since algorithm is correct
  assertEquals(2, lengthSymList(unreachedStates2));

  freeCoverageProblem(problem);
  freeCoverageProblem(problem2);
  freeWIdentList(unreachedStates);
  freeWIdentList(unreachedStates2);
  intPdsDelete(pds);
}

static void testStaticDriver()
{
  intPds *pds = irReadPds("staticdriver.rem", 0);
  intProblem *problem = analyseCoverage(pds, "de_uni_stuttgart_fmi_szs_jmoped"
					"_MethodWrapperTest_BubbleSort_"
					"driver_V", NULL);

  freeCoverageProblem(problem);
  intPdsDelete(pds);
}

static void testStaticMainDriver()
{
  intPds *pds = irReadPds("staticmaindriver.rem", 0);
  intProblem *problem = analyseCoverage(pds, "BubbleSort_main_ALjava_lang_String_V", NULL);
  wIdentList *states = getUnreachableStates(pds, 
					    (coverageContext*)problem->user);

  assertEquals(0, lengthSymList(states));
  
  freeWIdentList(states);
  freeCoverageProblem(problem);
  intPdsDelete(pds);
}

static wIdentList* createQuickSort3IgnoreList()
{
  wIdentList *list = NULL;
  insert(wIdentCreate("de_uni_stuttgart_fmi_szs_jmoped_coverage_"
		      "CoverageTesterTest_QuickSort_sort_AIII_V179"), &list);
  return list;
}

static void testQuickSortCoverage()
{
  intPds *pds = irReadPds("QuickSort3.rem", 0);
  wIdentList *ignored = createQuickSort3IgnoreList();
  intProblem *problem = analyseCoverage(pds, NULL, ignored);
  wIdentList *states = getUnreachableStates(pds,
					    (coverageContext*)problem->user);
  wIdentList *state;

  assertEquals(0, lengthSymList(states));
  for (state = states; state; state = state->next) {
    assert(isNotReachable(pds, state->wId));
  }

  freeWIdentList(ignored);
  freeWIdentList(states);
  freeCoverageProblem(problem);
  intPdsDelete(pds);

  pds = irReadPds("QuickSort2.rem", 0);
  problem = analyseCoverage(pds, NULL, NULL);
  states = getUnreachableStates(pds, (coverageContext*)problem->user);
  
/*   printList(states); */

  for (state = states; state; state = state->next) {
    assert(isNotReachable(pds, state->wId));
  }
  freeWIdentList(states);
  freeCoverageProblem(problem);
  intPdsDelete(pds);
}


int main(int argc, char** argv)
{
  TestTuple tests[] = {
    ADD_TEST(testNumberOfRules),
    ADD_TEST(testCountStackElements),
    ADD_TEST(testLabelsExist),
    ADD_TEST(testExcludeStacks),
    ADD_TEST(testCreateCoverageProblem),
    ADD_TEST(testIsCoverageSolved),
    ADD_TEST(testAanalyseCoverage),
    ADD_TEST(testBubbleCoverage),
    ADD_TEST(testStackPrefix),
    ADD_TEST(testStaticMainDriver),
    ADD_TEST(testStaticDriver),
    ADD_TEST(testQuickSortCoverage),
  };
  setUp();
  run_tests(sizeof(tests) / sizeof(TestTuple), tests, argc, argv);
  return 0;
}

/*
 * Local Variables:
 * compile-command: "make -k coveragetest"
 * End:
 */
