#include "test.h"

#include <coverage.h>
#include <int.h>
#include <ir.h>
#include <absRef.h>

extern int mc_irverbose;

/**
 * Returns the configuration with <code>label</code> or NULL if it can't be
 * found.
 */
static intConfig* getConfigAtLabel(intConfig *config, const char *label)
{
  intConfig *c;
  for (c = config; c; c = c->next_config) {
    if (strcmp(wIdentString(c->stack->stack), label) == 0) {
      return c;
    }
  }
  return NULL;
}

static intValuation* getValuationAt(intConfig *config, int index)
{
  int count;
  intPath *path;
  for (path = config->stack; path && count < index;
       path = path->rest_path, count++);
  if (path) {
    return path->values;
  }
}

static int getValue(intValuation *var)
{
  return var->value.integer;
}

static void testGetPostTrace()
{
  intPds *pds = irReadPds("PostTrace.rem", 0);
  intProblem problem = { .type = PROBLEM_REACHABILITY, 
			 .isSolved = &bddStateIsReached, 
			 .queue = NULL, .state = wIdentCreate("q"),
			 .stack = wIdentCreate("zeroloop") };
  intResult *result = intAnalyse(pds, &problem);
  intConfig *config;
  intValuation *valuation;

  assertEquals(RESULT_YES, result->yesno);
/*   intPrintResult(pds, result); */
  
  config = getConfigAtLabel(result->prefix, "test");
  assert(config != NULL);
  
  valuation = getValuationAt(config, 0);
  assert(valuation != NULL);
  assertEquals(0, strcmp(wIdentString(valuation->var->name), "v0"));
  assertEquals(0, getValue(valuation));

  intDeleteResult(result);

  problem.stack = wIdentCreate("oneloop");
  result = intAnalyse(pds, &problem);
/*   intPrintResult(pds, result); */
  
  config = getConfigAtLabel(result->prefix, "test");
  assert(config != NULL);
  
  valuation = getValuationAt(config, 0);
  assert(valuation != NULL);
  assertEquals(0, strcmp(wIdentString(valuation->var->name), "v0"));
  assertEquals(1, getValue(valuation));
  
  intDeleteResult(result);
  
  intPdsDelete(pds);
}

static wTrans **transitions;
static int size = 0;
static int count = 0;

static int collectTransitions(wTrans *transition, bddProblem *problem)
{
  if (transition->from == problem->state
      && matchPrefix(transition->name, NULL)
      && !bddEmpty(transition->label)) {
    if (count == size) {
      size = size > 0 ? size * 2 : 10;
      transitions = Realloc(transitions, size * sizeof(wTrans*));
    }
    transitions[count++] = transition;
  }
  return isCoverageSolved(transition, problem);
}

static void testGetPostTraceForAllTransitions()
{
  intPds *pds = irReadPds("PostTrace.rem", 0);  
  intProblem *problem = createCoverageProblem(pds, NULL, NULL);
  intResult *result;
  int i;
  wIdentList *states;

  Free(transitions);
  transitions = NULL;
  size = count = 0;
  
  problem->isSolved = &collectTransitions;
  result = intAnalyse(pds, problem);

  states = getUnreachableStates(pds, (coverageContext*)problem->user);
/*   printList(states); */

/*   intPrintResult(pds, result); */

  for (i = 0; i < count; i++) {
    bddResult bResult;
    intResult *iResult;
    bddGetPostTrace(pds->pds, result->sat->fa, transitions[i], &bResult);
/*     printf("\n\nresult for transition %s %s %d\n\n",  */
/* 	   wIdentString(transitions[i]->from), */
/* 	   wIdentString(transitions[i]->name), i); */
    iResult = intConvertResult(pds, &bResult);
    assert(iResult->prefix != NULL);
/*     intPrintResult(pds, iResult); */
  }

  freeCoverageProblem(problem);
  intPdsDelete(pds);
}

static void testGetPostTraceForFoundTransitions()
{
  intPds *pds = irReadPds("PostTrace.rem", 0);  
  intProblem *problem = createCoverageProblem(pds, NULL, NULL);
  intResult *result;
  int i;
  wIdentList *states;

  Free(transitions);
  transitions = NULL;
  size = count = 0;
  
  problem->isSolved = &collectTransitions;
  result = intAnalyse(pds, problem);

  states = getUnreachableStates(pds, (coverageContext*)problem->user);

  for (i = 0; i < count; i++) {
    bddResult bResult;
    intResult *iResult;
    wTrans *found = wFAFind(result->sat->fa, 
			    transitions[i]->from, transitions[i]->name,
			    transitions[i]->to);
    assert(found == transitions[i]);
    bddGetPostTrace(pds->pds, result->sat->fa, transitions[i], &bResult);
    printf("\n\nresult for transition %s %s %d\n\n", 
	   wIdentString(transitions[i]->from),
	   wIdentString(transitions[i]->name), i);
    iResult = intConvertResult(pds, &bResult);
    assert(iResult->prefix != NULL);
    intPrintResult(pds, iResult);
  }

  freeCoverageProblem(problem);
  intPdsDelete(pds);
}

int main(int argc, char** argv)
{
  TestTuple tests[] = {
    ADD_TEST(testGetPostTrace),
    ADD_TEST(testGetPostTraceForAllTransitions),
    ADD_TEST(testGetPostTraceForFoundTransitions),
  };
  setUp();
  run_tests(sizeof(tests) / sizeof(TestTuple), tests, argc, argv);
  return 0;
}

/*
 * Local Variables:
 * compile-command: "make -k posttracetest"
 * End:
 */
