#include "test.h"

#include <coverage.h>
#include <int.h>
#include <ir.h>
#include <absRef.h>

typedef struct monitorData
{
  int worked;
  int canceled;
  const char *taskName;
  const char *subTaskName;
  int workAmount;
} monitorData;

static int canceled(progressMonitor *this)
{
  monitorData *data = (monitorData*)this->user;
  return data->canceled;
}

static void worked(int amount, progressMonitor *this)
{
  monitorData *data = (monitorData*)this->user;
  data->worked += amount;
}

static void beginTask(const char *name, int workAmount, progressMonitor *this)
{
  monitorData *data = (monitorData*)this->user;
  data->taskName = name;
  data->workAmount = workAmount;
}

static void subTask(const char *name, progressMonitor *this)
{
  monitorData *data = (monitorData*)this->user;
  data->subTaskName = name;
}

static void resetData(monitorData *data)
{
  data->worked = 0;
  data->canceled = 0;
  data->taskName = NULL;
  data->subTaskName = NULL;
  data->workAmount = 1000;
}

static void testSubProgressMonitor()
{
  monitorData data = { 0, 0, NULL, NULL, 1000 };
  progressMonitor monitor = { .isCanceled = &canceled, .worked = &worked,
			      .beginTask = &beginTask, .subTask = &subTask,
			      .user = &data };
  
  monitor.beginTask("mainTask", 1000, &monitor);
  progressMonitor *subMonitor = createSubProgressMonitor(500, &monitor);

  // test beginTask
  subMonitor->beginTask("task", 1000, subMonitor);
  assertEquals(0, strcmp("task", data.subTaskName));
  assertEquals(0, strcmp("mainTask", data.taskName));

  // test worked
  subMonitor->worked(1, subMonitor);
  assertEquals(0, data.worked);
  subMonitor->worked(1, subMonitor);
  assertEquals(1, data.worked);

  // test done
  subMonitor->done(subMonitor);
  assertEquals(500, data.worked);

  resetData(&data);

  // test canceled
  assertEquals(0, subMonitor->isCanceled(subMonitor));
  data.canceled = 1;
  assertEquals(1, subMonitor->isCanceled(subMonitor));

  
  resetData(&data);
  
  // test other scaling
  freeSubProgressMonitor(subMonitor);
  subMonitor = createSubProgressMonitor(500, &monitor);
  subMonitor->beginTask("task", 250, subMonitor);
  subMonitor->worked(1, subMonitor);
  assertEquals(2, data.worked);
}

struct times
{
  int queriesAfterWhichToCancel;
  int cancelQueries;
  clock_t startTime;
  clock_t startIsCanceled;
  clock_t ticksAfterWhichToCancel;
};

static int isQueryCanceled(progressMonitor *monitor)
{
  struct tms time;
  struct times *t = (struct times*)monitor->user;
  if (t->cancelQueries > t->queriesAfterWhichToCancel) {
    return 1;
  }
  else if (t->cancelQueries == t->queriesAfterWhichToCancel) {
    times(&time);
    t->startIsCanceled = time.tms_utime;
  }
  ++(t->cancelQueries);
  return 0;
}

static int isCanceled(progressMonitor *monitor)
{
  struct tms time;
  struct times *t = (struct times*)monitor->user;
  if (t->startIsCanceled) {
    return 1;
  }
  times(&time);
  if (time.tms_utime - t->startTime > t->ticksAfterWhichToCancel) {
    t->startIsCanceled = time.tms_utime;
  }
  return 0;
}

static void testRespEveryStep()
{
  char *files[] = { "progressmonitortest-data/small_3.rem",
		    "progressmonitortest-data/small_5.rem",
		    "progressmonitortest-data/large_3.rem",
		    "progressmonitortest-data/large_5.rem",
		    "progressmonitortest-data/quicksort_3.rem",
		    "progressmonitortest-data/quicksort_5.rem",
		    "progressmonitortest-data/huge_3.rem",
		    "progressmonitortest-data/huge_5.rem",
  };
  struct times t;
  struct tms time;
  struct progressMonitor monitor = { .isCanceled = &isQueryCanceled, .user = &t,
				     .done = NULL, .worked = NULL, 
				     .beginTask = NULL, .subTask = NULL };
  int i;
  long accumlatedResponseTimes = 0;
  int count = 0;
  int index = 7;

  // += 7
  for (i = 0; i < 1000 * 1000; i += 23) {
    if (i > 0) {
      wInit();
      bddInit();
    }
    else {
       i = 71 * 1000;
    }
    t.startIsCanceled = 0;
    t.cancelQueries = 0;
    t.queriesAfterWhichToCancel = i;
    intPds *pds = irReadPdsMonitored(files[index], 0, &monitor);
    times(&time);
    if (t.startIsCanceled && pds == NULL) {
      printf("Response cycles/time %d %d ms for %d\n", 
	     time.tms_utime - t.startIsCanceled,
	     ((time.tms_utime - t.startIsCanceled) * 1000) / HZ, i);
      count++;
      accumlatedResponseTimes += time.tms_utime - t.startIsCanceled;
      /* average response time should be smaller than half a second */
      /* assert(((accumlatedResponseTimes * 1000) / (count * HZ)) < 500); */
    }
    else if (pds) {
      intPdsDelete(pds);
      break;
    }
    wFinish();
    bddFinish();
  }
  printf("Average %d\n", ((accumlatedResponseTimes * 1000) / (count * HZ)));
}

static void testResponsiveness()
{
  char *files[] = { "progressmonitortest-data/small_3.rem",
		    "progressmonitortest-data/small_5.rem",
		    "progressmonitortest-data/large_3.rem",
		    "progressmonitortest-data/large_5.rem",
		    "progressmonitortest-data/quicksort_3.rem",
		    "progressmonitortest-data/quicksort_5.rem",
		    "progressmonitortest-data/huge_3.rem",
		    "progressmonitortest-data/huge_5.rem",
  };
  struct times t;
  struct tms time;
  struct progressMonitor monitor = { .isCanceled = &isCanceled, .user = &t,
				     .done = NULL, .worked = NULL, 
				     .beginTask = NULL, .subTask = NULL };
  clock_t i;
  long accumlatedResponseTimes = 0;
  int count = 0;
  int index = 1;

  for (i = 0; i < 5 * 1000; i += (index % 2 == 0) ? 50 : 100) {
    times(&time);
    t.startTime = time.tms_utime;
    t.startIsCanceled = 0;
    t.ticksAfterWhichToCancel = i;
    intPds *pds = irReadPdsMonitored(files[index], 0, &monitor);
    times(&time);
    if (t.startIsCanceled && pds == NULL) {
      printf("Response time %d ms\n", 
	     ((time.tms_utime - t.startIsCanceled) * 1000) / HZ);
      count++;
      accumlatedResponseTimes += time.tms_utime - t.startIsCanceled;
      /* average response time should be smaller than half a second */
      assert(((accumlatedResponseTimes * 1000) / (count * HZ)) < 500);
    }
    else if (pds) {
      intPdsDelete(pds);
      break;
    }
  }
  printf("Average %d\n", ((accumlatedResponseTimes * 1000) / (count * HZ)));
}

int main(int argc, char** argv)
{
  TestTuple tests[] = {
    ADD_TEST(testSubProgressMonitor),
    ADD_LONG_TEST(testResponsiveness),
    ADD_LONG_TEST(testRespEveryStep),
  };
  setUp();
  bddSetMaxMemory(256 * 1024 * 1024);
  run_tests(sizeof(tests) / sizeof(TestTuple), tests, argc, argv);
  return 0;
}

/*
 * Local Variables:
 * compile-command: "make -k progressmonitortest"
 * End:
 */
